/* ----------------------------------------------------------------------------
 * This file was automatically generated by SWIG (http://www.swig.org).
 * Version 1.3.31
 * 
 * This file is not intended to be easily readable and contains a number of 
 * coding conventions designed to improve portability and efficiency. Do not make
 * changes to this file unless you know what you are doing--modify the SWIG 
 * interface file instead. 
 * ----------------------------------------------------------------------------- */

#define SWIGPYTHON
#define SWIG_PYTHON_DIRECTOR_NO_VTABLE

#ifdef __cplusplus
template<class T> class SwigValueWrapper {
    T *tt;
public:
    SwigValueWrapper() : tt(0) { }
    SwigValueWrapper(const SwigValueWrapper<T>& rhs) : tt(new T(*rhs.tt)) { }
    SwigValueWrapper(const T& t) : tt(new T(t)) { }
    ~SwigValueWrapper() { delete tt; } 
    SwigValueWrapper& operator=(const T& t) { delete tt; tt = new T(t); return *this; }
    operator T&() const { return *tt; }
    T *operator&() { return tt; }
private:
    SwigValueWrapper& operator=(const SwigValueWrapper<T>& rhs);
};
#endif

/* -----------------------------------------------------------------------------
 *  This section contains generic SWIG labels for method/variable
 *  declarations/attributes, and other compiler dependent labels.
 * ----------------------------------------------------------------------------- */

/* template workaround for compilers that cannot correctly implement the C++ standard */
#ifndef SWIGTEMPLATEDISAMBIGUATOR
# if defined(__SUNPRO_CC)
#   if (__SUNPRO_CC <= 0x560)
#     define SWIGTEMPLATEDISAMBIGUATOR template
#   else
#     define SWIGTEMPLATEDISAMBIGUATOR 
#   endif
# else
#   define SWIGTEMPLATEDISAMBIGUATOR 
# endif
#endif

/* inline attribute */
#ifndef SWIGINLINE
# if defined(__cplusplus) || (defined(__GNUC__) && !defined(__STRICT_ANSI__))
#   define SWIGINLINE inline
# else
#   define SWIGINLINE
# endif
#endif

/* attribute recognised by some compilers to avoid 'unused' warnings */
#ifndef SWIGUNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define SWIGUNUSED __attribute__ ((__unused__)) 
#   else
#     define SWIGUNUSED
#   endif
# elif defined(__ICC)
#   define SWIGUNUSED __attribute__ ((__unused__)) 
# else
#   define SWIGUNUSED 
# endif
#endif

#ifndef SWIGUNUSEDPARM
# ifdef __cplusplus
#   define SWIGUNUSEDPARM(p)
# else
#   define SWIGUNUSEDPARM(p) p SWIGUNUSED 
# endif
#endif

/* internal SWIG method */
#ifndef SWIGINTERN
# define SWIGINTERN static SWIGUNUSED
#endif

/* internal inline SWIG method */
#ifndef SWIGINTERNINLINE
# define SWIGINTERNINLINE SWIGINTERN SWIGINLINE
#endif

/* exporting methods */
#if (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)
#  ifndef GCC_HASCLASSVISIBILITY
#    define GCC_HASCLASSVISIBILITY
#  endif
#endif

#ifndef SWIGEXPORT
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   if defined(STATIC_LINKED)
#     define SWIGEXPORT
#   else
#     define SWIGEXPORT __declspec(dllexport)
#   endif
# else
#   if defined(__GNUC__) && defined(GCC_HASCLASSVISIBILITY)
#     define SWIGEXPORT __attribute__ ((visibility("default")))
#   else
#     define SWIGEXPORT
#   endif
# endif
#endif

/* calling conventions for Windows */
#ifndef SWIGSTDCALL
# if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#   define SWIGSTDCALL __stdcall
# else
#   define SWIGSTDCALL
# endif 
#endif

/* Deal with Microsoft's attempt at deprecating C standard runtime functions */
#if !defined(SWIG_NO_CRT_SECURE_NO_DEPRECATE) && defined(_MSC_VER) && !defined(_CRT_SECURE_NO_DEPRECATE)
# define _CRT_SECURE_NO_DEPRECATE
#endif


/* Python.h has to appear first */
#include <Python.h>

/* -----------------------------------------------------------------------------
 * swigrun.swg
 *
 * This file contains generic CAPI SWIG runtime support for pointer
 * type checking.
 * ----------------------------------------------------------------------------- */

/* This should only be incremented when either the layout of swig_type_info changes,
   or for whatever reason, the runtime changes incompatibly */
#define SWIG_RUNTIME_VERSION "3"

/* define SWIG_TYPE_TABLE_NAME as "SWIG_TYPE_TABLE" */
#ifdef SWIG_TYPE_TABLE
# define SWIG_QUOTE_STRING(x) #x
# define SWIG_EXPAND_AND_QUOTE_STRING(x) SWIG_QUOTE_STRING(x)
# define SWIG_TYPE_TABLE_NAME SWIG_EXPAND_AND_QUOTE_STRING(SWIG_TYPE_TABLE)
#else
# define SWIG_TYPE_TABLE_NAME
#endif

/*
  You can use the SWIGRUNTIME and SWIGRUNTIMEINLINE macros for
  creating a static or dynamic library from the swig runtime code.
  In 99.9% of the cases, swig just needs to declare them as 'static'.
  
  But only do this if is strictly necessary, ie, if you have problems
  with your compiler or so.
*/

#ifndef SWIGRUNTIME
# define SWIGRUNTIME SWIGINTERN
#endif

#ifndef SWIGRUNTIMEINLINE
# define SWIGRUNTIMEINLINE SWIGRUNTIME SWIGINLINE
#endif

/*  Generic buffer size */
#ifndef SWIG_BUFFER_SIZE
# define SWIG_BUFFER_SIZE 1024
#endif

/* Flags for pointer conversions */
#define SWIG_POINTER_DISOWN        0x1

/* Flags for new pointer objects */
#define SWIG_POINTER_OWN           0x1


/* 
   Flags/methods for returning states.
   
   The swig conversion methods, as ConvertPtr, return and integer 
   that tells if the conversion was successful or not. And if not,
   an error code can be returned (see swigerrors.swg for the codes).
   
   Use the following macros/flags to set or process the returning
   states.
   
   In old swig versions, you usually write code as:

     if (SWIG_ConvertPtr(obj,vptr,ty.flags) != -1) {
       // success code
     } else {
       //fail code
     }

   Now you can be more explicit as:

    int res = SWIG_ConvertPtr(obj,vptr,ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
    } else {
      // fail code
    }

   that seems to be the same, but now you can also do

    Type *ptr;
    int res = SWIG_ConvertPtr(obj,(void **)(&ptr),ty.flags);
    if (SWIG_IsOK(res)) {
      // success code
      if (SWIG_IsNewObj(res) {
        ...
	delete *ptr;
      } else {
        ...
      }
    } else {
      // fail code
    }
    
   I.e., now SWIG_ConvertPtr can return new objects and you can
   identify the case and take care of the deallocation. Of course that
   requires also to SWIG_ConvertPtr to return new result values, as

      int SWIG_ConvertPtr(obj, ptr,...) {         
        if (<obj is ok>) {			       
          if (<need new object>) {		       
            *ptr = <ptr to new allocated object>; 
            return SWIG_NEWOBJ;		       
          } else {				       
            *ptr = <ptr to old object>;	       
            return SWIG_OLDOBJ;		       
          } 				       
        } else {				       
          return SWIG_BADOBJ;		       
        }					       
      }

   Of course, returning the plain '0(success)/-1(fail)' still works, but you can be
   more explicit by returning SWIG_BADOBJ, SWIG_ERROR or any of the
   swig errors code.

   Finally, if the SWIG_CASTRANK_MODE is enabled, the result code
   allows to return the 'cast rank', for example, if you have this

       int food(double)
       int fooi(int);

   and you call
 
      food(1)   // cast rank '1'  (1 -> 1.0)
      fooi(1)   // cast rank '0'

   just use the SWIG_AddCast()/SWIG_CheckState()


 */
#define SWIG_OK                    (0) 
#define SWIG_ERROR                 (-1)
#define SWIG_IsOK(r)               (r >= 0)
#define SWIG_ArgError(r)           ((r != SWIG_ERROR) ? r : SWIG_TypeError)  

/* The CastRankLimit says how many bits are used for the cast rank */
#define SWIG_CASTRANKLIMIT         (1 << 8)
/* The NewMask denotes the object was created (using new/malloc) */
#define SWIG_NEWOBJMASK            (SWIG_CASTRANKLIMIT  << 1)
/* The TmpMask is for in/out typemaps that use temporal objects */
#define SWIG_TMPOBJMASK            (SWIG_NEWOBJMASK << 1)
/* Simple returning values */
#define SWIG_BADOBJ                (SWIG_ERROR)
#define SWIG_OLDOBJ                (SWIG_OK)
#define SWIG_NEWOBJ                (SWIG_OK | SWIG_NEWOBJMASK)
#define SWIG_TMPOBJ                (SWIG_OK | SWIG_TMPOBJMASK)
/* Check, add and del mask methods */
#define SWIG_AddNewMask(r)         (SWIG_IsOK(r) ? (r | SWIG_NEWOBJMASK) : r)
#define SWIG_DelNewMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_NEWOBJMASK) : r)
#define SWIG_IsNewObj(r)           (SWIG_IsOK(r) && (r & SWIG_NEWOBJMASK))
#define SWIG_AddTmpMask(r)         (SWIG_IsOK(r) ? (r | SWIG_TMPOBJMASK) : r)
#define SWIG_DelTmpMask(r)         (SWIG_IsOK(r) ? (r & ~SWIG_TMPOBJMASK) : r)
#define SWIG_IsTmpObj(r)           (SWIG_IsOK(r) && (r & SWIG_TMPOBJMASK))


/* Cast-Rank Mode */
#if defined(SWIG_CASTRANK_MODE)
#  ifndef SWIG_TypeRank
#    define SWIG_TypeRank             unsigned long
#  endif
#  ifndef SWIG_MAXCASTRANK            /* Default cast allowed */
#    define SWIG_MAXCASTRANK          (2)
#  endif
#  define SWIG_CASTRANKMASK          ((SWIG_CASTRANKLIMIT) -1)
#  define SWIG_CastRank(r)           (r & SWIG_CASTRANKMASK)
SWIGINTERNINLINE int SWIG_AddCast(int r) { 
  return SWIG_IsOK(r) ? ((SWIG_CastRank(r) < SWIG_MAXCASTRANK) ? (r + 1) : SWIG_ERROR) : r;
}
SWIGINTERNINLINE int SWIG_CheckState(int r) { 
  return SWIG_IsOK(r) ? SWIG_CastRank(r) + 1 : 0; 
}
#else /* no cast-rank mode */
#  define SWIG_AddCast
#  define SWIG_CheckState(r) (SWIG_IsOK(r) ? 1 : 0)
#endif




#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

typedef void *(*swig_converter_func)(void *);
typedef struct swig_type_info *(*swig_dycast_func)(void **);

/* Structure to store inforomation on one type */
typedef struct swig_type_info {
  const char             *name;			/* mangled name of this type */
  const char             *str;			/* human readable name of this type */
  swig_dycast_func        dcast;		/* dynamic cast function down a hierarchy */
  struct swig_cast_info  *cast;			/* linked list of types that can cast into this type */
  void                   *clientdata;		/* language specific type data */
  int                    owndata;		/* flag if the structure owns the clientdata */
} swig_type_info;

/* Structure to store a type and conversion function used for casting */
typedef struct swig_cast_info {
  swig_type_info         *type;			/* pointer to type that is equivalent to this type */
  swig_converter_func     converter;		/* function to cast the void pointers */
  struct swig_cast_info  *next;			/* pointer to next cast in linked list */
  struct swig_cast_info  *prev;			/* pointer to the previous cast */
} swig_cast_info;

/* Structure used to store module information
 * Each module generates one structure like this, and the runtime collects
 * all of these structures and stores them in a circularly linked list.*/
typedef struct swig_module_info {
  swig_type_info         **types;		/* Array of pointers to swig_type_info structures that are in this module */
  size_t                 size;		        /* Number of types in this module */
  struct swig_module_info *next;		/* Pointer to next element in circularly linked list */
  swig_type_info         **type_initial;	/* Array of initially generated type structures */
  swig_cast_info         **cast_initial;	/* Array of initially generated casting structures */
  void                    *clientdata;		/* Language specific module data */
} swig_module_info;

/* 
  Compare two type names skipping the space characters, therefore
  "char*" == "char *" and "Class<int>" == "Class<int >", etc.

  Return 0 when the two name types are equivalent, as in
  strncmp, but skipping ' '.
*/
SWIGRUNTIME int
SWIG_TypeNameComp(const char *f1, const char *l1,
		  const char *f2, const char *l2) {
  for (;(f1 != l1) && (f2 != l2); ++f1, ++f2) {
    while ((*f1 == ' ') && (f1 != l1)) ++f1;
    while ((*f2 == ' ') && (f2 != l2)) ++f2;
    if (*f1 != *f2) return (*f1 > *f2) ? 1 : -1;
  }
  return (l1 - f1) - (l2 - f2);
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if not equal, 1 if equal
*/
SWIGRUNTIME int
SWIG_TypeEquiv(const char *nb, const char *tb) {
  int equiv = 0;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (!equiv && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = (SWIG_TypeNameComp(nb, ne, tb, te) == 0) ? 1 : 0;
    if (*ne) ++ne;
  }
  return equiv;
}

/*
  Check type equivalence in a name list like <name1>|<name2>|...
  Return 0 if equal, -1 if nb < tb, 1 if nb > tb
*/
SWIGRUNTIME int
SWIG_TypeCompare(const char *nb, const char *tb) {
  int equiv = 0;
  const char* te = tb + strlen(tb);
  const char* ne = nb;
  while (!equiv && *ne) {
    for (nb = ne; *ne; ++ne) {
      if (*ne == '|') break;
    }
    equiv = (SWIG_TypeNameComp(nb, ne, tb, te) == 0) ? 1 : 0;
    if (*ne) ++ne;
  }
  return equiv;
}


/* think of this as a c++ template<> or a scheme macro */
#define SWIG_TypeCheck_Template(comparison, ty)         \
  if (ty) {                                             \
    swig_cast_info *iter = ty->cast;                    \
    while (iter) {                                      \
      if (comparison) {                                 \
        if (iter == ty->cast) return iter;              \
        /* Move iter to the top of the linked list */   \
        iter->prev->next = iter->next;                  \
        if (iter->next)                                 \
          iter->next->prev = iter->prev;                \
        iter->next = ty->cast;                          \
        iter->prev = 0;                                 \
        if (ty->cast) ty->cast->prev = iter;            \
        ty->cast = iter;                                \
        return iter;                                    \
      }                                                 \
      iter = iter->next;                                \
    }                                                   \
  }                                                     \
  return 0

/*
  Check the typename
*/
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheck(const char *c, swig_type_info *ty) {
  SWIG_TypeCheck_Template(strcmp(iter->type->name, c) == 0, ty);
}

/* Same as previous function, except strcmp is replaced with a pointer comparison */
SWIGRUNTIME swig_cast_info *
SWIG_TypeCheckStruct(swig_type_info *from, swig_type_info *into) {
  SWIG_TypeCheck_Template(iter->type == from, into);
}

/*
  Cast a pointer up an inheritance hierarchy
*/
SWIGRUNTIMEINLINE void *
SWIG_TypeCast(swig_cast_info *ty, void *ptr) {
  return ((!ty) || (!ty->converter)) ? ptr : (*ty->converter)(ptr);
}

/* 
   Dynamic pointer casting. Down an inheritance hierarchy
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeDynamicCast(swig_type_info *ty, void **ptr) {
  swig_type_info *lastty = ty;
  if (!ty || !ty->dcast) return ty;
  while (ty && (ty->dcast)) {
    ty = (*ty->dcast)(ptr);
    if (ty) lastty = ty;
  }
  return lastty;
}

/*
  Return the name associated with this type
*/
SWIGRUNTIMEINLINE const char *
SWIG_TypeName(const swig_type_info *ty) {
  return ty->name;
}

/*
  Return the pretty name associated with this type,
  that is an unmangled type name in a form presentable to the user.
*/
SWIGRUNTIME const char *
SWIG_TypePrettyName(const swig_type_info *type) {
  /* The "str" field contains the equivalent pretty names of the
     type, separated by vertical-bar characters.  We choose
     to print the last name, as it is often (?) the most
     specific. */
  if (!type) return NULL;
  if (type->str != NULL) {
    const char *last_name = type->str;
    const char *s;
    for (s = type->str; *s; s++)
      if (*s == '|') last_name = s+1;
    return last_name;
  }
  else
    return type->name;
}

/* 
   Set the clientdata field for a type
*/
SWIGRUNTIME void
SWIG_TypeClientData(swig_type_info *ti, void *clientdata) {
  swig_cast_info *cast = ti->cast;
  /* if (ti->clientdata == clientdata) return; */
  ti->clientdata = clientdata;
  
  while (cast) {
    if (!cast->converter) {
      swig_type_info *tc = cast->type;
      if (!tc->clientdata) {
	SWIG_TypeClientData(tc, clientdata);
      }
    }    
    cast = cast->next;
  }
}
SWIGRUNTIME void
SWIG_TypeNewClientData(swig_type_info *ti, void *clientdata) {
  SWIG_TypeClientData(ti, clientdata);
  ti->owndata = 1;
}
  
/*
  Search for a swig_type_info structure only by mangled name
  Search is a O(log #types)
  
  We start searching at module start, and finish searching when start == end.  
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_MangledTypeQueryModule(swig_module_info *start, 
                            swig_module_info *end, 
		            const char *name) {
  swig_module_info *iter = start;
  do {
    if (iter->size) {
      register size_t l = 0;
      register size_t r = iter->size - 1;
      do {
	/* since l+r >= 0, we can (>> 1) instead (/ 2) */
	register size_t i = (l + r) >> 1; 
	const char *iname = iter->types[i]->name;
	if (iname) {
	  register int compare = strcmp(name, iname);
	  if (compare == 0) {	    
	    return iter->types[i];
	  } else if (compare < 0) {
	    if (i) {
	      r = i - 1;
	    } else {
	      break;
	    }
	  } else if (compare > 0) {
	    l = i + 1;
	  }
	} else {
	  break; /* should never happen */
	}
      } while (l <= r);
    }
    iter = iter->next;
  } while (iter != end);
  return 0;
}

/*
  Search for a swig_type_info structure for either a mangled name or a human readable name.
  It first searches the mangled names of the types, which is a O(log #types)
  If a type is not found it then searches the human readable names, which is O(#types).
  
  We start searching at module start, and finish searching when start == end.  
  Note: if start == end at the beginning of the function, we go all the way around
  the circular list.
*/
SWIGRUNTIME swig_type_info *
SWIG_TypeQueryModule(swig_module_info *start, 
                     swig_module_info *end, 
		     const char *name) {
  /* STEP 1: Search the name field using binary search */
  swig_type_info *ret = SWIG_MangledTypeQueryModule(start, end, name);
  if (ret) {
    return ret;
  } else {
    /* STEP 2: If the type hasn't been found, do a complete search
       of the str field (the human readable name) */
    swig_module_info *iter = start;
    do {
      register size_t i = 0;
      for (; i < iter->size; ++i) {
	if (iter->types[i]->str && (SWIG_TypeEquiv(iter->types[i]->str, name)))
	  return iter->types[i];
      }
      iter = iter->next;
    } while (iter != end);
  }
  
  /* neither found a match */
  return 0;
}

/* 
   Pack binary data into a string
*/
SWIGRUNTIME char *
SWIG_PackData(char *c, void *ptr, size_t sz) {
  static const char hex[17] = "0123456789abcdef";
  register const unsigned char *u = (unsigned char *) ptr;
  register const unsigned char *eu =  u + sz;
  for (; u != eu; ++u) {
    register unsigned char uu = *u;
    *(c++) = hex[(uu & 0xf0) >> 4];
    *(c++) = hex[uu & 0xf];
  }
  return c;
}

/* 
   Unpack binary data from a string
*/
SWIGRUNTIME const char *
SWIG_UnpackData(const char *c, void *ptr, size_t sz) {
  register unsigned char *u = (unsigned char *) ptr;
  register const unsigned char *eu = u + sz;
  for (; u != eu; ++u) {
    register char d = *(c++);
    register unsigned char uu;
    if ((d >= '0') && (d <= '9'))
      uu = ((d - '0') << 4);
    else if ((d >= 'a') && (d <= 'f'))
      uu = ((d - ('a'-10)) << 4);
    else 
      return (char *) 0;
    d = *(c++);
    if ((d >= '0') && (d <= '9'))
      uu |= (d - '0');
    else if ((d >= 'a') && (d <= 'f'))
      uu |= (d - ('a'-10));
    else 
      return (char *) 0;
    *u = uu;
  }
  return c;
}

/* 
   Pack 'void *' into a string buffer.
*/
SWIGRUNTIME char *
SWIG_PackVoidPtr(char *buff, void *ptr, const char *name, size_t bsz) {
  char *r = buff;
  if ((2*sizeof(void *) + 2) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,&ptr,sizeof(void *));
  if (strlen(name) + 1 > (bsz - (r - buff))) return 0;
  strcpy(r,name);
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackVoidPtr(const char *c, void **ptr, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      *ptr = (void *) 0;
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sizeof(void *));
}

SWIGRUNTIME char *
SWIG_PackDataName(char *buff, void *ptr, size_t sz, const char *name, size_t bsz) {
  char *r = buff;
  size_t lname = (name ? strlen(name) : 0);
  if ((2*sz + 2 + lname) > bsz) return 0;
  *(r++) = '_';
  r = SWIG_PackData(r,ptr,sz);
  if (lname) {
    strncpy(r,name,lname+1);
  } else {
    *r = 0;
  }
  return buff;
}

SWIGRUNTIME const char *
SWIG_UnpackDataName(const char *c, void *ptr, size_t sz, const char *name) {
  if (*c != '_') {
    if (strcmp(c,"NULL") == 0) {
      memset(ptr,0,sz);
      return name;
    } else {
      return 0;
    }
  }
  return SWIG_UnpackData(++c,ptr,sz);
}

#ifdef __cplusplus
}
#endif

/*  Errors in SWIG */
#define  SWIG_UnknownError    	   -1 
#define  SWIG_IOError        	   -2 
#define  SWIG_RuntimeError   	   -3 
#define  SWIG_IndexError     	   -4 
#define  SWIG_TypeError      	   -5 
#define  SWIG_DivisionByZero 	   -6 
#define  SWIG_OverflowError  	   -7 
#define  SWIG_SyntaxError    	   -8 
#define  SWIG_ValueError     	   -9 
#define  SWIG_SystemError    	   -10
#define  SWIG_AttributeError 	   -11
#define  SWIG_MemoryError    	   -12 
#define  SWIG_NullReferenceError   -13




/* Add PyOS_snprintf for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# if defined(_MSC_VER) || defined(__BORLANDC__) || defined(_WATCOM)
#  define PyOS_snprintf _snprintf
# else
#  define PyOS_snprintf snprintf
# endif
#endif

/* A crude PyString_FromFormat implementation for old Pythons */
#if PY_VERSION_HEX < 0x02020000

#ifndef SWIG_PYBUFFER_SIZE
# define SWIG_PYBUFFER_SIZE 1024
#endif

static PyObject *
PyString_FromFormat(const char *fmt, ...) {
  va_list ap;
  char buf[SWIG_PYBUFFER_SIZE * 2];
  int res;
  va_start(ap, fmt);
  res = vsnprintf(buf, sizeof(buf), fmt, ap);
  va_end(ap);
  return (res < 0 || res >= (int)sizeof(buf)) ? 0 : PyString_FromString(buf);
}
#endif

/* Add PyObject_Del for old Pythons */
#if PY_VERSION_HEX < 0x01060000
# define PyObject_Del(op) PyMem_DEL((op))
#endif
#ifndef PyObject_DEL
# define PyObject_DEL PyObject_Del
#endif

/* A crude PyExc_StopIteration exception for old Pythons */
#if PY_VERSION_HEX < 0x02020000
# ifndef PyExc_StopIteration
#  define PyExc_StopIteration PyExc_RuntimeError
# endif
# ifndef PyObject_GenericGetAttr
#  define PyObject_GenericGetAttr 0
# endif
#endif
/* Py_NotImplemented is defined in 2.1 and up. */
#if PY_VERSION_HEX < 0x02010000
# ifndef Py_NotImplemented
#  define Py_NotImplemented PyExc_RuntimeError
# endif
#endif


/* A crude PyString_AsStringAndSize implementation for old Pythons */
#if PY_VERSION_HEX < 0x02010000
# ifndef PyString_AsStringAndSize
#  define PyString_AsStringAndSize(obj, s, len) {*s = PyString_AsString(obj); *len = *s ? strlen(*s) : 0;}
# endif
#endif

/* PySequence_Size for old Pythons */
#if PY_VERSION_HEX < 0x02000000
# ifndef PySequence_Size
#  define PySequence_Size PySequence_Length
# endif
#endif


/* PyBool_FromLong for old Pythons */
#if PY_VERSION_HEX < 0x02030000
static
PyObject *PyBool_FromLong(long ok)
{
  PyObject *result = ok ? Py_True : Py_False;
  Py_INCREF(result);
  return result;
}
#endif

/* Py_ssize_t for old Pythons */
/* This code is as recommended by: */
/* http://www.python.org/dev/peps/pep-0353/#conversion-guidelines */
#if PY_VERSION_HEX < 0x02050000 && !defined(PY_SSIZE_T_MIN)
typedef int Py_ssize_t;
# define PY_SSIZE_T_MAX INT_MAX
# define PY_SSIZE_T_MIN INT_MIN
#endif

/* -----------------------------------------------------------------------------
 * error manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIME PyObject*
SWIG_Python_ErrorType(int code) {
  PyObject* type = 0;
  switch(code) {
  case SWIG_MemoryError:
    type = PyExc_MemoryError;
    break;
  case SWIG_IOError:
    type = PyExc_IOError;
    break;
  case SWIG_RuntimeError:
    type = PyExc_RuntimeError;
    break;
  case SWIG_IndexError:
    type = PyExc_IndexError;
    break;
  case SWIG_TypeError:
    type = PyExc_TypeError;
    break;
  case SWIG_DivisionByZero:
    type = PyExc_ZeroDivisionError;
    break;
  case SWIG_OverflowError:
    type = PyExc_OverflowError;
    break;
  case SWIG_SyntaxError:
    type = PyExc_SyntaxError;
    break;
  case SWIG_ValueError:
    type = PyExc_ValueError;
    break;
  case SWIG_SystemError:
    type = PyExc_SystemError;
    break;
  case SWIG_AttributeError:
    type = PyExc_AttributeError;
    break;
  default:
    type = PyExc_RuntimeError;
  }
  return type;
}


SWIGRUNTIME void
SWIG_Python_AddErrorMsg(const char* mesg)
{
  PyObject *type = 0;
  PyObject *value = 0;
  PyObject *traceback = 0;

  if (PyErr_Occurred()) PyErr_Fetch(&type, &value, &traceback);
  if (value) {
    PyObject *old_str = PyObject_Str(value);
    PyErr_Clear();
    Py_XINCREF(type);
    PyErr_Format(type, "%s %s", PyString_AsString(old_str), mesg);
    Py_DECREF(old_str);
    Py_DECREF(value);
  } else {
    PyErr_Format(PyExc_RuntimeError, mesg);
  }
}



#if defined(SWIG_PYTHON_NO_THREADS)
#  if defined(SWIG_PYTHON_THREADS)
#    undef SWIG_PYTHON_THREADS
#  endif
#endif
#if defined(SWIG_PYTHON_THREADS) /* Threading support is enabled */
#  if !defined(SWIG_PYTHON_USE_GIL) && !defined(SWIG_PYTHON_NO_USE_GIL)
#    if (PY_VERSION_HEX >= 0x02030000) /* For 2.3 or later, use the PyGILState calls */
#      define SWIG_PYTHON_USE_GIL
#    endif
#  endif
#  if defined(SWIG_PYTHON_USE_GIL) /* Use PyGILState threads calls */
#    ifndef SWIG_PYTHON_INITIALIZE_THREADS
#     define SWIG_PYTHON_INITIALIZE_THREADS  PyEval_InitThreads() 
#    endif
#    ifdef __cplusplus /* C++ code */
       class SWIG_Python_Thread_Block {
         bool status;
         PyGILState_STATE state;
       public:
         void end() { if (status) { PyGILState_Release(state); status = false;} }
         SWIG_Python_Thread_Block() : status(true), state(PyGILState_Ensure()) {}
         ~SWIG_Python_Thread_Block() { end(); }
       };
       class SWIG_Python_Thread_Allow {
         bool status;
         PyThreadState *save;
       public:
         void end() { if (status) { PyEval_RestoreThread(save); status = false; }}
         SWIG_Python_Thread_Allow() : status(true), save(PyEval_SaveThread()) {}
         ~SWIG_Python_Thread_Allow() { end(); }
       };
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   SWIG_Python_Thread_Block _swig_thread_block
#      define SWIG_PYTHON_THREAD_END_BLOCK     _swig_thread_block.end()
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   SWIG_Python_Thread_Allow _swig_thread_allow
#      define SWIG_PYTHON_THREAD_END_ALLOW     _swig_thread_allow.end()
#    else /* C code */
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK   PyGILState_STATE _swig_thread_block = PyGILState_Ensure()
#      define SWIG_PYTHON_THREAD_END_BLOCK     PyGILState_Release(_swig_thread_block)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW   PyThreadState *_swig_thread_allow = PyEval_SaveThread()
#      define SWIG_PYTHON_THREAD_END_ALLOW     PyEval_RestoreThread(_swig_thread_allow)
#    endif
#  else /* Old thread way, not implemented, user must provide it */
#    if !defined(SWIG_PYTHON_INITIALIZE_THREADS)
#      define SWIG_PYTHON_INITIALIZE_THREADS
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_BLOCK)
#      define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_BLOCK)
#      define SWIG_PYTHON_THREAD_END_BLOCK
#    endif
#    if !defined(SWIG_PYTHON_THREAD_BEGIN_ALLOW)
#      define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#    endif
#    if !defined(SWIG_PYTHON_THREAD_END_ALLOW)
#      define SWIG_PYTHON_THREAD_END_ALLOW
#    endif
#  endif
#else /* No thread support */
#  define SWIG_PYTHON_INITIALIZE_THREADS
#  define SWIG_PYTHON_THREAD_BEGIN_BLOCK
#  define SWIG_PYTHON_THREAD_END_BLOCK
#  define SWIG_PYTHON_THREAD_BEGIN_ALLOW
#  define SWIG_PYTHON_THREAD_END_ALLOW
#endif

/* -----------------------------------------------------------------------------
 * Python API portion that goes into the runtime
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* cc-mode */
#endif
#endif

/* -----------------------------------------------------------------------------
 * Constant declarations
 * ----------------------------------------------------------------------------- */

/* Constant Types */
#define SWIG_PY_POINTER 4
#define SWIG_PY_BINARY  5

/* Constant information structure */
typedef struct swig_const_info {
  int type;
  char *name;
  long lvalue;
  double dvalue;
  void   *pvalue;
  swig_type_info **ptype;
} swig_const_info;

#ifdef __cplusplus
#if 0
{ /* cc-mode */
#endif
}
#endif


/* -----------------------------------------------------------------------------
 * See the LICENSE file for information on copyright, usage and redistribution
 * of SWIG, and the README file for authors - http://www.swig.org/release.html.
 *
 * pyrun.swg
 *
 * This file contains the runtime support for Python modules
 * and includes code for managing global variables and pointer
 * type checking.
 *
 * ----------------------------------------------------------------------------- */

/* Common SWIG API */

/* for raw pointers */
#define SWIG_Python_ConvertPtr(obj, pptr, type, flags)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, 0)
#define SWIG_ConvertPtr(obj, pptr, type, flags)         SWIG_Python_ConvertPtr(obj, pptr, type, flags)
#define SWIG_ConvertPtrAndOwn(obj,pptr,type,flags,own)  SWIG_Python_ConvertPtrAndOwn(obj, pptr, type, flags, own)
#define SWIG_NewPointerObj(ptr, type, flags)            SWIG_Python_NewPointerObj(ptr, type, flags)
#define SWIG_CheckImplicit(ty)                          SWIG_Python_CheckImplicit(ty) 
#define SWIG_AcquirePtr(ptr, src)                       SWIG_Python_AcquirePtr(ptr, src)
#define swig_owntype                                    int

/* for raw packed data */
#define SWIG_ConvertPacked(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewPackedObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)

/* for class or struct pointers */
#define SWIG_ConvertInstance(obj, pptr, type, flags)    SWIG_ConvertPtr(obj, pptr, type, flags)
#define SWIG_NewInstanceObj(ptr, type, flags)           SWIG_NewPointerObj(ptr, type, flags)

/* for C or C++ function pointers */
#define SWIG_ConvertFunctionPtr(obj, pptr, type)        SWIG_Python_ConvertFunctionPtr(obj, pptr, type)
#define SWIG_NewFunctionPtrObj(ptr, type)               SWIG_Python_NewPointerObj(ptr, type, 0)

/* for C++ member pointers, ie, member methods */
#define SWIG_ConvertMember(obj, ptr, sz, ty)            SWIG_Python_ConvertPacked(obj, ptr, sz, ty)
#define SWIG_NewMemberObj(ptr, sz, type)                SWIG_Python_NewPackedObj(ptr, sz, type)


/* Runtime API */

#define SWIG_GetModule(clientdata)                      SWIG_Python_GetModule()
#define SWIG_SetModule(clientdata, pointer)             SWIG_Python_SetModule(pointer)
#define SWIG_NewClientData(obj)                         PySwigClientData_New(obj)

#define SWIG_SetErrorObj                                SWIG_Python_SetErrorObj                            
#define SWIG_SetErrorMsg                        	SWIG_Python_SetErrorMsg				   
#define SWIG_ErrorType(code)                    	SWIG_Python_ErrorType(code)                        
#define SWIG_Error(code, msg)            		SWIG_Python_SetErrorMsg(SWIG_ErrorType(code), msg) 
#define SWIG_fail                        		goto fail					   


/* Runtime API implementation */

/* Error manipulation */

SWIGINTERN void 
SWIG_Python_SetErrorObj(PyObject *errtype, PyObject *obj) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK; 
  PyErr_SetObject(errtype, obj);
  Py_DECREF(obj);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

SWIGINTERN void 
SWIG_Python_SetErrorMsg(PyObject *errtype, const char *msg) {
  SWIG_PYTHON_THREAD_BEGIN_BLOCK;
  PyErr_SetString(errtype, (char *) msg);
  SWIG_PYTHON_THREAD_END_BLOCK;
}

#define SWIG_Python_Raise(obj, type, desc)  SWIG_Python_SetErrorObj(SWIG_Python_ExceptionType(desc), obj)

/* Set a constant value */

SWIGINTERN void
SWIG_Python_SetConstant(PyObject *d, const char *name, PyObject *obj) {   
  PyDict_SetItemString(d, (char*) name, obj);
  Py_DECREF(obj);                            
}

/* Append a value to the result obj */

SWIGINTERN PyObject*
SWIG_Python_AppendOutput(PyObject* result, PyObject* obj) {
#if !defined(SWIG_PYTHON_OUTPUT_TUPLE)
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyList_Check(result)) {
      PyObject *o2 = result;
      result = PyList_New(1);
      PyList_SetItem(result, 0, o2);
    }
    PyList_Append(result,obj);
    Py_DECREF(obj);
  }
  return result;
#else
  PyObject*   o2;
  PyObject*   o3;
  if (!result) {
    result = obj;
  } else if (result == Py_None) {
    Py_DECREF(result);
    result = obj;
  } else {
    if (!PyTuple_Check(result)) {
      o2 = result;
      result = PyTuple_New(1);
      PyTuple_SET_ITEM(result, 0, o2);
    }
    o3 = PyTuple_New(1);
    PyTuple_SET_ITEM(o3, 0, obj);
    o2 = result;
    result = PySequence_Concat(o2, o3);
    Py_DECREF(o2);
    Py_DECREF(o3);
  }
  return result;
#endif
}

/* Unpack the argument tuple */

SWIGINTERN int
SWIG_Python_UnpackTuple(PyObject *args, const char *name, int min, int max, PyObject **objs)
{
  if (!args) {
    if (!min && !max) {
      return 1;
    } else {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got none", 
		   name, (min == max ? "" : "at least "), min);
      return 0;
    }
  }  
  if (!PyTuple_Check(args)) {
    PyErr_SetString(PyExc_SystemError, "UnpackTuple() argument list is not a tuple");
    return 0;
  } else {
    register int l = PyTuple_GET_SIZE(args);
    if (l < min) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at least "), min, l);
      return 0;
    } else if (l > max) {
      PyErr_Format(PyExc_TypeError, "%s expected %s%d arguments, got %d", 
		   name, (min == max ? "" : "at most "), max, l);
      return 0;
    } else {
      register int i;
      for (i = 0; i < l; ++i) {
	objs[i] = PyTuple_GET_ITEM(args, i);
      }
      for (; l < max; ++l) {
	objs[l] = 0;
      }
      return i + 1;
    }    
  }
}

/* A functor is a function object with one single object argument */
#if PY_VERSION_HEX >= 0x02020000
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunctionObjArgs(functor, obj, NULL);
#else
#define SWIG_Python_CallFunctor(functor, obj)	        PyObject_CallFunction(functor, "O", obj);
#endif

/*
  Helper for static pointer initialization for both C and C++ code, for example
  static PyObject *SWIG_STATIC_POINTER(MyVar) = NewSomething(...);
*/
#ifdef __cplusplus
#define SWIG_STATIC_POINTER(var)  var
#else
#define SWIG_STATIC_POINTER(var)  var = 0; if (!var) var
#endif

/* -----------------------------------------------------------------------------
 * Pointer declarations
 * ----------------------------------------------------------------------------- */

/* Flags for new pointer objects */
#define SWIG_POINTER_NOSHADOW       (SWIG_POINTER_OWN      << 1)
#define SWIG_POINTER_NEW            (SWIG_POINTER_NOSHADOW | SWIG_POINTER_OWN)

#define SWIG_POINTER_IMPLICIT_CONV  (SWIG_POINTER_DISOWN   << 1)

#ifdef __cplusplus
extern "C" {
#if 0
} /* cc-mode */
#endif
#endif

/*  How to access Py_None */
#if defined(_WIN32) || defined(__WIN32__) || defined(__CYGWIN__)
#  ifndef SWIG_PYTHON_NO_BUILD_NONE
#    ifndef SWIG_PYTHON_BUILD_NONE
#      define SWIG_PYTHON_BUILD_NONE
#    endif
#  endif
#endif

#ifdef SWIG_PYTHON_BUILD_NONE
#  ifdef Py_None
#   undef Py_None
#   define Py_None SWIG_Py_None()
#  endif
SWIGRUNTIMEINLINE PyObject * 
_SWIG_Py_None(void)
{
  PyObject *none = Py_BuildValue((char*)"");
  Py_DECREF(none);
  return none;
}
SWIGRUNTIME PyObject * 
SWIG_Py_None(void)
{
  static PyObject *SWIG_STATIC_POINTER(none) = _SWIG_Py_None();
  return none;
}
#endif

/* The python void return value */

SWIGRUNTIMEINLINE PyObject * 
SWIG_Py_Void(void)
{
  PyObject *none = Py_None;
  Py_INCREF(none);
  return none;
}

/* PySwigClientData */

typedef struct {
  PyObject *klass;
  PyObject *newraw;
  PyObject *newargs;
  PyObject *destroy;
  int delargs;
  int implicitconv;
} PySwigClientData;

SWIGRUNTIMEINLINE int 
SWIG_Python_CheckImplicit(swig_type_info *ty)
{
  PySwigClientData *data = (PySwigClientData *)ty->clientdata;
  return data ? data->implicitconv : 0;
}

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_ExceptionType(swig_type_info *desc) {
  PySwigClientData *data = desc ? (PySwigClientData *) desc->clientdata : 0;
  PyObject *klass = data ? data->klass : 0;
  return (klass ? klass : PyExc_RuntimeError);
}


SWIGRUNTIME PySwigClientData * 
PySwigClientData_New(PyObject* obj)
{
  if (!obj) {
    return 0;
  } else {
    PySwigClientData *data = (PySwigClientData *)malloc(sizeof(PySwigClientData));
    /* the klass element */
    data->klass = obj;
    Py_INCREF(data->klass);
    /* the newraw method and newargs arguments used to create a new raw instance */
    if (PyClass_Check(obj)) {
      data->newraw = 0;
      data->newargs = obj;
      Py_INCREF(obj);
    } else {
#if (PY_VERSION_HEX < 0x02020000)
      data->newraw = 0;
#else
      data->newraw = PyObject_GetAttrString(data->klass, (char *)"__new__");
#endif
      if (data->newraw) {
	Py_INCREF(data->newraw);
	data->newargs = PyTuple_New(1);
	PyTuple_SetItem(data->newargs, 0, obj);
      } else {
	data->newargs = obj;
      }
      Py_INCREF(data->newargs);
    }
    /* the destroy method, aka as the C++ delete method */
    data->destroy = PyObject_GetAttrString(data->klass, (char *)"__swig_destroy__");
    if (PyErr_Occurred()) {
      PyErr_Clear();
      data->destroy = 0;
    }
    if (data->destroy) {
      int flags;
      Py_INCREF(data->destroy);
      flags = PyCFunction_GET_FLAGS(data->destroy);
#ifdef METH_O
      data->delargs = !(flags & (METH_O));
#else
      data->delargs = 0;
#endif
    } else {
      data->delargs = 0;
    }
    data->implicitconv = 0;
    return data;
  }
}

SWIGRUNTIME void 
PySwigClientData_Del(PySwigClientData* data)
{
  Py_XDECREF(data->newraw);
  Py_XDECREF(data->newargs);
  Py_XDECREF(data->destroy);
}

/* =============== PySwigObject =====================*/

typedef struct {
  PyObject_HEAD
  void *ptr;
  swig_type_info *ty;
  int own;
  PyObject *next;
} PySwigObject;

SWIGRUNTIME PyObject *
PySwigObject_long(PySwigObject *v)
{
  return PyLong_FromVoidPtr(v->ptr);
}

SWIGRUNTIME PyObject *
PySwigObject_format(const char* fmt, PySwigObject *v)
{
  PyObject *res = NULL;
  PyObject *args = PyTuple_New(1);
  if (args) {
    if (PyTuple_SetItem(args, 0, PySwigObject_long(v)) == 0) {
      PyObject *ofmt = PyString_FromString(fmt);
      if (ofmt) {
	res = PyString_Format(ofmt,args);
	Py_DECREF(ofmt);
      }
      Py_DECREF(args);
    }
  }
  return res;
}

SWIGRUNTIME PyObject *
PySwigObject_oct(PySwigObject *v)
{
  return PySwigObject_format("%o",v);
}

SWIGRUNTIME PyObject *
PySwigObject_hex(PySwigObject *v)
{
  return PySwigObject_format("%x",v);
}

SWIGRUNTIME PyObject *
#ifdef METH_NOARGS
PySwigObject_repr(PySwigObject *v)
#else
PySwigObject_repr(PySwigObject *v, PyObject *args)
#endif
{
  const char *name = SWIG_TypePrettyName(v->ty);
  PyObject *hex = PySwigObject_hex(v);    
  PyObject *repr = PyString_FromFormat("<Swig Object of type '%s' at 0x%s>", name, PyString_AsString(hex));
  Py_DECREF(hex);
  if (v->next) {
#ifdef METH_NOARGS
    PyObject *nrep = PySwigObject_repr((PySwigObject *)v->next);
#else
    PyObject *nrep = PySwigObject_repr((PySwigObject *)v->next, args);
#endif
    PyString_ConcatAndDel(&repr,nrep);
  }
  return repr;  
}

SWIGRUNTIME int
PySwigObject_print(PySwigObject *v, FILE *fp, int SWIGUNUSEDPARM(flags))
{
#ifdef METH_NOARGS
  PyObject *repr = PySwigObject_repr(v);
#else
  PyObject *repr = PySwigObject_repr(v, NULL);
#endif
  if (repr) {
    fputs(PyString_AsString(repr), fp);
    Py_DECREF(repr);
    return 0; 
  } else {
    return 1; 
  }
}

SWIGRUNTIME PyObject *
PySwigObject_str(PySwigObject *v)
{
  char result[SWIG_BUFFER_SIZE];
  return SWIG_PackVoidPtr(result, v->ptr, v->ty->name, sizeof(result)) ?
    PyString_FromString(result) : 0;
}

SWIGRUNTIME int
PySwigObject_compare(PySwigObject *v, PySwigObject *w)
{
  void *i = v->ptr;
  void *j = w->ptr;
  return (i < j) ? -1 : ((i > j) ? 1 : 0);
}

SWIGRUNTIME PyTypeObject* _PySwigObject_type(void);

SWIGRUNTIME PyTypeObject*
PySwigObject_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = _PySwigObject_type();
  return type;
}

SWIGRUNTIMEINLINE int
PySwigObject_Check(PyObject *op) {
  return ((op)->ob_type == PySwigObject_type())
    || (strcmp((op)->ob_type->tp_name,"PySwigObject") == 0);
}

SWIGRUNTIME PyObject *
PySwigObject_New(void *ptr, swig_type_info *ty, int own);

SWIGRUNTIME void
PySwigObject_dealloc(PyObject *v)
{
  PySwigObject *sobj = (PySwigObject *) v;
  PyObject *next = sobj->next;
  if (sobj->own) {
    swig_type_info *ty = sobj->ty;
    PySwigClientData *data = ty ? (PySwigClientData *) ty->clientdata : 0;
    PyObject *destroy = data ? data->destroy : 0;
    if (destroy) {
      /* destroy is always a VARARGS method */
      PyObject *res;
      if (data->delargs) {
	/* we need to create a temporal object to carry the destroy operation */
	PyObject *tmp = PySwigObject_New(sobj->ptr, ty, 0);
	res = SWIG_Python_CallFunctor(destroy, tmp);
	Py_DECREF(tmp);
      } else {
	PyCFunction meth = PyCFunction_GET_FUNCTION(destroy);
	PyObject *mself = PyCFunction_GET_SELF(destroy);
	res = ((*meth)(mself, v));
      }
      Py_XDECREF(res);
    } else {
      const char *name = SWIG_TypePrettyName(ty);
#if !defined(SWIG_PYTHON_SILENT_MEMLEAK)
      printf("swig/python detected a memory leak of type '%s', no destructor found.\n", name);
#endif
    }
  } 
  Py_XDECREF(next);
  PyObject_DEL(v);
}

SWIGRUNTIME PyObject* 
PySwigObject_append(PyObject* v, PyObject* next)
{
  PySwigObject *sobj = (PySwigObject *) v;
#ifndef METH_O
  PyObject *tmp = 0;
  if (!PyArg_ParseTuple(next,(char *)"O:append", &tmp)) return NULL;
  next = tmp;
#endif
  if (!PySwigObject_Check(next)) {
    return NULL;
  }
  sobj->next = next;
  Py_INCREF(next);
  return SWIG_Py_Void();
}

SWIGRUNTIME PyObject* 
#ifdef METH_NOARGS
PySwigObject_next(PyObject* v)
#else
PySwigObject_next(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  PySwigObject *sobj = (PySwigObject *) v;
  if (sobj->next) {    
    Py_INCREF(sobj->next);
    return sobj->next;
  } else {
    return SWIG_Py_Void();
  }
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
PySwigObject_disown(PyObject *v)
#else
PySwigObject_disown(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  PySwigObject *sobj = (PySwigObject *)v;
  sobj->own = 0;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
#ifdef METH_NOARGS
PySwigObject_acquire(PyObject *v)
#else
PySwigObject_acquire(PyObject* v, PyObject *SWIGUNUSEDPARM(args))
#endif
{
  PySwigObject *sobj = (PySwigObject *)v;
  sobj->own = SWIG_POINTER_OWN;
  return SWIG_Py_Void();
}

SWIGINTERN PyObject*
PySwigObject_own(PyObject *v, PyObject *args)
{
  PyObject *val = 0;
#if (PY_VERSION_HEX < 0x02020000)
  if (!PyArg_ParseTuple(args,(char *)"|O:own",&val))
#else
  if (!PyArg_UnpackTuple(args, (char *)"own", 0, 1, &val)) 
#endif
    {
      return NULL;
    } 
  else
    {
      PySwigObject *sobj = (PySwigObject *)v;
      PyObject *obj = PyBool_FromLong(sobj->own);
      if (val) {
#ifdef METH_NOARGS
	if (PyObject_IsTrue(val)) {
	  PySwigObject_acquire(v);
	} else {
	  PySwigObject_disown(v);
	}
#else
	if (PyObject_IsTrue(val)) {
	  PySwigObject_acquire(v,args);
	} else {
	  PySwigObject_disown(v,args);
	}
#endif
      } 
      return obj;
    }
}

#ifdef METH_O
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)PySwigObject_disown,  METH_NOARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)PySwigObject_acquire, METH_NOARGS,  (char *)"aquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)PySwigObject_own,     METH_VARARGS, (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)PySwigObject_append,  METH_O,       (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)PySwigObject_next,    METH_NOARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)PySwigObject_repr,    METH_NOARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#else
static PyMethodDef
swigobject_methods[] = {
  {(char *)"disown",  (PyCFunction)PySwigObject_disown,  METH_VARARGS,  (char *)"releases ownership of the pointer"},
  {(char *)"acquire", (PyCFunction)PySwigObject_acquire, METH_VARARGS,  (char *)"aquires ownership of the pointer"},
  {(char *)"own",     (PyCFunction)PySwigObject_own,     METH_VARARGS,  (char *)"returns/sets ownership of the pointer"},
  {(char *)"append",  (PyCFunction)PySwigObject_append,  METH_VARARGS,  (char *)"appends another 'this' object"},
  {(char *)"next",    (PyCFunction)PySwigObject_next,    METH_VARARGS,  (char *)"returns the next 'this' object"},
  {(char *)"__repr__",(PyCFunction)PySwigObject_repr,   METH_VARARGS,  (char *)"returns object representation"},
  {0, 0, 0, 0}  
};
#endif

#if PY_VERSION_HEX < 0x02020000
SWIGINTERN PyObject *
PySwigObject_getattr(PySwigObject *sobj,char *name)
{
  return Py_FindMethod(swigobject_methods, (PyObject *)sobj, name);
}
#endif

SWIGRUNTIME PyTypeObject*
_PySwigObject_type(void) {
  static char swigobject_doc[] = "Swig object carries a C/C++ instance pointer";
  
  static PyNumberMethods PySwigObject_as_number = {
    (binaryfunc)0, /*nb_add*/
    (binaryfunc)0, /*nb_subtract*/
    (binaryfunc)0, /*nb_multiply*/
    (binaryfunc)0, /*nb_divide*/
    (binaryfunc)0, /*nb_remainder*/
    (binaryfunc)0, /*nb_divmod*/
    (ternaryfunc)0,/*nb_power*/
    (unaryfunc)0,  /*nb_negative*/
    (unaryfunc)0,  /*nb_positive*/
    (unaryfunc)0,  /*nb_absolute*/
    (inquiry)0,    /*nb_nonzero*/
    0,		   /*nb_invert*/
    0,		   /*nb_lshift*/
    0,		   /*nb_rshift*/
    0,		   /*nb_and*/
    0,		   /*nb_xor*/
    0,		   /*nb_or*/
    (coercion)0,   /*nb_coerce*/
    (unaryfunc)PySwigObject_long, /*nb_int*/
    (unaryfunc)PySwigObject_long, /*nb_long*/
    (unaryfunc)0,                 /*nb_float*/
    (unaryfunc)PySwigObject_oct,  /*nb_oct*/
    (unaryfunc)PySwigObject_hex,  /*nb_hex*/
#if PY_VERSION_HEX >= 0x02020000
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_true_divide */ 
#elif PY_VERSION_HEX >= 0x02000000
    0,0,0,0,0,0,0,0,0,0,0 /* nb_inplace_add -> nb_inplace_or */
#endif
  };

  static PyTypeObject pyswigobject_type;  
  static int type_init = 0;
  if (!type_init) {
    const PyTypeObject tmp
      = {
	PyObject_HEAD_INIT(NULL)
	0,				    /* ob_size */
	(char *)"PySwigObject",		    /* tp_name */
	sizeof(PySwigObject),		    /* tp_basicsize */
	0,			            /* tp_itemsize */
	(destructor)PySwigObject_dealloc,   /* tp_dealloc */
	(printfunc)PySwigObject_print,	    /* tp_print */
#if PY_VERSION_HEX < 0x02020000
	(getattrfunc)PySwigObject_getattr,  /* tp_getattr */ 
#else
	(getattrfunc)0,			    /* tp_getattr */ 
#endif
	(setattrfunc)0,			    /* tp_setattr */ 
	(cmpfunc)PySwigObject_compare,	    /* tp_compare */ 
	(reprfunc)PySwigObject_repr,	    /* tp_repr */    
	&PySwigObject_as_number,	    /* tp_as_number */
	0,				    /* tp_as_sequence */
	0,				    /* tp_as_mapping */
	(hashfunc)0,			    /* tp_hash */
	(ternaryfunc)0,			    /* tp_call */
	(reprfunc)PySwigObject_str,	    /* tp_str */
	PyObject_GenericGetAttr,            /* tp_getattro */
	0,				    /* tp_setattro */
	0,		                    /* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	            /* tp_flags */
	swigobject_doc, 	            /* tp_doc */        
	0,                                  /* tp_traverse */
	0,                                  /* tp_clear */
	0,                                  /* tp_richcompare */
	0,                                  /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
	0,                                  /* tp_iter */
	0,                                  /* tp_iternext */
	swigobject_methods,		    /* tp_methods */ 
	0,			            /* tp_members */
	0,				    /* tp_getset */	    	
	0,			            /* tp_base */	        
	0,				    /* tp_dict */	    	
	0,				    /* tp_descr_get */  	
	0,				    /* tp_descr_set */  	
	0,				    /* tp_dictoffset */ 	
	0,				    /* tp_init */	    	
	0,				    /* tp_alloc */	    	
	0,			            /* tp_new */	    	
	0,	                            /* tp_free */	   
        0,                                  /* tp_is_gc */  
	0,				    /* tp_bases */   
	0,				    /* tp_mro */
	0,				    /* tp_cache */   
 	0,				    /* tp_subclasses */
	0,				    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
	0,                                  /* tp_del */
#endif
#ifdef COUNT_ALLOCS
	0,0,0,0                             /* tp_alloc -> tp_next */
#endif
      };
    pyswigobject_type = tmp;
    pyswigobject_type.ob_type = &PyType_Type;
    type_init = 1;
  }
  return &pyswigobject_type;
}

SWIGRUNTIME PyObject *
PySwigObject_New(void *ptr, swig_type_info *ty, int own)
{
  PySwigObject *sobj = PyObject_NEW(PySwigObject, PySwigObject_type());
  if (sobj) {
    sobj->ptr  = ptr;
    sobj->ty   = ty;
    sobj->own  = own;
    sobj->next = 0;
  }
  return (PyObject *)sobj;
}

/* -----------------------------------------------------------------------------
 * Implements a simple Swig Packed type, and use it instead of string
 * ----------------------------------------------------------------------------- */

typedef struct {
  PyObject_HEAD
  void *pack;
  swig_type_info *ty;
  size_t size;
} PySwigPacked;

SWIGRUNTIME int
PySwigPacked_print(PySwigPacked *v, FILE *fp, int SWIGUNUSEDPARM(flags))
{
  char result[SWIG_BUFFER_SIZE];
  fputs("<Swig Packed ", fp); 
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    fputs("at ", fp); 
    fputs(result, fp); 
  }
  fputs(v->ty->name,fp); 
  fputs(">", fp);
  return 0; 
}
  
SWIGRUNTIME PyObject *
PySwigPacked_repr(PySwigPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))) {
    return PyString_FromFormat("<Swig Packed at %s%s>", result, v->ty->name);
  } else {
    return PyString_FromFormat("<Swig Packed %s>", v->ty->name);
  }  
}

SWIGRUNTIME PyObject *
PySwigPacked_str(PySwigPacked *v)
{
  char result[SWIG_BUFFER_SIZE];
  if (SWIG_PackDataName(result, v->pack, v->size, 0, sizeof(result))){
    return PyString_FromFormat("%s%s", result, v->ty->name);
  } else {
    return PyString_FromString(v->ty->name);
  }  
}

SWIGRUNTIME int
PySwigPacked_compare(PySwigPacked *v, PySwigPacked *w)
{
  size_t i = v->size;
  size_t j = w->size;
  int s = (i < j) ? -1 : ((i > j) ? 1 : 0);
  return s ? s : strncmp((char *)v->pack, (char *)w->pack, 2*v->size);
}

SWIGRUNTIME PyTypeObject* _PySwigPacked_type(void);

SWIGRUNTIME PyTypeObject*
PySwigPacked_type(void) {
  static PyTypeObject *SWIG_STATIC_POINTER(type) = _PySwigPacked_type();
  return type;
}

SWIGRUNTIMEINLINE int
PySwigPacked_Check(PyObject *op) {
  return ((op)->ob_type == _PySwigPacked_type()) 
    || (strcmp((op)->ob_type->tp_name,"PySwigPacked") == 0);
}

SWIGRUNTIME void
PySwigPacked_dealloc(PyObject *v)
{
  if (PySwigPacked_Check(v)) {
    PySwigPacked *sobj = (PySwigPacked *) v;
    free(sobj->pack);
  }
  PyObject_DEL(v);
}

SWIGRUNTIME PyTypeObject*
_PySwigPacked_type(void) {
  static char swigpacked_doc[] = "Swig object carries a C/C++ instance pointer";
  static PyTypeObject pyswigpacked_type;
  static int type_init = 0;  
  if (!type_init) {
    const PyTypeObject tmp
      = {
	PyObject_HEAD_INIT(NULL)
	0,				    /* ob_size */	
	(char *)"PySwigPacked",		    /* tp_name */	
	sizeof(PySwigPacked),		    /* tp_basicsize */	
	0,				    /* tp_itemsize */	
	(destructor)PySwigPacked_dealloc,   /* tp_dealloc */	
	(printfunc)PySwigPacked_print,	    /* tp_print */   	
	(getattrfunc)0,			    /* tp_getattr */ 	
	(setattrfunc)0,			    /* tp_setattr */ 	
	(cmpfunc)PySwigPacked_compare,	    /* tp_compare */ 	
	(reprfunc)PySwigPacked_repr,	    /* tp_repr */    	
	0,	                            /* tp_as_number */	
	0,				    /* tp_as_sequence */
	0,				    /* tp_as_mapping */	
	(hashfunc)0,			    /* tp_hash */	
	(ternaryfunc)0,			    /* tp_call */	
	(reprfunc)PySwigPacked_str,	    /* tp_str */	
	PyObject_GenericGetAttr,            /* tp_getattro */
	0,				    /* tp_setattro */
	0,		                    /* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,	            /* tp_flags */
	swigpacked_doc, 	            /* tp_doc */
	0,                                  /* tp_traverse */
	0,                                  /* tp_clear */
	0,                                  /* tp_richcompare */
	0,                                  /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
	0,                                  /* tp_iter */
	0,                                  /* tp_iternext */
	0,		                    /* tp_methods */ 
	0,			            /* tp_members */
	0,				    /* tp_getset */	    	
	0,			            /* tp_base */	        
	0,				    /* tp_dict */	    	
	0,				    /* tp_descr_get */  	
	0,				    /* tp_descr_set */  	
	0,				    /* tp_dictoffset */ 	
	0,				    /* tp_init */	    	
	0,				    /* tp_alloc */	    	
	0,			            /* tp_new */	    	
	0, 	                            /* tp_free */	   
        0,                                  /* tp_is_gc */  
	0,				    /* tp_bases */   
	0,				    /* tp_mro */
	0,				    /* tp_cache */   
 	0,				    /* tp_subclasses */
	0,				    /* tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
	0,                                  /* tp_del */
#endif
#ifdef COUNT_ALLOCS
	0,0,0,0                             /* tp_alloc -> tp_next */
#endif
      };
    pyswigpacked_type = tmp;
    pyswigpacked_type.ob_type = &PyType_Type;
    type_init = 1;
  }
  return &pyswigpacked_type;
}

SWIGRUNTIME PyObject *
PySwigPacked_New(void *ptr, size_t size, swig_type_info *ty)
{
  PySwigPacked *sobj = PyObject_NEW(PySwigPacked, PySwigPacked_type());
  if (sobj) {
    void *pack = malloc(size);
    if (pack) {
      memcpy(pack, ptr, size);
      sobj->pack = pack;
      sobj->ty   = ty;
      sobj->size = size;
    } else {
      PyObject_DEL((PyObject *) sobj);
      sobj = 0;
    }
  }
  return (PyObject *) sobj;
}

SWIGRUNTIME swig_type_info *
PySwigPacked_UnpackData(PyObject *obj, void *ptr, size_t size)
{
  if (PySwigPacked_Check(obj)) {
    PySwigPacked *sobj = (PySwigPacked *)obj;
    if (sobj->size != size) return 0;
    memcpy(ptr, sobj->pack, size);
    return sobj->ty;
  } else {
    return 0;
  }
}

/* -----------------------------------------------------------------------------
 * pointers/data manipulation
 * ----------------------------------------------------------------------------- */

SWIGRUNTIMEINLINE PyObject *
_SWIG_This(void)
{
  return PyString_FromString("this");
}

SWIGRUNTIME PyObject *
SWIG_This(void)
{
  static PyObject *SWIG_STATIC_POINTER(swig_this) = _SWIG_This();
  return swig_this;
}

/* #define SWIG_PYTHON_SLOW_GETSET_THIS */

SWIGRUNTIME PySwigObject *
SWIG_Python_GetSwigThis(PyObject *pyobj) 
{
  if (PySwigObject_Check(pyobj)) {
    return (PySwigObject *) pyobj;
  } else {
    PyObject *obj = 0;
#if (!defined(SWIG_PYTHON_SLOW_GETSET_THIS) && (PY_VERSION_HEX >= 0x02030000))
    if (PyInstance_Check(pyobj)) {
      obj = _PyInstance_Lookup(pyobj, SWIG_This());      
    } else {
      PyObject **dictptr = _PyObject_GetDictPtr(pyobj);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	obj = dict ? PyDict_GetItem(dict, SWIG_This()) : 0;
      } else {
#ifdef PyWeakref_CheckProxy
	if (PyWeakref_CheckProxy(pyobj)) {
	  PyObject *wobj = PyWeakref_GET_OBJECT(pyobj);
	  return wobj ? SWIG_Python_GetSwigThis(wobj) : 0;
	}
#endif
	obj = PyObject_GetAttr(pyobj,SWIG_This());
	if (obj) {
	  Py_DECREF(obj);
	} else {
	  if (PyErr_Occurred()) PyErr_Clear();
	  return 0;
	}
      }
    }
#else
    obj = PyObject_GetAttr(pyobj,SWIG_This());
    if (obj) {
      Py_DECREF(obj);
    } else {
      if (PyErr_Occurred()) PyErr_Clear();
      return 0;
    }
#endif
    if (obj && !PySwigObject_Check(obj)) {
      /* a PyObject is called 'this', try to get the 'real this'
	 PySwigObject from it */ 
      return SWIG_Python_GetSwigThis(obj);
    }
    return (PySwigObject *)obj;
  }
}

/* Acquire a pointer value */

SWIGRUNTIME int
SWIG_Python_AcquirePtr(PyObject *obj, int own) {
  if (own) {
    PySwigObject *sobj = SWIG_Python_GetSwigThis(obj);
    if (sobj) {
      int oldown = sobj->own;
      sobj->own = own;
      return oldown;
    }
  }
  return 0;
}

/* Convert a pointer value */

SWIGRUNTIME int
SWIG_Python_ConvertPtrAndOwn(PyObject *obj, void **ptr, swig_type_info *ty, int flags, int *own) {
  if (!obj) return SWIG_ERROR;
  if (obj == Py_None) {
    if (ptr) *ptr = 0;
    return SWIG_OK;
  } else {
    PySwigObject *sobj = SWIG_Python_GetSwigThis(obj);
    while (sobj) {
      void *vptr = sobj->ptr;
      if (ty) {
	swig_type_info *to = sobj->ty;
	if (to == ty) {
	  /* no type cast needed */
	  if (ptr) *ptr = vptr;
	  break;
	} else {
	  swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
	  if (!tc) {
	    sobj = (PySwigObject *)sobj->next;
	  } else {
	    if (ptr) *ptr = SWIG_TypeCast(tc,vptr);
	    break;
	  }
	}
      } else {
	if (ptr) *ptr = vptr;
	break;
      }
    }
    if (sobj) {
      if (own) *own = sobj->own;
      if (flags & SWIG_POINTER_DISOWN) {
	sobj->own = 0;
      }
      return SWIG_OK;
    } else {
      int res = SWIG_ERROR;
      if (flags & SWIG_POINTER_IMPLICIT_CONV) {
	PySwigClientData *data = ty ? (PySwigClientData *) ty->clientdata : 0;
	if (data && !data->implicitconv) {
	  PyObject *klass = data->klass;
	  if (klass) {
	    PyObject *impconv;
	    data->implicitconv = 1; /* avoid recursion and call 'explicit' constructors*/
	    impconv = SWIG_Python_CallFunctor(klass, obj);
	    data->implicitconv = 0;
	    if (PyErr_Occurred()) {
	      PyErr_Clear();
	      impconv = 0;
	    }
	    if (impconv) {
	      PySwigObject *iobj = SWIG_Python_GetSwigThis(impconv);
	      if (iobj) {
		void *vptr;
		res = SWIG_Python_ConvertPtrAndOwn((PyObject*)iobj, &vptr, ty, 0, 0);
		if (SWIG_IsOK(res)) {
		  if (ptr) {
		    *ptr = vptr;
		    /* transfer the ownership to 'ptr' */
		    iobj->own = 0;
		    res = SWIG_AddCast(res);
		    res = SWIG_AddNewMask(res);
		  } else {
		    res = SWIG_AddCast(res);		    
		  }
		}
	      }
	      Py_DECREF(impconv);
	    }
	  }
	}
      }
      return res;
    }
  }
}

/* Convert a function ptr value */

SWIGRUNTIME int
SWIG_Python_ConvertFunctionPtr(PyObject *obj, void **ptr, swig_type_info *ty) {
  if (!PyCFunction_Check(obj)) {
    return SWIG_ConvertPtr(obj, ptr, ty, 0);
  } else {
    void *vptr = 0;
    
    /* here we get the method pointer for callbacks */
    const char *doc = (((PyCFunctionObject *)obj) -> m_ml -> ml_doc);
    const char *desc = doc ? strstr(doc, "swig_ptr: ") : 0;
    if (desc) {
      desc = ty ? SWIG_UnpackVoidPtr(desc + 10, &vptr, ty->name) : 0;
      if (!desc) return SWIG_ERROR;
    }
    if (ty) {
      swig_cast_info *tc = SWIG_TypeCheck(desc,ty);
      if (!tc) return SWIG_ERROR;
      *ptr = SWIG_TypeCast(tc,vptr);
    } else {
      *ptr = vptr;
    }
    return SWIG_OK;
  }
}

/* Convert a packed value value */

SWIGRUNTIME int
SWIG_Python_ConvertPacked(PyObject *obj, void *ptr, size_t sz, swig_type_info *ty) {
  swig_type_info *to = PySwigPacked_UnpackData(obj, ptr, sz);
  if (!to) return SWIG_ERROR;
  if (ty) {
    if (to != ty) {
      /* check type cast? */
      swig_cast_info *tc = SWIG_TypeCheck(to->name,ty);
      if (!tc) return SWIG_ERROR;
    }
  }
  return SWIG_OK;
}  

/* -----------------------------------------------------------------------------
 * Create a new pointer object
 * ----------------------------------------------------------------------------- */

/*
  Create a new instance object, whitout calling __init__, and set the
  'this' attribute.
*/

SWIGRUNTIME PyObject* 
SWIG_Python_NewShadowInstance(PySwigClientData *data, PyObject *swig_this)
{
#if (PY_VERSION_HEX >= 0x02020000)
  PyObject *inst = 0;
  PyObject *newraw = data->newraw;
  if (newraw) {
    inst = PyObject_Call(newraw, data->newargs, NULL);
    if (inst) {
#if !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
      PyObject **dictptr = _PyObject_GetDictPtr(inst);
      if (dictptr != NULL) {
	PyObject *dict = *dictptr;
	if (dict == NULL) {
	  dict = PyDict_New();
	  *dictptr = dict;
	  PyDict_SetItem(dict, SWIG_This(), swig_this);
	}
      }
#else
      PyObject *key = SWIG_This();
      PyObject_SetAttr(inst, key, swig_this);
#endif
    }
  } else {
    PyObject *dict = PyDict_New();
    PyDict_SetItem(dict, SWIG_This(), swig_this);
    inst = PyInstance_NewRaw(data->newargs, dict);
    Py_DECREF(dict);
  }
  return inst;
#else
#if (PY_VERSION_HEX >= 0x02010000)
  PyObject *inst;
  PyObject *dict = PyDict_New();
  PyDict_SetItem(dict, SWIG_This(), swig_this);
  inst = PyInstance_NewRaw(data->newargs, dict);
  Py_DECREF(dict);
  return (PyObject *) inst;
#else
  PyInstanceObject *inst = PyObject_NEW(PyInstanceObject, &PyInstance_Type);
  if (inst == NULL) {
    return NULL;
  }
  inst->in_class = (PyClassObject *)data->newargs;
  Py_INCREF(inst->in_class);
  inst->in_dict = PyDict_New();
  if (inst->in_dict == NULL) {
    Py_DECREF(inst);
    return NULL;
  }
#ifdef Py_TPFLAGS_HAVE_WEAKREFS
  inst->in_weakreflist = NULL;
#endif
#ifdef Py_TPFLAGS_GC
  PyObject_GC_Init(inst);
#endif
  PyDict_SetItem(inst->in_dict, SWIG_This(), swig_this);
  return (PyObject *) inst;
#endif
#endif
}

SWIGRUNTIME void
SWIG_Python_SetSwigThis(PyObject *inst, PyObject *swig_this)
{
 PyObject *dict;
#if (PY_VERSION_HEX >= 0x02020000) && !defined(SWIG_PYTHON_SLOW_GETSET_THIS)
 PyObject **dictptr = _PyObject_GetDictPtr(inst);
 if (dictptr != NULL) {
   dict = *dictptr;
   if (dict == NULL) {
     dict = PyDict_New();
     *dictptr = dict;
   }
   PyDict_SetItem(dict, SWIG_This(), swig_this);
   return;
 }
#endif
 dict = PyObject_GetAttrString(inst, (char*)"__dict__");
 PyDict_SetItem(dict, SWIG_This(), swig_this);
 Py_DECREF(dict);
} 


SWIGINTERN PyObject *
SWIG_Python_InitShadowInstance(PyObject *args) {
  PyObject *obj[2];
  if (!SWIG_Python_UnpackTuple(args,(char*)"swiginit", 2, 2, obj)) {
    return NULL;
  } else {
    PySwigObject *sthis = SWIG_Python_GetSwigThis(obj[0]);
    if (sthis) {
      PySwigObject_append((PyObject*) sthis, obj[1]);
    } else {
      SWIG_Python_SetSwigThis(obj[0], obj[1]);
    }
    return SWIG_Py_Void();
  }
}

/* Create a new pointer object */

SWIGRUNTIME PyObject *
SWIG_Python_NewPointerObj(void *ptr, swig_type_info *type, int flags) {
  if (!ptr) {
    return SWIG_Py_Void();
  } else {
    int own = (flags & SWIG_POINTER_OWN) ? SWIG_POINTER_OWN : 0;
    PyObject *robj = PySwigObject_New(ptr, type, own);
    PySwigClientData *clientdata = type ? (PySwigClientData *)(type->clientdata) : 0;
    if (clientdata && !(flags & SWIG_POINTER_NOSHADOW)) {
      PyObject *inst = SWIG_Python_NewShadowInstance(clientdata, robj);
      if (inst) {
	Py_DECREF(robj);
	robj = inst;
      }
    }
    return robj;
  }
}

/* Create a new packed object */

SWIGRUNTIMEINLINE PyObject *
SWIG_Python_NewPackedObj(void *ptr, size_t sz, swig_type_info *type) {
  return ptr ? PySwigPacked_New((void *) ptr, sz, type) : SWIG_Py_Void();
}

/* -----------------------------------------------------------------------------*
 *  Get type list 
 * -----------------------------------------------------------------------------*/

#ifdef SWIG_LINK_RUNTIME
void *SWIG_ReturnGlobalTypeList(void *);
#endif

SWIGRUNTIME swig_module_info *
SWIG_Python_GetModule(void) {
  static void *type_pointer = (void *)0;
  /* first check if module already created */
  if (!type_pointer) {
#ifdef SWIG_LINK_RUNTIME
    type_pointer = SWIG_ReturnGlobalTypeList((void *)0);
#else
    type_pointer = PyCObject_Import((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION,
				    (char*)"type_pointer" SWIG_TYPE_TABLE_NAME);
    if (PyErr_Occurred()) {
      PyErr_Clear();
      type_pointer = (void *)0;
    }
#endif
  }
  return (swig_module_info *) type_pointer;
}

#if PY_MAJOR_VERSION < 2
/* PyModule_AddObject function was introduced in Python 2.0.  The following function
   is copied out of Python/modsupport.c in python version 2.3.4 */
SWIGINTERN int
PyModule_AddObject(PyObject *m, char *name, PyObject *o)
{
  PyObject *dict;
  if (!PyModule_Check(m)) {
    PyErr_SetString(PyExc_TypeError,
		    "PyModule_AddObject() needs module as first arg");
    return SWIG_ERROR;
  }
  if (!o) {
    PyErr_SetString(PyExc_TypeError,
		    "PyModule_AddObject() needs non-NULL value");
    return SWIG_ERROR;
  }
  
  dict = PyModule_GetDict(m);
  if (dict == NULL) {
    /* Internal error -- modules must have a dict! */
    PyErr_Format(PyExc_SystemError, "module '%s' has no __dict__",
		 PyModule_GetName(m));
    return SWIG_ERROR;
  }
  if (PyDict_SetItemString(dict, name, o))
    return SWIG_ERROR;
  Py_DECREF(o);
  return SWIG_OK;
}
#endif

SWIGRUNTIME void
SWIG_Python_DestroyModule(void *vptr)
{
  swig_module_info *swig_module = (swig_module_info *) vptr;
  swig_type_info **types = swig_module->types;
  size_t i;
  for (i =0; i < swig_module->size; ++i) {
    swig_type_info *ty = types[i];
    if (ty->owndata) {
      PySwigClientData *data = (PySwigClientData *) ty->clientdata;
      if (data) PySwigClientData_Del(data);
    }
  }
  Py_DECREF(SWIG_This());
}

SWIGRUNTIME void
SWIG_Python_SetModule(swig_module_info *swig_module) {
  static PyMethodDef swig_empty_runtime_method_table[] = { {NULL, NULL, 0, NULL} };/* Sentinel */

  PyObject *module = Py_InitModule((char*)"swig_runtime_data" SWIG_RUNTIME_VERSION,
				   swig_empty_runtime_method_table);
  PyObject *pointer = PyCObject_FromVoidPtr((void *) swig_module, SWIG_Python_DestroyModule);
  if (pointer && module) {
    PyModule_AddObject(module, (char*)"type_pointer" SWIG_TYPE_TABLE_NAME, pointer);
  } else {
    Py_XDECREF(pointer);
  }
}

/* The python cached type query */
SWIGRUNTIME PyObject *
SWIG_Python_TypeCache(void) {
  static PyObject *SWIG_STATIC_POINTER(cache) = PyDict_New();
  return cache;
}

SWIGRUNTIME swig_type_info *
SWIG_Python_TypeQuery(const char *type)
{
  PyObject *cache = SWIG_Python_TypeCache();
  PyObject *key = PyString_FromString(type); 
  PyObject *obj = PyDict_GetItem(cache, key);
  swig_type_info *descriptor;
  if (obj) {
    descriptor = (swig_type_info *) PyCObject_AsVoidPtr(obj);
  } else {
    swig_module_info *swig_module = SWIG_Python_GetModule();
    descriptor = SWIG_TypeQueryModule(swig_module, swig_module, type);
    if (descriptor) {
      obj = PyCObject_FromVoidPtr(descriptor, NULL);
      PyDict_SetItem(cache, key, obj);
      Py_DECREF(obj);
    }
  }
  Py_DECREF(key);
  return descriptor;
}

/* 
   For backward compatibility only
*/
#define SWIG_POINTER_EXCEPTION  0
#define SWIG_arg_fail(arg)      SWIG_Python_ArgFail(arg)
#define SWIG_MustGetPtr(p, type, argnum, flags)  SWIG_Python_MustGetPtr(p, type, argnum, flags)

SWIGRUNTIME int
SWIG_Python_AddErrMesg(const char* mesg, int infront)
{
  if (PyErr_Occurred()) {
    PyObject *type = 0;
    PyObject *value = 0;
    PyObject *traceback = 0;
    PyErr_Fetch(&type, &value, &traceback);
    if (value) {
      PyObject *old_str = PyObject_Str(value);
      Py_XINCREF(type);
      PyErr_Clear();
      if (infront) {
	PyErr_Format(type, "%s %s", mesg, PyString_AsString(old_str));
      } else {
	PyErr_Format(type, "%s %s", PyString_AsString(old_str), mesg);
      }
      Py_DECREF(old_str);
    }
    return 1;
  } else {
    return 0;
  }
}
  
SWIGRUNTIME int
SWIG_Python_ArgFail(int argnum)
{
  if (PyErr_Occurred()) {
    /* add information about failing argument */
    char mesg[256];
    PyOS_snprintf(mesg, sizeof(mesg), "argument number %d:", argnum);
    return SWIG_Python_AddErrMesg(mesg, 1);
  } else {
    return 0;
  }
}

SWIGRUNTIMEINLINE const char *
PySwigObject_GetDesc(PyObject *self)
{
  PySwigObject *v = (PySwigObject *)self;
  swig_type_info *ty = v ? v->ty : 0;
  return ty ? ty->str : (char*)"";
}

SWIGRUNTIME void
SWIG_Python_TypeError(const char *type, PyObject *obj)
{
  if (type) {
#if defined(SWIG_COBJECT_TYPES)
    if (obj && PySwigObject_Check(obj)) {
      const char *otype = (const char *) PySwigObject_GetDesc(obj);
      if (otype) {
	PyErr_Format(PyExc_TypeError, "a '%s' is expected, 'PySwigObject(%s)' is received",
		     type, otype);
	return;
      }
    } else 
#endif      
    {
      const char *otype = (obj ? obj->ob_type->tp_name : 0); 
      if (otype) {
	PyObject *str = PyObject_Str(obj);
	const char *cstr = str ? PyString_AsString(str) : 0;
	if (cstr) {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s(%s)' is received",
		       type, otype, cstr);
	} else {
	  PyErr_Format(PyExc_TypeError, "a '%s' is expected, '%s' is received",
		       type, otype);
	}
	Py_XDECREF(str);
	return;
      }
    }   
    PyErr_Format(PyExc_TypeError, "a '%s' is expected", type);
  } else {
    PyErr_Format(PyExc_TypeError, "unexpected type is received");
  }
}


/* Convert a pointer value, signal an exception on a type mismatch */
SWIGRUNTIME void *
SWIG_Python_MustGetPtr(PyObject *obj, swig_type_info *ty, int argnum, int flags) {
  void *result;
  if (SWIG_Python_ConvertPtr(obj, &result, ty, flags) == -1) {
    PyErr_Clear();
    if (flags & SWIG_POINTER_EXCEPTION) {
      SWIG_Python_TypeError(SWIG_TypePrettyName(ty), obj);
      SWIG_Python_ArgFail(argnum);
    }
  }
  return result;
}


#ifdef __cplusplus
#if 0
{ /* cc-mode */
#endif
}
#endif



#define SWIG_exception_fail(code, msg) do { SWIG_Error(code, msg); SWIG_fail; } while(0) 

#define SWIG_contract_assert(expr, msg) if (!(expr)) { SWIG_Error(SWIG_RuntimeError, msg); SWIG_fail; } else 



  #define SWIG_exception(code, msg) do { SWIG_Error(code, msg); SWIG_fail;; } while(0) 


/* -------- TYPES TABLE (BEGIN) -------- */

#define SWIGTYPE_p_allocator_type swig_types[0]
#define SWIGTYPE_p_char swig_types[1]
#define SWIGTYPE_p_const_reference swig_types[2]
#define SWIGTYPE_p_difference_type swig_types[3]
#define SWIGTYPE_p_key_type swig_types[4]
#define SWIGTYPE_p_p_PyObject swig_types[5]
#define SWIGTYPE_p_p_steps__tetmesh__TmPatch swig_types[6]
#define SWIGTYPE_p_reference swig_types[7]
#define SWIGTYPE_p_size_type swig_types[8]
#define SWIGTYPE_p_std__invalid_argument swig_types[9]
#define SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t swig_types[10]
#define SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type swig_types[11]
#define SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t swig_types[12]
#define SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type swig_types[13]
#define SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type swig_types[14]
#define SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type swig_types[15]
#define SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t swig_types[16]
#define SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type swig_types[17]
#define SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t swig_types[18]
#define SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type swig_types[19]
#define SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t swig_types[20]
#define SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t__allocator_type swig_types[21]
#define SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t swig_types[22]
#define SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type swig_types[23]
#define SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t swig_types[24]
#define SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type swig_types[25]
#define SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t swig_types[26]
#define SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type swig_types[27]
#define SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t swig_types[28]
#define SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type swig_types[29]
#define SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t swig_types[30]
#define SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type swig_types[31]
#define SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t swig_types[32]
#define SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type swig_types[33]
#define SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t swig_types[34]
#define SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type swig_types[35]
#define SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t swig_types[36]
#define SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type swig_types[37]
#define SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t swig_types[38]
#define SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type swig_types[39]
#define SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t swig_types[40]
#define SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type swig_types[41]
#define SWIGTYPE_p_steps__ArgErr swig_types[42]
#define SWIGTYPE_p_steps__Err swig_types[43]
#define SWIGTYPE_p_steps__NotImplErr swig_types[44]
#define SWIGTYPE_p_steps__model__Diff swig_types[45]
#define SWIGTYPE_p_steps__model__Model swig_types[46]
#define SWIGTYPE_p_steps__model__Reac swig_types[47]
#define SWIGTYPE_p_steps__model__SReac swig_types[48]
#define SWIGTYPE_p_steps__model__Spec swig_types[49]
#define SWIGTYPE_p_steps__model__Surfsys swig_types[50]
#define SWIGTYPE_p_steps__model__Volsys swig_types[51]
#define SWIGTYPE_p_steps__rng__RNG swig_types[52]
#define SWIGTYPE_p_steps__solver__API swig_types[53]
#define SWIGTYPE_p_steps__tetexact__Tetexact swig_types[54]
#define SWIGTYPE_p_steps__tetmesh__DiffBoundary swig_types[55]
#define SWIGTYPE_p_steps__tetmesh__Tetmesh swig_types[56]
#define SWIGTYPE_p_steps__tetmesh__TmComp swig_types[57]
#define SWIGTYPE_p_steps__tetmesh__TmPatch swig_types[58]
#define SWIGTYPE_p_steps__wm__Comp swig_types[59]
#define SWIGTYPE_p_steps__wm__Geom swig_types[60]
#define SWIGTYPE_p_steps__wm__Patch swig_types[61]
#define SWIGTYPE_p_steps__wmdirect__Wmdirect swig_types[62]
#define SWIGTYPE_p_steps__wmrk4__Wmrk4 swig_types[63]
#define SWIGTYPE_p_swig__PySwigIterator swig_types[64]
#define SWIGTYPE_p_unsigned_char swig_types[65]
#define SWIGTYPE_p_unsigned_int swig_types[66]
#define SWIGTYPE_p_unsigned_long swig_types[67]
#define SWIGTYPE_p_unsigned_short swig_types[68]
#define SWIGTYPE_p_value_type swig_types[69]
#define SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type swig_types[70]
#define SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type swig_types[71]
#define SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type swig_types[72]
#define SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type swig_types[73]
#define SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type swig_types[74]
#define SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type swig_types[75]
#define SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type swig_types[76]
#define SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type swig_types[77]
#define SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type swig_types[78]
#define SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type swig_types[79]
static swig_type_info *swig_types[81];
static swig_module_info swig_module = {swig_types, 80, 0, 0, 0, 0};
#define SWIG_TypeQuery(name) SWIG_TypeQueryModule(&swig_module, &swig_module, name)
#define SWIG_MangledTypeQuery(name) SWIG_MangledTypeQueryModule(&swig_module, &swig_module, name)

/* -------- TYPES TABLE (END) -------- */

#if (PY_VERSION_HEX <= 0x02000000)
# if !defined(SWIG_PYTHON_CLASSIC)
#  error "This python version requires swig to be run with the '-classic' option"
# endif
#endif

/*-----------------------------------------------
              @(target):= _steps_swig.so
  ------------------------------------------------*/
#define SWIG_init    init_steps_swig

#define SWIG_name    "_steps_swig"

#define SWIGVERSION 0x010331 
#define SWIG_VERSION SWIGVERSION


#define SWIG_as_voidptr(a) const_cast< void * >(static_cast< const void * >(a)) 
#define SWIG_as_voidptrptr(a) ((void)SWIG_as_voidptr(*a),reinterpret_cast< void** >(a)) 


#include <stdexcept>


namespace swig {
  class PyObject_ptr {
  protected:
    PyObject *_obj;

  public:
    PyObject_ptr() :_obj(0)
    {
    }

    PyObject_ptr(const PyObject_ptr& item) : _obj(item._obj)
    {
      Py_XINCREF(_obj);      
    }
    
    PyObject_ptr(PyObject *obj, bool initial_ref = true) :_obj(obj)
    {
      if (initial_ref) Py_XINCREF(_obj);
    }
    
    PyObject_ptr & operator=(const PyObject_ptr& item) 
    {
      Py_XINCREF(item._obj);
      Py_XDECREF(_obj);
      _obj = item._obj;
      return *this;      
    }
    
    ~PyObject_ptr() 
    {
      Py_XDECREF(_obj);
    }
    
    operator PyObject *() const
    {
      return _obj;
    }

    PyObject *operator->() const
    {
      return _obj;
    }
  };
}


namespace swig {
  struct PyObject_var : PyObject_ptr {
    PyObject_var(PyObject* obj = 0) : PyObject_ptr(obj, false) { }
    
    PyObject_var & operator = (PyObject* obj)
    {
      Py_XDECREF(_obj);
      _obj = obj;
      return *this;      
    }
  };
}


#include <stdexcept>


  
#if defined(__SUNPRO_CC) && defined(_RWSTD_VER)
#  define SWIG_STD_NOASSIGN_STL
#  define SWIG_STD_NOINSERT_TEMPLATE_STL
#  define SWIG_STD_NOITERATOR_TRAITS_STL
#endif

#if defined(__GNUC__)
#  if __GNUC__ == 2 && __GNUC_MINOR <= 96
#     define SWIG_STD_NOMODERN_STL
#  endif
#endif




#include <string>
#include <stdexcept>


#include <utility>


#include <iostream>

  
namespace swig {
  struct stop_iteration {
  };

  struct PySwigIterator {
  private:
    PyObject_ptr _seq;

  protected:
    PySwigIterator(PyObject *seq) : _seq(seq)
    {
    }
      
  public:
    virtual ~PySwigIterator() {}

    // Access iterator method, required by Python
    virtual PyObject *value() const = 0;

    // Forward iterator method, required by Python
    virtual PySwigIterator *incr(size_t n = 1) = 0;
    
    // Backward iterator method, very common in C++, but not required in Python
    virtual PySwigIterator *decr(size_t n = 1)
    {
      throw stop_iteration();
    }

    // Random access iterator methods, but not required in Python
    virtual ptrdiff_t distance(const PySwigIterator &x) const
    {
      throw std::invalid_argument("operation not supported");
    }

    virtual bool equal (const PySwigIterator &x) const
    {
      throw std::invalid_argument("operation not supported");
    }
    
    // C++ common/needed methods
    virtual PySwigIterator *copy() const = 0;

    PyObject *next()
    {
      PyObject *obj = value();
      incr();
      return obj;
    }

    PyObject *previous()
    {
      decr();
      return value();
    }

    PySwigIterator *advance(ptrdiff_t n)
    {
      return  (n > 0) ?  incr(n) : decr(-n);
    }
      
    bool operator == (const PySwigIterator& x)  const
    {
      return equal(x);
    }
      
    bool operator != (const PySwigIterator& x) const
    {
      return ! operator==(x);
    }
      
    PySwigIterator& operator += (ptrdiff_t n)
    {
      return *advance(n);
    }

    PySwigIterator& operator -= (ptrdiff_t n)
    {
      return *advance(-n);
    }
      
    PySwigIterator* operator + (ptrdiff_t n) const
    {
      return copy()->advance(n);
    }

    PySwigIterator* operator - (ptrdiff_t n) const
    {
      return copy()->advance(-n);
    }
      
    ptrdiff_t operator - (const PySwigIterator& x) const
    {
      return x.distance(*this);
    }
      
    static swig_type_info* descriptor() {
      static int init = 0;
      static swig_type_info* desc = 0;
      if (!init) {
	desc = SWIG_TypeQuery("swig::PySwigIterator *");
	init = 1;
      }	
      return desc;
    }    
  };
}


SWIGINTERN int
SWIG_AsVal_double (PyObject *obj, double *val)
{
  int res = SWIG_TypeError;
  if (PyFloat_Check(obj)) {
    if (val) *val = PyFloat_AsDouble(obj);
    return SWIG_OK;
  } else if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else if (PyLong_Check(obj)) {
    double v = PyLong_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    double d = PyFloat_AsDouble(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = d;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      long v = PyLong_AsLong(obj);
      if (!PyErr_Occurred()) {
	if (val) *val = v;
	return SWIG_AddCast(SWIG_AddCast(SWIG_OK));
      } else {
	PyErr_Clear();
      }
    }
  }
#endif
  return res;
}


#include <float.h>


#include <math.h>


SWIGINTERNINLINE int
SWIG_CanCastAsInteger(double *d, double min, double max) {
  double x = *d;
  if ((min <= x && x <= max)) {
   double fx = floor(x);
   double cx = ceil(x);
   double rd =  ((x - fx) < 0.5) ? fx : cx; /* simple rint */
   if ((errno == EDOM) || (errno == ERANGE)) {
     errno = 0;
   } else {
     double summ, reps, diff;
     if (rd < x) {
       diff = x - rd;
     } else if (rd > x) {
       diff = rd - x;
     } else {
       return 1;
     }
     summ = rd + x;
     reps = diff/summ;
     if (reps < 8*DBL_EPSILON) {
       *d = rd;
       return 1;
     }
   }
  }
  return 0;
}


SWIGINTERN int
SWIG_AsVal_unsigned_SS_long (PyObject *obj, unsigned long *val) 
{
  if (PyInt_Check(obj)) {
    long v = PyInt_AsLong(obj);
    if (v >= 0) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      return SWIG_OverflowError;
    }
  } else if (PyLong_Check(obj)) {
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    unsigned long v = PyLong_AsUnsignedLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, 0, ULONG_MAX)) {
	if (val) *val = (unsigned long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE int
SWIG_AsVal_size_t (PyObject * obj, size_t *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, val ? &v : 0);
  if (SWIG_IsOK(res) && val) *val = static_cast< size_t >(v);
  return res;
}


  #define SWIG_From_long   PyInt_FromLong 


SWIGINTERNINLINE PyObject *
SWIG_From_ptrdiff_t  (ptrdiff_t value)
{    
  return SWIG_From_long  (static_cast< long >(value));
}


SWIGINTERNINLINE PyObject*
  SWIG_From_bool  (bool value)
{
  return PyBool_FromLong(value ? 1 : 0);
}


SWIGINTERN int
SWIG_AsVal_long (PyObject *obj, long* val)
{
  if (PyInt_Check(obj)) {
    if (val) *val = PyInt_AsLong(obj);
    return SWIG_OK;
  } else if (PyLong_Check(obj)) {
    long v = PyLong_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_OK;
    } else {
      PyErr_Clear();
    }
  }
#ifdef SWIG_PYTHON_CAST_MODE
  {
    int dispatch = 0;
    long v = PyInt_AsLong(obj);
    if (!PyErr_Occurred()) {
      if (val) *val = v;
      return SWIG_AddCast(SWIG_OK);
    } else {
      PyErr_Clear();
    }
    if (!dispatch) {
      double d;
      int res = SWIG_AddCast(SWIG_AsVal_double (obj,&d));
      if (SWIG_IsOK(res) && SWIG_CanCastAsInteger(&d, LONG_MIN, LONG_MAX)) {
	if (val) *val = (long)(d);
	return res;
      }
    }
  }
#endif
  return SWIG_TypeError;
}


SWIGINTERNINLINE int
SWIG_AsVal_ptrdiff_t (PyObject * obj, ptrdiff_t *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, val ? &v : 0);
  if (SWIG_IsOK(res) && val) *val = static_cast< ptrdiff_t >(v);
  return res;
}


#include <stdexcept>


#include <algorithm>


#include <map>
#include <algorithm>
#include <stdexcept>


#include <string>


#include <vector>


// Autotools definitions.
#include "../cpp/error.hpp"


SWIGINTERN swig_type_info*
SWIG_pchar_descriptor(void)
{
  static int init = 0;
  static swig_type_info* info = 0;
  if (!init) {
    info = SWIG_TypeQuery("_p_char");
    init = 1;
  }
  return info;
}


SWIGINTERN int
SWIG_AsCharPtrAndSize(PyObject *obj, char** cptr, size_t* psize, int *alloc)
{
  if (PyString_Check(obj)) {
    char *cstr; Py_ssize_t len;
    PyString_AsStringAndSize(obj, &cstr, &len);
    if (cptr)  {
      if (alloc) {
	/* 
	   In python the user should not be able to modify the inner
	   string representation. To warranty that, if you define
	   SWIG_PYTHON_SAFE_CSTRINGS, a new/copy of the python string
	   buffer is always returned.

	   The default behavior is just to return the pointer value,
	   so, be careful.
	*/ 
#if defined(SWIG_PYTHON_SAFE_CSTRINGS)
	if (*alloc != SWIG_OLDOBJ) 
#else
	if (*alloc == SWIG_NEWOBJ) 
#endif
	  {
	    *cptr = reinterpret_cast< char* >(memcpy((new char[len + 1]), cstr, sizeof(char)*(len + 1)));
	    *alloc = SWIG_NEWOBJ;
	  }
	else {
	  *cptr = cstr;
	  *alloc = SWIG_OLDOBJ;
	}
      } else {
	*cptr = PyString_AsString(obj);
      }
    }
    if (psize) *psize = len + 1;
    return SWIG_OK;
  } else {
    swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
    if (pchar_descriptor) {
      void* vptr = 0;
      if (SWIG_ConvertPtr(obj, &vptr, pchar_descriptor, 0) == SWIG_OK) {
	if (cptr) *cptr = (char *) vptr;
	if (psize) *psize = vptr ? (strlen((char *)vptr) + 1) : 0;
	if (alloc) *alloc = SWIG_OLDOBJ;
	return SWIG_OK;
      }
    }
  }
  return SWIG_TypeError;
}


SWIGINTERN int
SWIG_AsPtr_std_string (PyObject * obj, std::string **val) 
{
  char* buf = 0 ; size_t size = 0; int alloc = SWIG_OLDOBJ;
  if (SWIG_IsOK((SWIG_AsCharPtrAndSize(obj, &buf, &size, &alloc)))) {
    if (buf) {
      if (val) *val = new std::string(buf, size - 1);
      if (alloc == SWIG_NEWOBJ) delete[] buf;
      return SWIG_NEWOBJ;
    } else {
      if (val) *val = 0;
      return SWIG_OLDOBJ;
    }
  } else {
    static int init = 0;
    static swig_type_info* descriptor = 0;
    if (!init) {
      descriptor = SWIG_TypeQuery("std::string" " *");
      init = 1;
    }
    if (descriptor) {
      std::string *vptr;
      int res = SWIG_ConvertPtr(obj, (void**)&vptr, descriptor, 0);
      if (SWIG_IsOK(res) && val) *val = vptr;
      return res;
    }
  }
  return SWIG_ERROR;
}


SWIGINTERNINLINE PyObject *
SWIG_FromCharPtrAndSize(const char* carray, size_t size)
{
  if (carray) {
    if (size > INT_MAX) {
      swig_type_info* pchar_descriptor = SWIG_pchar_descriptor();
      return pchar_descriptor ? 
	SWIG_NewPointerObj(const_cast< char * >(carray), pchar_descriptor, 0) : SWIG_Py_Void();
    } else {
      return PyString_FromStringAndSize(carray, static_cast< int >(size));
    }
  } else {
    return SWIG_Py_Void();
  }
}


SWIGINTERNINLINE PyObject * 
SWIG_FromCharPtr(const char *cptr)
{ 
  return SWIG_FromCharPtrAndSize(cptr, (cptr ? strlen(cptr) : 0));
}


#include "../cpp/model/model.hpp"
#include "../cpp/model/diff.hpp"
#include "../cpp/model/reac.hpp"
#include "../cpp/model/spec.hpp"
#include "../cpp/model/sreac.hpp"
#include "../cpp/model/surfsys.hpp"
#include "../cpp/model/volsys.hpp"


namespace swig {  
  template <class Type>
  struct noconst_traits {
    typedef Type noconst_type;
  };

  template <class Type>
  struct noconst_traits<const Type> {
    typedef Type noconst_type;
  };

  /*
    type categories
  */
  struct pointer_category { };  
  struct value_category { };

  /*
    General traits that provides type_name and type_info
  */
  template <class Type> struct traits { };

  template <class Type>
  inline const char* type_name() {
    return traits<typename noconst_traits<Type >::noconst_type >::type_name();
  }

  template <class Type> 
  struct traits_info {
    static swig_type_info *type_query(std::string name) {
      name += " *";
      return SWIG_TypeQuery(name.c_str());
    }    
    static swig_type_info *type_info() {
      static swig_type_info *info = type_query(type_name<Type>());
      return info;
    }
  };

  template <class Type>
  inline swig_type_info *type_info() {
    return traits_info<Type>::type_info();
  }

  /*
    Partial specialization for pointers
  */
  template <class Type> struct traits <Type *> {
    typedef pointer_category category;
    static std::string make_ptr_name(const char* name) {
      std::string ptrname = name;
      ptrname += " *";
      return ptrname;
    }    
    static const char* type_name() {
      static std::string name = make_ptr_name(swig::type_name<Type>());
      return name.c_str();
    }
  };

  template <class Type, class Category> 
  struct traits_as { };
 
  template <class Type, class Category> 
  struct traits_check { };

}


namespace swig {  
  /*
    Traits that provides the from method
  */
  template <class Type> struct traits_from_ptr {
    static PyObject *from(Type *val, int owner = 0) {
      return SWIG_NewPointerObj(val, type_info<Type>(), owner);
    }
  };

  template <class Type> struct traits_from {
    static PyObject *from(const Type& val) {
      return traits_from_ptr<Type>::from(new Type(val), 1);
    }
  };

  template <class Type> struct traits_from<Type *> {
    static PyObject *from(Type* val) {
      return traits_from_ptr<Type>::from(val, 0);
    }
  };

  template <class Type> struct traits_from<const Type *> {
    static PyObject *from(const Type* val) {
      return traits_from_ptr<Type>::from(const_cast<Type*>(val), 0);
    }
  };


  template <class Type>
  inline PyObject *from(const Type& val) {
    return traits_from<Type>::from(val);
  }

  template <class Type>
  inline PyObject *from_ptr(Type* val, int owner) {
    return traits_from_ptr<Type>::from(val, owner);
  }

  /*
    Traits that provides the asval/as/check method
  */
  template <class Type>
  struct traits_asptr {   
    static int asptr(PyObject *obj, Type **val) {
      Type *p;
      int res = (SWIG_ConvertPtr(obj, (void**)&p, type_info<Type>(), 0) == SWIG_OK) ? SWIG_OLDOBJ : 0;
      if (SWIG_IsOK(res)) {
	if (val) *val = p;
      }
      return res;
    }
  }; 

  template <class Type>
  inline int asptr(PyObject *obj, Type **vptr) {
    return traits_asptr<Type>::asptr(obj, vptr);
  }

  template <class Type> 
  struct traits_asval {
    static int asval(PyObject *obj, Type *val) {
      if (val) {
	Type *p = 0;
	int res = traits_asptr<Type>::asptr(obj, &p);
	if (!SWIG_IsOK(res)) return res;	
	if (p) {
	  typedef typename noconst_traits<Type>::noconst_type noconst_type;
	  *(const_cast<noconst_type*>(val)) = *p;
	  if (SWIG_IsNewObj(res)){
	    delete p;
	    res = SWIG_DelNewMask(res);
	  }
	  return res;
	} else {
	  return SWIG_ERROR;
	}
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };

  template <class Type> struct traits_asval<Type*> {
    static int asval(PyObject *obj, Type **val) {
      if (val) {
        typedef typename noconst_traits<Type>::noconst_type noconst_type;
        noconst_type *p = 0;
        int res = traits_asptr<noconst_type>::asptr(obj,  &p);
        if (SWIG_IsOK(res)) {
          *(const_cast<noconst_type**>(val)) = p;
	}
	return res;
      } else {
	return traits_asptr<Type>::asptr(obj, (Type **)(0));
      }
    }
  };
  
  template <class Type>
  inline int asval(PyObject *obj, Type *val) {
    return traits_asval<Type>::asval(obj, val);
  }

  template <class Type> 
  struct traits_as<Type, value_category> {
    static Type as(PyObject *obj, bool throw_error) {
      Type v;
      int res = asval(obj, &v);
      if (!obj || !SWIG_IsOK(res)) {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	if (throw_error) throw std::invalid_argument("bad type");
      }
      return v;
    }
  };

  template <class Type> 
  struct traits_as<Type, pointer_category> {
    static Type as(PyObject *obj, bool throw_error) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res) && v) {
	if (SWIG_IsNewObj(res)) {
	  Type r(*v);
	  delete v;
	  return r;
	} else {
	  return *v;
	}
      } else {
	// Uninitialized return value, no Type() constructor required.
	static Type *v_def = (Type*) malloc(sizeof(Type));
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	if (throw_error) throw std::invalid_argument("bad type");
	memset(v_def,0,sizeof(Type));
	return *v_def;
      }
    }
  };

  template <class Type> 
  struct traits_as<Type*, pointer_category> {
    static Type* as(PyObject *obj, bool throw_error) {
      Type *v = 0;      
      int res = (obj ? traits_asptr<Type>::asptr(obj, &v) : SWIG_ERROR);
      if (SWIG_IsOK(res)) {
	return v;
      } else {
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<Type>());
	}
	if (throw_error) throw std::invalid_argument("bad type");
	return 0;
      }
    }
  };
    
  template <class Type>
  inline Type as(PyObject *obj, bool te = false) {
    return traits_as<Type, typename traits<Type>::category>::as(obj, te);
  }

  template <class Type> 
  struct traits_check<Type, value_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asval(obj, (Type *)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type> 
  struct traits_check<Type, pointer_category> {
    static bool check(PyObject *obj) {
      int res = obj ? asptr(obj, (Type **)(0)) : SWIG_ERROR;
      return SWIG_IsOK(res) ? true : false;
    }
  };

  template <class Type>
  inline bool check(PyObject *obj) {
    return traits_check<Type, typename traits<Type>::category>::check(obj);
  }
}


#include <functional>

namespace std {
  template <>
  struct less <PyObject *>: public binary_function<PyObject *, PyObject *, bool>
  {
    bool
    operator()(PyObject * v, PyObject *w) const
    { 
      bool res;
      SWIG_PYTHON_THREAD_BEGIN_BLOCK;
      res = PyObject_Compare(v, w) < 0;
      SWIG_PYTHON_THREAD_END_BLOCK;
      return res;
    }
  };

  template <>
  struct less <swig::PyObject_ptr>: public binary_function<swig::PyObject_ptr, swig::PyObject_ptr, bool>
  {
    bool
    operator()(const swig::PyObject_ptr& v, const swig::PyObject_ptr& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

  template <>
  struct less <swig::PyObject_var>: public binary_function<swig::PyObject_var, swig::PyObject_var, bool>
  {
    bool
    operator()(const swig::PyObject_var& v, const swig::PyObject_var& w) const
    {
      return std::less<PyObject *>()(v, w);
    }
  };

}

namespace swig {
  template <> struct traits<PyObject *> {
    typedef value_category category;
    static const char* type_name() { return "PyObject *"; }
  };  

  template <>  struct traits_asval<PyObject * > {   
    typedef PyObject * value_type;
    static int asval(PyObject *obj, value_type *val) {
      if (val) *val = obj;
      return SWIG_OK;
    }
  };

  template <> 
  struct traits_check<PyObject *, value_category> {
    static bool check(PyObject *) {
      return true;
    }
  };

  template <>  struct traits_from<PyObject *> {
    typedef PyObject * value_type;
    static PyObject *from(const value_type& val) {
      Py_XINCREF(val);
      return val;
    }
  };
  
}

namespace swig {
  inline size_t
  check_index(ptrdiff_t i, size_t size, bool insert = false) {
    if ( i < 0 ) {
      if ((size_t) (-i) <= size)
	return (size_t) (i + size);
    } else if ( (size_t) i < size ) {
      return (size_t) i;
    } else if (insert && ((size_t) i == size)) {
      return size;
    }
    
    throw std::out_of_range("index out of range");
  }

  inline size_t
  slice_index(ptrdiff_t i, size_t size) {
    if ( i < 0 ) {
      if ((size_t) (-i) <= size) {
	return (size_t) (i + size);
      } else {
	throw std::out_of_range("index out of range");
      }
    } else {
      return ( (size_t) i < size ) ? ((size_t) i) : size;
    }
  }

  template <class Sequence, class Difference>
  inline typename Sequence::iterator
  getpos(Sequence* self, Difference i)  {
    typename Sequence::iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence, class Difference>
  inline typename Sequence::const_iterator
  cgetpos(const Sequence* self, Difference i)  {
    typename Sequence::const_iterator pos = self->begin();
    std::advance(pos, check_index(i,self->size()));
    return pos;
  }

  template <class Sequence, class Difference>
  inline Sequence*
  getslice(const Sequence* self, Difference i, Difference j) {
    typename Sequence::size_type size = self->size();
    typename Sequence::size_type ii = swig::check_index(i, size);
    typename Sequence::size_type jj = swig::slice_index(j, size);

    if (jj > ii) {
      typename Sequence::const_iterator vb = self->begin();
      typename Sequence::const_iterator ve = self->begin();
      std::advance(vb,ii);
      std::advance(ve,jj);
      return new Sequence(vb, ve);
    } else {
      return new Sequence();
    }
  }

  template <class Sequence, class Difference, class InputSeq>
  inline void
  setslice(Sequence* self, Difference i, Difference j, const InputSeq& v) {
    typename Sequence::size_type size = self->size();
    typename Sequence::size_type ii = swig::check_index(i, size, true);
    typename Sequence::size_type jj = swig::slice_index(j, size);
    if (jj < ii) jj = ii;
    size_t ssize = jj - ii;
    if (ssize <= v.size()) {
      typename Sequence::iterator sb = self->begin();
      typename InputSeq::const_iterator vmid = v.begin();
      std::advance(sb,ii);
      std::advance(vmid, jj - ii);
      self->insert(std::copy(v.begin(), vmid, sb), vmid, v.end());
    } else {
      typename Sequence::iterator sb = self->begin();
      typename Sequence::iterator se = self->begin();
      std::advance(sb,ii);
      std::advance(se,jj);
      self->erase(sb,se);
      self->insert(sb, v.begin(), v.end());
    }
  }

  template <class Sequence, class Difference>
  inline void
  delslice(Sequence* self, Difference i, Difference j) {
    typename Sequence::size_type size = self->size();
    typename Sequence::size_type ii = swig::check_index(i, size, true);
    typename Sequence::size_type jj = swig::slice_index(j, size);
    if (jj > ii) {
      typename Sequence::iterator sb = self->begin();
      typename Sequence::iterator se = self->begin();
      std::advance(sb,ii);
      std::advance(se,jj);
      self->erase(sb,se);
    }
  }
}


#if !defined(SWIG_STD_NOITERATOR_TRAITS_STL)
#include <iterator>
#else
namespace std  {
  template <class Iterator>
  struct iterator_traits {
    typedef ptrdiff_t difference_type;
    typedef typename Iterator::value_type value_type;
  };

#if defined(__SUNPRO_CC) && defined(_RWSTD_VER)
  template <class Iterator, class Category,class T, class Reference, class Pointer, class Distance>
  struct iterator_traits<__reverse_bi_iterator<Iterator,Category,T,Reference,Pointer,Distance> > {
    typedef Distance difference_type;
    typedef T value_type;
  };
#endif  

  template <class T>
  struct iterator_traits<T*> {
    typedef T value_type;
    typedef ptrdiff_t difference_type;
  };

  template<typename _InputIterator>
  inline typename iterator_traits<_InputIterator>::difference_type
  distance(_InputIterator __first, _InputIterator __last)
  {
    typename iterator_traits<_InputIterator>::difference_type __n = 0;
    while (__first != __last) {
      ++__first; ++__n;
    }
    return __n;
  }

} 
#endif


namespace swig {
  template<typename OutIterator>
  class PySwigIterator_T :  public PySwigIterator
  {
  public:
    typedef OutIterator out_iterator;
    typedef typename std::iterator_traits<out_iterator>::value_type value_type;    
    typedef PySwigIterator_T<out_iterator> self_type;

    PySwigIterator_T(out_iterator curr, PyObject *seq)
      : PySwigIterator(seq), current(curr)
    {
    }

    const out_iterator& get_current() const
    {
      return current;
    }

    
    bool equal (const PySwigIterator &iter) const
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return (current == iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }
    
    ptrdiff_t distance(const PySwigIterator &iter) const
    {
      const self_type *iters = dynamic_cast<const self_type *>(&iter);
      if (iters) {
	return std::distance(current, iters->get_current());
      } else {
	throw std::invalid_argument("bad iterator type");
      }
    }    
    
  protected:
    out_iterator current;
  };
  
  template <class ValueType>
  struct from_oper 
  {
    typedef const ValueType& argument_type;
    typedef PyObject *result_type;
    result_type operator()(argument_type v) const
    {
      return swig::from(v);
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class PySwigIteratorOpen_T :  public PySwigIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef PySwigIterator_T<out_iterator>  base;
    typedef PySwigIteratorOpen_T<OutIterator, ValueType, FromOper> self_type;
    
    PySwigIteratorOpen_T(out_iterator curr, PyObject *seq)
      : PySwigIterator_T<OutIterator>(curr, seq)
    {
    }
    
    PyObject *value() const {
      return from(static_cast<const value_type&>(*(base::current)));
    }
    
    PySwigIterator *copy() const
    {
      return new self_type(*this);
    }

    PySwigIterator *incr(size_t n = 1)
    {
      while (n--) {
	++base::current;
      }
      return this;
    }

    PySwigIterator *decr(size_t n = 1)
    {
      while (n--) {
	--base::current;
      }
      return this;
    }
  };

  template<typename OutIterator, 
	   typename ValueType = typename std::iterator_traits<OutIterator>::value_type,
	   typename FromOper = from_oper<ValueType> >
  class PySwigIteratorClosed_T :  public PySwigIterator_T<OutIterator>
  {
  public:
    FromOper from;
    typedef OutIterator out_iterator;
    typedef ValueType value_type;
    typedef PySwigIterator_T<out_iterator>  base;    
    typedef PySwigIteratorClosed_T<OutIterator, ValueType, FromOper> self_type;
    
    PySwigIteratorClosed_T(out_iterator curr, out_iterator first, out_iterator last, PyObject *seq)
      : PySwigIterator_T<OutIterator>(curr, seq), begin(first), end(last)
    {
    }
    
    PyObject *value() const {
      if (base::current == end) {
	throw stop_iteration();
      } else {
	return from(static_cast<const value_type&>(*(base::current)));
      }
    }
    
    PySwigIterator *copy() const
    {
      return new self_type(*this);
    }

    PySwigIterator *incr(size_t n = 1)
    {
      while (n--) {
	if (base::current == end) {
	  throw stop_iteration();
	} else {
	  ++base::current;
	}
      }
      return this;
    }

    PySwigIterator *decr(size_t n = 1)
    {
      while (n--) {
	if (base::current == begin) {
	  throw stop_iteration();
	} else {
	  --base::current;
	}
      }
      return this;
    }

  private:
    out_iterator begin;
    out_iterator end;
  };

  template<typename OutIter>
  inline PySwigIterator*
  make_output_iterator(const OutIter& current, const OutIter& begin,const OutIter& end, PyObject *seq = 0)
  {
    return new PySwigIteratorClosed_T<OutIter>(current, begin, end, seq);
  }

  template<typename OutIter>
  inline PySwigIterator*
  make_output_iterator(const OutIter& current, PyObject *seq = 0)
  {
    return new PySwigIteratorOpen_T<OutIter>(current, seq);
  }
}


namespace swig
{
  template <class T>
  struct PySequence_Ref
  {
    PySequence_Ref(PyObject* seq, int index)
      : _seq(seq), _index(index)
    {
    }
    
    operator T () const
    {
      swig::PyObject_var item = PySequence_GetItem(_seq, _index);
      try {
	return swig::as<T>(item, true);
      } catch (std::exception& e) {
	char msg[1024];
	sprintf(msg, "in sequence element %d ", _index);
	if (!PyErr_Occurred()) {
	  SWIG_Error(SWIG_TypeError,  swig::type_name<T>());
	}
	SWIG_Python_AddErrorMsg(msg);
	SWIG_Python_AddErrorMsg(e.what());
	throw;
      }
    }

    PySequence_Ref& operator=(const T& v)
    {
      PySequence_SetItem(_seq, _index, swig::from<T>(v));
      return *this;
    }

  private:
    PyObject* _seq;
    int _index;
  };

  template <class T>
  struct PySequence_ArrowProxy
  {
    PySequence_ArrowProxy(const T& x): m_value(x) {}
    const T* operator->() const { return &m_value; }
    operator const T*() const { return &m_value; }
    T m_value;
  };

  template <class T, class Reference >
  struct PySequence_InputIterator
  {
    typedef PySequence_InputIterator<T, Reference > self;

    typedef std::random_access_iterator_tag iterator_category;
    typedef Reference reference;
    typedef T value_type;
    typedef T* pointer;
    typedef int difference_type;

    PySequence_InputIterator()
    {
    }

    PySequence_InputIterator(PyObject* seq, int index)
      : _seq(seq), _index(index)
    {
    }

    reference operator*() const
    {
      return reference(_seq, _index);
    }

    PySequence_ArrowProxy<T>
    operator->() const {
      return PySequence_ArrowProxy<T>(operator*());
    }

    bool operator==(const self& ri) const
    {
      return (_index == ri._index) && (_seq == ri._seq);
    }

    bool operator!=(const self& ri) const
    {
      return !(operator==(ri));
    }

    self& operator ++ ()
    {
      ++_index;
      return *this;
    }

    self& operator -- ()
    {
      --_index;
      return *this;
    }

    self& operator += (difference_type n)
    {
      _index += n;
      return *this;
    }

    self operator +(difference_type n) const
    {
      return self(_seq, _index + n);
    }

    self& operator -= (difference_type n)
    {
      _index -= n;
      return *this;
    }

    self operator -(difference_type n) const
    {
      return self(_seq, _index - n);
    }

    difference_type operator - (const self& ri) const
    {
      return _index - ri._index;
    }

    bool operator < (const self& ri) const
    {
      return _index < ri._index;
    }

    reference
    operator[](difference_type n) const
    {
      return reference(_seq, _index + n);
    }

  private:
    PyObject* _seq;
    difference_type _index;
  };

  template <class T>
  struct PySequence_Cont
  {
    typedef PySequence_Ref<T> reference;
    typedef const PySequence_Ref<T> const_reference;
    typedef T value_type;
    typedef T* pointer;
    typedef int difference_type;
    typedef int size_type;
    typedef const pointer const_pointer;
    typedef PySequence_InputIterator<T, reference> iterator;
    typedef PySequence_InputIterator<T, const_reference> const_iterator;

    PySequence_Cont(PyObject* seq) : _seq(0)
    {
      if (!PySequence_Check(seq)) {
	throw std::invalid_argument("a sequence is expected");
      }
      _seq = seq;
      Py_INCREF(_seq);
    }

    ~PySequence_Cont()
    {
      if (_seq) Py_DECREF(_seq);
    }

    size_type size() const
    {
      return PySequence_Size(_seq);
    }

    bool empty() const
    {
      return size() == 0;
    }

    iterator begin()
    {
      return iterator(_seq, 0);
    }

    const_iterator begin() const
    {
      return const_iterator(_seq, 0);
    }

    iterator end()
    {
      return iterator(_seq, size());
    }

    const_iterator end() const
    {
      return const_iterator(_seq, size());
    }

    reference operator[](difference_type n)
    {
      return reference(_seq, n);
    }

    const_reference operator[](difference_type n)  const
    {
      return const_reference(_seq, n);
    }

    bool check(bool set_err = true) const
    {
      int s = size();
      for (int i = 0; i < s; ++i) {
	swig::PyObject_var item = PySequence_GetItem(_seq, i);
	if (!swig::check<value_type>(item)) {
	  if (set_err) {
	    char msg[1024];
	    sprintf(msg, "in sequence element %d", i);
	    SWIG_Error(SWIG_RuntimeError, msg);
	  }
	  return false;
	}
      }
      return true;
    }

  private:
    PyObject* _seq;
  };

}


SWIGINTERN int
SWIG_AsVal_std_string (PyObject * obj, std::string *val)
{
  std::string* v = (std::string *) 0;
  int res = SWIG_AsPtr_std_string (obj, &v);
  if (!SWIG_IsOK(res)) return res;
  if (v) {
    if (val) *val = *v;
    if (SWIG_IsNewObj(res)) {
      delete v;
      res = SWIG_DelNewMask(res);
    }
    return res;
  }
  return SWIG_ERROR;
}


SWIGINTERNINLINE PyObject *
SWIG_From_std_string  (const std::string& s)
{
  if (s.size()) {
    return SWIG_FromCharPtrAndSize(s.data(), s.size());
  } else {
    return SWIG_FromCharPtrAndSize(s.c_str(), 0);
  }
}


namespace swig {
  template <> struct traits<std::string > {
    typedef value_category category;
    static const char* type_name() { return"std::string"; }
  };  
  template <>  struct traits_asval<std::string > {   
    typedef std::string value_type;
    static int asval(PyObject *obj, value_type *val) { 
      return SWIG_AsVal_std_string (obj, val);
    }
  };
  template <>  struct traits_from<std::string > {
    typedef std::string value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_std_string  (val);
    }
  };
}


namespace swig {
  template <class PySeq, class Seq>
  inline void
  assign(const PySeq& pyseq, Seq* seq) {
#ifdef SWIG_STD_NOASSIGN_STL
    typedef typename PySeq::value_type value_type;
    typename PySeq::const_iterator it = pyseq.begin();
    for (;it != pyseq.end(); ++it) {
      seq->insert(seq->end(),(value_type)(*it));
    }
#else
    seq->assign(pyseq.begin(), pyseq.end());
#endif
  }

  template <class Seq, class T = typename Seq::value_type >
  struct traits_asptr_stdseq {
    typedef Seq sequence;
    typedef T value_type;

    static int asptr(PyObject *obj, sequence **seq) {
      if (PySequence_Check(obj)) {
	try {
	  PySequence_Cont<value_type> pyseq(obj);
	  if (seq) {
	    sequence *pseq = new sequence();
	    assign(pyseq, pseq);
	    *seq = pseq;
	    return SWIG_NEWOBJ;
	  } else {
	    return pyseq.check() ? SWIG_OK : SWIG_ERROR;
	  }
	} catch (std::exception& e) {
	  if (seq) {
	    if (!PyErr_Occurred()) {
	      PyErr_SetString(PyExc_TypeError, e.what());
	    }
	  }
	  return SWIG_ERROR;
	}
      } else {
	sequence *p;
	if (SWIG_ConvertPtr(obj,(void**)&p,
			    swig::type_info<sequence>(),0) == SWIG_OK) {
	  if (seq) *seq = p;
	  return SWIG_OLDOBJ;
	}
      }
      return SWIG_ERROR;
    }
  };

  template <class Seq, class T = typename Seq::value_type >
  struct traits_from_stdseq {
    typedef Seq sequence;
    typedef T value_type;
    typedef typename Seq::size_type size_type;
    typedef typename sequence::const_iterator const_iterator;

    static PyObject *from(const sequence& seq) {






      size_type size = seq.size();
      if (size <= (size_type)INT_MAX) {
	PyObject *obj = PyTuple_New((int)size);
	int i = 0;
	for (const_iterator it = seq.begin();
	     it != seq.end(); ++it, ++i) {
	  PyTuple_SetItem(obj,i,swig::from<value_type>(*it));
	}
	return obj;
      } else {
	PyErr_SetString(PyExc_OverflowError,"sequence size not valid in python");
	return NULL;
      }
    }
  };
}


  namespace swig {
    template <class T>
    struct traits_asptr<std::vector<T> >  {
      static int asptr(PyObject *obj, std::vector<T> **vec) {
	return traits_asptr_stdseq<std::vector<T> >::asptr(obj, vec);
      }
    };
    
    template <class T>
    struct traits_from<std::vector<T> > {
      static PyObject *from(const std::vector<T>& vec) {
	return traits_from_stdseq<std::vector<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<std::string, std::allocator<std::string > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "std::string" "," "std::allocator<std::string >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_std_string_Sg__iterator(std::vector<std::string > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_std_string_Sg____nonzero__(std::vector<std::string > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<std::string >::size_type std_vector_Sl_std_string_Sg____len__(std::vector<std::string > const *self){
      return self->size();
    }

SWIGINTERNINLINE PyObject* 
SWIG_From_unsigned_SS_long  (unsigned long value)
{
  return (value > LONG_MAX) ?
    PyLong_FromUnsignedLong(value) : PyInt_FromLong(static_cast< long >(value)); 
}


SWIGINTERNINLINE PyObject *
SWIG_From_size_t  (size_t value)
{    
  return SWIG_From_unsigned_SS_long  (static_cast< unsigned long >(value));
}

SWIGINTERN std::vector<std::string >::value_type std_vector_Sl_std_string_Sg__pop(std::vector<std::string > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<std::string,std::allocator<std::string > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<std::string,std::allocator<std::string > > *std_vector_Sl_std_string_Sg____getslice__(std::vector<std::string > *self,std::vector<std::string >::difference_type i,std::vector<std::string >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setslice__(std::vector<std::string > *self,std::vector<std::string >::difference_type i,std::vector<std::string >::difference_type j,std::vector<std::string,std::allocator<std::string > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delslice__(std::vector<std::string > *self,std::vector<std::string >::difference_type i,std::vector<std::string >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____delitem__(std::vector<std::string > *self,std::vector<std::string >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<std::string >::value_type const &std_vector_Sl_std_string_Sg____getitem__(std::vector<std::string > const *self,std::vector<std::string >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_std_string_Sg____setitem__(std::vector<std::string > *self,std::vector<std::string >::difference_type i,std::vector<std::string >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_std_string_Sg__append(std::vector<std::string > *self,std::vector<std::string >::value_type const &x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::model::Spec > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::model::Spec"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::model::Spec*, std::allocator<steps::model::Spec * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::model::Spec" " *," "std::allocator<steps::model::Spec * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_model_Spec_Sm__Sg__iterator(std::vector<steps::model::Spec * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_model_Spec_Sm__Sg____nonzero__(std::vector<steps::model::Spec * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::model::Spec * >::size_type std_vector_Sl_steps_model_Spec_Sm__Sg____len__(std::vector<steps::model::Spec * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::model::Spec * >::value_type std_vector_Sl_steps_model_Spec_Sm__Sg__pop(std::vector<steps::model::Spec * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *std_vector_Sl_steps_model_Spec_Sm__Sg____getslice__(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::difference_type i,std::vector<steps::model::Spec * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Spec_Sm__Sg____setslice__(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::difference_type i,std::vector<steps::model::Spec * >::difference_type j,std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_model_Spec_Sm__Sg____delslice__(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::difference_type i,std::vector<steps::model::Spec * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Spec_Sm__Sg____delitem__(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::model::Spec * >::value_type std_vector_Sl_steps_model_Spec_Sm__Sg____getitem__(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_model_Spec_Sm__Sg____setitem__(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::difference_type i,std::vector<steps::model::Spec * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_model_Spec_Sm__Sg__append(std::vector<steps::model::Spec * > *self,std::vector<steps::model::Spec * >::value_type x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::model::Reac > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::model::Reac"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::model::Reac*, std::allocator<steps::model::Reac * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::model::Reac" " *," "std::allocator<steps::model::Reac * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_model_Reac_Sm__Sg__iterator(std::vector<steps::model::Reac * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_model_Reac_Sm__Sg____nonzero__(std::vector<steps::model::Reac * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::model::Reac * >::size_type std_vector_Sl_steps_model_Reac_Sm__Sg____len__(std::vector<steps::model::Reac * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::model::Reac * >::value_type std_vector_Sl_steps_model_Reac_Sm__Sg__pop(std::vector<steps::model::Reac * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > *std_vector_Sl_steps_model_Reac_Sm__Sg____getslice__(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::difference_type i,std::vector<steps::model::Reac * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Reac_Sm__Sg____setslice__(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::difference_type i,std::vector<steps::model::Reac * >::difference_type j,std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_model_Reac_Sm__Sg____delslice__(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::difference_type i,std::vector<steps::model::Reac * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Reac_Sm__Sg____delitem__(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::model::Reac * >::value_type std_vector_Sl_steps_model_Reac_Sm__Sg____getitem__(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_model_Reac_Sm__Sg____setitem__(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::difference_type i,std::vector<steps::model::Reac * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_model_Reac_Sm__Sg__append(std::vector<steps::model::Reac * > *self,std::vector<steps::model::Reac * >::value_type x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::model::SReac > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::model::SReac"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::model::SReac*, std::allocator<steps::model::SReac * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::model::SReac" " *," "std::allocator<steps::model::SReac * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_model_SReac_Sm__Sg__iterator(std::vector<steps::model::SReac * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_model_SReac_Sm__Sg____nonzero__(std::vector<steps::model::SReac * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::model::SReac * >::size_type std_vector_Sl_steps_model_SReac_Sm__Sg____len__(std::vector<steps::model::SReac * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::model::SReac * >::value_type std_vector_Sl_steps_model_SReac_Sm__Sg__pop(std::vector<steps::model::SReac * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > *std_vector_Sl_steps_model_SReac_Sm__Sg____getslice__(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::difference_type i,std::vector<steps::model::SReac * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_SReac_Sm__Sg____setslice__(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::difference_type i,std::vector<steps::model::SReac * >::difference_type j,std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_model_SReac_Sm__Sg____delslice__(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::difference_type i,std::vector<steps::model::SReac * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_SReac_Sm__Sg____delitem__(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::model::SReac * >::value_type std_vector_Sl_steps_model_SReac_Sm__Sg____getitem__(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_model_SReac_Sm__Sg____setitem__(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::difference_type i,std::vector<steps::model::SReac * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_model_SReac_Sm__Sg__append(std::vector<steps::model::SReac * > *self,std::vector<steps::model::SReac * >::value_type x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::model::Diff > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::model::Diff"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::model::Diff*, std::allocator<steps::model::Diff * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::model::Diff" " *," "std::allocator<steps::model::Diff * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_model_Diff_Sm__Sg__iterator(std::vector<steps::model::Diff * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_model_Diff_Sm__Sg____nonzero__(std::vector<steps::model::Diff * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::model::Diff * >::size_type std_vector_Sl_steps_model_Diff_Sm__Sg____len__(std::vector<steps::model::Diff * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::model::Diff * >::value_type std_vector_Sl_steps_model_Diff_Sm__Sg__pop(std::vector<steps::model::Diff * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > *std_vector_Sl_steps_model_Diff_Sm__Sg____getslice__(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::difference_type i,std::vector<steps::model::Diff * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Diff_Sm__Sg____setslice__(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::difference_type i,std::vector<steps::model::Diff * >::difference_type j,std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_model_Diff_Sm__Sg____delslice__(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::difference_type i,std::vector<steps::model::Diff * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Diff_Sm__Sg____delitem__(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::model::Diff * >::value_type std_vector_Sl_steps_model_Diff_Sm__Sg____getitem__(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_model_Diff_Sm__Sg____setitem__(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::difference_type i,std::vector<steps::model::Diff * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_model_Diff_Sm__Sg__append(std::vector<steps::model::Diff * > *self,std::vector<steps::model::Diff * >::value_type x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::model::Volsys > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::model::Volsys"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::model::Volsys*, std::allocator<steps::model::Volsys * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::model::Volsys" " *," "std::allocator<steps::model::Volsys * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_model_Volsys_Sm__Sg__iterator(std::vector<steps::model::Volsys * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_model_Volsys_Sm__Sg____nonzero__(std::vector<steps::model::Volsys * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::model::Volsys * >::size_type std_vector_Sl_steps_model_Volsys_Sm__Sg____len__(std::vector<steps::model::Volsys * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::model::Volsys * >::value_type std_vector_Sl_steps_model_Volsys_Sm__Sg__pop(std::vector<steps::model::Volsys * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > *std_vector_Sl_steps_model_Volsys_Sm__Sg____getslice__(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::difference_type i,std::vector<steps::model::Volsys * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Volsys_Sm__Sg____setslice__(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::difference_type i,std::vector<steps::model::Volsys * >::difference_type j,std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_model_Volsys_Sm__Sg____delslice__(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::difference_type i,std::vector<steps::model::Volsys * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Volsys_Sm__Sg____delitem__(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::model::Volsys * >::value_type std_vector_Sl_steps_model_Volsys_Sm__Sg____getitem__(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_model_Volsys_Sm__Sg____setitem__(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::difference_type i,std::vector<steps::model::Volsys * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_model_Volsys_Sm__Sg__append(std::vector<steps::model::Volsys * > *self,std::vector<steps::model::Volsys * >::value_type x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::model::Surfsys > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::model::Surfsys"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::model::Surfsys*, std::allocator<steps::model::Surfsys * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::model::Surfsys" " *," "std::allocator<steps::model::Surfsys * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_model_Surfsys_Sm__Sg__iterator(std::vector<steps::model::Surfsys * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_model_Surfsys_Sm__Sg____nonzero__(std::vector<steps::model::Surfsys * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::model::Surfsys * >::size_type std_vector_Sl_steps_model_Surfsys_Sm__Sg____len__(std::vector<steps::model::Surfsys * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::model::Surfsys * >::value_type std_vector_Sl_steps_model_Surfsys_Sm__Sg__pop(std::vector<steps::model::Surfsys * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > *std_vector_Sl_steps_model_Surfsys_Sm__Sg____getslice__(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::difference_type i,std::vector<steps::model::Surfsys * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Surfsys_Sm__Sg____setslice__(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::difference_type i,std::vector<steps::model::Surfsys * >::difference_type j,std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_model_Surfsys_Sm__Sg____delslice__(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::difference_type i,std::vector<steps::model::Surfsys * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_model_Surfsys_Sm__Sg____delitem__(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::model::Surfsys * >::value_type std_vector_Sl_steps_model_Surfsys_Sm__Sg____getitem__(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_model_Surfsys_Sm__Sg____setitem__(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::difference_type i,std::vector<steps::model::Surfsys * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_model_Surfsys_Sm__Sg__append(std::vector<steps::model::Surfsys * > *self,std::vector<steps::model::Surfsys * >::value_type x){
      self->push_back(x);
    }

  #define SWIG_From_double   PyFloat_FromDouble 


SWIGINTERNINLINE PyObject *
SWIG_From_int  (int value)
{    
  return SWIG_From_long  (value);
}


#include <set>


#include "../cpp/geom/geom.hpp"
#include "../cpp//geom/comp.hpp"
#include "../cpp/geom/patch.hpp"
#include "../cpp/geom/tmcomp.hpp"
#include "../cpp/geom/tmpatch.hpp"
#include "../cpp/geom/tet.hpp"
#include "../cpp/geom/tetmesh_rw.hpp"
#include "../cpp/geom/tetmesh.hpp"
#include "../cpp/geom/tri.hpp"
#include "../cpp/error.hpp"
#include "../cpp/geom/diffboundary.hpp"


  namespace swig {
    template <class PySeq, class T> 
    inline void 
    assign(const PySeq& pyseq, std::set<T>* seq) {
#ifdef SWIG_STD_NOINSERT_TEMPLATE_STL
      typedef typename PySeq::value_type value_type;
      typename PySeq::const_iterator it = pyseq.begin();
      for (;it != pyseq.end(); ++it) {
	seq->insert(seq->end(),(value_type)(*it));
      }
#else
      seq->insert(pyseq.begin(), pyseq.end());
#endif
    }

    template <class T>
    struct traits_asptr<std::set<T> >  {
      static int asptr(PyObject *obj, std::set<T> **s) {
	return traits_asptr_stdseq<std::set<T> >::asptr(obj, s);
      }
    };

    template <class T>
    struct traits_from<std::set<T> > {
      static PyObject *from(const std::set<T>& vec) {
	return traits_from_stdseq<std::set<T> >::from(vec);
      }
    };
  }


      namespace swig {
	template <>  struct traits<std::set<std::string, std::less<std::string >, std::allocator<std::string > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::set<" "std::string" "," "std::allocator<std::string >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_set_Sl_std_string_Sg__iterator(std::set<std::string > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_set_Sl_std_string_Sg____nonzero__(std::set<std::string > const *self){
      return !(self->empty());
    }
SWIGINTERN std::set<std::string >::size_type std_set_Sl_std_string_Sg____len__(std::set<std::string > const *self){
      return self->size();
    }
SWIGINTERN void std_set_Sl_std_string_Sg__append(std::set<std::string > *self,std::set<std::string >::value_type x){
       self->insert(x);
     }
SWIGINTERN bool std_set_Sl_std_string_Sg____contains__(std::set<std::string > *self,std::set<std::string >::value_type x){
       return self->find(x) != self->end();
     }
SWIGINTERN std::set<std::string >::value_type std_set_Sl_std_string_Sg____getitem__(std::set<std::string > const *self,std::set<std::string >::difference_type i){
       return *(swig::cgetpos(self, i));
     }


  namespace swig {
    template <>  struct traits<steps::wm::Patch * > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::wm::Patch *"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::set<steps::wm::Patch *, std::less<steps::wm::Patch * >, std::allocator<steps::wm::Patch * > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::set<" "steps::wm::Patch *" "," "std::allocator<steps::wm::Patch * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_set_Sl_steps_wm_Patch_Sm__Sg__iterator(std::set<steps::wm::Patch * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_set_Sl_steps_wm_Patch_Sm__Sg____nonzero__(std::set<steps::wm::Patch * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::set<steps::wm::Patch * >::size_type std_set_Sl_steps_wm_Patch_Sm__Sg____len__(std::set<steps::wm::Patch * > const *self){
      return self->size();
    }
SWIGINTERN void std_set_Sl_steps_wm_Patch_Sm__Sg__append(std::set<steps::wm::Patch * > *self,std::set<steps::wm::Patch * >::value_type x){
       self->insert(x);
     }
SWIGINTERN bool std_set_Sl_steps_wm_Patch_Sm__Sg____contains__(std::set<steps::wm::Patch * > *self,std::set<steps::wm::Patch * >::value_type x){
       return self->find(x) != self->end();
     }
SWIGINTERN std::set<steps::wm::Patch * >::value_type std_set_Sl_steps_wm_Patch_Sm__Sg____getitem__(std::set<steps::wm::Patch * > const *self,std::set<steps::wm::Patch * >::difference_type i){
       return *(swig::cgetpos(self, i));
     }

  namespace swig {
    template <>  struct traits<steps::wm::Patch > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::wm::Patch"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::wm::Patch*, std::allocator<steps::wm::Patch * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::wm::Patch" " *," "std::allocator<steps::wm::Patch * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_wm_Patch_Sm__Sg__iterator(std::vector<steps::wm::Patch * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_wm_Patch_Sm__Sg____nonzero__(std::vector<steps::wm::Patch * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::wm::Patch * >::size_type std_vector_Sl_steps_wm_Patch_Sm__Sg____len__(std::vector<steps::wm::Patch * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::wm::Patch * >::value_type std_vector_Sl_steps_wm_Patch_Sm__Sg__pop(std::vector<steps::wm::Patch * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > *std_vector_Sl_steps_wm_Patch_Sm__Sg____getslice__(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::difference_type i,std::vector<steps::wm::Patch * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_wm_Patch_Sm__Sg____setslice__(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::difference_type i,std::vector<steps::wm::Patch * >::difference_type j,std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_wm_Patch_Sm__Sg____delslice__(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::difference_type i,std::vector<steps::wm::Patch * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_wm_Patch_Sm__Sg____delitem__(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::wm::Patch * >::value_type std_vector_Sl_steps_wm_Patch_Sm__Sg____getitem__(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_wm_Patch_Sm__Sg____setitem__(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::difference_type i,std::vector<steps::wm::Patch * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_wm_Patch_Sm__Sg__append(std::vector<steps::wm::Patch * > *self,std::vector<steps::wm::Patch * >::value_type x){
      self->push_back(x);
    }

  namespace swig {
    template <>  struct traits<steps::wm::Comp > {
      typedef pointer_category category;
      static const char* type_name() { return"steps::wm::Comp"; }
    };
  }


      namespace swig {
	template <>  struct traits<std::vector<steps::wm::Comp*, std::allocator<steps::wm::Comp * > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<" "steps::wm::Comp" " *," "std::allocator<steps::wm::Comp * >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_steps_wm_Comp_Sm__Sg__iterator(std::vector<steps::wm::Comp * > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_steps_wm_Comp_Sm__Sg____nonzero__(std::vector<steps::wm::Comp * > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<steps::wm::Comp * >::size_type std_vector_Sl_steps_wm_Comp_Sm__Sg____len__(std::vector<steps::wm::Comp * > const *self){
      return self->size();
    }
SWIGINTERN std::vector<steps::wm::Comp * >::value_type std_vector_Sl_steps_wm_Comp_Sm__Sg__pop(std::vector<steps::wm::Comp * > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > *std_vector_Sl_steps_wm_Comp_Sm__Sg____getslice__(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::difference_type i,std::vector<steps::wm::Comp * >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_wm_Comp_Sm__Sg____setslice__(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::difference_type i,std::vector<steps::wm::Comp * >::difference_type j,std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_steps_wm_Comp_Sm__Sg____delslice__(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::difference_type i,std::vector<steps::wm::Comp * >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_steps_wm_Comp_Sm__Sg____delitem__(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<steps::wm::Comp * >::value_type std_vector_Sl_steps_wm_Comp_Sm__Sg____getitem__(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_steps_wm_Comp_Sm__Sg____setitem__(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::difference_type i,std::vector<steps::wm::Comp * >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_steps_wm_Comp_Sm__Sg__append(std::vector<steps::wm::Comp * > *self,std::vector<steps::wm::Comp * >::value_type x){
      self->push_back(x);
    }

#include <limits.h>
#ifndef LLONG_MIN
# define LLONG_MIN	LONG_LONG_MIN
#endif
#ifndef LLONG_MAX
# define LLONG_MAX	LONG_LONG_MAX
#endif
#ifndef ULLONG_MAX
# define ULLONG_MAX	ULONG_LONG_MAX
#endif


SWIGINTERN int
SWIG_AsVal_int (PyObject * obj, int *val)
{
  long v;
  int res = SWIG_AsVal_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < INT_MIN || v > INT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< int >(v);
    }
  }  
  return res;
}


namespace swig {
  template <> struct traits<int > {
    typedef value_category category;
    static const char* type_name() { return"int"; }
  };  
  template <>  struct traits_asval<int > {   
    typedef int value_type;
    static int asval(PyObject *obj, value_type *val) { 
      return SWIG_AsVal_int (obj, val);
    }
  };
  template <>  struct traits_from<int > {
    typedef int value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_int  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector<int, std::allocator<int > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "int" "," "std::allocator<int >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_int_Sg__iterator(std::vector<int > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_int_Sg____nonzero__(std::vector<int > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<int >::size_type std_vector_Sl_int_Sg____len__(std::vector<int > const *self){
      return self->size();
    }
SWIGINTERN std::vector<int >::value_type std_vector_Sl_int_Sg__pop(std::vector<int > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<int,std::allocator<int > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<int,std::allocator<int > > *std_vector_Sl_int_Sg____getslice__(std::vector<int > *self,std::vector<int >::difference_type i,std::vector<int >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_int_Sg____setslice__(std::vector<int > *self,std::vector<int >::difference_type i,std::vector<int >::difference_type j,std::vector<int,std::allocator<int > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_int_Sg____delslice__(std::vector<int > *self,std::vector<int >::difference_type i,std::vector<int >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_int_Sg____delitem__(std::vector<int > *self,std::vector<int >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<int >::value_type const &std_vector_Sl_int_Sg____getitem__(std::vector<int > const *self,std::vector<int >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_int_Sg____setitem__(std::vector<int > *self,std::vector<int >::difference_type i,std::vector<int >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_int_Sg__append(std::vector<int > *self,std::vector<int >::value_type const &x){
      self->push_back(x);
    }

SWIGINTERN int
SWIG_AsVal_unsigned_SS_int (PyObject * obj, unsigned int *val)
{
  unsigned long v;
  int res = SWIG_AsVal_unsigned_SS_long (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v > UINT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< unsigned int >(v);
    }
  }  
  return res;
}


SWIGINTERNINLINE PyObject *
SWIG_From_unsigned_SS_int  (unsigned int value)
{    
  return SWIG_From_unsigned_SS_long  (value);
}


namespace swig {
  template <> struct traits<unsigned int > {
    typedef value_category category;
    static const char* type_name() { return"unsigned int"; }
  };  
  template <>  struct traits_asval<unsigned int > {   
    typedef unsigned int value_type;
    static int asval(PyObject *obj, value_type *val) { 
      return SWIG_AsVal_unsigned_SS_int (obj, val);
    }
  };
  template <>  struct traits_from<unsigned int > {
    typedef unsigned int value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_unsigned_SS_int  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector<unsigned int, std::allocator<unsigned int > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "unsigned int" "," "std::allocator<unsigned int >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_unsigned_SS_int_Sg__iterator(std::vector<unsigned int > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_unsigned_SS_int_Sg____nonzero__(std::vector<unsigned int > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<unsigned int >::size_type std_vector_Sl_unsigned_SS_int_Sg____len__(std::vector<unsigned int > const *self){
      return self->size();
    }
SWIGINTERN std::vector<unsigned int >::value_type std_vector_Sl_unsigned_SS_int_Sg__pop(std::vector<unsigned int > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<unsigned int,std::allocator<unsigned int > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<unsigned int,std::allocator<unsigned int > > *std_vector_Sl_unsigned_SS_int_Sg____getslice__(std::vector<unsigned int > *self,std::vector<unsigned int >::difference_type i,std::vector<unsigned int >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setslice__(std::vector<unsigned int > *self,std::vector<unsigned int >::difference_type i,std::vector<unsigned int >::difference_type j,std::vector<unsigned int,std::allocator<unsigned int > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____delslice__(std::vector<unsigned int > *self,std::vector<unsigned int >::difference_type i,std::vector<unsigned int >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____delitem__(std::vector<unsigned int > *self,std::vector<unsigned int >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<unsigned int >::value_type const &std_vector_Sl_unsigned_SS_int_Sg____getitem__(std::vector<unsigned int > const *self,std::vector<unsigned int >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg____setitem__(std::vector<unsigned int > *self,std::vector<unsigned int >::difference_type i,std::vector<unsigned int >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_unsigned_SS_int_Sg__append(std::vector<unsigned int > *self,std::vector<unsigned int >::value_type const &x){
      self->push_back(x);
    }

namespace swig {
  template <> struct traits<double > {
    typedef value_category category;
    static const char* type_name() { return"double"; }
  };  
  template <>  struct traits_asval<double > {   
    typedef double value_type;
    static int asval(PyObject *obj, value_type *val) { 
      return SWIG_AsVal_double (obj, val);
    }
  };
  template <>  struct traits_from<double > {
    typedef double value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_double  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector<double, std::allocator<double > > > {
	  typedef pointer_category category;
	  static const char* type_name() {
	    return "std::vector<" "double" "," "std::allocator<double >" " >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_double_Sg__iterator(std::vector<double > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_double_Sg____nonzero__(std::vector<double > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<double >::size_type std_vector_Sl_double_Sg____len__(std::vector<double > const *self){
      return self->size();
    }
SWIGINTERN std::vector<double >::value_type std_vector_Sl_double_Sg__pop(std::vector<double > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<double,std::allocator<double > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<double,std::allocator<double > > *std_vector_Sl_double_Sg____getslice__(std::vector<double > *self,std::vector<double >::difference_type i,std::vector<double >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_double_Sg____setslice__(std::vector<double > *self,std::vector<double >::difference_type i,std::vector<double >::difference_type j,std::vector<double,std::allocator<double > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delslice__(std::vector<double > *self,std::vector<double >::difference_type i,std::vector<double >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_double_Sg____delitem__(std::vector<double > *self,std::vector<double >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<double >::value_type const &std_vector_Sl_double_Sg____getitem__(std::vector<double > const *self,std::vector<double >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_double_Sg____setitem__(std::vector<double > *self,std::vector<double >::difference_type i,std::vector<double >::value_type const &x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_double_Sg__append(std::vector<double > *self,std::vector<double >::value_type const &x){
      self->push_back(x);
    }

SWIGINTERN int
SWIG_AsVal_bool (PyObject *obj, bool *val)
{
  if (obj == Py_True) {
    if (val) *val = true;
    return SWIG_OK;
  } else if (obj == Py_False) {
    if (val) *val = false;
    return SWIG_OK;
  } else {
    long v = 0;
    int res = SWIG_AddCast(SWIG_AsVal_long (obj, val ? &v : 0));
    if (SWIG_IsOK(res) && val) *val = v ? true : false;
    return res;
  }
}


namespace swig {
  template <> struct traits<bool > {
    typedef value_category category;
    static const char* type_name() { return"bool"; }
  };  
  template <>  struct traits_asval<bool > {   
    typedef bool value_type;
    static int asval(PyObject *obj, value_type *val) { 
      return SWIG_AsVal_bool (obj, val);
    }
  };
  template <>  struct traits_from<bool > {
    typedef bool value_type;
    static PyObject *from(const value_type& val) {
      return SWIG_From_bool  (val);
    }
  };
}


      namespace swig {
	template <>  struct traits<std::vector<bool, std::allocator<bool > > > {
	  typedef value_category category;
	  static const char* type_name() {
	    return "std::vector<bool, std::allocator<bool > >";
	  }
	};
      }
    
SWIGINTERN swig::PySwigIterator *std_vector_Sl_bool_Sg__iterator(std::vector<bool > *self,PyObject **PYTHON_SELF){
      return swig::make_output_iterator(self->begin(), self->begin(), self->end(), *PYTHON_SELF);
    }
SWIGINTERN bool std_vector_Sl_bool_Sg____nonzero__(std::vector<bool > const *self){
      return !(self->empty());
    }
SWIGINTERN std::vector<bool >::size_type std_vector_Sl_bool_Sg____len__(std::vector<bool > const *self){
      return self->size();
    }
SWIGINTERN std::vector<bool >::value_type std_vector_Sl_bool_Sg__pop(std::vector<bool > *self){
      if (self->size() == 0)
	throw std::out_of_range("pop from empty container");
      std::vector<bool,std::allocator<bool > >::value_type x = self->back();
      self->pop_back();
      return x;
    }
SWIGINTERN std::vector<bool,std::allocator<bool > > *std_vector_Sl_bool_Sg____getslice__(std::vector<bool > *self,std::vector<bool >::difference_type i,std::vector<bool >::difference_type j){
      return swig::getslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_bool_Sg____setslice__(std::vector<bool > *self,std::vector<bool >::difference_type i,std::vector<bool >::difference_type j,std::vector<bool,std::allocator<bool > > const &v){
      swig::setslice(self, i, j, v);
    }
SWIGINTERN void std_vector_Sl_bool_Sg____delslice__(std::vector<bool > *self,std::vector<bool >::difference_type i,std::vector<bool >::difference_type j){
      swig::delslice(self, i, j);
    }
SWIGINTERN void std_vector_Sl_bool_Sg____delitem__(std::vector<bool > *self,std::vector<bool >::difference_type i){
      self->erase(swig::getpos(self,i));
    }
SWIGINTERN std::vector<bool >::value_type std_vector_Sl_bool_Sg____getitem__(std::vector<bool > *self,std::vector<bool >::difference_type i){
      return *(swig::cgetpos(self, i));
    }
SWIGINTERN void std_vector_Sl_bool_Sg____setitem__(std::vector<bool > *self,std::vector<bool >::difference_type i,std::vector<bool >::value_type x){
      *(swig::getpos(self,i)) = x;
    }
SWIGINTERN void std_vector_Sl_bool_Sg__append(std::vector<bool > *self,std::vector<bool >::value_type x){
      self->push_back(x);
    }
	
#include "../cpp/rng/rng.hpp"


SWIGINTERNINLINE PyObject *
SWIG_From_float  (float value)
{    
  return SWIG_From_double  (value);
}


SWIGINTERN int
SWIG_AsVal_float (PyObject * obj, float *val)
{
  double v;
  int res = SWIG_AsVal_double (obj, &v);
  if (SWIG_IsOK(res)) {
    if ((v < -FLT_MAX || v > FLT_MAX)) {
      return SWIG_OverflowError;
    } else {
      if (val) *val = static_cast< float >(v);
    }
  }  
  return res;
}


#include "../cpp//solver/api.hpp"
#include "../cpp/solver/statedef.hpp"
#include "../cpp/wmrk4/wmrk4.hpp"
#include "../cpp/wmdirect/wmdirect.hpp"
#include "../cpp/tetexact/tetexact.hpp"
#include "../cpp/error.hpp"

#ifdef __cplusplus
extern "C" {
#endif
SWIGINTERN PyObject *_wrap_delete_PySwigIterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_PySwigIterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_PySwigIterator" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  delete arg1;
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_value(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  PyObject *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:PySwigIterator_value",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_value" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  try {
    result = (PyObject *)((swig::PySwigIterator const *)arg1)->value();
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_incr__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  size_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator_incr",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_incr" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator_incr" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = static_cast< size_t >(val2);
  try {
    result = (swig::PySwigIterator *)(arg1)->incr(arg2);
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_incr__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:PySwigIterator_incr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_incr" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  try {
    result = (swig::PySwigIterator *)(arg1)->incr();
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_incr(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__PySwigIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_PySwigIterator_incr__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__PySwigIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PySwigIterator_incr__SWIG_0(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'PySwigIterator_incr'.\n  Possible C/C++ prototypes are:\n    incr(size_t)\n    incr()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_decr__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  size_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator_decr",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_decr" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator_decr" "', argument " "2"" of type '" "size_t""'");
  } 
  arg2 = static_cast< size_t >(val2);
  try {
    result = (swig::PySwigIterator *)(arg1)->decr(arg2);
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_decr__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:PySwigIterator_decr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_decr" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  try {
    result = (swig::PySwigIterator *)(arg1)->decr();
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_decr(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__PySwigIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_PySwigIterator_decr__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__PySwigIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PySwigIterator_decr__SWIG_0(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'PySwigIterator_decr'.\n  Possible C/C++ prototypes are:\n    decr(size_t)\n    decr()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_distance(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *arg2 = 0 ;
  ptrdiff_t result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator_distance",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_distance" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__PySwigIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PySwigIterator_distance" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PySwigIterator_distance" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::PySwigIterator * >(argp2);
  try {
    result = ((swig::PySwigIterator const *)arg1)->distance((swig::PySwigIterator const &)*arg2);
  }
  catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail;
  }
  
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_equal(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *arg2 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator_equal",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_equal" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__PySwigIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PySwigIterator_equal" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PySwigIterator_equal" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::PySwigIterator * >(argp2);
  try {
    result = (bool)((swig::PySwigIterator const *)arg1)->equal((swig::PySwigIterator const &)*arg2);
  }
  catch(std::invalid_argument &_e) {
    SWIG_Python_Raise(SWIG_NewPointerObj((new std::invalid_argument(static_cast< const std::invalid_argument& >(_e))),SWIGTYPE_p_std__invalid_argument,SWIG_POINTER_OWN), "std::invalid_argument", SWIGTYPE_p_std__invalid_argument); SWIG_fail;
  }
  
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_copy(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:PySwigIterator_copy",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_copy" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  result = (swig::PySwigIterator *)((swig::PySwigIterator const *)arg1)->copy();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_next(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  PyObject *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:PySwigIterator_next",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_next" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->next();
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_previous(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  PyObject *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:PySwigIterator_previous",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_previous" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  try {
    result = (PyObject *)(arg1)->previous();
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = result;
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator_advance(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  ptrdiff_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator_advance",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator_advance" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator_advance" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::PySwigIterator *)(arg1)->advance(arg2);
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___eq__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *arg2 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___eq__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___eq__" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__PySwigIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PySwigIterator___eq__" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PySwigIterator___eq__" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::PySwigIterator * >(argp2);
  result = (bool)((swig::PySwigIterator const *)arg1)->operator ==((swig::PySwigIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___ne__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *arg2 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___ne__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___ne__" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__PySwigIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PySwigIterator___ne__" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PySwigIterator___ne__" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::PySwigIterator * >(argp2);
  result = (bool)((swig::PySwigIterator const *)arg1)->operator !=((swig::PySwigIterator const &)*arg2);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___iadd__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  ptrdiff_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___iadd__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___iadd__" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator___iadd__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    {
      swig::PySwigIterator &_result_ref = (arg1)->operator +=(arg2);
      result = (swig::PySwigIterator *) &_result_ref;
    }
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___isub__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  ptrdiff_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___isub__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___isub__" "', argument " "1"" of type '" "swig::PySwigIterator *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator___isub__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    {
      swig::PySwigIterator &_result_ref = (arg1)->operator -=(arg2);
      result = (swig::PySwigIterator *) &_result_ref;
    }
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___add__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  ptrdiff_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___add__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___add__" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator___add__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::PySwigIterator *)((swig::PySwigIterator const *)arg1)->operator +(arg2);
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___sub____SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  ptrdiff_t arg2 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___sub__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___sub__" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "PySwigIterator___sub__" "', argument " "2"" of type '" "ptrdiff_t""'");
  } 
  arg2 = static_cast< ptrdiff_t >(val2);
  try {
    result = (swig::PySwigIterator *)((swig::PySwigIterator const *)arg1)->operator -(arg2);
  }
  catch(swig::stop_iteration &_e) {
    {
      (void)_e;
      SWIG_SetErrorObj(PyExc_StopIteration, SWIG_Py_Void());
      SWIG_fail;
    }
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___sub____SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  swig::PySwigIterator *arg1 = (swig::PySwigIterator *) 0 ;
  swig::PySwigIterator *arg2 = 0 ;
  ptrdiff_t result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:PySwigIterator___sub__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_swig__PySwigIterator, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "PySwigIterator___sub__" "', argument " "1"" of type '" "swig::PySwigIterator const *""'"); 
  }
  arg1 = reinterpret_cast< swig::PySwigIterator * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_swig__PySwigIterator,  0  | 0);
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "PySwigIterator___sub__" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "PySwigIterator___sub__" "', argument " "2"" of type '" "swig::PySwigIterator const &""'"); 
  }
  arg2 = reinterpret_cast< swig::PySwigIterator * >(argp2);
  result = ((swig::PySwigIterator const *)arg1)->operator -((swig::PySwigIterator const &)*arg2);
  resultobj = SWIG_From_ptrdiff_t(static_cast< ptrdiff_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_PySwigIterator___sub__(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__PySwigIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_ConvertPtr(argv[1], 0, SWIGTYPE_p_swig__PySwigIterator, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_PySwigIterator___sub____SWIG_1(self, args);
      }
    }
  }
  if (argc == 2) {
    int _v;
    void *vptr = 0;
    int res = SWIG_ConvertPtr(argv[0], &vptr, SWIGTYPE_p_swig__PySwigIterator, 0);
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_ptrdiff_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_PySwigIterator___sub____SWIG_0(self, args);
      }
    }
  }
  
fail:
  Py_INCREF(Py_NotImplemented);
  return Py_NotImplemented;
}


SWIGINTERN PyObject *PySwigIterator_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_swig__PySwigIterator, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Err__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::Err *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_Err",&obj0)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Err" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Err" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  result = (steps::Err *)new steps::Err((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__Err, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Err__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::Err *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_Err")) SWIG_fail;
  result = (steps::Err *)new steps::Err();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__Err, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Err(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_Err__SWIG_1(self, args);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_Err__SWIG_0(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_Err'.\n  Possible C/C++ prototypes are:\n    steps::Err(std::string const &)\n    steps::Err()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_Err_getMsg(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::Err *arg1 = (steps::Err *) 0 ;
  char *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Err_getMsg",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__Err, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Err_getMsg" "', argument " "1"" of type '" "steps::Err *""'"); 
  }
  arg1 = reinterpret_cast< steps::Err * >(argp1);
  result = (char *)(arg1)->getMsg();
  resultobj = SWIG_FromCharPtr((const char *)result);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Err(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::Err *arg1 = (steps::Err *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Err",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__Err, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Err" "', argument " "1"" of type '" "steps::Err *""'"); 
  }
  arg1 = reinterpret_cast< steps::Err * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Err_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__Err, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_ArgErr__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::ArgErr *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_ArgErr",&obj0)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_ArgErr" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_ArgErr" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  result = (steps::ArgErr *)new steps::ArgErr((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__ArgErr, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_ArgErr__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::ArgErr *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_ArgErr")) SWIG_fail;
  result = (steps::ArgErr *)new steps::ArgErr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__ArgErr, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_ArgErr(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_ArgErr__SWIG_1(self, args);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_ArgErr__SWIG_0(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_ArgErr'.\n  Possible C/C++ prototypes are:\n    steps::ArgErr(std::string const &)\n    steps::ArgErr()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_ArgErr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::ArgErr *arg1 = (steps::ArgErr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_ArgErr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__ArgErr, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_ArgErr" "', argument " "1"" of type '" "steps::ArgErr *""'"); 
  }
  arg1 = reinterpret_cast< steps::ArgErr * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *ArgErr_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__ArgErr, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_NotImplErr__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::NotImplErr *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_NotImplErr",&obj0)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_NotImplErr" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_NotImplErr" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  result = (steps::NotImplErr *)new steps::NotImplErr((std::string const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__NotImplErr, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_NotImplErr__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::NotImplErr *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_NotImplErr")) SWIG_fail;
  result = (steps::NotImplErr *)new steps::NotImplErr();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__NotImplErr, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_NotImplErr(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_NotImplErr__SWIG_1(self, args);
  }
  if (argc == 1) {
    int _v;
    int res = SWIG_AsPtr_std_string(argv[0], (std::string**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_NotImplErr__SWIG_0(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_NotImplErr'.\n  Possible C/C++ prototypes are:\n    steps::NotImplErr(std::string const &)\n    steps::NotImplErr()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_NotImplErr(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::NotImplErr *arg1 = (steps::NotImplErr *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_NotImplErr",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__NotImplErr, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_NotImplErr" "', argument " "1"" of type '" "steps::NotImplErr *""'"); 
  }
  arg1 = reinterpret_cast< steps::NotImplErr * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *NotImplErr_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__NotImplErr, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_str_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_iterator" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_std_string_Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___nonzero__" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (bool)std_vector_Sl_std_string_Sg____nonzero__((std::vector<std::string > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___len__" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = std_vector_Sl_std_string_Sg____len__((std::vector<std::string > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_pop" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  try {
    result = std_vector_Sl_std_string_Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::difference_type arg2 ;
  std::vector<std::string >::difference_type arg3 ;
  std::vector<std::string,std::allocator<std::string > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___getslice__" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str___getslice__" "', argument " "2"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_str___getslice__" "', argument " "3"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<std::string >::difference_type >(val3);
  try {
    result = (std::vector<std::string,std::allocator<std::string > > *)std_vector_Sl_std_string_Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::difference_type arg2 ;
  std::vector<std::string >::difference_type arg3 ;
  std::vector<std::string,std::allocator<std::string > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_str___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___setslice__" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str___setslice__" "', argument " "2"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_str___setslice__" "', argument " "3"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<std::string >::difference_type >(val3);
  {
    std::vector<std::string,std::allocator<std::string > > *ptr = (std::vector<std::string,std::allocator<std::string > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_str___setslice__" "', argument " "4"" of type '" "std::vector<std::string,std::allocator<std::string > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str___setslice__" "', argument " "4"" of type '" "std::vector<std::string,std::allocator<std::string > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_std_string_Sg____setslice__(arg1,arg2,arg3,(std::vector<std::string,std::allocator<std::string > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::difference_type arg2 ;
  std::vector<std::string >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___delslice__" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str___delslice__" "', argument " "2"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_str___delslice__" "', argument " "3"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<std::string >::difference_type >(val3);
  try {
    std_vector_Sl_std_string_Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___delitem__" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str___delitem__" "', argument " "2"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::difference_type >(val2);
  try {
    std_vector_Sl_std_string_Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::difference_type arg2 ;
  std::vector<std::string >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___getitem__" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str___getitem__" "', argument " "2"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::difference_type >(val2);
  try {
    {
      std::vector<std::string >::value_type const &_result_ref = std_vector_Sl_std_string_Sg____getitem__((std::vector<std::string > const *)arg1,arg2);
      result = (std::vector<std::string >::value_type *) &_result_ref;
    }
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::difference_type arg2 ;
  std::vector<std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str___setitem__" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str___setitem__" "', argument " "2"" of type '" "std::vector<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::difference_type >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_str___setitem__" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str___setitem__" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  try {
    std_vector_Sl_std_string_Sg____setitem__(arg1,arg2,(std::string const &)*arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_append" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_str_append" "', argument " "2"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_append" "', argument " "2"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  std_vector_Sl_std_string_Sg__append(arg1,(std::string const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_str__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_str")) SWIG_fail;
  result = (std::vector<std::string > *)new std::vector<std::string >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_str__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = 0 ;
  std::vector<std::string > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_str",&obj0)) SWIG_fail;
  {
    std::vector<std::string,std::allocator<std::string > > *ptr = (std::vector<std::string,std::allocator<std::string > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_str" "', argument " "1"" of type '" "std::vector<std::string > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_str" "', argument " "1"" of type '" "std::vector<std::string > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<std::string > *)new std::vector<std::string >((std::vector<std::string > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_empty" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (bool)((std::vector<std::string > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_size" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_clear" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_swap" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_str_swap" "', argument " "2"" of type '" "std::vector<std::string > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_swap" "', argument " "2"" of type '" "std::vector<std::string > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<std::string > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  SwigValueWrapper<std::allocator<std::string > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_get_allocator" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<std::string >::allocator_type(static_cast< const std::vector<std::string >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_begin" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_begin" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_end" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_end" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_rbegin" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_rbegin" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_rend" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_rend" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_str_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_str__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string >::size_type arg1 ;
  std::vector<std::string > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_str",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_str" "', argument " "1"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg1 = static_cast< std::vector<std::string >::size_type >(val1);
  result = (std::vector<std::string > *)new std::vector<std::string >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_pop_back" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_resize" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str_resize" "', argument " "2"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::iterator arg2 ;
  std::vector<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_erase" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_erase" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_erase" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::iterator arg2 ;
  std::vector<std::string >::iterator arg3 ;
  std::vector<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_erase" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_erase" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_erase" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_erase" "', argument " "3"" of type '" "std::vector<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_erase" "', argument " "3"" of type '" "std::vector<std::string >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_str_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_str_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<std::string >::iterator)\n    erase(std::vector<std::string >::iterator,std::vector<std::string >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_str__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string >::size_type arg1 ;
  std::vector<std::string >::value_type *arg2 = 0 ;
  std::vector<std::string > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_str",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_str" "', argument " "1"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg1 = static_cast< std::vector<std::string >::size_type >(val1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_str" "', argument " "2"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_str" "', argument " "2"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  result = (std::vector<std::string > *)new std::vector<std::string >(arg1,(std::vector<std::string >::value_type const &)*arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_str(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_str__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_str__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_str__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_str__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_str'.\n  Possible C/C++ prototypes are:\n    std::vector<(std::string)>()\n    std::vector<(std::string)>(std::vector<std::string > const &)\n    std::vector<(std::string)>(std::vector<std::string >::size_type)\n    std::vector<(std::string)>(std::vector<std::string >::size_type,std::vector<std::string >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_push_back" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_str_push_back" "', argument " "2"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_push_back" "', argument " "2"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  (arg1)->push_back((std::vector<std::string >::value_type const &)*arg2);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_front" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  {
    std::vector<std::string >::value_type const &_result_ref = ((std::vector<std::string > const *)arg1)->front();
    result = (std::vector<std::string >::value_type *) &_result_ref;
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_back" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  {
    std::vector<std::string >::value_type const &_result_ref = ((std::vector<std::string > const *)arg1)->back();
    result = (std::vector<std::string >::value_type *) &_result_ref;
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type arg2 ;
  std::vector<std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_assign" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str_assign" "', argument " "2"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::size_type >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_str_assign" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_assign" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->assign(arg2,(std::vector<std::string >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type arg2 ;
  std::vector<std::string >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_resize" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str_resize" "', argument " "2"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::size_type >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_str_resize" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_resize" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  (arg1)->resize(arg2,(std::vector<std::string >::value_type const &)*arg3);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_str_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        int res = SWIG_AsPtr_std_string(argv[2], (std::string**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_str_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<std::string >::size_type)\n    resize(std::vector<std::string >::size_type,std::vector<std::string >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::iterator arg2 ;
  std::vector<std::string >::value_type *arg3 = 0 ;
  std::vector<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_str_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_insert" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_insert" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_insert" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
    }
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_str_insert" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_insert" "', argument " "3"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg3 = ptr;
  }
  result = (arg1)->insert(arg2,(std::vector<std::string >::value_type const &)*arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::iterator arg2 ;
  std::vector<std::string >::size_type arg3 ;
  std::vector<std::string >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_str_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_insert" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_insert" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_str_insert" "', argument " "2"" of type '" "std::vector<std::string >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_str_insert" "', argument " "3"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg3 = static_cast< std::vector<std::string >::size_type >(val3);
  {
    std::string *ptr = (std::string *)0;
    res4 = SWIG_AsPtr_std_string(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_str_insert" "', argument " "4"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_str_insert" "', argument " "4"" of type '" "std::vector<std::string >::value_type const &""'"); 
    }
    arg4 = ptr;
  }
  (arg1)->insert(arg2,arg3,(std::vector<std::string >::value_type const &)*arg4);
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter) != 0));
      if (_v) {
        int res = SWIG_AsPtr_std_string(argv[2], (std::string**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_str_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<std::string,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<std::string >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          int res = SWIG_AsPtr_std_string(argv[3], (std::string**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_str_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_str_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<std::string >::iterator,std::vector<std::string >::value_type const &)\n    insert(std::vector<std::string >::iterator,std::vector<std::string >::size_type,std::vector<std::string >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_str_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_reserve" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_str_reserve" "', argument " "2"" of type '" "std::vector<std::string >::size_type""'");
  } 
  arg2 = static_cast< std::vector<std::string >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_str_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  std::vector<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_str_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_str_capacity" "', argument " "1"" of type '" "std::vector<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  result = ((std::vector<std::string > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_str(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<std::string > *arg1 = (std::vector<std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_str",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_str" "', argument " "1"" of type '" "std::vector<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<std::string > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_str_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTstd__string_std__allocatorTstd__string_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_spc_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_iterator" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_steps_model_Spec_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___nonzero__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (bool)std_vector_Sl_steps_model_Spec_Sm__Sg____nonzero__((std::vector<steps::model::Spec * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___len__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = std_vector_Sl_steps_model_Spec_Sm__Sg____len__((std::vector<steps::model::Spec * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_pop" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  try {
    result = (std::vector<steps::model::Spec * >::value_type)std_vector_Sl_steps_model_Spec_Sm__Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::difference_type arg2 ;
  std::vector<steps::model::Spec * >::difference_type arg3 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___getslice__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc___getslice__" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_spc___getslice__" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val3);
  try {
    result = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *)std_vector_Sl_steps_model_Spec_Sm__Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::difference_type arg2 ;
  std::vector<steps::model::Spec * >::difference_type arg3 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_spc___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___setslice__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc___setslice__" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_spc___setslice__" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val3);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_spc___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_spc___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_steps_model_Spec_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::difference_type arg2 ;
  std::vector<steps::model::Spec * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___delslice__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc___delslice__" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_spc___delslice__" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val3);
  try {
    std_vector_Sl_steps_model_Spec_Sm__Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___delitem__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc___delitem__" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val2);
  try {
    std_vector_Sl_steps_model_Spec_Sm__Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::difference_type arg2 ;
  std::vector<steps::model::Spec * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___getitem__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc___getitem__" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val2);
  try {
    result = (std::vector<steps::model::Spec * >::value_type)std_vector_Sl_steps_model_Spec_Sm__Sg____getitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::difference_type arg2 ;
  std::vector<steps::model::Spec * >::value_type arg3 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc___setitem__" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc___setitem__" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_spc___setitem__" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp3);
  try {
    std_vector_Sl_steps_model_Spec_Sm__Sg____setitem__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::value_type arg2 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_append" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_spc_append" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp2);
  std_vector_Sl_steps_model_Spec_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_spc__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_spc")) SWIG_fail;
  result = (std::vector<steps::model::Spec * > *)new std::vector<steps::model::Spec * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_spc__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = 0 ;
  std::vector<steps::model::Spec * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_spc",&obj0)) SWIG_fail;
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_spc" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_spc" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<steps::model::Spec * > *)new std::vector<steps::model::Spec * >((std::vector<steps::model::Spec * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_empty" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (bool)((std::vector<steps::model::Spec * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_size" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_clear" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_swap" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_spc_swap" "', argument " "2"" of type '" "std::vector<steps::model::Spec * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_spc_swap" "', argument " "2"" of type '" "std::vector<steps::model::Spec * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_get_allocator" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<steps::model::Spec * >::allocator_type(static_cast< const std::vector<steps::model::Spec * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_begin" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_begin" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_end" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_end" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_rend" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_rend" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_spc_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_spc__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * >::size_type arg1 ;
  std::vector<steps::model::Spec * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_spc",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_spc" "', argument " "1"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Spec * >::size_type >(val1);
  result = (std::vector<steps::model::Spec * > *)new std::vector<steps::model::Spec * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_pop_back" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_resize" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc_resize" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::iterator arg2 ;
  std::vector<steps::model::Spec * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_erase" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_erase" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_erase" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::iterator arg2 ;
  std::vector<steps::model::Spec * >::iterator arg3 ;
  std::vector<steps::model::Spec * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_erase" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_erase" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_erase" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_erase" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_erase" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_spc_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_spc_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::model::Spec * >::iterator)\n    erase(std::vector<steps::model::Spec * >::iterator,std::vector<steps::model::Spec * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_spc__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * >::size_type arg1 ;
  std::vector<steps::model::Spec * >::value_type arg2 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  std::vector<steps::model::Spec * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_spc",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_spc" "', argument " "1"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Spec * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_spc" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp2);
  result = (std::vector<steps::model::Spec * > *)new std::vector<steps::model::Spec * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_spc(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_spc__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_spc__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_spc__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_spc__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_spc'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::model::Spec)>()\n    std::vector<(p.steps::model::Spec)>(std::vector<steps::model::Spec * > const &)\n    std::vector<(p.steps::model::Spec)>(std::vector<steps::model::Spec * >::size_type)\n    std::vector<(p.steps::model::Spec)>(std::vector<steps::model::Spec * >::size_type,std::vector<steps::model::Spec * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::value_type arg2 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_push_back" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_spc_push_back" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_front" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (std::vector<steps::model::Spec * >::value_type)((std::vector<steps::model::Spec * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_back" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = (std::vector<steps::model::Spec * >::value_type)((std::vector<steps::model::Spec * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type arg2 ;
  std::vector<steps::model::Spec * >::value_type arg3 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_assign" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc_assign" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_spc_assign" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type arg2 ;
  std::vector<steps::model::Spec * >::value_type arg3 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_resize" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc_resize" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_spc_resize" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_spc_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_spc_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::model::Spec * >::size_type)\n    resize(std::vector<steps::model::Spec * >::size_type,std::vector<steps::model::Spec * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::iterator arg2 ;
  std::vector<steps::model::Spec * >::value_type arg3 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  std::vector<steps::model::Spec * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_spc_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_insert" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_insert" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_insert" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_spc_insert" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp3);
  result = (arg1)->insert(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Spec * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::iterator arg2 ;
  std::vector<steps::model::Spec * >::size_type arg3 ;
  std::vector<steps::model::Spec * >::value_type arg4 = (std::vector<steps::model::Spec * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_spc_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_insert" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_insert" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_spc_insert" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_spc_insert" "', argument " "3"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Spec * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_spc_insert" "', argument " "4"" of type '" "std::vector<steps::model::Spec * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::model::Spec * >::value_type >(argp4);
  (arg1)->insert(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_spc_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Spec * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_spc_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_spc_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::model::Spec * >::iterator,std::vector<steps::model::Spec * >::value_type)\n    insert(std::vector<steps::model::Spec * >::iterator,std::vector<steps::model::Spec * >::size_type,std::vector<steps::model::Spec * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_spc_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_reserve" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_spc_reserve" "', argument " "2"" of type '" "std::vector<steps::model::Spec * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Spec * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_spc_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  std::vector<steps::model::Spec * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_spc_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_spc_capacity" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  result = ((std::vector<steps::model::Spec * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_spc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Spec * > *arg1 = (std::vector<steps::model::Spec * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_spc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_spc" "', argument " "1"" of type '" "std::vector<steps::model::Spec * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Spec * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_spc_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_rec_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_iterator" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_steps_model_Reac_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___nonzero__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (bool)std_vector_Sl_steps_model_Reac_Sm__Sg____nonzero__((std::vector<steps::model::Reac * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___len__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = std_vector_Sl_steps_model_Reac_Sm__Sg____len__((std::vector<steps::model::Reac * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_pop" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  try {
    result = (std::vector<steps::model::Reac * >::value_type)std_vector_Sl_steps_model_Reac_Sm__Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::difference_type arg2 ;
  std::vector<steps::model::Reac * >::difference_type arg3 ;
  std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___getslice__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec___getslice__" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_rec___getslice__" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val3);
  try {
    result = (std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > *)std_vector_Sl_steps_model_Reac_Sm__Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::difference_type arg2 ;
  std::vector<steps::model::Reac * >::difference_type arg3 ;
  std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_rec___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___setslice__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec___setslice__" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_rec___setslice__" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val3);
  {
    std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > > *ptr = (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_rec___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_rec___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_steps_model_Reac_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::difference_type arg2 ;
  std::vector<steps::model::Reac * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___delslice__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec___delslice__" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_rec___delslice__" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val3);
  try {
    std_vector_Sl_steps_model_Reac_Sm__Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___delitem__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec___delitem__" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val2);
  try {
    std_vector_Sl_steps_model_Reac_Sm__Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::difference_type arg2 ;
  std::vector<steps::model::Reac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___getitem__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec___getitem__" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val2);
  try {
    result = (std::vector<steps::model::Reac * >::value_type)std_vector_Sl_steps_model_Reac_Sm__Sg____getitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::difference_type arg2 ;
  std::vector<steps::model::Reac * >::value_type arg3 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec___setitem__" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec___setitem__" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_rec___setitem__" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp3);
  try {
    std_vector_Sl_steps_model_Reac_Sm__Sg____setitem__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::value_type arg2 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_append" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_rec_append" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp2);
  std_vector_Sl_steps_model_Reac_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_rec__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_rec")) SWIG_fail;
  result = (std::vector<steps::model::Reac * > *)new std::vector<steps::model::Reac * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_rec__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = 0 ;
  std::vector<steps::model::Reac * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_rec",&obj0)) SWIG_fail;
  {
    std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > > *ptr = (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_rec" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_rec" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<steps::model::Reac * > *)new std::vector<steps::model::Reac * >((std::vector<steps::model::Reac * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_empty" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (bool)((std::vector<steps::model::Reac * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_size" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_clear" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_swap" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_rec_swap" "', argument " "2"" of type '" "std::vector<steps::model::Reac * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_rec_swap" "', argument " "2"" of type '" "std::vector<steps::model::Reac * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::model::Reac * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_get_allocator" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<steps::model::Reac * >::allocator_type(static_cast< const std::vector<steps::model::Reac * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_begin" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_begin" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_end" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_end" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_rend" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_rend" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_rec_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_rec__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * >::size_type arg1 ;
  std::vector<steps::model::Reac * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_rec",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_rec" "', argument " "1"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Reac * >::size_type >(val1);
  result = (std::vector<steps::model::Reac * > *)new std::vector<steps::model::Reac * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_pop_back" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_resize" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec_resize" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::iterator arg2 ;
  std::vector<steps::model::Reac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_erase" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_erase" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_erase" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::iterator arg2 ;
  std::vector<steps::model::Reac * >::iterator arg3 ;
  std::vector<steps::model::Reac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_erase" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_erase" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_erase" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_erase" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_erase" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_rec_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_rec_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::model::Reac * >::iterator)\n    erase(std::vector<steps::model::Reac * >::iterator,std::vector<steps::model::Reac * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_rec__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * >::size_type arg1 ;
  std::vector<steps::model::Reac * >::value_type arg2 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  std::vector<steps::model::Reac * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_rec",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_rec" "', argument " "1"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Reac * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_rec" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp2);
  result = (std::vector<steps::model::Reac * > *)new std::vector<steps::model::Reac * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_rec(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_rec__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_rec__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_rec__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_rec__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_rec'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::model::Reac)>()\n    std::vector<(p.steps::model::Reac)>(std::vector<steps::model::Reac * > const &)\n    std::vector<(p.steps::model::Reac)>(std::vector<steps::model::Reac * >::size_type)\n    std::vector<(p.steps::model::Reac)>(std::vector<steps::model::Reac * >::size_type,std::vector<steps::model::Reac * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::value_type arg2 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_push_back" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_rec_push_back" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_front" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (std::vector<steps::model::Reac * >::value_type)((std::vector<steps::model::Reac * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_back" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = (std::vector<steps::model::Reac * >::value_type)((std::vector<steps::model::Reac * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type arg2 ;
  std::vector<steps::model::Reac * >::value_type arg3 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_assign" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec_assign" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_rec_assign" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type arg2 ;
  std::vector<steps::model::Reac * >::value_type arg3 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_resize" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec_resize" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_rec_resize" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_rec_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_rec_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::model::Reac * >::size_type)\n    resize(std::vector<steps::model::Reac * >::size_type,std::vector<steps::model::Reac * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::iterator arg2 ;
  std::vector<steps::model::Reac * >::value_type arg3 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  std::vector<steps::model::Reac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_rec_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_insert" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_insert" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_insert" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_rec_insert" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp3);
  result = (arg1)->insert(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Reac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::iterator arg2 ;
  std::vector<steps::model::Reac * >::size_type arg3 ;
  std::vector<steps::model::Reac * >::value_type arg4 = (std::vector<steps::model::Reac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_rec_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_insert" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_insert" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_rec_insert" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_rec_insert" "', argument " "3"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Reac * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_rec_insert" "', argument " "4"" of type '" "std::vector<steps::model::Reac * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::model::Reac * >::value_type >(argp4);
  (arg1)->insert(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_rec_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Reac * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_rec_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_rec_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::model::Reac * >::iterator,std::vector<steps::model::Reac * >::value_type)\n    insert(std::vector<steps::model::Reac * >::iterator,std::vector<steps::model::Reac * >::size_type,std::vector<steps::model::Reac * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_rec_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_reserve" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_rec_reserve" "', argument " "2"" of type '" "std::vector<steps::model::Reac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Reac * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_rec_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  std::vector<steps::model::Reac * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_rec_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_rec_capacity" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  result = ((std::vector<steps::model::Reac * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_rec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Reac * > *arg1 = (std::vector<steps::model::Reac * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_rec",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_rec" "', argument " "1"" of type '" "std::vector<steps::model::Reac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Reac * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_rec_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_src_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_iterator" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_steps_model_SReac_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___nonzero__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (bool)std_vector_Sl_steps_model_SReac_Sm__Sg____nonzero__((std::vector<steps::model::SReac * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___len__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = std_vector_Sl_steps_model_SReac_Sm__Sg____len__((std::vector<steps::model::SReac * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_pop" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  try {
    result = (std::vector<steps::model::SReac * >::value_type)std_vector_Sl_steps_model_SReac_Sm__Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::difference_type arg2 ;
  std::vector<steps::model::SReac * >::difference_type arg3 ;
  std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___getslice__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src___getslice__" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_src___getslice__" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val3);
  try {
    result = (std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > *)std_vector_Sl_steps_model_SReac_Sm__Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::difference_type arg2 ;
  std::vector<steps::model::SReac * >::difference_type arg3 ;
  std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_src___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___setslice__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src___setslice__" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_src___setslice__" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val3);
  {
    std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > > *ptr = (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_src___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_src___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_steps_model_SReac_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::difference_type arg2 ;
  std::vector<steps::model::SReac * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___delslice__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src___delslice__" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_src___delslice__" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val3);
  try {
    std_vector_Sl_steps_model_SReac_Sm__Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___delitem__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src___delitem__" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val2);
  try {
    std_vector_Sl_steps_model_SReac_Sm__Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::difference_type arg2 ;
  std::vector<steps::model::SReac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___getitem__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src___getitem__" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val2);
  try {
    result = (std::vector<steps::model::SReac * >::value_type)std_vector_Sl_steps_model_SReac_Sm__Sg____getitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::difference_type arg2 ;
  std::vector<steps::model::SReac * >::value_type arg3 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src___setitem__" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src___setitem__" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_src___setitem__" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp3);
  try {
    std_vector_Sl_steps_model_SReac_Sm__Sg____setitem__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::value_type arg2 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_append" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_src_append" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp2);
  std_vector_Sl_steps_model_SReac_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_src__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_src")) SWIG_fail;
  result = (std::vector<steps::model::SReac * > *)new std::vector<steps::model::SReac * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_src__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = 0 ;
  std::vector<steps::model::SReac * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_src",&obj0)) SWIG_fail;
  {
    std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > > *ptr = (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_src" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_src" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<steps::model::SReac * > *)new std::vector<steps::model::SReac * >((std::vector<steps::model::SReac * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_empty" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (bool)((std::vector<steps::model::SReac * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_size" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_clear" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_swap" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_src_swap" "', argument " "2"" of type '" "std::vector<steps::model::SReac * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_src_swap" "', argument " "2"" of type '" "std::vector<steps::model::SReac * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::model::SReac * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_get_allocator" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<steps::model::SReac * >::allocator_type(static_cast< const std::vector<steps::model::SReac * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_begin" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_begin" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_end" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_end" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_rend" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_rend" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_src_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_src__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * >::size_type arg1 ;
  std::vector<steps::model::SReac * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_src",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_src" "', argument " "1"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::SReac * >::size_type >(val1);
  result = (std::vector<steps::model::SReac * > *)new std::vector<steps::model::SReac * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_pop_back" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_resize" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src_resize" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::iterator arg2 ;
  std::vector<steps::model::SReac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_erase" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_erase" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_erase" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::iterator arg2 ;
  std::vector<steps::model::SReac * >::iterator arg3 ;
  std::vector<steps::model::SReac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_erase" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_erase" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_erase" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_erase" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_erase" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_src_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_src_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::model::SReac * >::iterator)\n    erase(std::vector<steps::model::SReac * >::iterator,std::vector<steps::model::SReac * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_src__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * >::size_type arg1 ;
  std::vector<steps::model::SReac * >::value_type arg2 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  std::vector<steps::model::SReac * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_src",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_src" "', argument " "1"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::SReac * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_src" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp2);
  result = (std::vector<steps::model::SReac * > *)new std::vector<steps::model::SReac * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_src(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_src__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_src__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_src__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_src__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_src'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::model::SReac)>()\n    std::vector<(p.steps::model::SReac)>(std::vector<steps::model::SReac * > const &)\n    std::vector<(p.steps::model::SReac)>(std::vector<steps::model::SReac * >::size_type)\n    std::vector<(p.steps::model::SReac)>(std::vector<steps::model::SReac * >::size_type,std::vector<steps::model::SReac * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::value_type arg2 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_push_back" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_src_push_back" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_front" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (std::vector<steps::model::SReac * >::value_type)((std::vector<steps::model::SReac * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_back" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = (std::vector<steps::model::SReac * >::value_type)((std::vector<steps::model::SReac * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type arg2 ;
  std::vector<steps::model::SReac * >::value_type arg3 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_assign" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src_assign" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_src_assign" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type arg2 ;
  std::vector<steps::model::SReac * >::value_type arg3 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_resize" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src_resize" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_src_resize" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_src_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_src_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::model::SReac * >::size_type)\n    resize(std::vector<steps::model::SReac * >::size_type,std::vector<steps::model::SReac * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::iterator arg2 ;
  std::vector<steps::model::SReac * >::value_type arg3 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  std::vector<steps::model::SReac * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_src_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_insert" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_insert" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_insert" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_src_insert" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp3);
  result = (arg1)->insert(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::SReac * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::iterator arg2 ;
  std::vector<steps::model::SReac * >::size_type arg3 ;
  std::vector<steps::model::SReac * >::value_type arg4 = (std::vector<steps::model::SReac * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_src_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_insert" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_insert" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_src_insert" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_src_insert" "', argument " "3"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::SReac * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_src_insert" "', argument " "4"" of type '" "std::vector<steps::model::SReac * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::model::SReac * >::value_type >(argp4);
  (arg1)->insert(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_src_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::SReac * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_src_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_src_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::model::SReac * >::iterator,std::vector<steps::model::SReac * >::value_type)\n    insert(std::vector<steps::model::SReac * >::iterator,std::vector<steps::model::SReac * >::size_type,std::vector<steps::model::SReac * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_src_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_reserve" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_src_reserve" "', argument " "2"" of type '" "std::vector<steps::model::SReac * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::SReac * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_src_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  std::vector<steps::model::SReac * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_src_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_src_capacity" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  result = ((std::vector<steps::model::SReac * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_src(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::SReac * > *arg1 = (std::vector<steps::model::SReac * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_src",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_src" "', argument " "1"" of type '" "std::vector<steps::model::SReac * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::SReac * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_src_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_dif_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_iterator" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_steps_model_Diff_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___nonzero__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (bool)std_vector_Sl_steps_model_Diff_Sm__Sg____nonzero__((std::vector<steps::model::Diff * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___len__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = std_vector_Sl_steps_model_Diff_Sm__Sg____len__((std::vector<steps::model::Diff * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_pop" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  try {
    result = (std::vector<steps::model::Diff * >::value_type)std_vector_Sl_steps_model_Diff_Sm__Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::difference_type arg2 ;
  std::vector<steps::model::Diff * >::difference_type arg3 ;
  std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___getslice__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif___getslice__" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dif___getslice__" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val3);
  try {
    result = (std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > *)std_vector_Sl_steps_model_Diff_Sm__Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::difference_type arg2 ;
  std::vector<steps::model::Diff * >::difference_type arg3 ;
  std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_dif___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___setslice__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif___setslice__" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dif___setslice__" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val3);
  {
    std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > > *ptr = (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_dif___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_dif___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_steps_model_Diff_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::difference_type arg2 ;
  std::vector<steps::model::Diff * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___delslice__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif___delslice__" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dif___delslice__" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val3);
  try {
    std_vector_Sl_steps_model_Diff_Sm__Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___delitem__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif___delitem__" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val2);
  try {
    std_vector_Sl_steps_model_Diff_Sm__Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::difference_type arg2 ;
  std::vector<steps::model::Diff * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___getitem__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif___getitem__" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val2);
  try {
    result = (std::vector<steps::model::Diff * >::value_type)std_vector_Sl_steps_model_Diff_Sm__Sg____getitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::difference_type arg2 ;
  std::vector<steps::model::Diff * >::value_type arg3 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif___setitem__" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif___setitem__" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_dif___setitem__" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp3);
  try {
    std_vector_Sl_steps_model_Diff_Sm__Sg____setitem__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::value_type arg2 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_append" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_dif_append" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp2);
  std_vector_Sl_steps_model_Diff_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dif__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_dif")) SWIG_fail;
  result = (std::vector<steps::model::Diff * > *)new std::vector<steps::model::Diff * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dif__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = 0 ;
  std::vector<steps::model::Diff * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_dif",&obj0)) SWIG_fail;
  {
    std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > > *ptr = (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_dif" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_dif" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<steps::model::Diff * > *)new std::vector<steps::model::Diff * >((std::vector<steps::model::Diff * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_empty" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (bool)((std::vector<steps::model::Diff * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_size" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_clear" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_swap" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_dif_swap" "', argument " "2"" of type '" "std::vector<steps::model::Diff * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_dif_swap" "', argument " "2"" of type '" "std::vector<steps::model::Diff * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::model::Diff * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_get_allocator" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<steps::model::Diff * >::allocator_type(static_cast< const std::vector<steps::model::Diff * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_begin" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_begin" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_end" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_end" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_rend" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_rend" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dif_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dif__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * >::size_type arg1 ;
  std::vector<steps::model::Diff * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_dif",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_dif" "', argument " "1"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Diff * >::size_type >(val1);
  result = (std::vector<steps::model::Diff * > *)new std::vector<steps::model::Diff * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_pop_back" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_resize" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif_resize" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::iterator arg2 ;
  std::vector<steps::model::Diff * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_erase" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_erase" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_erase" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::iterator arg2 ;
  std::vector<steps::model::Diff * >::iterator arg3 ;
  std::vector<steps::model::Diff * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_erase" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_erase" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_erase" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_erase" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_erase" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_dif_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_dif_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::model::Diff * >::iterator)\n    erase(std::vector<steps::model::Diff * >::iterator,std::vector<steps::model::Diff * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dif__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * >::size_type arg1 ;
  std::vector<steps::model::Diff * >::value_type arg2 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  std::vector<steps::model::Diff * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_dif",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_dif" "', argument " "1"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Diff * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_dif" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp2);
  result = (std::vector<steps::model::Diff * > *)new std::vector<steps::model::Diff * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dif(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_dif__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_dif__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_dif__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_dif__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_dif'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::model::Diff)>()\n    std::vector<(p.steps::model::Diff)>(std::vector<steps::model::Diff * > const &)\n    std::vector<(p.steps::model::Diff)>(std::vector<steps::model::Diff * >::size_type)\n    std::vector<(p.steps::model::Diff)>(std::vector<steps::model::Diff * >::size_type,std::vector<steps::model::Diff * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::value_type arg2 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_push_back" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_dif_push_back" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_front" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (std::vector<steps::model::Diff * >::value_type)((std::vector<steps::model::Diff * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_back" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = (std::vector<steps::model::Diff * >::value_type)((std::vector<steps::model::Diff * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type arg2 ;
  std::vector<steps::model::Diff * >::value_type arg3 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_assign" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif_assign" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_dif_assign" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type arg2 ;
  std::vector<steps::model::Diff * >::value_type arg3 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_resize" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif_resize" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_dif_resize" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_dif_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_dif_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::model::Diff * >::size_type)\n    resize(std::vector<steps::model::Diff * >::size_type,std::vector<steps::model::Diff * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::iterator arg2 ;
  std::vector<steps::model::Diff * >::value_type arg3 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  std::vector<steps::model::Diff * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dif_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_insert" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_insert" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_insert" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_dif_insert" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp3);
  result = (arg1)->insert(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Diff * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::iterator arg2 ;
  std::vector<steps::model::Diff * >::size_type arg3 ;
  std::vector<steps::model::Diff * >::value_type arg4 = (std::vector<steps::model::Diff * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_dif_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_insert" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_insert" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dif_insert" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dif_insert" "', argument " "3"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Diff * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_dif_insert" "', argument " "4"" of type '" "std::vector<steps::model::Diff * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::model::Diff * >::value_type >(argp4);
  (arg1)->insert(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_dif_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Diff * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_dif_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dif_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::model::Diff * >::iterator,std::vector<steps::model::Diff * >::value_type)\n    insert(std::vector<steps::model::Diff * >::iterator,std::vector<steps::model::Diff * >::size_type,std::vector<steps::model::Diff * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dif_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_reserve" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dif_reserve" "', argument " "2"" of type '" "std::vector<steps::model::Diff * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Diff * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dif_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  std::vector<steps::model::Diff * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dif_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dif_capacity" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  result = ((std::vector<steps::model::Diff * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_dif(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Diff * > *arg1 = (std::vector<steps::model::Diff * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_dif",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_dif" "', argument " "1"" of type '" "std::vector<steps::model::Diff * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Diff * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_dif_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_vsys_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_iterator" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_steps_model_Volsys_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___nonzero__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (bool)std_vector_Sl_steps_model_Volsys_Sm__Sg____nonzero__((std::vector<steps::model::Volsys * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___len__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = std_vector_Sl_steps_model_Volsys_Sm__Sg____len__((std::vector<steps::model::Volsys * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_pop" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  try {
    result = (std::vector<steps::model::Volsys * >::value_type)std_vector_Sl_steps_model_Volsys_Sm__Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::difference_type arg2 ;
  std::vector<steps::model::Volsys * >::difference_type arg3 ;
  std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___getslice__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys___getslice__" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_vsys___getslice__" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val3);
  try {
    result = (std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > *)std_vector_Sl_steps_model_Volsys_Sm__Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::difference_type arg2 ;
  std::vector<steps::model::Volsys * >::difference_type arg3 ;
  std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_vsys___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___setslice__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys___setslice__" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_vsys___setslice__" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val3);
  {
    std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > > *ptr = (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_vsys___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_vsys___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_steps_model_Volsys_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::difference_type arg2 ;
  std::vector<steps::model::Volsys * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___delslice__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys___delslice__" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_vsys___delslice__" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val3);
  try {
    std_vector_Sl_steps_model_Volsys_Sm__Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___delitem__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys___delitem__" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val2);
  try {
    std_vector_Sl_steps_model_Volsys_Sm__Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::difference_type arg2 ;
  std::vector<steps::model::Volsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___getitem__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys___getitem__" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val2);
  try {
    result = (std::vector<steps::model::Volsys * >::value_type)std_vector_Sl_steps_model_Volsys_Sm__Sg____getitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::difference_type arg2 ;
  std::vector<steps::model::Volsys * >::value_type arg3 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys___setitem__" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys___setitem__" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_vsys___setitem__" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp3);
  try {
    std_vector_Sl_steps_model_Volsys_Sm__Sg____setitem__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::value_type arg2 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_append" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_vsys_append" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp2);
  std_vector_Sl_steps_model_Volsys_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_vsys__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_vsys")) SWIG_fail;
  result = (std::vector<steps::model::Volsys * > *)new std::vector<steps::model::Volsys * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_vsys__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = 0 ;
  std::vector<steps::model::Volsys * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_vsys",&obj0)) SWIG_fail;
  {
    std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > > *ptr = (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_vsys" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_vsys" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<steps::model::Volsys * > *)new std::vector<steps::model::Volsys * >((std::vector<steps::model::Volsys * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_empty" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (bool)((std::vector<steps::model::Volsys * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_size" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_clear" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_swap" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_vsys_swap" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_vsys_swap" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::model::Volsys * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_get_allocator" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<steps::model::Volsys * >::allocator_type(static_cast< const std::vector<steps::model::Volsys * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_begin" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_begin" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_end" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_end" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_rend" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_rend" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_vsys_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_vsys__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * >::size_type arg1 ;
  std::vector<steps::model::Volsys * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_vsys",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_vsys" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val1);
  result = (std::vector<steps::model::Volsys * > *)new std::vector<steps::model::Volsys * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_pop_back" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_resize" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys_resize" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::iterator arg2 ;
  std::vector<steps::model::Volsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_erase" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::iterator arg2 ;
  std::vector<steps::model::Volsys * >::iterator arg3 ;
  std::vector<steps::model::Volsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_erase" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_erase" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_erase" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_vsys_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_vsys_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::model::Volsys * >::iterator)\n    erase(std::vector<steps::model::Volsys * >::iterator,std::vector<steps::model::Volsys * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_vsys__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * >::size_type arg1 ;
  std::vector<steps::model::Volsys * >::value_type arg2 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  std::vector<steps::model::Volsys * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_vsys",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_vsys" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_vsys" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp2);
  result = (std::vector<steps::model::Volsys * > *)new std::vector<steps::model::Volsys * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_vsys(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_vsys__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_vsys__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_vsys__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_vsys__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_vsys'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::model::Volsys)>()\n    std::vector<(p.steps::model::Volsys)>(std::vector<steps::model::Volsys * > const &)\n    std::vector<(p.steps::model::Volsys)>(std::vector<steps::model::Volsys * >::size_type)\n    std::vector<(p.steps::model::Volsys)>(std::vector<steps::model::Volsys * >::size_type,std::vector<steps::model::Volsys * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::value_type arg2 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_push_back" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_vsys_push_back" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_front" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (std::vector<steps::model::Volsys * >::value_type)((std::vector<steps::model::Volsys * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_back" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = (std::vector<steps::model::Volsys * >::value_type)((std::vector<steps::model::Volsys * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type arg2 ;
  std::vector<steps::model::Volsys * >::value_type arg3 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_assign" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys_assign" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_vsys_assign" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type arg2 ;
  std::vector<steps::model::Volsys * >::value_type arg3 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_resize" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys_resize" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_vsys_resize" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_vsys_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_vsys_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::model::Volsys * >::size_type)\n    resize(std::vector<steps::model::Volsys * >::size_type,std::vector<steps::model::Volsys * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::iterator arg2 ;
  std::vector<steps::model::Volsys * >::value_type arg3 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  std::vector<steps::model::Volsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_vsys_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_insert" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_vsys_insert" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp3);
  result = (arg1)->insert(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Volsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::iterator arg2 ;
  std::vector<steps::model::Volsys * >::size_type arg3 ;
  std::vector<steps::model::Volsys * >::value_type arg4 = (std::vector<steps::model::Volsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_vsys_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_insert" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_vsys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_vsys_insert" "', argument " "3"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_vsys_insert" "', argument " "4"" of type '" "std::vector<steps::model::Volsys * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::model::Volsys * >::value_type >(argp4);
  (arg1)->insert(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_vsys_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Volsys * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_vsys_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_vsys_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::model::Volsys * >::iterator,std::vector<steps::model::Volsys * >::value_type)\n    insert(std::vector<steps::model::Volsys * >::iterator,std::vector<steps::model::Volsys * >::size_type,std::vector<steps::model::Volsys * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_vsys_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_reserve" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_vsys_reserve" "', argument " "2"" of type '" "std::vector<steps::model::Volsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Volsys * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_vsys_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  std::vector<steps::model::Volsys * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_vsys_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_vsys_capacity" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  result = ((std::vector<steps::model::Volsys * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_vsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Volsys * > *arg1 = (std::vector<steps::model::Volsys * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_vsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_vsys" "', argument " "1"" of type '" "std::vector<steps::model::Volsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Volsys * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_vsys_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_ssys_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_iterator" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (swig::PySwigIterator *)std_vector_Sl_steps_model_Surfsys_Sm__Sg__iterator(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___nonzero__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (bool)std_vector_Sl_steps_model_Surfsys_Sm__Sg____nonzero__((std::vector<steps::model::Surfsys * > const *)arg1);
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___len__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = std_vector_Sl_steps_model_Surfsys_Sm__Sg____len__((std::vector<steps::model::Surfsys * > const *)arg1);
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_pop" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  try {
    result = (std::vector<steps::model::Surfsys * >::value_type)std_vector_Sl_steps_model_Surfsys_Sm__Sg__pop(arg1);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::difference_type arg2 ;
  std::vector<steps::model::Surfsys * >::difference_type arg3 ;
  std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___getslice__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys___getslice__" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ssys___getslice__" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val3);
  try {
    result = (std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > *)std_vector_Sl_steps_model_Surfsys_Sm__Sg____getslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::difference_type arg2 ;
  std::vector<steps::model::Surfsys * >::difference_type arg3 ;
  std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_ssys___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___setslice__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys___setslice__" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ssys___setslice__" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val3);
  {
    std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > > *ptr = (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_ssys___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_ssys___setslice__" "', argument " "4"" of type '" "std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > const &""'"); 
    }
    arg4 = ptr;
  }
  try {
    std_vector_Sl_steps_model_Surfsys_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > const &)*arg4);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  catch(std::invalid_argument &_e) {
    SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::difference_type arg2 ;
  std::vector<steps::model::Surfsys * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___delslice__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys___delslice__" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ssys___delslice__" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val3);
  try {
    std_vector_Sl_steps_model_Surfsys_Sm__Sg____delslice__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___delitem__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys___delitem__" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val2);
  try {
    std_vector_Sl_steps_model_Surfsys_Sm__Sg____delitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::difference_type arg2 ;
  std::vector<steps::model::Surfsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___getitem__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys___getitem__" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val2);
  try {
    result = (std::vector<steps::model::Surfsys * >::value_type)std_vector_Sl_steps_model_Surfsys_Sm__Sg____getitem__(arg1,arg2);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::difference_type arg2 ;
  std::vector<steps::model::Surfsys * >::value_type arg3 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys___setitem__" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys___setitem__" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ssys___setitem__" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp3);
  try {
    std_vector_Sl_steps_model_Surfsys_Sm__Sg____setitem__(arg1,arg2,arg3);
  }
  catch(std::out_of_range &_e) {
    SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
  }
  
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::value_type arg2 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_append" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_ssys_append" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp2);
  std_vector_Sl_steps_model_Surfsys_Sm__Sg__append(arg1,arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ssys__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_ssys")) SWIG_fail;
  result = (std::vector<steps::model::Surfsys * > *)new std::vector<steps::model::Surfsys * >();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ssys__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = 0 ;
  std::vector<steps::model::Surfsys * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_ssys",&obj0)) SWIG_fail;
  {
    std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > > *ptr = (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_ssys" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_ssys" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const &""'"); 
    }
    arg1 = ptr;
  }
  result = (std::vector<steps::model::Surfsys * > *)new std::vector<steps::model::Surfsys * >((std::vector<steps::model::Surfsys * > const &)*arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_empty" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (bool)((std::vector<steps::model::Surfsys * > const *)arg1)->empty();
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_size" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->size();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_clear" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  (arg1)->clear();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_swap" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_ssys_swap" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_ssys_swap" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp2);
  (arg1)->swap(*arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::model::Surfsys * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_get_allocator" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->get_allocator();
  resultobj = SWIG_NewPointerObj((new std::vector<steps::model::Surfsys * >::allocator_type(static_cast< const std::vector<steps::model::Surfsys * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_begin" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_begin" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->begin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_end" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_end" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->end();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_rbegin" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->rbegin();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_rend" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_rend" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->rend();
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ssys_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ssys__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * >::size_type arg1 ;
  std::vector<steps::model::Surfsys * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_ssys",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_ssys" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val1);
  result = (std::vector<steps::model::Surfsys * > *)new std::vector<steps::model::Surfsys * >(arg1);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_pop_back" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  (arg1)->pop_back();
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_resize" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys_resize" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val2);
  (arg1)->resize(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::iterator arg2 ;
  std::vector<steps::model::Surfsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_erase" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::iterator arg2 ;
  std::vector<steps::model::Surfsys * >::iterator arg3 ;
  std::vector<steps::model::Surfsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_erase" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_erase" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_erase" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_erase" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
    }
  }
  result = (arg1)->erase(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_ssys_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_ssys_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::model::Surfsys * >::iterator)\n    erase(std::vector<steps::model::Surfsys * >::iterator,std::vector<steps::model::Surfsys * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ssys__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * >::size_type arg1 ;
  std::vector<steps::model::Surfsys * >::value_type arg2 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  std::vector<steps::model::Surfsys * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_ssys",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_ssys" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_ssys" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp2);
  result = (std::vector<steps::model::Surfsys * > *)new std::vector<steps::model::Surfsys * >(arg1,arg2);
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ssys(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_ssys__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_ssys__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_ssys__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_ssys__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_ssys'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::model::Surfsys)>()\n    std::vector<(p.steps::model::Surfsys)>(std::vector<steps::model::Surfsys * > const &)\n    std::vector<(p.steps::model::Surfsys)>(std::vector<steps::model::Surfsys * >::size_type)\n    std::vector<(p.steps::model::Surfsys)>(std::vector<steps::model::Surfsys * >::size_type,std::vector<steps::model::Surfsys * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::value_type arg2 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_push_back" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_ssys_push_back" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp2);
  (arg1)->push_back(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_front" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (std::vector<steps::model::Surfsys * >::value_type)((std::vector<steps::model::Surfsys * > const *)arg1)->front();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_back" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = (std::vector<steps::model::Surfsys * >::value_type)((std::vector<steps::model::Surfsys * > const *)arg1)->back();
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type arg2 ;
  std::vector<steps::model::Surfsys * >::value_type arg3 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_assign" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys_assign" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ssys_assign" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp3);
  (arg1)->assign(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type arg2 ;
  std::vector<steps::model::Surfsys * >::value_type arg3 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_resize" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys_resize" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ssys_resize" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp3);
  (arg1)->resize(arg2,arg3);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_ssys_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_ssys_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::model::Surfsys * >::size_type)\n    resize(std::vector<steps::model::Surfsys * >::size_type,std::vector<steps::model::Surfsys * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::iterator arg2 ;
  std::vector<steps::model::Surfsys * >::value_type arg3 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  std::vector<steps::model::Surfsys * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ssys_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_insert" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ssys_insert" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp3);
  result = (arg1)->insert(arg2,arg3);
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::model::Surfsys * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::iterator arg2 ;
  std::vector<steps::model::Surfsys * >::size_type arg3 ;
  std::vector<steps::model::Surfsys * >::value_type arg4 = (std::vector<steps::model::Surfsys * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_ssys_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_insert" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ssys_insert" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ssys_insert" "', argument " "3"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_ssys_insert" "', argument " "4"" of type '" "std::vector<steps::model::Surfsys * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::model::Surfsys * >::value_type >(argp4);
  (arg1)->insert(arg2,arg3,arg4);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_ssys_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::model::Surfsys * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_ssys_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ssys_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::model::Surfsys * >::iterator,std::vector<steps::model::Surfsys * >::value_type)\n    insert(std::vector<steps::model::Surfsys * >::iterator,std::vector<steps::model::Surfsys * >::size_type,std::vector<steps::model::Surfsys * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ssys_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_reserve" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ssys_reserve" "', argument " "2"" of type '" "std::vector<steps::model::Surfsys * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::model::Surfsys * >::size_type >(val2);
  (arg1)->reserve(arg2);
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ssys_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  std::vector<steps::model::Surfsys * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ssys_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ssys_capacity" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  result = ((std::vector<steps::model::Surfsys * > const *)arg1)->capacity();
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_ssys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::model::Surfsys * > *arg1 = (std::vector<steps::model::Surfsys * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_ssys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_ssys" "', argument " "1"" of type '" "std::vector<steps::model::Surfsys * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::model::Surfsys * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_ssys_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_isValidID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  bool result;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:isValidID",&obj0)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "isValidID" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "isValidID" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (bool)steps::model::isValidID((std::string const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_checkID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:checkID",&obj0)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "checkID" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "checkID" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      steps::model::checkID((std::string const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Model(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_Model")) SWIG_fail;
  {
    try {
      result = (steps::model::Model *)new steps::model::Model();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Model(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Model",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Model" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_getSpec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::string *arg2 = 0 ;
  steps::model::Spec *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Model_getSpec",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_getSpec" "', argument " "1"" of type '" "steps::model::Model const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Model_getSpec" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Model_getSpec" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::model::Spec *)((steps::model::Model const *)arg1)->getSpec((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_delSpec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Model_delSpec",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_delSpec" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Model_delSpec" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Model_delSpec" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delSpec((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_getAllSpecs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Model_getAllSpecs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_getAllSpecs" "', argument " "1"" of type '" "steps::model::Model const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    try {
      result = ((steps::model::Model const *)arg1)->getAllSpecs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_getVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::string *arg2 = 0 ;
  steps::model::Volsys *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Model_getVolsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_getVolsys" "', argument " "1"" of type '" "steps::model::Model const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Model_getVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Model_getVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::model::Volsys *)((steps::model::Model const *)arg1)->getVolsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_delVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Model_delVolsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_delVolsys" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Model_delVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Model_delVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delVolsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_getAllVolsyss(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Model_getAllVolsyss",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_getAllVolsyss" "', argument " "1"" of type '" "steps::model::Model const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    try {
      result = ((steps::model::Model const *)arg1)->getAllVolsyss();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Volsys*,std::allocator<steps::model::Volsys * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_getSurfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::string *arg2 = 0 ;
  steps::model::Surfsys *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Model_getSurfsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_getSurfsys" "', argument " "1"" of type '" "steps::model::Model const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Model_getSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Model_getSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::model::Surfsys *)((steps::model::Model const *)arg1)->getSurfsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_delSurfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Model_delSurfsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_delSurfsys" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Model_delSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Model_delSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delSurfsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Model_getAllSurfsyss(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Model_getAllSurfsyss",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Model_getAllSurfsyss" "', argument " "1"" of type '" "steps::model::Model const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  {
    try {
      result = ((steps::model::Model const *)arg1)->getAllSurfsyss();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Surfsys*,std::allocator<steps::model::Surfsys * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Model_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__Model, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Spec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::model::Model *arg2 = (steps::model::Model *) 0 ;
  steps::model::Spec *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_Spec",&obj0,&obj1)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Spec" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Spec" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Spec" "', argument " "2"" of type '" "steps::model::Model *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Model * >(argp2);
  {
    try {
      result = (steps::model::Spec *)new steps::model::Spec((std::string const &)*arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Spec, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Spec(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Spec *arg1 = (steps::model::Spec *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Spec",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Spec, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Spec" "', argument " "1"" of type '" "steps::model::Spec *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Spec * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Spec_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Spec *arg1 = (steps::model::Spec *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Spec_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Spec_getID" "', argument " "1"" of type '" "steps::model::Spec const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Spec * >(argp1);
  {
    try {
      result = ((steps::model::Spec const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Spec_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Spec *arg1 = (steps::model::Spec *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Spec_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Spec_setID" "', argument " "1"" of type '" "steps::model::Spec *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Spec * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Spec_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Spec_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Spec_getModel(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Spec *arg1 = (steps::model::Spec *) 0 ;
  steps::model::Model *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Spec_getModel",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Spec_getModel" "', argument " "1"" of type '" "steps::model::Spec const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Spec * >(argp1);
  {
    try {
      result = (steps::model::Model *)((steps::model::Spec const *)arg1)->getModel();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Spec_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__Spec, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Surfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::model::Model *arg2 = (steps::model::Model *) 0 ;
  steps::model::Surfsys *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_Surfsys",&obj0,&obj1)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Surfsys" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Surfsys" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Surfsys" "', argument " "2"" of type '" "steps::model::Model *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Model * >(argp2);
  {
    try {
      result = (steps::model::Surfsys *)new steps::model::Surfsys((std::string const &)*arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Surfsys, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Surfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Surfsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Surfsys" "', argument " "1"" of type '" "steps::model::Surfsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Surfsys_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_getID" "', argument " "1"" of type '" "steps::model::Surfsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    try {
      result = ((steps::model::Surfsys const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Surfsys_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_setID" "', argument " "1"" of type '" "steps::model::Surfsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Surfsys_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Surfsys_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_getModel(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  steps::model::Model *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Surfsys_getModel",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_getModel" "', argument " "1"" of type '" "steps::model::Surfsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    try {
      result = (steps::model::Model *)((steps::model::Surfsys const *)arg1)->getModel();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_getSReac(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  std::string *arg2 = 0 ;
  steps::model::SReac *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Surfsys_getSReac",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_getSReac" "', argument " "1"" of type '" "steps::model::Surfsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Surfsys_getSReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Surfsys_getSReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::model::SReac *)((steps::model::Surfsys const *)arg1)->getSReac((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_delSReac(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Surfsys_delSReac",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_delSReac" "', argument " "1"" of type '" "steps::model::Surfsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Surfsys_delSReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Surfsys_delSReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delSReac((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_getAllSReacs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Surfsys_getAllSReacs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_getAllSReacs" "', argument " "1"" of type '" "steps::model::Surfsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    try {
      result = ((steps::model::Surfsys const *)arg1)->getAllSReacs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::SReac*,std::allocator<steps::model::SReac * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Surfsys_getAllSpecs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Surfsys *arg1 = (steps::model::Surfsys *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Surfsys_getAllSpecs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Surfsys_getAllSpecs" "', argument " "1"" of type '" "steps::model::Surfsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Surfsys * >(argp1);
  {
    try {
      result = ((steps::model::Surfsys const *)arg1)->getAllSpecs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Surfsys_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__Surfsys, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Volsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::model::Model *arg2 = (steps::model::Model *) 0 ;
  steps::model::Volsys *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_Volsys",&obj0,&obj1)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Volsys" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Volsys" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Volsys" "', argument " "2"" of type '" "steps::model::Model *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Model * >(argp2);
  {
    try {
      result = (steps::model::Volsys *)new steps::model::Volsys((std::string const &)*arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Volsys, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Volsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Volsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Volsys" "', argument " "1"" of type '" "steps::model::Volsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Volsys_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getID" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    try {
      result = ((steps::model::Volsys const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Volsys_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_setID" "', argument " "1"" of type '" "steps::model::Volsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Volsys_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Volsys_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getModel(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  steps::model::Model *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Volsys_getModel",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getModel" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    try {
      result = (steps::model::Model *)((steps::model::Volsys const *)arg1)->getModel();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getReac(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::string *arg2 = 0 ;
  steps::model::Reac *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Volsys_getReac",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getReac" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Volsys_getReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Volsys_getReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::model::Reac *)((steps::model::Volsys const *)arg1)->getReac((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_delReac(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Volsys_delReac",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_delReac" "', argument " "1"" of type '" "steps::model::Volsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Volsys_delReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Volsys_delReac" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delReac((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getAllReacs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Volsys_getAllReacs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getAllReacs" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    try {
      result = ((steps::model::Volsys const *)arg1)->getAllReacs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Reac*,std::allocator<steps::model::Reac * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getDiff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::string *arg2 = 0 ;
  steps::model::Diff *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Volsys_getDiff",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getDiff" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Volsys_getDiff" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Volsys_getDiff" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::model::Diff *)((steps::model::Volsys const *)arg1)->getDiff((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_delDiff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Volsys_delDiff",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_delDiff" "', argument " "1"" of type '" "steps::model::Volsys *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Volsys_delDiff" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Volsys_delDiff" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delDiff((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getAllDiffs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Volsys_getAllDiffs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getAllDiffs" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    try {
      result = ((steps::model::Volsys const *)arg1)->getAllDiffs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Diff*,std::allocator<steps::model::Diff * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Volsys_getAllSpecs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Volsys *arg1 = (steps::model::Volsys *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Volsys_getAllSpecs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Volsys_getAllSpecs" "', argument " "1"" of type '" "steps::model::Volsys const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Volsys * >(argp1);
  {
    try {
      result = ((steps::model::Volsys const *)arg1)->getAllSpecs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Volsys_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__Volsys, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Diff(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::model::Volsys *arg2 = (steps::model::Volsys *) 0 ;
  steps::model::Spec *arg3 = (steps::model::Spec *) 0 ;
  double arg4 = (double) 0.0 ;
  steps::model::Diff *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "volsys",(char *) "lig",(char *) "dcst", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OOO|O:new_Diff",kwnames,&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Diff" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Diff" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Diff" "', argument " "2"" of type '" "steps::model::Volsys *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Volsys * >(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Diff" "', argument " "3"" of type '" "steps::model::Spec *""'"); 
  }
  arg3 = reinterpret_cast< steps::model::Spec * >(argp3);
  if (obj3) {
    ecode4 = SWIG_AsVal_double(obj3, &val4);
    if (!SWIG_IsOK(ecode4)) {
      SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "new_Diff" "', argument " "4"" of type '" "double""'");
    } 
    arg4 = static_cast< double >(val4);
  }
  {
    try {
      result = (steps::model::Diff *)new steps::model::Diff((std::string const &)*arg1,arg2,arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Diff, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Diff(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Diff",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Diff" "', argument " "1"" of type '" "steps::model::Diff *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Diff_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_getID" "', argument " "1"" of type '" "steps::model::Diff const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    try {
      result = ((steps::model::Diff const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Diff_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_setID" "', argument " "1"" of type '" "steps::model::Diff *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Diff_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Diff_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_getVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  steps::model::Volsys *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Diff_getVolsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_getVolsys" "', argument " "1"" of type '" "steps::model::Diff const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    try {
      result = (steps::model::Volsys *)((steps::model::Diff const *)arg1)->getVolsys();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_getModel(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  steps::model::Model *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Diff_getModel",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_getModel" "', argument " "1"" of type '" "steps::model::Diff const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    try {
      result = (steps::model::Model *)((steps::model::Diff const *)arg1)->getModel();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_getLig(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  steps::model::Spec *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Diff_getLig",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_getLig" "', argument " "1"" of type '" "steps::model::Diff const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    try {
      result = (steps::model::Spec *)((steps::model::Diff const *)arg1)->getLig();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_setLig(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  steps::model::Spec *arg2 = (steps::model::Spec *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Diff_setLig",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_setLig" "', argument " "1"" of type '" "steps::model::Diff *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Spec, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Diff_setLig" "', argument " "2"" of type '" "steps::model::Spec *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Spec * >(argp2);
  {
    try {
      (arg1)->setLig(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_getDcst(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Diff_getDcst",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_getDcst" "', argument " "1"" of type '" "steps::model::Diff const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  {
    try {
      result = (double)((steps::model::Diff const *)arg1)->getDcst();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Diff_setDcst(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Diff *arg1 = (steps::model::Diff *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Diff_setDcst",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Diff, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Diff_setDcst" "', argument " "1"" of type '" "steps::model::Diff *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Diff * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Diff_setDcst" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setDcst(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Diff_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__Diff, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Reac(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::model::Volsys *arg2 = (steps::model::Volsys *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > arg3 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > >) (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > >)std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > arg4 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > >) (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > >)std::vector<steps::model::Spec * >() ;
  double arg5 = (double) 0.0 ;
  steps::model::Reac *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  double val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "volsys",(char *) "lhs",(char *) "rhs",(char *) "kcst", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OO|OOO:new_Reac",kwnames,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Reac" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Reac" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Reac" "', argument " "2"" of type '" "steps::model::Volsys *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Volsys * >(argp2);
  if (obj2) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      int res = swig::asptr(obj2, &ptr);
      if (!SWIG_IsOK(res) || !ptr) {
        SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_Reac" "', argument " "3"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const""'"); 
      }
      arg3 = *ptr;
      if (SWIG_IsNewObj(res)) delete ptr;
    }
  }
  if (obj3) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      int res = swig::asptr(obj3, &ptr);
      if (!SWIG_IsOK(res) || !ptr) {
        SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "new_Reac" "', argument " "4"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const""'"); 
      }
      arg4 = *ptr;
      if (SWIG_IsNewObj(res)) delete ptr;
    }
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_double(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_Reac" "', argument " "5"" of type '" "double""'");
    } 
    arg5 = static_cast< double >(val5);
  }
  {
    try {
      result = (steps::model::Reac *)new steps::model::Reac((std::string const &)*arg1,arg2,arg3,arg4,arg5);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Reac, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Reac(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Reac",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Reac" "', argument " "1"" of type '" "steps::model::Reac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getID" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = ((steps::model::Reac const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Reac_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_setID" "', argument " "1"" of type '" "steps::model::Reac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Reac_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Reac_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  steps::model::Volsys *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getVolsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getVolsys" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = (steps::model::Volsys *)((steps::model::Reac const *)arg1)->getVolsys();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Volsys, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getModel(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  steps::model::Model *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getModel",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getModel" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = (steps::model::Model *)((steps::model::Reac const *)arg1)->getModel();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getLHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getLHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getLHS" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = ((steps::model::Reac const *)arg1)->getLHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_setLHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Reac_setLHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_setLHS" "', argument " "1"" of type '" "steps::model::Reac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Reac_setLHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Reac_setLHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setLHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getRHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getRHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getRHS" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = ((steps::model::Reac const *)arg1)->getRHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_setRHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Reac_setRHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_setRHS" "', argument " "1"" of type '" "steps::model::Reac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Reac_setRHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Reac_setRHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setRHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getOrder(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getOrder",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getOrder" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = (int)((steps::model::Reac const *)arg1)->getOrder();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getKcst(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getKcst",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getKcst" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = (double)((steps::model::Reac const *)arg1)->getKcst();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_setKcst(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Reac_setKcst",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_setKcst" "', argument " "1"" of type '" "steps::model::Reac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Reac_setKcst" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setKcst(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Reac_getAllSpecs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Reac *arg1 = (steps::model::Reac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Reac_getAllSpecs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Reac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Reac_getAllSpecs" "', argument " "1"" of type '" "steps::model::Reac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Reac * >(argp1);
  {
    try {
      result = ((steps::model::Reac const *)arg1)->getAllSpecs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Reac_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__Reac, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_SReac(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::model::Surfsys *arg2 = (steps::model::Surfsys *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &arg3_defvalue = std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg3 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *) &arg3_defvalue ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &arg4_defvalue = std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg4 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *) &arg4_defvalue ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &arg5_defvalue = std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg5 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *) &arg5_defvalue ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &arg6_defvalue = std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg6 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *) &arg6_defvalue ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &arg7_defvalue = std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg7 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *) &arg7_defvalue ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &arg8_defvalue = std::vector<steps::model::Spec * >() ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg8 = (std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *) &arg8_defvalue ;
  double arg9 = (double) 0.0 ;
  steps::model::SReac *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  int res6 = SWIG_OLDOBJ ;
  int res7 = SWIG_OLDOBJ ;
  int res8 = SWIG_OLDOBJ ;
  double val9 ;
  int ecode9 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "surfsys",(char *) "olhs",(char *) "ilhs",(char *) "slhs",(char *) "irhs",(char *) "srhs",(char *) "orhs",(char *) "kcst", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OO|OOOOOOO:new_SReac",kwnames,&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_SReac" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_SReac" "', argument " "2"" of type '" "steps::model::Surfsys *""'"); 
  }
  arg2 = reinterpret_cast< steps::model::Surfsys * >(argp2);
  if (obj2) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      res3 = swig::asptr(obj2, &ptr);
      if (!SWIG_IsOK(res3)) {
        SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_SReac" "', argument " "3"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      if (!ptr) {
        SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "3"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      arg3 = ptr;
    }
  }
  if (obj3) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      res4 = swig::asptr(obj3, &ptr);
      if (!SWIG_IsOK(res4)) {
        SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_SReac" "', argument " "4"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      if (!ptr) {
        SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "4"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      arg4 = ptr;
    }
  }
  if (obj4) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      res5 = swig::asptr(obj4, &ptr);
      if (!SWIG_IsOK(res5)) {
        SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_SReac" "', argument " "5"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      if (!ptr) {
        SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "5"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      arg5 = ptr;
    }
  }
  if (obj5) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      res6 = swig::asptr(obj5, &ptr);
      if (!SWIG_IsOK(res6)) {
        SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "new_SReac" "', argument " "6"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      if (!ptr) {
        SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "6"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      arg6 = ptr;
    }
  }
  if (obj6) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      res7 = swig::asptr(obj6, &ptr);
      if (!SWIG_IsOK(res7)) {
        SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "new_SReac" "', argument " "7"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      if (!ptr) {
        SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "7"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      arg7 = ptr;
    }
  }
  if (obj7) {
    {
      std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
      res8 = swig::asptr(obj7, &ptr);
      if (!SWIG_IsOK(res8)) {
        SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "new_SReac" "', argument " "8"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      if (!ptr) {
        SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_SReac" "', argument " "8"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
      }
      arg8 = ptr;
    }
  }
  if (obj8) {
    ecode9 = SWIG_AsVal_double(obj8, &val9);
    if (!SWIG_IsOK(ecode9)) {
      SWIG_exception_fail(SWIG_ArgError(ecode9), "in method '" "new_SReac" "', argument " "9"" of type '" "double""'");
    } 
    arg9 = static_cast< double >(val9);
  }
  {
    try {
      result = (steps::model::SReac *)new steps::model::SReac((std::string const &)*arg1,arg2,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg3,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg4,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg5,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg6,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg7,(std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg8,arg9);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__SReac, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  if (SWIG_IsNewObj(res6)) delete arg6;
  if (SWIG_IsNewObj(res7)) delete arg7;
  if (SWIG_IsNewObj(res8)) delete arg8;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  if (SWIG_IsNewObj(res6)) delete arg6;
  if (SWIG_IsNewObj(res7)) delete arg7;
  if (SWIG_IsNewObj(res8)) delete arg8;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_SReac(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_SReac",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_SReac" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getID" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setID" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getSurfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  steps::model::Surfsys *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getSurfsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getSurfsys" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = (steps::model::Surfsys *)((steps::model::SReac const *)arg1)->getSurfsys();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Surfsys, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getModel(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  steps::model::Model *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getModel",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getModel" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = (steps::model::Model *)((steps::model::SReac const *)arg1)->getModel();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__model__Model, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getInner(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getInner",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getInner" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = (bool)((steps::model::SReac const *)arg1)->getInner();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getOuter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getOuter",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getOuter" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = (bool)((steps::model::SReac const *)arg1)->getOuter();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getOLHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getOLHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getOLHS" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getOLHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setOLHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setOLHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setOLHS" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setOLHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setOLHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setOLHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getILHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getILHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getILHS" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getILHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setILHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setILHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setILHS" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setILHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setILHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setILHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getSLHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getSLHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getSLHS" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getSLHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setSLHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setSLHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setSLHS" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setSLHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setSLHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setSLHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getIRHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getIRHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getIRHS" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getIRHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setIRHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setIRHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setIRHS" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setIRHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setIRHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setIRHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getSRHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getSRHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getSRHS" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getSRHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setSRHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setSRHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setSRHS" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setSRHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setSRHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setSRHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getORHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getORHS",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getORHS" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getORHS();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setORHS(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setORHS",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setORHS" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *ptr = (std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "SReac_setORHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "SReac_setORHS" "', argument " "2"" of type '" "std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setORHS((std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getOrder(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getOrder",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getOrder" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = (int)((steps::model::SReac const *)arg1)->getOrder();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getKcst(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getKcst",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getKcst" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = (double)((steps::model::SReac const *)arg1)->getKcst();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_setKcst(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:SReac_setKcst",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_setKcst" "', argument " "1"" of type '" "steps::model::SReac *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "SReac_setKcst" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setKcst(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_SReac_getAllSpecs(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::SReac *arg1 = (steps::model::SReac *) 0 ;
  std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:SReac_getAllSpecs",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__SReac, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "SReac_getAllSpecs" "', argument " "1"" of type '" "steps::model::SReac const *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::SReac * >(argp1);
  {
    try {
      result = ((steps::model::SReac const *)arg1)->getAllSpecs();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::model::Spec*,std::allocator<steps::model::Spec * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *SReac_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__model__SReac, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_set_str_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_iterator" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_set_Sl_std_string_Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str___nonzero__" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (bool)std_set_Sl_std_string_Sg____nonzero__((std::set<std::string > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str___len__" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = std_set_Sl_std_string_Sg____len__((std::set<std::string > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_append" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "set_str_append" "', argument " "2"" of type '" "std::set<std::string >::value_type""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    try {
      std_set_Sl_std_string_Sg__append(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str___contains__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::value_type arg2 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str___contains__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str___contains__" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "set_str___contains__" "', argument " "2"" of type '" "std::set<std::string >::value_type""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    try {
      result = (bool)std_set_Sl_std_string_Sg____contains__(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::difference_type arg2 ;
  std::set<std::string >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str___getitem__" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "set_str___getitem__" "', argument " "2"" of type '" "std::set<std::string >::difference_type""'");
  } 
  arg2 = static_cast< std::set<std::string >::difference_type >(val2);
  {
    try {
      try {
        result = std_set_Sl_std_string_Sg____getitem__((std::set<std::string > const *)arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_set_str__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_set_str")) SWIG_fail;
  {
    try {
      result = (std::set<std::string > *)new std::set<std::string >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_set_str__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = 0 ;
  std::set<std::string > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_set_str",&obj0)) SWIG_fail;
  {
    std::set<std::string,std::less<std::string >,std::allocator<std::string > > *ptr = (std::set<std::string,std::less<std::string >,std::allocator<std::string > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_set_str" "', argument " "1"" of type '" "std::set<std::string > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_set_str" "', argument " "1"" of type '" "std::set<std::string > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::set<std::string > *)new std::set<std::string >((std::set<std::string > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_set_str(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_set_str__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_set_str__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_set_str'.\n  Possible C/C++ prototypes are:\n    std::set<(std::string)>()\n    std::set<(std::string)>(std::set<std::string > const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_empty" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (bool)((std::set<std::string > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_size" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = ((std::set<std::string > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_clear" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_swap" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_swap" "', argument " "2"" of type '" "std::set<std::string > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_swap" "', argument " "2"" of type '" "std::set<std::string > &""'"); 
  }
  arg2 = reinterpret_cast< std::set<std::string > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  SwigValueWrapper<std::allocator<std::string > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_get_allocator" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = ((std::set<std::string > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::set<std::string >::allocator_type(static_cast< const std::set<std::string >::allocator_type& >(result))), SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_begin" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_begin" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = ((std::set<std::string > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_str_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_end" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_end" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = ((std::set<std::string > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_str_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_rbegin" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_rbegin" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = ((std::set<std::string > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_str_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_rend" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_str_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_rend" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      result = ((std::set<std::string > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_str_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_str_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::key_type *arg2 = 0 ;
  std::set<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_erase" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_erase" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_erase" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (arg1)->erase((std::set<std::string >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_count(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::key_type *arg2 = 0 ;
  std::set<std::string >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_count",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_count" "', argument " "1"" of type '" "std::set<std::string > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_count" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_count" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = ((std::set<std::string > const *)arg1)->count((std::set<std::string >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_erase" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_str_erase" "', argument " "2"" of type '" "std::set<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::set<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::set<std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_str_erase" "', argument " "2"" of type '" "std::set<std::string >::iterator""'");
    }
  }
  {
    try {
      (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_erase__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::iterator arg2 ;
  std::set<std::string >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:set_str_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_erase" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_str_erase" "', argument " "2"" of type '" "std::set<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::set<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::set<std::string >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_str_erase" "', argument " "2"" of type '" "std::set<std::string >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_str_erase" "', argument " "3"" of type '" "std::set<std::string >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::set<std::string >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::set<std::string >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_str_erase" "', argument " "3"" of type '" "std::set<std::string >::iterator""'");
    }
  }
  {
    try {
      (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::set<std::string >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_set_str_erase__SWIG_1(self, args);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = SWIG_AsPtr_std_string(argv[1], (std::string**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_set_str_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<std::string,std::less<std::string >,std::allocator<std::string > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::set<std::string >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::set<std::string >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_set_str_erase__SWIG_2(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_str_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::set<std::string >::key_type const &)\n    erase(std::set<std::string >::iterator)\n    erase(std::set<std::string >::iterator,std::set<std::string >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_find(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::key_type *arg2 = 0 ;
  std::set<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_find",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_find" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_find" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_find" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (arg1)->find((std::set<std::string >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_lower_bound(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::key_type *arg2 = 0 ;
  std::set<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_lower_bound",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_lower_bound" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_lower_bound" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_lower_bound" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (arg1)->lower_bound((std::set<std::string >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_upper_bound(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::key_type *arg2 = 0 ;
  std::set<std::string >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_upper_bound",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_upper_bound" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_upper_bound" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_upper_bound" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (arg1)->upper_bound((std::set<std::string >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<std::string >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_equal_range(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::key_type *arg2 = 0 ;
  SwigValueWrapper<std::pair<std::set<std::string >::iterator,std::set<std::string >::iterator > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_equal_range",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_equal_range" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_equal_range" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_equal_range" "', argument " "2"" of type '" "std::set<std::string >::key_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (arg1)->equal_range((std::set<std::string >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair<std::set<std::string >::iterator,std::set<std::string >::iterator > & >(result).first),
      swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN));
  PyTuple_SetItem(resultobj,1,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair<std::set<std::string >::iterator,std::set<std::string >::iterator > & >(result).second),
      swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_str_insert(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  std::set<std::string >::value_type *arg2 = 0 ;
  SwigValueWrapper<std::pair<std::set<std::string >::iterator,bool > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_str_insert",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_str_insert" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_str_insert" "', argument " "2"" of type '" "std::set<std::string >::value_type const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_str_insert" "', argument " "2"" of type '" "std::set<std::string >::value_type const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (arg1)->insert((std::set<std::string >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair<std::set<std::string >::iterator,bool > & >(result).first),
      swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN));    
  PyTuple_SetItem(resultobj,1,SWIG_From_bool  (static_cast< const std::pair<std::set<std::string >::iterator,bool > & >(result).second));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_set_str(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<std::string > *arg1 = (std::set<std::string > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_set_str",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_set_str" "', argument " "1"" of type '" "std::set<std::string > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<std::string > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *set_str_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_set_ptc_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_iterator" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_set_Sl_steps_wm_Patch_Sm__Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc___nonzero__" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (bool)std_set_Sl_steps_wm_Patch_Sm__Sg____nonzero__((std::set<steps::wm::Patch * > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc___len__" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = std_set_Sl_steps_wm_Patch_Sm__Sg____len__((std::set<steps::wm::Patch * > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::value_type arg2 = (std::set<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_append" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_append" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::set<steps::wm::Patch * >::value_type >(argp2);
  {
    try {
      std_set_Sl_steps_wm_Patch_Sm__Sg__append(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc___contains__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::value_type arg2 = (std::set<steps::wm::Patch * >::value_type) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc___contains__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc___contains__" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc___contains__" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::set<steps::wm::Patch * >::value_type >(argp2);
  {
    try {
      result = (bool)std_set_Sl_steps_wm_Patch_Sm__Sg____contains__(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::difference_type arg2 ;
  std::set<steps::wm::Patch * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc___getitem__" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "set_ptc___getitem__" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::set<steps::wm::Patch * >::difference_type >(val2);
  {
    try {
      try {
        result = (std::set<steps::wm::Patch * >::value_type)std_set_Sl_steps_wm_Patch_Sm__Sg____getitem__((std::set<steps::wm::Patch * > const *)arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_set_ptc__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_set_ptc")) SWIG_fail;
  {
    try {
      result = (std::set<steps::wm::Patch * > *)new std::set<steps::wm::Patch * >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_set_ptc__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = 0 ;
  std::set<steps::wm::Patch * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_set_ptc",&obj0)) SWIG_fail;
  {
    std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > *ptr = (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_set_ptc" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_set_ptc" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::set<steps::wm::Patch * > *)new std::set<steps::wm::Patch * >((std::set<steps::wm::Patch * > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_set_ptc(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_set_ptc__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_set_ptc__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_set_ptc'.\n  Possible C/C++ prototypes are:\n    std::set<(p.steps::wm::Patch)>()\n    std::set<(p.steps::wm::Patch)>(std::set<steps::wm::Patch * > const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_empty" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (bool)((std::set<steps::wm::Patch * > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_size" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_clear" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_swap" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_swap" "', argument " "2"" of type '" "std::set<steps::wm::Patch * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "set_ptc_swap" "', argument " "2"" of type '" "std::set<steps::wm::Patch * > &""'"); 
  }
  arg2 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::wm::Patch * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_get_allocator" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::set<steps::wm::Patch * >::allocator_type(static_cast< const std::set<steps::wm::Patch * >::allocator_type& >(result))), SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_begin" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_begin" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_ptc_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_end" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_end" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_ptc_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_rbegin" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_rbegin" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_ptc_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_rend" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:set_ptc_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_rend" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_set_ptc_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_ptc_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::key_type *arg2 = 0 ;
  std::set<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::key_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_erase" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_erase" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::key_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::key_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = (arg1)->erase((std::set<steps::wm::Patch * >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_count(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::key_type *arg2 = 0 ;
  std::set<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::key_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_count",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_count" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_count" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::key_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::key_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = ((std::set<steps::wm::Patch * > const *)arg1)->count((std::set<steps::wm::Patch * >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::iterator arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_erase" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_ptc_erase" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_ptc_erase" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::iterator""'");
    }
  }
  {
    try {
      (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_erase__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::iterator arg2 ;
  std::set<steps::wm::Patch * >::iterator arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:set_ptc_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_erase" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_ptc_erase" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_ptc_erase" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_ptc_erase" "', argument " "3"" of type '" "std::set<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "set_ptc_erase" "', argument " "3"" of type '" "std::set<steps::wm::Patch * >::iterator""'");
    }
  }
  {
    try {
      (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_set_ptc_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_set_ptc_erase__SWIG_1(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::set<steps::wm::Patch * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_set_ptc_erase__SWIG_2(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'set_ptc_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::set<steps::wm::Patch * >::key_type const &)\n    erase(std::set<steps::wm::Patch * >::iterator)\n    erase(std::set<steps::wm::Patch * >::iterator,std::set<steps::wm::Patch * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_find(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::key_type *arg2 = 0 ;
  std::set<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::key_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_find",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_find" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_find" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::key_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::key_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = (arg1)->find((std::set<steps::wm::Patch * >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_lower_bound(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::key_type *arg2 = 0 ;
  std::set<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::key_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_lower_bound",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_lower_bound" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_lower_bound" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::key_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::key_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = (arg1)->lower_bound((std::set<steps::wm::Patch * >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_upper_bound(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::key_type *arg2 = 0 ;
  std::set<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::key_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_upper_bound",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_upper_bound" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_upper_bound" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::key_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::key_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = (arg1)->upper_bound((std::set<steps::wm::Patch * >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::set<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_equal_range(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::key_type *arg2 = 0 ;
  SwigValueWrapper<std::pair<std::set<steps::wm::Patch * >::iterator,std::set<steps::wm::Patch * >::iterator > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::key_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_equal_range",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_equal_range" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_equal_range" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::key_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::key_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = (arg1)->equal_range((std::set<steps::wm::Patch * >::key_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair<std::set<steps::wm::Patch * >::iterator,std::set<steps::wm::Patch * >::iterator > & >(result).first),
      swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN));
  PyTuple_SetItem(resultobj,1,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair<std::set<steps::wm::Patch * >::iterator,std::set<steps::wm::Patch * >::iterator > & >(result).second),
      swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_set_ptc_insert(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  std::set<steps::wm::Patch * >::value_type *arg2 = 0 ;
  SwigValueWrapper<std::pair<std::set<steps::wm::Patch * >::iterator,bool > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  std::set<steps::wm::Patch * >::value_type temp2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:set_ptc_insert",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "set_ptc_insert" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "set_ptc_insert" "', argument " "2"" of type '" "std::set<steps::wm::Patch * >::value_type""'"); 
  }
  temp2 = reinterpret_cast< std::set<steps::wm::Patch * >::value_type >(argp2);
  arg2 = &temp2;
  {
    try {
      result = (arg1)->insert((std::set<steps::wm::Patch * >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = PyTuple_New(2);
  PyTuple_SetItem(resultobj,0,SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::pair<std::set<steps::wm::Patch * >::iterator,bool > & >(result).first),
      swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN));    
  PyTuple_SetItem(resultobj,1,SWIG_From_bool  (static_cast< const std::pair<std::set<steps::wm::Patch * >::iterator,bool > & >(result).second));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_set_ptc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::set<steps::wm::Patch * > *arg1 = (std::set<steps::wm::Patch * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_set_ptc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_set_ptc" "', argument " "1"" of type '" "std::set<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::set<steps::wm::Patch * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *set_ptc_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_ptc_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_iterator" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_vector_Sl_steps_wm_Patch_Sm__Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___nonzero__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (bool)std_vector_Sl_steps_wm_Patch_Sm__Sg____nonzero__((std::vector<steps::wm::Patch * > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___len__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = std_vector_Sl_steps_wm_Patch_Sm__Sg____len__((std::vector<steps::wm::Patch * > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_pop" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      try {
        result = (std::vector<steps::wm::Patch * >::value_type)std_vector_Sl_steps_wm_Patch_Sm__Sg__pop(arg1);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::difference_type arg2 ;
  std::vector<steps::wm::Patch * >::difference_type arg3 ;
  std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___getslice__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc___getslice__" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ptc___getslice__" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val3);
  {
    try {
      try {
        result = (std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > *)std_vector_Sl_steps_wm_Patch_Sm__Sg____getslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::difference_type arg2 ;
  std::vector<steps::wm::Patch * >::difference_type arg3 ;
  std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_ptc___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___setslice__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc___setslice__" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ptc___setslice__" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val3);
  {
    std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > > *ptr = (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_ptc___setslice__" "', argument " "4"" of type '" "std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_ptc___setslice__" "', argument " "4"" of type '" "std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    try {
      try {
        std_vector_Sl_steps_wm_Patch_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > const &)*arg4);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      catch(std::invalid_argument &_e) {
        SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::difference_type arg2 ;
  std::vector<steps::wm::Patch * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___delslice__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc___delslice__" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ptc___delslice__" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val3);
  {
    try {
      try {
        std_vector_Sl_steps_wm_Patch_Sm__Sg____delslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___delitem__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc___delitem__" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val2);
  {
    try {
      try {
        std_vector_Sl_steps_wm_Patch_Sm__Sg____delitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::difference_type arg2 ;
  std::vector<steps::wm::Patch * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___getitem__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc___getitem__" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val2);
  {
    try {
      try {
        result = (std::vector<steps::wm::Patch * >::value_type)std_vector_Sl_steps_wm_Patch_Sm__Sg____getitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::difference_type arg2 ;
  std::vector<steps::wm::Patch * >::value_type arg3 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc___setitem__" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc___setitem__" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ptc___setitem__" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp3);
  {
    try {
      try {
        std_vector_Sl_steps_wm_Patch_Sm__Sg____setitem__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::value_type arg2 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_append" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_ptc_append" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp2);
  {
    try {
      std_vector_Sl_steps_wm_Patch_Sm__Sg__append(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ptc__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_ptc")) SWIG_fail;
  {
    try {
      result = (std::vector<steps::wm::Patch * > *)new std::vector<steps::wm::Patch * >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ptc__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = 0 ;
  std::vector<steps::wm::Patch * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_ptc",&obj0)) SWIG_fail;
  {
    std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > > *ptr = (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_ptc" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_ptc" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::vector<steps::wm::Patch * > *)new std::vector<steps::wm::Patch * >((std::vector<steps::wm::Patch * > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_empty" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (bool)((std::vector<steps::wm::Patch * > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_size" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_clear" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_swap" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_ptc_swap" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_ptc_swap" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::wm::Patch * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_get_allocator" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::vector<steps::wm::Patch * >::allocator_type(static_cast< const std::vector<steps::wm::Patch * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_begin" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_begin" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_end" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_end" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_rbegin" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_rbegin" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_rend" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_rend" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_ptc_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ptc__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * >::size_type arg1 ;
  std::vector<steps::wm::Patch * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_ptc",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_ptc" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val1);
  {
    try {
      result = (std::vector<steps::wm::Patch * > *)new std::vector<steps::wm::Patch * >(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_pop_back" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      (arg1)->pop_back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_resize" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc_resize" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val2);
  {
    try {
      (arg1)->resize(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::iterator arg2 ;
  std::vector<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_erase" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::iterator arg2 ;
  std::vector<steps::wm::Patch * >::iterator arg3 ;
  std::vector<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_erase" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_erase" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_erase" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_ptc_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_ptc_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::wm::Patch * >::iterator)\n    erase(std::vector<steps::wm::Patch * >::iterator,std::vector<steps::wm::Patch * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ptc__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * >::size_type arg1 ;
  std::vector<steps::wm::Patch * >::value_type arg2 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  std::vector<steps::wm::Patch * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_ptc",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_ptc" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_ptc" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp2);
  {
    try {
      result = (std::vector<steps::wm::Patch * > *)new std::vector<steps::wm::Patch * >(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_ptc(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_ptc__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_ptc__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_ptc__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_ptc__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_ptc'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::wm::Patch)>()\n    std::vector<(p.steps::wm::Patch)>(std::vector<steps::wm::Patch * > const &)\n    std::vector<(p.steps::wm::Patch)>(std::vector<steps::wm::Patch * >::size_type)\n    std::vector<(p.steps::wm::Patch)>(std::vector<steps::wm::Patch * >::size_type,std::vector<steps::wm::Patch * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::value_type arg2 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_push_back" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_ptc_push_back" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp2);
  {
    try {
      (arg1)->push_back(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_front" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (std::vector<steps::wm::Patch * >::value_type)((std::vector<steps::wm::Patch * > const *)arg1)->front();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_back" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = (std::vector<steps::wm::Patch * >::value_type)((std::vector<steps::wm::Patch * > const *)arg1)->back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type arg2 ;
  std::vector<steps::wm::Patch * >::value_type arg3 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_assign" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc_assign" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ptc_assign" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp3);
  {
    try {
      (arg1)->assign(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type arg2 ;
  std::vector<steps::wm::Patch * >::value_type arg3 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_resize" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc_resize" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ptc_resize" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp3);
  {
    try {
      (arg1)->resize(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_ptc_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_ptc_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::wm::Patch * >::size_type)\n    resize(std::vector<steps::wm::Patch * >::size_type,std::vector<steps::wm::Patch * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::iterator arg2 ;
  std::vector<steps::wm::Patch * >::value_type arg3 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  std::vector<steps::wm::Patch * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_ptc_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_insert" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_ptc_insert" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp3);
  {
    try {
      result = (arg1)->insert(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Patch * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::iterator arg2 ;
  std::vector<steps::wm::Patch * >::size_type arg3 ;
  std::vector<steps::wm::Patch * >::value_type arg4 = (std::vector<steps::wm::Patch * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_ptc_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_insert" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_ptc_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_ptc_insert" "', argument " "3"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_ptc_insert" "', argument " "4"" of type '" "std::vector<steps::wm::Patch * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::wm::Patch * >::value_type >(argp4);
  {
    try {
      (arg1)->insert(arg2,arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_ptc_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Patch * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_ptc_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_ptc_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::wm::Patch * >::iterator,std::vector<steps::wm::Patch * >::value_type)\n    insert(std::vector<steps::wm::Patch * >::iterator,std::vector<steps::wm::Patch * >::size_type,std::vector<steps::wm::Patch * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_ptc_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_reserve" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_ptc_reserve" "', argument " "2"" of type '" "std::vector<steps::wm::Patch * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Patch * >::size_type >(val2);
  {
    try {
      (arg1)->reserve(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_ptc_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  std::vector<steps::wm::Patch * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_ptc_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_ptc_capacity" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Patch * > const *)arg1)->capacity();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_ptc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Patch * > *arg1 = (std::vector<steps::wm::Patch * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_ptc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_ptc" "', argument " "1"" of type '" "std::vector<steps::wm::Patch * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Patch * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_ptc_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_cmp_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_iterator" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_vector_Sl_steps_wm_Comp_Sm__Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___nonzero__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (bool)std_vector_Sl_steps_wm_Comp_Sm__Sg____nonzero__((std::vector<steps::wm::Comp * > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___len__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = std_vector_Sl_steps_wm_Comp_Sm__Sg____len__((std::vector<steps::wm::Comp * > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_pop" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      try {
        result = (std::vector<steps::wm::Comp * >::value_type)std_vector_Sl_steps_wm_Comp_Sm__Sg__pop(arg1);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::difference_type arg2 ;
  std::vector<steps::wm::Comp * >::difference_type arg3 ;
  std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___getslice__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp___getslice__" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_cmp___getslice__" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val3);
  {
    try {
      try {
        result = (std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > *)std_vector_Sl_steps_wm_Comp_Sm__Sg____getslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::difference_type arg2 ;
  std::vector<steps::wm::Comp * >::difference_type arg3 ;
  std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_cmp___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___setslice__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp___setslice__" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_cmp___setslice__" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val3);
  {
    std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > > *ptr = (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_cmp___setslice__" "', argument " "4"" of type '" "std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_cmp___setslice__" "', argument " "4"" of type '" "std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    try {
      try {
        std_vector_Sl_steps_wm_Comp_Sm__Sg____setslice__(arg1,arg2,arg3,(std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > const &)*arg4);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      catch(std::invalid_argument &_e) {
        SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::difference_type arg2 ;
  std::vector<steps::wm::Comp * >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___delslice__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp___delslice__" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_cmp___delslice__" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val3);
  {
    try {
      try {
        std_vector_Sl_steps_wm_Comp_Sm__Sg____delslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___delitem__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp___delitem__" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val2);
  {
    try {
      try {
        std_vector_Sl_steps_wm_Comp_Sm__Sg____delitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::difference_type arg2 ;
  std::vector<steps::wm::Comp * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___getitem__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp___getitem__" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val2);
  {
    try {
      try {
        result = (std::vector<steps::wm::Comp * >::value_type)std_vector_Sl_steps_wm_Comp_Sm__Sg____getitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::difference_type arg2 ;
  std::vector<steps::wm::Comp * >::value_type arg3 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp___setitem__" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp___setitem__" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::difference_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_cmp___setitem__" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp3);
  {
    try {
      try {
        std_vector_Sl_steps_wm_Comp_Sm__Sg____setitem__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::value_type arg2 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_append" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_cmp_append" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp2);
  {
    try {
      std_vector_Sl_steps_wm_Comp_Sm__Sg__append(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_cmp__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_cmp")) SWIG_fail;
  {
    try {
      result = (std::vector<steps::wm::Comp * > *)new std::vector<steps::wm::Comp * >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_cmp__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = 0 ;
  std::vector<steps::wm::Comp * > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_cmp",&obj0)) SWIG_fail;
  {
    std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > > *ptr = (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_cmp" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_cmp" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::vector<steps::wm::Comp * > *)new std::vector<steps::wm::Comp * >((std::vector<steps::wm::Comp * > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_empty" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (bool)((std::vector<steps::wm::Comp * > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_size" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_clear" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_swap" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_cmp_swap" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_cmp_swap" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  SwigValueWrapper<std::allocator<steps::wm::Comp * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_get_allocator" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::vector<steps::wm::Comp * >::allocator_type(static_cast< const std::vector<steps::wm::Comp * >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_begin" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_begin" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_end" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_end" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_rbegin" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_rbegin" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_rend" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_rend" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_cmp_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_cmp__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * >::size_type arg1 ;
  std::vector<steps::wm::Comp * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_cmp",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_cmp" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val1);
  {
    try {
      result = (std::vector<steps::wm::Comp * > *)new std::vector<steps::wm::Comp * >(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_pop_back" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      (arg1)->pop_back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_resize" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp_resize" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val2);
  {
    try {
      (arg1)->resize(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::iterator arg2 ;
  std::vector<steps::wm::Comp * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_erase" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::iterator arg2 ;
  std::vector<steps::wm::Comp * >::iterator arg3 ;
  std::vector<steps::wm::Comp * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_erase" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_erase" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_erase" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_erase" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_cmp_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_cmp_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<steps::wm::Comp * >::iterator)\n    erase(std::vector<steps::wm::Comp * >::iterator,std::vector<steps::wm::Comp * >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_cmp__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * >::size_type arg1 ;
  std::vector<steps::wm::Comp * >::value_type arg2 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  std::vector<steps::wm::Comp * > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_cmp",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_cmp" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg1 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_vector_cmp" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp2);
  {
    try {
      result = (std::vector<steps::wm::Comp * > *)new std::vector<steps::wm::Comp * >(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_cmp(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_cmp__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_cmp__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_cmp__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      void *vptr = 0;
      int res = SWIG_ConvertPtr(argv[1], &vptr, SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0);
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_vector_cmp__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_cmp'.\n  Possible C/C++ prototypes are:\n    std::vector<(p.steps::wm::Comp)>()\n    std::vector<(p.steps::wm::Comp)>(std::vector<steps::wm::Comp * > const &)\n    std::vector<(p.steps::wm::Comp)>(std::vector<steps::wm::Comp * >::size_type)\n    std::vector<(p.steps::wm::Comp)>(std::vector<steps::wm::Comp * >::size_type,std::vector<steps::wm::Comp * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::value_type arg2 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_push_back" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_cmp_push_back" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg2 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp2);
  {
    try {
      (arg1)->push_back(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_front" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (std::vector<steps::wm::Comp * >::value_type)((std::vector<steps::wm::Comp * > const *)arg1)->front();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_back" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = (std::vector<steps::wm::Comp * >::value_type)((std::vector<steps::wm::Comp * > const *)arg1)->back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type arg2 ;
  std::vector<steps::wm::Comp * >::value_type arg3 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_assign" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp_assign" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_cmp_assign" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp3);
  {
    try {
      (arg1)->assign(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type arg2 ;
  std::vector<steps::wm::Comp * >::value_type arg3 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_resize" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp_resize" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_cmp_resize" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp3);
  {
    try {
      (arg1)->resize(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_cmp_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_cmp_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<steps::wm::Comp * >::size_type)\n    resize(std::vector<steps::wm::Comp * >::size_type,std::vector<steps::wm::Comp * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::iterator arg2 ;
  std::vector<steps::wm::Comp * >::value_type arg3 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  std::vector<steps::wm::Comp * >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_cmp_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_insert" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "vector_cmp_insert" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg3 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp3);
  {
    try {
      result = (arg1)->insert(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<steps::wm::Comp * >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::iterator arg2 ;
  std::vector<steps::wm::Comp * >::size_type arg3 ;
  std::vector<steps::wm::Comp * >::value_type arg4 = (std::vector<steps::wm::Comp * >::value_type) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_cmp_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_insert" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_cmp_insert" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_cmp_insert" "', argument " "3"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg3 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val3);
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_cmp_insert" "', argument " "4"" of type '" "std::vector<steps::wm::Comp * >::value_type""'"); 
  }
  arg4 = reinterpret_cast< std::vector<steps::wm::Comp * >::value_type >(argp4);
  {
    try {
      (arg1)->insert(arg2,arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter) != 0));
      if (_v) {
        void *vptr = 0;
        int res = SWIG_ConvertPtr(argv[2], &vptr, SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0);
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_vector_cmp_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<steps::wm::Comp * >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          void *vptr = 0;
          int res = SWIG_ConvertPtr(argv[3], &vptr, SWIGTYPE_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0);
          _v = SWIG_CheckState(res);
          if (_v) {
            return _wrap_vector_cmp_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_cmp_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<steps::wm::Comp * >::iterator,std::vector<steps::wm::Comp * >::value_type)\n    insert(std::vector<steps::wm::Comp * >::iterator,std::vector<steps::wm::Comp * >::size_type,std::vector<steps::wm::Comp * >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_cmp_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_reserve" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_cmp_reserve" "', argument " "2"" of type '" "std::vector<steps::wm::Comp * >::size_type""'");
  } 
  arg2 = static_cast< std::vector<steps::wm::Comp * >::size_type >(val2);
  {
    try {
      (arg1)->reserve(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_cmp_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  std::vector<steps::wm::Comp * >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_cmp_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_cmp_capacity" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      result = ((std::vector<steps::wm::Comp * > const *)arg1)->capacity();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_cmp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<steps::wm::Comp * > *arg1 = (std::vector<steps::wm::Comp * > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_cmp",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_cmp" "', argument " "1"" of type '" "std::vector<steps::wm::Comp * > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<steps::wm::Comp * > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_cmp_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_int_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_iterator" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_vector_Sl_int_Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___nonzero__" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (bool)std_vector_Sl_int_Sg____nonzero__((std::vector<int > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___len__" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = std_vector_Sl_int_Sg____len__((std::vector<int > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_pop" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      try {
        result = (std::vector<int >::value_type)std_vector_Sl_int_Sg__pop(arg1);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::difference_type arg2 ;
  std::vector<int >::difference_type arg3 ;
  std::vector<int,std::allocator<int > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___getslice__" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int___getslice__" "', argument " "2"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int___getslice__" "', argument " "3"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<int >::difference_type >(val3);
  {
    try {
      try {
        result = (std::vector<int,std::allocator<int > > *)std_vector_Sl_int_Sg____getslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::difference_type arg2 ;
  std::vector<int >::difference_type arg3 ;
  std::vector<int,std::allocator<int > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_int___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___setslice__" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int___setslice__" "', argument " "2"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int___setslice__" "', argument " "3"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<int >::difference_type >(val3);
  {
    std::vector<int,std::allocator<int > > *ptr = (std::vector<int,std::allocator<int > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_int___setslice__" "', argument " "4"" of type '" "std::vector<int,std::allocator<int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_int___setslice__" "', argument " "4"" of type '" "std::vector<int,std::allocator<int > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    try {
      try {
        std_vector_Sl_int_Sg____setslice__(arg1,arg2,arg3,(std::vector<int,std::allocator<int > > const &)*arg4);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      catch(std::invalid_argument &_e) {
        SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::difference_type arg2 ;
  std::vector<int >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___delslice__" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int___delslice__" "', argument " "2"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int___delslice__" "', argument " "3"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<int >::difference_type >(val3);
  {
    try {
      try {
        std_vector_Sl_int_Sg____delslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___delitem__" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int___delitem__" "', argument " "2"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<int >::difference_type >(val2);
  {
    try {
      try {
        std_vector_Sl_int_Sg____delitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::difference_type arg2 ;
  std::vector<int >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___getitem__" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int___getitem__" "', argument " "2"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<int >::difference_type >(val2);
  {
    try {
      try {
        {
          std::vector<int >::value_type const &_result_ref = std_vector_Sl_int_Sg____getitem__((std::vector<int > const *)arg1,arg2);
          result = (std::vector<int >::value_type *) &_result_ref;
        }
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::difference_type arg2 ;
  std::vector<int >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector<int >::value_type temp3 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int___setitem__" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int___setitem__" "', argument " "2"" of type '" "std::vector<int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int___setitem__" "', argument " "3"" of type '" "std::vector<int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      try {
        std_vector_Sl_int_Sg____setitem__(arg1,arg2,(int const &)*arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector<int >::value_type temp2 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_append" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int_append" "', argument " "2"" of type '" "std::vector<int >::value_type""'");
  } 
  temp2 = static_cast< std::vector<int >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      std_vector_Sl_int_Sg__append(arg1,(int const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_int__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_int")) SWIG_fail;
  {
    try {
      result = (std::vector<int > *)new std::vector<int >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_int__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = 0 ;
  std::vector<int > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_int",&obj0)) SWIG_fail;
  {
    std::vector<int,std::allocator<int > > *ptr = (std::vector<int,std::allocator<int > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_int" "', argument " "1"" of type '" "std::vector<int > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_int" "', argument " "1"" of type '" "std::vector<int > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::vector<int > *)new std::vector<int >((std::vector<int > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_empty" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (bool)((std::vector<int > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_size" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_clear" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_swap" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_int_swap" "', argument " "2"" of type '" "std::vector<int > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_int_swap" "', argument " "2"" of type '" "std::vector<int > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<int > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  SwigValueWrapper<std::allocator<int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_get_allocator" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::vector<int >::allocator_type(static_cast< const std::vector<int >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_begin" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_begin" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_end" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_end" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_rbegin" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_rbegin" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_rend" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_rend" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_int_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_int__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int >::size_type arg1 ;
  std::vector<int > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_int",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_int" "', argument " "1"" of type '" "std::vector<int >::size_type""'");
  } 
  arg1 = static_cast< std::vector<int >::size_type >(val1);
  {
    try {
      result = (std::vector<int > *)new std::vector<int >(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_pop_back" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      (arg1)->pop_back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_resize" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int_resize" "', argument " "2"" of type '" "std::vector<int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<int >::size_type >(val2);
  {
    try {
      (arg1)->resize(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::iterator arg2 ;
  std::vector<int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_erase" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_erase" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_erase" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::iterator arg2 ;
  std::vector<int >::iterator arg3 ;
  std::vector<int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_erase" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_erase" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_erase" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_erase" "', argument " "3"" of type '" "std::vector<int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_erase" "', argument " "3"" of type '" "std::vector<int >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_int_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_int_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<int >::iterator)\n    erase(std::vector<int >::iterator,std::vector<int >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_int__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int >::size_type arg1 ;
  std::vector<int >::value_type *arg2 = 0 ;
  std::vector<int > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector<int >::value_type temp2 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_int",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_int" "', argument " "1"" of type '" "std::vector<int >::size_type""'");
  } 
  arg1 = static_cast< std::vector<int >::size_type >(val1);
  ecode2 = SWIG_AsVal_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vector_int" "', argument " "2"" of type '" "std::vector<int >::value_type""'");
  } 
  temp2 = static_cast< std::vector<int >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      result = (std::vector<int > *)new std::vector<int >(arg1,(std::vector<int >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_int(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_int__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_int__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_int__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_vector_int__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_int'.\n  Possible C/C++ prototypes are:\n    std::vector<(int)>()\n    std::vector<(int)>(std::vector<int > const &)\n    std::vector<(int)>(std::vector<int >::size_type)\n    std::vector<(int)>(std::vector<int >::size_type,std::vector<int >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector<int >::value_type temp2 ;
  int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_push_back" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int_push_back" "', argument " "2"" of type '" "std::vector<int >::value_type""'");
  } 
  temp2 = static_cast< std::vector<int >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      (arg1)->push_back((std::vector<int >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_front" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      {
        std::vector<int >::value_type const &_result_ref = ((std::vector<int > const *)arg1)->front();
        result = (std::vector<int >::value_type *) &_result_ref;
      }
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_back" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      {
        std::vector<int >::value_type const &_result_ref = ((std::vector<int > const *)arg1)->back();
        result = (std::vector<int >::value_type *) &_result_ref;
      }
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type arg2 ;
  std::vector<int >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector<int >::value_type temp3 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_assign" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int_assign" "', argument " "2"" of type '" "std::vector<int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<int >::size_type >(val2);
  ecode3 = SWIG_AsVal_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int_assign" "', argument " "3"" of type '" "std::vector<int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      (arg1)->assign(arg2,(std::vector<int >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type arg2 ;
  std::vector<int >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector<int >::value_type temp3 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_resize" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int_resize" "', argument " "2"" of type '" "std::vector<int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<int >::size_type >(val2);
  ecode3 = SWIG_AsVal_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int_resize" "', argument " "3"" of type '" "std::vector<int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      (arg1)->resize(arg2,(std::vector<int >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_int_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_int_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<int >::size_type)\n    resize(std::vector<int >::size_type,std::vector<int >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::iterator arg2 ;
  std::vector<int >::value_type *arg3 = 0 ;
  std::vector<int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  std::vector<int >::value_type temp3 ;
  int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_int_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_insert" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_insert" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_insert" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int_insert" "', argument " "3"" of type '" "std::vector<int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      result = (arg1)->insert(arg2,(std::vector<int >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::iterator arg2 ;
  std::vector<int >::size_type arg3 ;
  std::vector<int >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  std::vector<int >::value_type temp4 ;
  int val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_int_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_insert" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_insert" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_int_insert" "', argument " "2"" of type '" "std::vector<int >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_int_insert" "', argument " "3"" of type '" "std::vector<int >::size_type""'");
  } 
  arg3 = static_cast< std::vector<int >::size_type >(val3);
  ecode4 = SWIG_AsVal_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vector_int_insert" "', argument " "4"" of type '" "std::vector<int >::value_type""'");
  } 
  temp4 = static_cast< std::vector<int >::value_type >(val4);
  arg4 = &temp4;
  {
    try {
      (arg1)->insert(arg2,arg3,(std::vector<int >::value_type const &)*arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_int_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<int,std::allocator<int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<int >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_vector_int_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_int_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<int >::iterator,std::vector<int >::value_type const &)\n    insert(std::vector<int >::iterator,std::vector<int >::size_type,std::vector<int >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_int_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_reserve" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_int_reserve" "', argument " "2"" of type '" "std::vector<int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<int >::size_type >(val2);
  {
    try {
      (arg1)->reserve(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_int_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  std::vector<int >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_int_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_int_capacity" "', argument " "1"" of type '" "std::vector<int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      result = ((std::vector<int > const *)arg1)->capacity();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_int(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<int > *arg1 = (std::vector<int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_int",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_int" "', argument " "1"" of type '" "std::vector<int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<int > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_int_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTint_std__allocatorTint_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_uint_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_iterator" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_vector_Sl_unsigned_SS_int_Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___nonzero__" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (bool)std_vector_Sl_unsigned_SS_int_Sg____nonzero__((std::vector<unsigned int > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___len__" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = std_vector_Sl_unsigned_SS_int_Sg____len__((std::vector<unsigned int > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_pop" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      try {
        result = (std::vector<unsigned int >::value_type)std_vector_Sl_unsigned_SS_int_Sg__pop(arg1);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::difference_type arg2 ;
  std::vector<unsigned int >::difference_type arg3 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___getslice__" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint___getslice__" "', argument " "2"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint___getslice__" "', argument " "3"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<unsigned int >::difference_type >(val3);
  {
    try {
      try {
        result = (std::vector<unsigned int,std::allocator<unsigned int > > *)std_vector_Sl_unsigned_SS_int_Sg____getslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::difference_type arg2 ;
  std::vector<unsigned int >::difference_type arg3 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_uint___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___setslice__" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint___setslice__" "', argument " "2"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint___setslice__" "', argument " "3"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<unsigned int >::difference_type >(val3);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_uint___setslice__" "', argument " "4"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_uint___setslice__" "', argument " "4"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    try {
      try {
        std_vector_Sl_unsigned_SS_int_Sg____setslice__(arg1,arg2,arg3,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg4);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      catch(std::invalid_argument &_e) {
        SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::difference_type arg2 ;
  std::vector<unsigned int >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___delslice__" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint___delslice__" "', argument " "2"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint___delslice__" "', argument " "3"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<unsigned int >::difference_type >(val3);
  {
    try {
      try {
        std_vector_Sl_unsigned_SS_int_Sg____delslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___delitem__" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint___delitem__" "', argument " "2"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::difference_type >(val2);
  {
    try {
      try {
        std_vector_Sl_unsigned_SS_int_Sg____delitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::difference_type arg2 ;
  std::vector<unsigned int >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___getitem__" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint___getitem__" "', argument " "2"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::difference_type >(val2);
  {
    try {
      try {
        {
          std::vector<unsigned int >::value_type const &_result_ref = std_vector_Sl_unsigned_SS_int_Sg____getitem__((std::vector<unsigned int > const *)arg1,arg2);
          result = (std::vector<unsigned int >::value_type *) &_result_ref;
        }
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::difference_type arg2 ;
  std::vector<unsigned int >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector<unsigned int >::value_type temp3 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint___setitem__" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint___setitem__" "', argument " "2"" of type '" "std::vector<unsigned int >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::difference_type >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint___setitem__" "', argument " "3"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<unsigned int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      try {
        std_vector_Sl_unsigned_SS_int_Sg____setitem__(arg1,arg2,(unsigned int const &)*arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector<unsigned int >::value_type temp2 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_append" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint_append" "', argument " "2"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp2 = static_cast< std::vector<unsigned int >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      std_vector_Sl_unsigned_SS_int_Sg__append(arg1,(unsigned int const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_uint__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_uint")) SWIG_fail;
  {
    try {
      result = (std::vector<unsigned int > *)new std::vector<unsigned int >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_uint__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = 0 ;
  std::vector<unsigned int > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_uint",&obj0)) SWIG_fail;
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_uint" "', argument " "1"" of type '" "std::vector<unsigned int > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_uint" "', argument " "1"" of type '" "std::vector<unsigned int > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::vector<unsigned int > *)new std::vector<unsigned int >((std::vector<unsigned int > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_empty" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (bool)((std::vector<unsigned int > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_size" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_clear" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_swap" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_uint_swap" "', argument " "2"" of type '" "std::vector<unsigned int > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_uint_swap" "', argument " "2"" of type '" "std::vector<unsigned int > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<unsigned int > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  SwigValueWrapper<std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_get_allocator" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::vector<unsigned int >::allocator_type(static_cast< const std::vector<unsigned int >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_begin" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_begin" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_end" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_end" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_rbegin" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_rbegin" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_rend" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_rend" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_uint_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_uint__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int >::size_type arg1 ;
  std::vector<unsigned int > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_uint",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_uint" "', argument " "1"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg1 = static_cast< std::vector<unsigned int >::size_type >(val1);
  {
    try {
      result = (std::vector<unsigned int > *)new std::vector<unsigned int >(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_pop_back" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      (arg1)->pop_back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_resize" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint_resize" "', argument " "2"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::size_type >(val2);
  {
    try {
      (arg1)->resize(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::iterator arg2 ;
  std::vector<unsigned int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_erase" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_erase" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_erase" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::iterator arg2 ;
  std::vector<unsigned int >::iterator arg3 ;
  std::vector<unsigned int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_erase" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_erase" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_erase" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_erase" "', argument " "3"" of type '" "std::vector<unsigned int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_erase" "', argument " "3"" of type '" "std::vector<unsigned int >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_uint_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_uint_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<unsigned int >::iterator)\n    erase(std::vector<unsigned int >::iterator,std::vector<unsigned int >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_uint__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int >::size_type arg1 ;
  std::vector<unsigned int >::value_type *arg2 = 0 ;
  std::vector<unsigned int > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector<unsigned int >::value_type temp2 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_uint",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_uint" "', argument " "1"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg1 = static_cast< std::vector<unsigned int >::size_type >(val1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vector_uint" "', argument " "2"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp2 = static_cast< std::vector<unsigned int >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      result = (std::vector<unsigned int > *)new std::vector<unsigned int >(arg1,(std::vector<unsigned int >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_uint(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_uint__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_uint__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_uint__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_vector_uint__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_uint'.\n  Possible C/C++ prototypes are:\n    std::vector<(unsigned int)>()\n    std::vector<(unsigned int)>(std::vector<unsigned int > const &)\n    std::vector<(unsigned int)>(std::vector<unsigned int >::size_type)\n    std::vector<(unsigned int)>(std::vector<unsigned int >::size_type,std::vector<unsigned int >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector<unsigned int >::value_type temp2 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_push_back" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint_push_back" "', argument " "2"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp2 = static_cast< std::vector<unsigned int >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      (arg1)->push_back((std::vector<unsigned int >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_front" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      {
        std::vector<unsigned int >::value_type const &_result_ref = ((std::vector<unsigned int > const *)arg1)->front();
        result = (std::vector<unsigned int >::value_type *) &_result_ref;
      }
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_back" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      {
        std::vector<unsigned int >::value_type const &_result_ref = ((std::vector<unsigned int > const *)arg1)->back();
        result = (std::vector<unsigned int >::value_type *) &_result_ref;
      }
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type arg2 ;
  std::vector<unsigned int >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector<unsigned int >::value_type temp3 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_assign" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint_assign" "', argument " "2"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::size_type >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint_assign" "', argument " "3"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<unsigned int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      (arg1)->assign(arg2,(std::vector<unsigned int >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type arg2 ;
  std::vector<unsigned int >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector<unsigned int >::value_type temp3 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_resize" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint_resize" "', argument " "2"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::size_type >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint_resize" "', argument " "3"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<unsigned int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      (arg1)->resize(arg2,(std::vector<unsigned int >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_uint_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_uint_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<unsigned int >::size_type)\n    resize(std::vector<unsigned int >::size_type,std::vector<unsigned int >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::iterator arg2 ;
  std::vector<unsigned int >::value_type *arg3 = 0 ;
  std::vector<unsigned int >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  std::vector<unsigned int >::value_type temp3 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_uint_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_insert" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_insert" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_insert" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint_insert" "', argument " "3"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp3 = static_cast< std::vector<unsigned int >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      result = (arg1)->insert(arg2,(std::vector<unsigned int >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<unsigned int >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::iterator arg2 ;
  std::vector<unsigned int >::size_type arg3 ;
  std::vector<unsigned int >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  std::vector<unsigned int >::value_type temp4 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_uint_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_insert" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_insert" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_uint_insert" "', argument " "2"" of type '" "std::vector<unsigned int >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_uint_insert" "', argument " "3"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg3 = static_cast< std::vector<unsigned int >::size_type >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vector_uint_insert" "', argument " "4"" of type '" "std::vector<unsigned int >::value_type""'");
  } 
  temp4 = static_cast< std::vector<unsigned int >::value_type >(val4);
  arg4 = &temp4;
  {
    try {
      (arg1)->insert(arg2,arg3,(std::vector<unsigned int >::value_type const &)*arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_uint_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<unsigned int >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_unsigned_SS_int(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_vector_uint_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_uint_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<unsigned int >::iterator,std::vector<unsigned int >::value_type const &)\n    insert(std::vector<unsigned int >::iterator,std::vector<unsigned int >::size_type,std::vector<unsigned int >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_uint_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_reserve" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_uint_reserve" "', argument " "2"" of type '" "std::vector<unsigned int >::size_type""'");
  } 
  arg2 = static_cast< std::vector<unsigned int >::size_type >(val2);
  {
    try {
      (arg1)->reserve(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_uint_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  std::vector<unsigned int >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_uint_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_uint_capacity" "', argument " "1"" of type '" "std::vector<unsigned int > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      result = ((std::vector<unsigned int > const *)arg1)->capacity();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_uint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<unsigned int > *arg1 = (std::vector<unsigned int > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_uint",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_uint" "', argument " "1"" of type '" "std::vector<unsigned int > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<unsigned int > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_uint_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_dbl_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_iterator" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_vector_Sl_double_Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___nonzero__" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (bool)std_vector_Sl_double_Sg____nonzero__((std::vector<double > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___len__" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = std_vector_Sl_double_Sg____len__((std::vector<double > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_pop" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      try {
        result = (std::vector<double >::value_type)std_vector_Sl_double_Sg__pop(arg1);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::difference_type arg2 ;
  std::vector<double >::difference_type arg3 ;
  std::vector<double,std::allocator<double > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___getslice__" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl___getslice__" "', argument " "2"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl___getslice__" "', argument " "3"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<double >::difference_type >(val3);
  {
    try {
      try {
        result = (std::vector<double,std::allocator<double > > *)std_vector_Sl_double_Sg____getslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::difference_type arg2 ;
  std::vector<double >::difference_type arg3 ;
  std::vector<double,std::allocator<double > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_dbl___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___setslice__" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl___setslice__" "', argument " "2"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl___setslice__" "', argument " "3"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<double >::difference_type >(val3);
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_dbl___setslice__" "', argument " "4"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_dbl___setslice__" "', argument " "4"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    try {
      try {
        std_vector_Sl_double_Sg____setslice__(arg1,arg2,arg3,(std::vector<double,std::allocator<double > > const &)*arg4);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      catch(std::invalid_argument &_e) {
        SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::difference_type arg2 ;
  std::vector<double >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___delslice__" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl___delslice__" "', argument " "2"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl___delslice__" "', argument " "3"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<double >::difference_type >(val3);
  {
    try {
      try {
        std_vector_Sl_double_Sg____delslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___delitem__" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl___delitem__" "', argument " "2"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<double >::difference_type >(val2);
  {
    try {
      try {
        std_vector_Sl_double_Sg____delitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::difference_type arg2 ;
  std::vector<double >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___getitem__" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl___getitem__" "', argument " "2"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<double >::difference_type >(val2);
  {
    try {
      try {
        {
          std::vector<double >::value_type const &_result_ref = std_vector_Sl_double_Sg____getitem__((std::vector<double > const *)arg1,arg2);
          result = (std::vector<double >::value_type *) &_result_ref;
        }
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::difference_type arg2 ;
  std::vector<double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  std::vector<double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl___setitem__" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl___setitem__" "', argument " "2"" of type '" "std::vector<double >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<double >::difference_type >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl___setitem__" "', argument " "3"" of type '" "std::vector<double >::value_type""'");
  } 
  temp3 = static_cast< std::vector<double >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      try {
        std_vector_Sl_double_Sg____setitem__(arg1,arg2,(double const &)*arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector<double >::value_type temp2 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_append" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl_append" "', argument " "2"" of type '" "std::vector<double >::value_type""'");
  } 
  temp2 = static_cast< std::vector<double >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      std_vector_Sl_double_Sg__append(arg1,(double const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dbl__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_dbl")) SWIG_fail;
  {
    try {
      result = (std::vector<double > *)new std::vector<double >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dbl__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = 0 ;
  std::vector<double > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_dbl",&obj0)) SWIG_fail;
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_dbl" "', argument " "1"" of type '" "std::vector<double > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_dbl" "', argument " "1"" of type '" "std::vector<double > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::vector<double > *)new std::vector<double >((std::vector<double > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_empty" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (bool)((std::vector<double > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_size" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_clear" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_swap" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_dbl_swap" "', argument " "2"" of type '" "std::vector<double > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_dbl_swap" "', argument " "2"" of type '" "std::vector<double > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<double > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  SwigValueWrapper<std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_get_allocator" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::vector<double >::allocator_type(static_cast< const std::vector<double >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_begin" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_begin" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_end" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_end" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_rbegin" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_rbegin" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_rend" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_rend" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_dbl_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dbl__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double >::size_type arg1 ;
  std::vector<double > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_dbl",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_dbl" "', argument " "1"" of type '" "std::vector<double >::size_type""'");
  } 
  arg1 = static_cast< std::vector<double >::size_type >(val1);
  {
    try {
      result = (std::vector<double > *)new std::vector<double >(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_pop_back" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      (arg1)->pop_back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_resize" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl_resize" "', argument " "2"" of type '" "std::vector<double >::size_type""'");
  } 
  arg2 = static_cast< std::vector<double >::size_type >(val2);
  {
    try {
      (arg1)->resize(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::iterator arg2 ;
  std::vector<double >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_erase" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_erase" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<double >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_erase" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::iterator arg2 ;
  std::vector<double >::iterator arg3 ;
  std::vector<double >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_erase" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_erase" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<double >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_erase" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_erase" "', argument " "3"" of type '" "std::vector<double >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<double >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_erase" "', argument " "3"" of type '" "std::vector<double >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_dbl_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_dbl_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<double >::iterator)\n    erase(std::vector<double >::iterator,std::vector<double >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dbl__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double >::size_type arg1 ;
  std::vector<double >::value_type *arg2 = 0 ;
  std::vector<double > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  std::vector<double >::value_type temp2 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_dbl",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_dbl" "', argument " "1"" of type '" "std::vector<double >::size_type""'");
  } 
  arg1 = static_cast< std::vector<double >::size_type >(val1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vector_dbl" "', argument " "2"" of type '" "std::vector<double >::value_type""'");
  } 
  temp2 = static_cast< std::vector<double >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      result = (std::vector<double > *)new std::vector<double >(arg1,(std::vector<double >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_dbl(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_dbl__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_dbl__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_dbl__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_double(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_vector_dbl__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_dbl'.\n  Possible C/C++ prototypes are:\n    std::vector<(double)>()\n    std::vector<(double)>(std::vector<double > const &)\n    std::vector<(double)>(std::vector<double >::size_type)\n    std::vector<(double)>(std::vector<double >::size_type,std::vector<double >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::value_type *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  std::vector<double >::value_type temp2 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_push_back" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl_push_back" "', argument " "2"" of type '" "std::vector<double >::value_type""'");
  } 
  temp2 = static_cast< std::vector<double >::value_type >(val2);
  arg2 = &temp2;
  {
    try {
      (arg1)->push_back((std::vector<double >::value_type const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_front" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      {
        std::vector<double >::value_type const &_result_ref = ((std::vector<double > const *)arg1)->front();
        result = (std::vector<double >::value_type *) &_result_ref;
      }
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::value_type *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_back" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      {
        std::vector<double >::value_type const &_result_ref = ((std::vector<double > const *)arg1)->back();
        result = (std::vector<double >::value_type *) &_result_ref;
      }
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(*result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type arg2 ;
  std::vector<double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector<double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_assign" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl_assign" "', argument " "2"" of type '" "std::vector<double >::size_type""'");
  } 
  arg2 = static_cast< std::vector<double >::size_type >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl_assign" "', argument " "3"" of type '" "std::vector<double >::value_type""'");
  } 
  temp3 = static_cast< std::vector<double >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      (arg1)->assign(arg2,(std::vector<double >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type arg2 ;
  std::vector<double >::value_type *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  std::vector<double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_resize" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl_resize" "', argument " "2"" of type '" "std::vector<double >::size_type""'");
  } 
  arg2 = static_cast< std::vector<double >::size_type >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl_resize" "', argument " "3"" of type '" "std::vector<double >::value_type""'");
  } 
  temp3 = static_cast< std::vector<double >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      (arg1)->resize(arg2,(std::vector<double >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_dbl_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_dbl_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<double >::size_type)\n    resize(std::vector<double >::size_type,std::vector<double >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::iterator arg2 ;
  std::vector<double >::value_type *arg3 = 0 ;
  std::vector<double >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  std::vector<double >::value_type temp3 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_dbl_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_insert" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_insert" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<double >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_insert" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl_insert" "', argument " "3"" of type '" "std::vector<double >::value_type""'");
  } 
  temp3 = static_cast< std::vector<double >::value_type >(val3);
  arg3 = &temp3;
  {
    try {
      result = (arg1)->insert(arg2,(std::vector<double >::value_type const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<double >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::iterator arg2 ;
  std::vector<double >::size_type arg3 ;
  std::vector<double >::value_type *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  std::vector<double >::value_type temp4 ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_dbl_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_insert" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_insert" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<double >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_dbl_insert" "', argument " "2"" of type '" "std::vector<double >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_dbl_insert" "', argument " "3"" of type '" "std::vector<double >::size_type""'");
  } 
  arg3 = static_cast< std::vector<double >::size_type >(val3);
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vector_dbl_insert" "', argument " "4"" of type '" "std::vector<double >::value_type""'");
  } 
  temp4 = static_cast< std::vector<double >::value_type >(val4);
  arg4 = &temp4;
  {
    try {
      (arg1)->insert(arg2,arg3,(std::vector<double >::value_type const &)*arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_double(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_dbl_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<double >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_double(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_vector_dbl_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_dbl_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<double >::iterator,std::vector<double >::value_type const &)\n    insert(std::vector<double >::iterator,std::vector<double >::size_type,std::vector<double >::value_type const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_dbl_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_reserve" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_dbl_reserve" "', argument " "2"" of type '" "std::vector<double >::size_type""'");
  } 
  arg2 = static_cast< std::vector<double >::size_type >(val2);
  {
    try {
      (arg1)->reserve(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_dbl_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  std::vector<double >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_dbl_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_dbl_capacity" "', argument " "1"" of type '" "std::vector<double > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      result = ((std::vector<double > const *)arg1)->capacity();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_dbl(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double > *arg1 = (std::vector<double > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_dbl",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_dbl" "', argument " "1"" of type '" "std::vector<double > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<double > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_dbl_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTdouble_std__allocatorTdouble_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_vector_bool_iterator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  PyObject **arg2 = (PyObject **) 0 ;
  swig::PySwigIterator *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  arg2 = &obj0;
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_iterator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_iterator" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (swig::PySwigIterator *)std_vector_Sl_bool_Sg__iterator(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_swig__PySwigIterator, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___nonzero__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool___nonzero__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___nonzero__" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (bool)std_vector_Sl_bool_Sg____nonzero__((std::vector<bool > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___len__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool___len__",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___len__" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = std_vector_Sl_bool_Sg____len__((std::vector<bool > const *)arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_pop(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_pop",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_pop" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      try {
        result = (std::vector<bool >::value_type)std_vector_Sl_bool_Sg__pop(arg1);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___getslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::difference_type arg2 ;
  std::vector<bool >::difference_type arg3 ;
  std::vector<bool,std::allocator<bool > > *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool___getslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___getslice__" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool___getslice__" "', argument " "2"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool___getslice__" "', argument " "3"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::difference_type >(val3);
  {
    try {
      try {
        result = (std::vector<bool,std::allocator<bool > > *)std_vector_Sl_bool_Sg____getslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___setslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::difference_type arg2 ;
  std::vector<bool >::difference_type arg3 ;
  std::vector<bool,std::allocator<bool > > *arg4 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  int res4 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_bool___setslice__",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___setslice__" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool___setslice__" "', argument " "2"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool___setslice__" "', argument " "3"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::difference_type >(val3);
  {
    std::vector<bool,std::allocator<bool > > *ptr = (std::vector<bool,std::allocator<bool > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "vector_bool___setslice__" "', argument " "4"" of type '" "std::vector<bool,std::allocator<bool > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_bool___setslice__" "', argument " "4"" of type '" "std::vector<bool,std::allocator<bool > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    try {
      try {
        std_vector_Sl_bool_Sg____setslice__(arg1,arg2,arg3,(std::vector<bool,std::allocator<bool > > const &)*arg4);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      catch(std::invalid_argument &_e) {
        SWIG_exception_fail(SWIG_ValueError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res4)) delete arg4;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res4)) delete arg4;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___delslice__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::difference_type arg2 ;
  std::vector<bool >::difference_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  ptrdiff_t val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool___delslice__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___delslice__" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool___delslice__" "', argument " "2"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::difference_type >(val2);
  ecode3 = SWIG_AsVal_ptrdiff_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool___delslice__" "', argument " "3"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::difference_type >(val3);
  {
    try {
      try {
        std_vector_Sl_bool_Sg____delslice__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___delitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::difference_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool___delitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___delitem__" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool___delitem__" "', argument " "2"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::difference_type >(val2);
  {
    try {
      try {
        std_vector_Sl_bool_Sg____delitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___getitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::difference_type arg2 ;
  std::vector<bool >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool___getitem__",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___getitem__" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool___getitem__" "', argument " "2"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::difference_type >(val2);
  {
    try {
      try {
        result = (std::vector<bool >::value_type)std_vector_Sl_bool_Sg____getitem__(arg1,arg2);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool___setitem__(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::difference_type arg2 ;
  std::vector<bool >::value_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  ptrdiff_t val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool___setitem__",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool___setitem__" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_ptrdiff_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool___setitem__" "', argument " "2"" of type '" "std::vector<bool >::difference_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::difference_type >(val2);
  ecode3 = SWIG_AsVal_bool(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool___setitem__" "', argument " "3"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::value_type >(val3);
  {
    try {
      try {
        std_vector_Sl_bool_Sg____setitem__(arg1,arg2,arg3);
      }
      catch(std::out_of_range &_e) {
        SWIG_exception_fail(SWIG_IndexError, (&_e)->what());
      }
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_append(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool_append",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_append" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool_append" "', argument " "2"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::value_type >(val2);
  {
    try {
      std_vector_Sl_bool_Sg__append(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_bool__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_vector_bool")) SWIG_fail;
  {
    try {
      result = (std::vector<bool > *)new std::vector<bool >();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_bool__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = 0 ;
  std::vector<bool > *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_bool",&obj0)) SWIG_fail;
  {
    std::vector<bool,std::allocator<bool > > *ptr = (std::vector<bool,std::allocator<bool > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_vector_bool" "', argument " "1"" of type '" "std::vector<bool > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_vector_bool" "', argument " "1"" of type '" "std::vector<bool > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    try {
      result = (std::vector<bool > *)new std::vector<bool >((std::vector<bool > const &)*arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_empty(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_empty",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_empty" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (bool)((std::vector<bool > const *)arg1)->empty();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_size(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_size",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_size" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->size();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_clear(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_clear",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_clear" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      (arg1)->clear();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_swap(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool > *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool_swap",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_swap" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2, SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t,  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "vector_bool_swap" "', argument " "2"" of type '" "std::vector<bool > &""'"); 
  }
  if (!argp2) {
    SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "vector_bool_swap" "', argument " "2"" of type '" "std::vector<bool > &""'"); 
  }
  arg2 = reinterpret_cast< std::vector<bool > * >(argp2);
  {
    try {
      (arg1)->swap(*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_get_allocator(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  SwigValueWrapper<std::allocator<bool > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_get_allocator",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_get_allocator" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->get_allocator();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj((new std::vector<bool >::allocator_type(static_cast< const std::vector<bool >::allocator_type& >(result))), SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type, SWIG_POINTER_OWN |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_begin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_begin" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_begin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_begin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_begin" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->begin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_begin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_begin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_begin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_begin'.\n  Possible C/C++ prototypes are:\n    begin()\n    begin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_end__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_end" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_end__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::const_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_end",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_end" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->end();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::const_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_end(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_end__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_end__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_end'.\n  Possible C/C++ prototypes are:\n    end()\n    end()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_rbegin__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_rbegin" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_rbegin__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_rbegin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_rbegin" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->rbegin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_rbegin(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_rbegin__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_rbegin__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_rbegin'.\n  Possible C/C++ prototypes are:\n    rbegin()\n    rbegin()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_rend__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_rend" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_rend__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::const_reverse_iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_rend",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_rend" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->rend();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::const_reverse_iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_rend(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[2];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 1); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_rend__SWIG_0(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_vector_bool_rend__SWIG_1(self, args);
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_rend'.\n  Possible C/C++ prototypes are:\n    rend()\n    rend()\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_bool__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool >::size_type arg1 ;
  std::vector<bool > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:new_vector_bool",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_bool" "', argument " "1"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg1 = static_cast< std::vector<bool >::size_type >(val1);
  {
    try {
      result = (std::vector<bool > *)new std::vector<bool >(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_pop_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_pop_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_pop_back" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      (arg1)->pop_back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_resize__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool_resize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_resize" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool_resize" "', argument " "2"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::size_type >(val2);
  {
    try {
      (arg1)->resize(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_erase__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::iterator arg2 ;
  std::vector<bool >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool_erase",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_erase" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_erase" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<bool >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_erase" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_erase__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::iterator arg2 ;
  std::vector<bool >::iterator arg3 ;
  std::vector<bool >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  swig::PySwigIterator *iter3 = 0 ;
  int res3 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool_erase",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_erase" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_erase" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<bool >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_erase" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
    }
  }
  res3 = SWIG_ConvertPtr(obj2, SWIG_as_voidptrptr(&iter3), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res3) || !iter3) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_erase" "', argument " "3"" of type '" "std::vector<bool >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<bool >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter3);
    if (iter_t) {
      arg3 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_erase" "', argument " "3"" of type '" "std::vector<bool >::iterator""'");
    }
  }
  {
    try {
      result = (arg1)->erase(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_erase(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter) != 0));
      if (_v) {
        return _wrap_vector_bool_erase__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter) != 0));
      if (_v) {
        swig::PySwigIterator *iter = 0;
        int res = SWIG_ConvertPtr(argv[2], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
        _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter) != 0));
        if (_v) {
          return _wrap_vector_bool_erase__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_erase'.\n  Possible C/C++ prototypes are:\n    erase(std::vector<bool >::iterator)\n    erase(std::vector<bool >::iterator,std::vector<bool >::iterator)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_bool__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool >::size_type arg1 ;
  std::vector<bool >::value_type arg2 ;
  std::vector<bool > *result = 0 ;
  size_t val1 ;
  int ecode1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_vector_bool",&obj0,&obj1)) SWIG_fail;
  ecode1 = SWIG_AsVal_size_t(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_vector_bool" "', argument " "1"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg1 = static_cast< std::vector<bool >::size_type >(val1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_vector_bool" "', argument " "2"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::value_type >(val2);
  {
    try {
      result = (std::vector<bool > *)new std::vector<bool >(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_vector_bool(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[3];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 2); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 0) {
    return _wrap_new_vector_bool__SWIG_0(self, args);
  }
  if (argc == 1) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      return _wrap_new_vector_bool__SWIG_2(self, args);
    }
  }
  if (argc == 1) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      return _wrap_new_vector_bool__SWIG_1(self, args);
    }
  }
  if (argc == 2) {
    int _v;
    {
      int res = SWIG_AsVal_size_t(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_bool(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_new_vector_bool__SWIG_3(self, args);
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_vector_bool'.\n  Possible C/C++ prototypes are:\n    std::vector<(bool)>()\n    std::vector<(bool)>(std::vector<bool > const &)\n    std::vector<(bool)>(std::vector<bool >::size_type)\n    std::vector<(bool)>(std::vector<bool >::size_type,std::vector<bool >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_push_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::value_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  bool val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool_push_back",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_push_back" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_bool(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool_push_back" "', argument " "2"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::value_type >(val2);
  {
    try {
      (arg1)->push_back(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_front(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_front",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_front" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (std::vector<bool >::value_type)((std::vector<bool > const *)arg1)->front();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_back(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::value_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_back",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_back" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = (std::vector<bool >::value_type)((std::vector<bool > const *)arg1)->back();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_assign(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type arg2 ;
  std::vector<bool >::value_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool_assign",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_assign" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool_assign" "', argument " "2"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::size_type >(val2);
  ecode3 = SWIG_AsVal_bool(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool_assign" "', argument " "3"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::value_type >(val3);
  {
    try {
      (arg1)->assign(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_resize__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type arg2 ;
  std::vector<bool >::value_type arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool_resize",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_resize" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool_resize" "', argument " "2"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::size_type >(val2);
  ecode3 = SWIG_AsVal_bool(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool_resize" "', argument " "3"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::value_type >(val3);
  {
    try {
      (arg1)->resize(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_resize(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[4];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 3); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        return _wrap_vector_bool_resize__SWIG_0(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      {
        int res = SWIG_AsVal_size_t(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_bool_resize__SWIG_1(self, args);
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_resize'.\n  Possible C/C++ prototypes are:\n    resize(std::vector<bool >::size_type)\n    resize(std::vector<bool >::size_type,std::vector<bool >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_insert__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::iterator arg2 ;
  std::vector<bool >::value_type arg3 ;
  std::vector<bool >::iterator result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  bool val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:vector_bool_insert",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_insert" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_insert" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<bool >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_insert" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_bool(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool_insert" "', argument " "3"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::value_type >(val3);
  {
    try {
      result = (arg1)->insert(arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(swig::make_output_iterator(static_cast< const std::vector<bool >::iterator & >(result)),
    swig::PySwigIterator::descriptor(),SWIG_POINTER_OWN);
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_insert__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::iterator arg2 ;
  std::vector<bool >::size_type arg3 ;
  std::vector<bool >::value_type arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  swig::PySwigIterator *iter2 = 0 ;
  int res2 ;
  size_t val3 ;
  int ecode3 = 0 ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:vector_bool_insert",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_insert" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, SWIG_as_voidptrptr(&iter2), swig::PySwigIterator::descriptor(), 0);
  if (!SWIG_IsOK(res2) || !iter2) {
    SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_insert" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
  } else {
    swig::PySwigIterator_T<std::vector<bool >::iterator > *iter_t = dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter2);
    if (iter_t) {
      arg2 = iter_t->get_current();
    } else {
      SWIG_exception_fail(SWIG_ArgError(SWIG_TypeError), "in method '" "vector_bool_insert" "', argument " "2"" of type '" "std::vector<bool >::iterator""'");
    }
  }
  ecode3 = SWIG_AsVal_size_t(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "vector_bool_insert" "', argument " "3"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg3 = static_cast< std::vector<bool >::size_type >(val3);
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "vector_bool_insert" "', argument " "4"" of type '" "std::vector<bool >::value_type""'");
  } 
  arg4 = static_cast< std::vector<bool >::value_type >(val4);
  {
    try {
      (arg1)->insert(arg2,arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_insert(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[5];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 4); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_bool(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_vector_bool_insert__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 4) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<bool,std::allocator<bool > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      swig::PySwigIterator *iter = 0;
      int res = SWIG_ConvertPtr(argv[1], SWIG_as_voidptrptr(&iter), swig::PySwigIterator::descriptor(), 0);
      _v = (SWIG_IsOK(res) && iter && (dynamic_cast<swig::PySwigIterator_T<std::vector<bool >::iterator > *>(iter) != 0));
      if (_v) {
        {
          int res = SWIG_AsVal_size_t(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          {
            int res = SWIG_AsVal_bool(argv[3], NULL);
            _v = SWIG_CheckState(res);
          }
          if (_v) {
            return _wrap_vector_bool_insert__SWIG_1(self, args);
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'vector_bool_insert'.\n  Possible C/C++ prototypes are:\n    insert(std::vector<bool >::iterator,std::vector<bool >::value_type)\n    insert(std::vector<bool >::iterator,std::vector<bool >::size_type,std::vector<bool >::value_type)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_reserve(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  size_t val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:vector_bool_reserve",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_reserve" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  ecode2 = SWIG_AsVal_size_t(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "vector_bool_reserve" "', argument " "2"" of type '" "std::vector<bool >::size_type""'");
  } 
  arg2 = static_cast< std::vector<bool >::size_type >(val2);
  {
    try {
      (arg1)->reserve(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_vector_bool_capacity(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  std::vector<bool >::size_type result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:vector_bool_capacity",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "vector_bool_capacity" "', argument " "1"" of type '" "std::vector<bool > const *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      result = ((std::vector<bool > const *)arg1)->capacity();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_size_t(static_cast< size_t >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_vector_bool(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<bool > *arg1 = (std::vector<bool > *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_vector_bool",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_vector_bool" "', argument " "1"" of type '" "std::vector<bool > *""'"); 
  }
  arg1 = reinterpret_cast< std::vector<bool > * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *vector_bool_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_std__vectorTbool_std__allocatorTbool_t_t, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Geom(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *result = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)":new_Geom")) SWIG_fail;
  {
    try {
      result = (steps::wm::Geom *)new steps::wm::Geom();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Geom, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Geom(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Geom",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Geom" "', argument " "1"" of type '" "steps::wm::Geom *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Geom_getComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  std::string *arg2 = 0 ;
  steps::wm::Comp *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Geom_getComp",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Geom_getComp" "', argument " "1"" of type '" "steps::wm::Geom const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Geom_getComp" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Geom_getComp" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::wm::Comp *)((steps::wm::Geom const *)arg1)->getComp((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Geom_delComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Geom_delComp",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Geom_delComp" "', argument " "1"" of type '" "steps::wm::Geom *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Geom_delComp" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Geom_delComp" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delComp((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Geom_getAllComps(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Geom_getAllComps",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Geom_getAllComps" "', argument " "1"" of type '" "steps::wm::Geom const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    try {
      result = ((steps::wm::Geom const *)arg1)->getAllComps();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Geom_getPatch(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  std::string *arg2 = 0 ;
  steps::wm::Patch *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Geom_getPatch",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Geom_getPatch" "', argument " "1"" of type '" "steps::wm::Geom const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Geom_getPatch" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Geom_getPatch" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::wm::Patch *)((steps::wm::Geom const *)arg1)->getPatch((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Geom_delPatch(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Geom_delPatch",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Geom_delPatch" "', argument " "1"" of type '" "steps::wm::Geom *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Geom_delPatch" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Geom_delPatch" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delPatch((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Geom_getAllPatches(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Geom *arg1 = (steps::wm::Geom *) 0 ;
  std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Geom_getAllPatches",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Geom_getAllPatches" "', argument " "1"" of type '" "steps::wm::Geom const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Geom * >(argp1);
  {
    try {
      result = ((steps::wm::Geom const *)arg1)->getAllPatches();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::wm::Patch*,std::allocator<steps::wm::Patch * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Geom_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__wm__Geom, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Patch(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::wm::Geom *arg2 = (steps::wm::Geom *) 0 ;
  steps::wm::Comp *arg3 = (steps::wm::Comp *) 0 ;
  steps::wm::Comp *arg4 = (steps::wm::Comp *) 0 ;
  double arg5 = (double) 0.0 ;
  steps::wm::Patch *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  double val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "container",(char *) "icomp",(char *) "ocomp",(char *) "area", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OOO|OO:new_Patch",kwnames,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Patch" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Patch" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Patch" "', argument " "2"" of type '" "steps::wm::Geom *""'"); 
  }
  arg2 = reinterpret_cast< steps::wm::Geom * >(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Patch" "', argument " "3"" of type '" "steps::wm::Comp *""'"); 
  }
  arg3 = reinterpret_cast< steps::wm::Comp * >(argp3);
  if (obj3) {
    res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_Patch" "', argument " "4"" of type '" "steps::wm::Comp *""'"); 
    }
    arg4 = reinterpret_cast< steps::wm::Comp * >(argp4);
  }
  if (obj4) {
    ecode5 = SWIG_AsVal_double(obj4, &val5);
    if (!SWIG_IsOK(ecode5)) {
      SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "new_Patch" "', argument " "5"" of type '" "double""'");
    } 
    arg5 = static_cast< double >(val5);
  }
  {
    try {
      result = (steps::wm::Patch *)new steps::wm::Patch((std::string const &)*arg1,arg2,arg3,arg4,arg5);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Patch, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Patch(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Patch",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Patch" "', argument " "1"" of type '" "steps::wm::Patch *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Patch_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_getID" "', argument " "1"" of type '" "steps::wm::Patch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      result = ((steps::wm::Patch const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Patch_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_setID" "', argument " "1"" of type '" "steps::wm::Patch *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Patch_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Patch_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_getContainer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  steps::wm::Geom *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Patch_getContainer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_getContainer" "', argument " "1"" of type '" "steps::wm::Patch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      result = (steps::wm::Geom *)((steps::wm::Patch const *)arg1)->getContainer();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_getArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Patch_getArea",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_getArea" "', argument " "1"" of type '" "steps::wm::Patch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      result = (double)((steps::wm::Patch const *)arg1)->getArea();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_setArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Patch_setArea",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_setArea" "', argument " "1"" of type '" "steps::wm::Patch *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Patch_setArea" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setArea(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_addSurfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Patch_addSurfsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_addSurfsys" "', argument " "1"" of type '" "steps::wm::Patch *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Patch_addSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Patch_addSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->addSurfsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_getSurfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  std::set<std::string,std::less<std::string >,std::allocator<std::string > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Patch_getSurfsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_getSurfsys" "', argument " "1"" of type '" "steps::wm::Patch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      result = ((steps::wm::Patch const *)arg1)->getSurfsys();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::set<std::string,std::less<std::string >,std::allocator<std::string > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_delSurfsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Patch_delSurfsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_delSurfsys" "', argument " "1"" of type '" "steps::wm::Patch *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Patch_delSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Patch_delSurfsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delSurfsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_getIComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  steps::wm::Comp *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Patch_getIComp",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_getIComp" "', argument " "1"" of type '" "steps::wm::Patch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      result = (steps::wm::Comp *)((steps::wm::Patch const *)arg1)->getIComp();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Patch_getOComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Patch *arg1 = (steps::wm::Patch *) 0 ;
  steps::wm::Comp *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Patch_getOComp",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Patch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Patch_getOComp" "', argument " "1"" of type '" "steps::wm::Patch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Patch * >(argp1);
  {
    try {
      result = (steps::wm::Comp *)((steps::wm::Patch const *)arg1)->getOComp();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Patch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__wm__Patch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Comp(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::wm::Geom *arg2 = (steps::wm::Geom *) 0 ;
  double arg3 = (double) 0.0 ;
  steps::wm::Comp *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "container",(char *) "vol", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OO|O:new_Comp",kwnames,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Comp" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Comp" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Comp" "', argument " "2"" of type '" "steps::wm::Geom *""'"); 
  }
  arg2 = reinterpret_cast< steps::wm::Geom * >(argp2);
  if (obj2) {
    ecode3 = SWIG_AsVal_double(obj2, &val3);
    if (!SWIG_IsOK(ecode3)) {
      SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Comp" "', argument " "3"" of type '" "double""'");
    } 
    arg3 = static_cast< double >(val3);
  }
  {
    try {
      result = (steps::wm::Comp *)new steps::wm::Comp((std::string const &)*arg1,arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Comp, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Comp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Comp",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Comp" "', argument " "1"" of type '" "steps::wm::Comp *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Comp_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_getID" "', argument " "1"" of type '" "steps::wm::Comp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      result = ((steps::wm::Comp const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Comp_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_setID" "', argument " "1"" of type '" "steps::wm::Comp *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Comp_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Comp_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_getContainer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  steps::wm::Geom *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Comp_getContainer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_getContainer" "', argument " "1"" of type '" "steps::wm::Comp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      result = (steps::wm::Geom *)((steps::wm::Comp const *)arg1)->getContainer();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_getVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Comp_getVol",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_getVol" "', argument " "1"" of type '" "steps::wm::Comp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      result = (double)((steps::wm::Comp const *)arg1)->getVol();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_setVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Comp_setVol",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_setVol" "', argument " "1"" of type '" "steps::wm::Comp *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Comp_setVol" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setVol(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_addVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Comp_addVolsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_addVolsys" "', argument " "1"" of type '" "steps::wm::Comp *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Comp_addVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Comp_addVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->addVolsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_getVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::set<std::string,std::less<std::string >,std::allocator<std::string > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Comp_getVolsys",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_getVolsys" "', argument " "1"" of type '" "steps::wm::Comp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      result = ((steps::wm::Comp const *)arg1)->getVolsys();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::set<std::string,std::less<std::string >,std::allocator<std::string > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_delVolsys(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Comp_delVolsys",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_delVolsys" "', argument " "1"" of type '" "steps::wm::Comp *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "Comp_delVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "Comp_delVolsys" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->delVolsys((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_getIPatches(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Comp_getIPatches",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_getIPatches" "', argument " "1"" of type '" "steps::wm::Comp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      result = ((steps::wm::Comp const *)arg1)->getIPatches();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Comp_getOPatches(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wm::Comp *arg1 = (steps::wm::Comp *) 0 ;
  std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Comp_getOPatches",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Comp_getOPatches" "', argument " "1"" of type '" "steps::wm::Comp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wm::Comp * >(argp1);
  {
    try {
      result = ((steps::wm::Comp const *)arg1)->getOPatches();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Comp_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__wm__Comp, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Tetmesh__SWIG_0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  steps::tetmesh::Tetmesh *result = 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:new_Tetmesh",&obj0,&obj1,&obj2)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "new_Tetmesh" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "new_Tetmesh" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "new_Tetmesh" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  {
    try {
      result = (steps::tetmesh::Tetmesh *)new steps::tetmesh::Tetmesh(arg1,arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__Tetmesh, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Tetmesh__SWIG_1(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double,std::allocator<double > > *arg1 = 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg2 = 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg3 = 0 ;
  steps::tetmesh::Tetmesh *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:new_Tetmesh",&obj0,&obj1,&obj2)) SWIG_fail;
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Tetmesh" "', argument " "1"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "1"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Tetmesh" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res3 = swig::asptr(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Tetmesh" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (steps::tetmesh::Tetmesh *)new steps::tetmesh::Tetmesh((std::vector<double,std::allocator<double > > const &)*arg1,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg2,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__Tetmesh, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Tetmesh__SWIG_2(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double,std::allocator<double > > *arg1 = 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg2 = 0 ;
  steps::tetmesh::Tetmesh *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:new_Tetmesh",&obj0,&obj1)) SWIG_fail;
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Tetmesh" "', argument " "1"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "1"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Tetmesh" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (steps::tetmesh::Tetmesh *)new steps::tetmesh::Tetmesh((std::vector<double,std::allocator<double > > const &)*arg1,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__Tetmesh, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Tetmesh__SWIG_3(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::vector<double,std::allocator<double > > *arg1 = 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg2 = 0 ;
  std::vector<double,std::allocator<double > > *arg3 = 0 ;
  std::vector<double,std::allocator<double > > *arg4 = 0 ;
  std::vector<int,std::allocator<int > > *arg5 = 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg6 = 0 ;
  std::vector<double,std::allocator<double > > *arg7 = 0 ;
  std::vector<double,std::allocator<double > > *arg8 = 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg9 = 0 ;
  std::vector<int,std::allocator<int > > *arg10 = 0 ;
  steps::tetmesh::Tetmesh *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  int res4 = SWIG_OLDOBJ ;
  int res5 = SWIG_OLDOBJ ;
  int res6 = SWIG_OLDOBJ ;
  int res7 = SWIG_OLDOBJ ;
  int res8 = SWIG_OLDOBJ ;
  int res9 = SWIG_OLDOBJ ;
  int res10 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  PyObject * obj6 = 0 ;
  PyObject * obj7 = 0 ;
  PyObject * obj8 = 0 ;
  PyObject * obj9 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOOOOOO:new_Tetmesh",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5,&obj6,&obj7,&obj8,&obj9)) SWIG_fail;
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res1 = swig::asptr(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Tetmesh" "', argument " "1"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "1"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg1 = ptr;
  }
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res2 = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Tetmesh" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res3 = swig::asptr(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Tetmesh" "', argument " "3"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "3"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg3 = ptr;
  }
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res4 = swig::asptr(obj3, &ptr);
    if (!SWIG_IsOK(res4)) {
      SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_Tetmesh" "', argument " "4"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "4"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg4 = ptr;
  }
  {
    std::vector<int,std::allocator<int > > *ptr = (std::vector<int,std::allocator<int > > *)0;
    res5 = swig::asptr(obj4, &ptr);
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_Tetmesh" "', argument " "5"" of type '" "std::vector<int,std::allocator<int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "5"" of type '" "std::vector<int,std::allocator<int > > const &""'"); 
    }
    arg5 = ptr;
  }
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res6 = swig::asptr(obj5, &ptr);
    if (!SWIG_IsOK(res6)) {
      SWIG_exception_fail(SWIG_ArgError(res6), "in method '" "new_Tetmesh" "', argument " "6"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "6"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg6 = ptr;
  }
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res7 = swig::asptr(obj6, &ptr);
    if (!SWIG_IsOK(res7)) {
      SWIG_exception_fail(SWIG_ArgError(res7), "in method '" "new_Tetmesh" "', argument " "7"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "7"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg7 = ptr;
  }
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    res8 = swig::asptr(obj7, &ptr);
    if (!SWIG_IsOK(res8)) {
      SWIG_exception_fail(SWIG_ArgError(res8), "in method '" "new_Tetmesh" "', argument " "8"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "8"" of type '" "std::vector<double,std::allocator<double > > const &""'"); 
    }
    arg8 = ptr;
  }
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res9 = swig::asptr(obj8, &ptr);
    if (!SWIG_IsOK(res9)) {
      SWIG_exception_fail(SWIG_ArgError(res9), "in method '" "new_Tetmesh" "', argument " "9"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "9"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg9 = ptr;
  }
  {
    std::vector<int,std::allocator<int > > *ptr = (std::vector<int,std::allocator<int > > *)0;
    res10 = swig::asptr(obj9, &ptr);
    if (!SWIG_IsOK(res10)) {
      SWIG_exception_fail(SWIG_ArgError(res10), "in method '" "new_Tetmesh" "', argument " "10"" of type '" "std::vector<int,std::allocator<int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_Tetmesh" "', argument " "10"" of type '" "std::vector<int,std::allocator<int > > const &""'"); 
    }
    arg10 = ptr;
  }
  {
    try {
      result = (steps::tetmesh::Tetmesh *)new steps::tetmesh::Tetmesh((std::vector<double,std::allocator<double > > const &)*arg1,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg2,(std::vector<double,std::allocator<double > > const &)*arg3,(std::vector<double,std::allocator<double > > const &)*arg4,(std::vector<int,std::allocator<int > > const &)*arg5,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg6,(std::vector<double,std::allocator<double > > const &)*arg7,(std::vector<double,std::allocator<double > > const &)*arg8,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg9,(std::vector<int,std::allocator<int > > const &)*arg10);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__Tetmesh, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  if (SWIG_IsNewObj(res6)) delete arg6;
  if (SWIG_IsNewObj(res7)) delete arg7;
  if (SWIG_IsNewObj(res8)) delete arg8;
  if (SWIG_IsNewObj(res9)) delete arg9;
  if (SWIG_IsNewObj(res10)) delete arg10;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  if (SWIG_IsNewObj(res4)) delete arg4;
  if (SWIG_IsNewObj(res5)) delete arg5;
  if (SWIG_IsNewObj(res6)) delete arg6;
  if (SWIG_IsNewObj(res7)) delete arg7;
  if (SWIG_IsNewObj(res8)) delete arg8;
  if (SWIG_IsNewObj(res9)) delete arg9;
  if (SWIG_IsNewObj(res10)) delete arg10;
  return NULL;
}


SWIGINTERN PyObject *_wrap_new_Tetmesh(PyObject *self, PyObject *args) {
  int argc;
  PyObject *argv[11];
  int ii;
  
  if (!PyTuple_Check(args)) SWIG_fail;
  argc = PyObject_Length(args);
  for (ii = 0; (ii < argc) && (ii < 10); ii++) {
    argv[ii] = PyTuple_GET_ITEM(args,ii);
  }
  if (argc == 2) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        return _wrap_new_Tetmesh__SWIG_2(self, args);
      }
    }
  }
  if (argc == 3) {
    int _v;
    {
      int res = SWIG_AsVal_unsigned_SS_int(argv[0], NULL);
      _v = SWIG_CheckState(res);
    }
    if (_v) {
      {
        int res = SWIG_AsVal_unsigned_SS_int(argv[1], NULL);
        _v = SWIG_CheckState(res);
      }
      if (_v) {
        {
          int res = SWIG_AsVal_unsigned_SS_int(argv[2], NULL);
          _v = SWIG_CheckState(res);
        }
        if (_v) {
          return _wrap_new_Tetmesh__SWIG_0(self, args);
        }
      }
    }
  }
  if (argc == 3) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          return _wrap_new_Tetmesh__SWIG_1(self, args);
        }
      }
    }
  }
  if (argc == 10) {
    int _v;
    int res = swig::asptr(argv[0], (std::vector<double,std::allocator<double > >**)(0));
    _v = SWIG_CheckState(res);
    if (_v) {
      int res = swig::asptr(argv[1], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
      _v = SWIG_CheckState(res);
      if (_v) {
        int res = swig::asptr(argv[2], (std::vector<double,std::allocator<double > >**)(0));
        _v = SWIG_CheckState(res);
        if (_v) {
          int res = swig::asptr(argv[3], (std::vector<double,std::allocator<double > >**)(0));
          _v = SWIG_CheckState(res);
          if (_v) {
            int res = swig::asptr(argv[4], (std::vector<int,std::allocator<int > >**)(0));
            _v = SWIG_CheckState(res);
            if (_v) {
              int res = swig::asptr(argv[5], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
              _v = SWIG_CheckState(res);
              if (_v) {
                int res = swig::asptr(argv[6], (std::vector<double,std::allocator<double > >**)(0));
                _v = SWIG_CheckState(res);
                if (_v) {
                  int res = swig::asptr(argv[7], (std::vector<double,std::allocator<double > >**)(0));
                  _v = SWIG_CheckState(res);
                  if (_v) {
                    int res = swig::asptr(argv[8], (std::vector<unsigned int,std::allocator<unsigned int > >**)(0));
                    _v = SWIG_CheckState(res);
                    if (_v) {
                      int res = swig::asptr(argv[9], (std::vector<int,std::allocator<int > >**)(0));
                      _v = SWIG_CheckState(res);
                      if (_v) {
                        return _wrap_new_Tetmesh__SWIG_3(self, args);
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  
fail:
  SWIG_SetErrorMsg(PyExc_NotImplementedError,"Wrong number of arguments for overloaded function 'new_Tetmesh'.\n  Possible C/C++ prototypes are:\n    steps::tetmesh::Tetmesh(unsigned int,unsigned int,unsigned int)\n    steps::tetmesh::Tetmesh(std::vector<double,std::allocator<double > > const &,std::vector<unsigned int,std::allocator<unsigned int > > const &,std::vector<unsigned int,std::allocator<unsigned int > > const &)\n    steps::tetmesh::Tetmesh(std::vector<double,std::allocator<double > > const &,std::vector<unsigned int,std::allocator<unsigned int > > const &)\n    steps::tetmesh::Tetmesh(std::vector<double,std::allocator<double > > const &,std::vector<unsigned int,std::allocator<unsigned int > > const &,std::vector<double,std::allocator<double > > const &,std::vector<double,std::allocator<double > > const &,std::vector<int,std::allocator<int > > const &,std::vector<unsigned int,std::allocator<unsigned int > > const &,std::vector<double,std::allocator<double > > const &,std::vector<double,std::allocator<double > > const &,std::vector<unsigned int,std::allocator<unsigned int > > const &,std::vector<int,std::allocator<int > > const &)\n");
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Tetmesh(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Tetmesh",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Tetmesh" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_setVertex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  double arg3 ;
  double arg4 ;
  double arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  double val3 ;
  int ecode3 = 0 ;
  double val4 ;
  int ecode4 = 0 ;
  double val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Tetmesh_setVertex",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_setVertex" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_setVertex" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tetmesh_setVertex" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tetmesh_setVertex" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  ecode5 = SWIG_AsVal_double(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Tetmesh_setVertex" "', argument " "5"" of type '" "double""'");
  } 
  arg5 = static_cast< double >(val5);
  {
    try {
      (arg1)->setVertex(arg2,arg3,arg4,arg5);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_setTri(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOO:Tetmesh_setTri",&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_setTri" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_setTri" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tetmesh_setTri" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tetmesh_setTri" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Tetmesh_setTri" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  {
    try {
      (arg1)->setTri(arg2,arg3,arg4,arg5);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_setTet(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  unsigned int arg3 ;
  unsigned int arg4 ;
  unsigned int arg5 ;
  unsigned int arg6 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  unsigned int val3 ;
  int ecode3 = 0 ;
  unsigned int val4 ;
  int ecode4 = 0 ;
  unsigned int val5 ;
  int ecode5 = 0 ;
  unsigned int val6 ;
  int ecode6 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  PyObject * obj5 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOOOO:Tetmesh_setTet",&obj0,&obj1,&obj2,&obj3,&obj4,&obj5)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_setTet" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_setTet" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  ecode3 = SWIG_AsVal_unsigned_SS_int(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "Tetmesh_setTet" "', argument " "3"" of type '" "unsigned int""'");
  } 
  arg3 = static_cast< unsigned int >(val3);
  ecode4 = SWIG_AsVal_unsigned_SS_int(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "Tetmesh_setTet" "', argument " "4"" of type '" "unsigned int""'");
  } 
  arg4 = static_cast< unsigned int >(val4);
  ecode5 = SWIG_AsVal_unsigned_SS_int(obj4, &val5);
  if (!SWIG_IsOK(ecode5)) {
    SWIG_exception_fail(SWIG_ArgError(ecode5), "in method '" "Tetmesh_setTet" "', argument " "5"" of type '" "unsigned int""'");
  } 
  arg5 = static_cast< unsigned int >(val5);
  ecode6 = SWIG_AsVal_unsigned_SS_int(obj5, &val6);
  if (!SWIG_IsOK(ecode6)) {
    SWIG_exception_fail(SWIG_ArgError(ecode6), "in method '" "Tetmesh_setTet" "', argument " "6"" of type '" "unsigned int""'");
  } 
  arg6 = static_cast< unsigned int >(val6);
  {
    try {
      (arg1)->setTet(arg2,arg3,arg4,arg5,arg6);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_setup(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_setup",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_setup" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      (arg1)->setup();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_isSetupDone(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_isSetupDone",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_isSetupDone" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = (bool)((steps::tetmesh::Tetmesh const *)arg1)->isSetupDone();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getVertex(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getVertex",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getVertex" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getVertex" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getVertex(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_countVertices(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_countVertices",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_countVertices" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = (unsigned int)((steps::tetmesh::Tetmesh const *)arg1)->countVertices();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTri(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTri",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTri" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTri" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTri(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_countTris(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_countTris",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_countTris" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = (unsigned int)((steps::tetmesh::Tetmesh const *)arg1)->countTris();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTriArea",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriArea" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTriArea" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((steps::tetmesh::Tetmesh const *)arg1)->getTriArea(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriBarycenter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTriBarycenter",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriBarycenter" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTriBarycenter" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTriBarycenter(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTetQualityRER(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTetQualityRER",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTetQualityRER" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTetQualityRER" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((steps::tetmesh::Tetmesh const *)arg1)->getTetQualityRER(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriNorm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTriNorm",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriNorm" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTriNorm" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTriNorm(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriPatch(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  steps::tetmesh::TmPatch *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTriPatch",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriPatch" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTriPatch" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (steps::tetmesh::TmPatch *)((steps::tetmesh::Tetmesh const *)arg1)->getTriPatch(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__TmPatch, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriDiffBoundary(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  steps::tetmesh::DiffBoundary *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTriDiffBoundary",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriDiffBoundary" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTriDiffBoundary" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (steps::tetmesh::DiffBoundary *)((steps::tetmesh::Tetmesh const *)arg1)->getTriDiffBoundary(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriTetNeighb(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<int,std::allocator<int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTriTetNeighb",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriTetNeighb" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTriTetNeighb" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTriTetNeighb(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<int,std::allocator<int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTriBoundary(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<int,std::allocator<int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_getTriBoundary",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTriBoundary" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTriBoundary();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<int,std::allocator<int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getSurfTris(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_getSurfTris",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getSurfTris" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getSurfTris();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTet(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTet",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTet" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTet" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTet(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_countTets(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_countTets",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_countTets" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = (unsigned int)((steps::tetmesh::Tetmesh const *)arg1)->countTets();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTetVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTetVol",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTetVol" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTetVol" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((steps::tetmesh::Tetmesh const *)arg1)->getTetVol(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTetBarycenter(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTetBarycenter",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTetBarycenter" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTetBarycenter" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTetBarycenter(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTetComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  steps::tetmesh::TmComp *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTetComp",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTetComp" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTetComp" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (steps::tetmesh::TmComp *)((steps::tetmesh::Tetmesh const *)arg1)->getTetComp(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTetTriNeighb(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTetTriNeighb",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTetTriNeighb" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTetTriNeighb" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTetTriNeighb(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getTetTetNeighb(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  unsigned int arg2 ;
  std::vector<int,std::allocator<int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_getTetTetNeighb",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getTetTetNeighb" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetmesh_getTetTetNeighb" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getTetTetNeighb(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<int,std::allocator<int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_findTetByPoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<double,std::allocator<double > > arg2 ;
  int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetmesh_findTetByPoint",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_findTetByPoint" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    std::vector<double,std::allocator<double > > *ptr = (std::vector<double,std::allocator<double > > *)0;
    int res = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "Tetmesh_findTetByPoint" "', argument " "2"" of type '" "std::vector<double,std::allocator<double > >""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    try {
      result = (int)((steps::tetmesh::Tetmesh const *)arg1)->findTetByPoint(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_int(static_cast< int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getBoundMin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_getBoundMin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getBoundMin" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getBoundMin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getBoundMax(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_getBoundMax",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getBoundMax" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = ((steps::tetmesh::Tetmesh const *)arg1)->getBoundMax();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetmesh_getMeshVolume(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::Tetmesh *arg1 = (steps::tetmesh::Tetmesh *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetmesh_getMeshVolume",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetmesh_getMeshVolume" "', argument " "1"" of type '" "steps::tetmesh::Tetmesh const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp1);
  {
    try {
      result = (double)((steps::tetmesh::Tetmesh const *)arg1)->getMeshVolume();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Tetmesh_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__tetmesh__Tetmesh, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TmComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::tetmesh::Tetmesh *arg2 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg3 = 0 ;
  steps::tetmesh::TmComp *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "container",(char *) "tets", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OOO:new_TmComp",kwnames,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TmComp" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TmComp" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_TmComp" "', argument " "2"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg2 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp2);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res3 = swig::asptr(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_TmComp" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TmComp" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (steps::tetmesh::TmComp *)new steps::tetmesh::TmComp((std::string const &)*arg1,arg2,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__TmComp, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TmComp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TmComp",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TmComp" "', argument " "1"" of type '" "steps::tetmesh::TmComp *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmComp_setVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TmComp_setVol",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmComp_setVol" "', argument " "1"" of type '" "steps::tetmesh::TmComp *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TmComp_setVol" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setVol(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmComp_getAllTetIndices(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TmComp_getAllTetIndices",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmComp_getAllTetIndices" "', argument " "1"" of type '" "steps::tetmesh::TmComp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  {
    try {
      result = ((steps::tetmesh::TmComp const *)arg1)->getAllTetIndices();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmComp_countTets(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TmComp_countTets",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmComp_countTets" "', argument " "1"" of type '" "steps::tetmesh::TmComp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  {
    try {
      result = (unsigned int)((steps::tetmesh::TmComp const *)arg1)->countTets();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmComp_isTetInside(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > arg2 ;
  std::vector<bool,std::allocator<bool > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TmComp_isTetInside",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmComp_isTetInside" "', argument " "1"" of type '" "steps::tetmesh::TmComp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    int res = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "TmComp_isTetInside" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > >""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    try {
      result = ((steps::tetmesh::TmComp const *)arg1)->isTetInside(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<bool,std::allocator<bool > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmComp_getBoundMin(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TmComp_getBoundMin",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmComp_getBoundMin" "', argument " "1"" of type '" "steps::tetmesh::TmComp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  {
    try {
      result = ((steps::tetmesh::TmComp const *)arg1)->getBoundMin();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmComp_getBoundMax(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmComp *arg1 = (steps::tetmesh::TmComp *) 0 ;
  std::vector<double,std::allocator<double > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TmComp_getBoundMax",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmComp, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmComp_getBoundMax" "', argument " "1"" of type '" "steps::tetmesh::TmComp const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmComp * >(argp1);
  {
    try {
      result = ((steps::tetmesh::TmComp const *)arg1)->getBoundMax();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<double,std::allocator<double > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TmComp_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__tetmesh__TmComp, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_TmPatch(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::tetmesh::Tetmesh *arg2 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg3 = 0 ;
  steps::wm::Comp *arg4 = (steps::wm::Comp *) 0 ;
  steps::wm::Comp *arg5 = (steps::wm::Comp *) 0 ;
  steps::tetmesh::TmPatch *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  void *argp4 = 0 ;
  int res4 = 0 ;
  void *argp5 = 0 ;
  int res5 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  PyObject * obj4 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "container",(char *) "tris",(char *) "icomp",(char *) "ocomp", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OOOO|O:new_TmPatch",kwnames,&obj0,&obj1,&obj2,&obj3,&obj4)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_TmPatch" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TmPatch" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_TmPatch" "', argument " "2"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg2 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp2);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res3 = swig::asptr(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_TmPatch" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_TmPatch" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg3 = ptr;
  }
  res4 = SWIG_ConvertPtr(obj3, &argp4,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
  if (!SWIG_IsOK(res4)) {
    SWIG_exception_fail(SWIG_ArgError(res4), "in method '" "new_TmPatch" "', argument " "4"" of type '" "steps::wm::Comp *""'"); 
  }
  arg4 = reinterpret_cast< steps::wm::Comp * >(argp4);
  if (obj4) {
    res5 = SWIG_ConvertPtr(obj4, &argp5,SWIGTYPE_p_steps__wm__Comp, 0 |  0 );
    if (!SWIG_IsOK(res5)) {
      SWIG_exception_fail(SWIG_ArgError(res5), "in method '" "new_TmPatch" "', argument " "5"" of type '" "steps::wm::Comp *""'"); 
    }
    arg5 = reinterpret_cast< steps::wm::Comp * >(argp5);
  }
  {
    try {
      result = (steps::tetmesh::TmPatch *)new steps::tetmesh::TmPatch((std::string const &)*arg1,arg2,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg3,arg4,arg5);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__TmPatch, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_TmPatch(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmPatch *arg1 = (steps::tetmesh::TmPatch *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_TmPatch",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmPatch, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_TmPatch" "', argument " "1"" of type '" "steps::tetmesh::TmPatch *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmPatch * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmPatch_setArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmPatch *arg1 = (steps::tetmesh::TmPatch *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TmPatch_setArea",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmPatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmPatch_setArea" "', argument " "1"" of type '" "steps::tetmesh::TmPatch *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmPatch * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "TmPatch_setArea" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setArea(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmPatch_getAllTriIndices(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmPatch *arg1 = (steps::tetmesh::TmPatch *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:TmPatch_getAllTriIndices",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmPatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmPatch_getAllTriIndices" "', argument " "1"" of type '" "steps::tetmesh::TmPatch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmPatch * >(argp1);
  {
    try {
      result = ((steps::tetmesh::TmPatch const *)arg1)->getAllTriIndices();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_TmPatch_isTriInside(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::TmPatch *arg1 = (steps::tetmesh::TmPatch *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > arg2 ;
  std::vector<bool,std::allocator<bool > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:TmPatch_isTriInside",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__TmPatch, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "TmPatch_isTriInside" "', argument " "1"" of type '" "steps::tetmesh::TmPatch const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::TmPatch * >(argp1);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    int res = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "TmPatch_isTriInside" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > >""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    try {
      result = ((steps::tetmesh::TmPatch const *)arg1)->isTriInside(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<bool,std::allocator<bool > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *TmPatch_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__tetmesh__TmPatch, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_DiffBoundary(PyObject *SWIGUNUSEDPARM(self), PyObject *args, PyObject *kwargs) {
  PyObject *resultobj = 0;
  std::string *arg1 = 0 ;
  steps::tetmesh::Tetmesh *arg2 = (steps::tetmesh::Tetmesh *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > *arg3 = 0 ;
  steps::tetmesh::DiffBoundary *result = 0 ;
  int res1 = SWIG_OLDOBJ ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  char *  kwnames[] = {
    (char *) "id",(char *) "container",(char *) "tris", NULL 
  };
  
  if (!PyArg_ParseTupleAndKeywords(args,kwargs,(char *)"OOO:new_DiffBoundary",kwnames,&obj0,&obj1,&obj2)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    res1 = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res1)) {
      SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_DiffBoundary" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_DiffBoundary" "', argument " "1"" of type '" "std::string const &""'"); 
    }
    arg1 = ptr;
  }
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_DiffBoundary" "', argument " "2"" of type '" "steps::tetmesh::Tetmesh *""'"); 
  }
  arg2 = reinterpret_cast< steps::tetmesh::Tetmesh * >(argp2);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    res3 = swig::asptr(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_DiffBoundary" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "new_DiffBoundary" "', argument " "3"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > > const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (steps::tetmesh::DiffBoundary *)new steps::tetmesh::DiffBoundary((std::string const &)*arg1,arg2,(std::vector<unsigned int,std::allocator<unsigned int > > const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__DiffBoundary, SWIG_POINTER_NEW |  0 );
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res1)) delete arg1;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_DiffBoundary(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_DiffBoundary",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_DiffBoundary" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DiffBoundary_getID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:DiffBoundary_getID",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DiffBoundary_getID" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    try {
      result = ((steps::tetmesh::DiffBoundary const *)arg1)->getID();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DiffBoundary_setID(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:DiffBoundary_setID",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DiffBoundary_setID" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "DiffBoundary_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "DiffBoundary_setID" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->setID((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_DiffBoundary_getContainer(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  steps::tetmesh::Tetmesh *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:DiffBoundary_getContainer",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DiffBoundary_getContainer" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    try {
      result = (steps::tetmesh::Tetmesh *)((steps::tetmesh::DiffBoundary const *)arg1)->getContainer();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetmesh__Tetmesh, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DiffBoundary_getComps(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:DiffBoundary_getComps",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DiffBoundary_getComps" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    try {
      result = ((steps::tetmesh::DiffBoundary const *)arg1)->getComps();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<steps::wm::Comp*,std::allocator<steps::wm::Comp * > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DiffBoundary_isTriInside(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > arg2 ;
  std::vector<bool,std::allocator<bool > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:DiffBoundary_isTriInside",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DiffBoundary_isTriInside" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    std::vector<unsigned int,std::allocator<unsigned int > > *ptr = (std::vector<unsigned int,std::allocator<unsigned int > > *)0;
    int res = swig::asptr(obj1, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "DiffBoundary_isTriInside" "', argument " "2"" of type '" "std::vector<unsigned int,std::allocator<unsigned int > >""'"); 
    }
    arg2 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  {
    try {
      result = ((steps::tetmesh::DiffBoundary const *)arg1)->isTriInside(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<bool,std::allocator<bool > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_DiffBoundary_getAllTriIndices(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetmesh::DiffBoundary *arg1 = (steps::tetmesh::DiffBoundary *) 0 ;
  std::vector<unsigned int,std::allocator<unsigned int > > result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:DiffBoundary_getAllTriIndices",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetmesh__DiffBoundary, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "DiffBoundary_getAllTriIndices" "', argument " "1"" of type '" "steps::tetmesh::DiffBoundary const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetmesh::DiffBoundary * >(argp1);
  {
    try {
      result = ((steps::tetmesh::DiffBoundary const *)arg1)->getAllTriIndices();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = swig::from(static_cast< std::vector<unsigned int,std::allocator<unsigned int > > >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *DiffBoundary_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__tetmesh__DiffBoundary, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_delete_RNG(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_RNG",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_RNG" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_initialize(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  unsigned long *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned long temp2 ;
  unsigned long val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:RNG_initialize",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_initialize" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_long(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RNG_initialize" "', argument " "2"" of type '" "unsigned long""'");
  } 
  temp2 = static_cast< unsigned long >(val2);
  arg2 = &temp2;
  {
    try {
      (arg1)->initialize((unsigned long const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_get(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_get",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_get" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (unsigned int)(arg1)->get();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getUnfII(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_getUnfII",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getUnfII" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (double)(arg1)->getUnfII();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getUnfIE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_getUnfIE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getUnfIE" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (double)(arg1)->getUnfIE();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getUnfEE(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_getUnfEE",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getUnfEE" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (double)(arg1)->getUnfEE();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getUnfIE53(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_getUnfIE53",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getUnfIE53" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (double)(arg1)->getUnfIE53();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getStdExp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  float result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_getStdExp",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getStdExp" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (float)(arg1)->getStdExp();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getExp(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  double arg2 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:RNG_getExp",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getExp" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RNG_getExp" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      result = (double)(arg1)->getExp(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getPsn(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  float arg2 ;
  long result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  float val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:RNG_getPsn",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getPsn" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  ecode2 = SWIG_AsVal_float(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "RNG_getPsn" "', argument " "2"" of type '" "float""'");
  } 
  arg2 = static_cast< float >(val2);
  {
    try {
      result = (long)(arg1)->getPsn(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_long(static_cast< long >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_RNG_getStdNrm(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::rng::RNG *arg1 = (steps::rng::RNG *) 0 ;
  float result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:RNG_getStdNrm",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "RNG_getStdNrm" "', argument " "1"" of type '" "steps::rng::RNG *""'"); 
  }
  arg1 = reinterpret_cast< steps::rng::RNG * >(argp1);
  {
    try {
      result = (float)(arg1)->getStdNrm();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_float(static_cast< float >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *RNG_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__rng__RNG, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_create_mt19937(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  unsigned int arg1 ;
  steps::rng::RNG *result = 0 ;
  unsigned int val1 ;
  int ecode1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:create_mt19937",&obj0)) SWIG_fail;
  ecode1 = SWIG_AsVal_unsigned_SS_int(obj0, &val1);
  if (!SWIG_IsOK(ecode1)) {
    SWIG_exception_fail(SWIG_ArgError(ecode1), "in method '" "create_mt19937" "', argument " "1"" of type '" "unsigned int""'");
  } 
  arg1 = static_cast< unsigned int >(val1);
  {
    try {
      result = (steps::rng::RNG *)steps::rng::create_mt19937(arg1);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_create(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  std::string arg1 ;
  unsigned int arg2 ;
  steps::rng::RNG *result = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:create",&obj0,&obj1)) SWIG_fail;
  {
    std::string *ptr = (std::string *)0;
    int res = SWIG_AsPtr_std_string(obj0, &ptr);
    if (!SWIG_IsOK(res) || !ptr) {
      SWIG_exception_fail(SWIG_ArgError((ptr ? res : SWIG_TypeError)), "in method '" "create" "', argument " "1"" of type '" "std::string""'"); 
    }
    arg1 = *ptr;
    if (SWIG_IsNewObj(res)) delete ptr;
  }
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "create" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (steps::rng::RNG *)steps::rng::create(arg1,arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_API(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_API",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_API" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getSolverName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getSolverName",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getSolverName" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = ((steps::solver::API const *)arg1)->getSolverName();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getSolverDesc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getSolverDesc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getSolverDesc" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = ((steps::solver::API const *)arg1)->getSolverDesc();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getSolverAuthors(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getSolverAuthors",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getSolverAuthors" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = ((steps::solver::API const *)arg1)->getSolverAuthors();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getSolverEmail(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getSolverEmail",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getSolverEmail" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = ((steps::solver::API const *)arg1)->getSolverEmail();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_reset(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_reset",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_reset" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      (arg1)->reset();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_checkpoint(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_checkpoint",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_checkpoint" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_checkpoint" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_checkpoint" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->checkpoint((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_restore(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_restore",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_restore" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_restore" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_restore" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      (arg1)->restore((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_run(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_run",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_run" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_run" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->run(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getTime",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTime" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTime();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_advance(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_advance",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_advance" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_advance" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->advance(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_step(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_step",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_step" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      (arg1)->step();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setRk4DT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_setRk4DT",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setRk4DT" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setRk4DT" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setRk4DT(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setDT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_setDT",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setDT" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setDT" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setDT(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getRk4DT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getRk4DT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getRk4DT" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getRk4DT();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getDT(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getDT",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getDT" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getDT();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_setTime",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTime" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTime" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->setTime(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getA0(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getA0",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getA0" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getA0();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getNSteps(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:API_getNSteps",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getNSteps" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    try {
      result = (unsigned int)((steps::solver::API const *)arg1)->getNSteps();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setNSteps(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_setNSteps",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setNSteps" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setNSteps" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      (arg1)->setNSteps(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_getCompVol",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompVol" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompVol" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompVol" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompVol((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_setCompVol",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompVol" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompVol" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompVol" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "API_setCompVol" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->setCompVol((std::string const &)*arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompCount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompCount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompCount((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompCount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompCount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompCount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setCompCount((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompAmount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompAmount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompAmount((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompAmount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompAmount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompAmount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setCompAmount((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompConc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompConc",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompConc" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompConc" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompConc" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompConc((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompConc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompConc",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompConc" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompConc" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompConc" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompConc" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setCompConc((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompClamped",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompClamped" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getCompClamped((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompClamped",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompClamped" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompClamped" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setCompClamped((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompReacK",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompReacK" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompReacK((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompReacK",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompReacK" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompReacK" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setCompReacK((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompReacActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompReacActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getCompReacActive((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompReacActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompReacActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompReacActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setCompReacActive((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompDiffD(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompDiffD",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompDiffD" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompDiffD" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompDiffD" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompDiffD((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompDiffD(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompDiffD",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompDiffD" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompDiffD" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompDiffD" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompDiffD" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setCompDiffD((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompDiffActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompDiffActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompDiffActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompDiffActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompDiffActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getCompDiffActive((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setCompDiffActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setCompDiffActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setCompDiffActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setCompDiffActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompDiffActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setCompDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setCompDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setCompDiffActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setCompDiffActive((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompReacC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompReacC",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompReacC" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompReacC" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacC" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompReacC((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompReacH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompReacH",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompReacH" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompReacH" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacH" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompReacH((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompReacA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompReacA",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompReacA" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompReacA" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacA" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getCompReacA((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getCompReacExtent(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getCompReacExtent",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getCompReacExtent" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getCompReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getCompReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getCompReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (unsigned int)((steps::solver::API const *)arg1)->getCompReacExtent((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_resetCompReacExtent(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_resetCompReacExtent",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_resetCompReacExtent" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_resetCompReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_resetCompReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_resetCompReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_resetCompReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      (arg1)->resetCompReacExtent((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetVol(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_getTetVol",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetVol" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetVol" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetVol(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetSpecDefined(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetSpecDefined",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetSpecDefined" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetSpecDefined" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetSpecDefined" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetSpecDefined" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTetSpecDefined(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetCount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetCount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetCount(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetCount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetCount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetCount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTetCount(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetAmount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetAmount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetAmount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetAmount(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetAmount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetAmount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetAmount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetAmount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTetAmount(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetConc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetConc",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetConc" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetConc" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetConc(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetConc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetConc",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetConc" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetConc" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetConc" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetConc" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTetConc(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetClamped",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetClamped" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetClamped" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTetClamped(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetClamped",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetClamped" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetClamped" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetClamped" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setTetClamped(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetReacK",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetReacK" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetReacK" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetReacK(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetReacK",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetReacK" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetReacK" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetReacK" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTetReacK(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetReacActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetReacActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetReacActive" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTetReacActive(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetReacActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetReacActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetReacActive" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetReacActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setTetReacActive(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetDiffD(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetDiffD",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetDiffD" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetDiffD" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetDiffD(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetDiffD(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetDiffD",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetDiffD" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetDiffD" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetDiffD" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetDiffD" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTetDiffD(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetDiffActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetDiffActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetDiffActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetDiffActive" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTetDiffActive(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTetDiffActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTetDiffActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTetDiffActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTetDiffActive" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTetDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTetDiffActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTetDiffActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setTetDiffActive(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetReacC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetReacC",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetReacC" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetReacC" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetReacC(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetReacH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetReacH",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetReacH" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetReacH" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetReacH(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetReacA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetReacA",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetReacA" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetReacA" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetReacA(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTetDiffA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTetDiffA",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTetDiffA" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTetDiffA" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTetDiffA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTetDiffA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTetDiffA(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_getPatchArea",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchArea" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchArea" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchArea" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchArea((std::string const &)*arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setPatchArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  double arg3 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  double val3 ;
  int ecode3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_setPatchArea",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setPatchArea" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setPatchArea" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchArea" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  ecode3 = SWIG_AsVal_double(obj2, &val3);
  if (!SWIG_IsOK(ecode3)) {
    SWIG_exception_fail(SWIG_ArgError(ecode3), "in method '" "API_setPatchArea" "', argument " "3"" of type '" "double""'");
  } 
  arg3 = static_cast< double >(val3);
  {
    try {
      (arg1)->setPatchArea((std::string const &)*arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchCount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchCount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchCount((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setPatchCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setPatchCount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setPatchCount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setPatchCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchCount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setPatchCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setPatchCount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setPatchCount((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchAmount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchAmount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchAmount((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setPatchAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setPatchAmount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setPatchAmount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setPatchAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchAmount" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setPatchAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setPatchAmount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setPatchAmount((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchClamped",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchClamped" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getPatchClamped((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setPatchClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setPatchClamped",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setPatchClamped" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setPatchClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchClamped" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setPatchClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setPatchClamped" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setPatchClamped((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchSReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchSReacK",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchSReacK" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchSReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchSReacK((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setPatchSReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setPatchSReacK",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setPatchSReacK" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setPatchSReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchSReacK" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setPatchSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setPatchSReacK" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setPatchSReacK((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchSReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchSReacActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchSReacActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchSReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getPatchSReacActive((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setPatchSReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setPatchSReacActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setPatchSReacActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setPatchSReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchSReacActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setPatchSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setPatchSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setPatchSReacActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setPatchSReacActive((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchSReacC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchSReacC",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchSReacC" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchSReacC" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacC" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchSReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchSReacC((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchSReacH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchSReacH",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchSReacH" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchSReacH" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacH" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchSReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchSReacH((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchSReacA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchSReacA",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchSReacA" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchSReacA" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacA" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchSReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getPatchSReacA((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getPatchSReacExtent(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  unsigned int result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getPatchSReacExtent",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getPatchSReacExtent" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getPatchSReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getPatchSReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getPatchSReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (unsigned int)((steps::solver::API const *)arg1)->getPatchSReacExtent((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_unsigned_SS_int(static_cast< unsigned int >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_resetPatchSReacExtent(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_resetPatchSReacExtent",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_resetPatchSReacExtent" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_resetPatchSReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_resetPatchSReacExtent" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_resetPatchSReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_resetPatchSReacExtent" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      (arg1)->resetPatchSReacExtent((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setDiffBoundaryDiffusionActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setDiffBoundaryDiffusionActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setDiffBoundaryDiffusionActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_setDiffBoundaryDiffusionActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setDiffBoundaryDiffusionActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setDiffBoundaryDiffusionActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setDiffBoundaryDiffusionActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setDiffBoundaryDiffusionActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setDiffBoundaryDiffusionActive((std::string const &)*arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getDiffBoundaryDiffusionActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  std::string *arg2 = 0 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  int res2 = SWIG_OLDOBJ ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getDiffBoundaryDiffusionActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getDiffBoundaryDiffusionActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  {
    std::string *ptr = (std::string *)0;
    res2 = SWIG_AsPtr_std_string(obj1, &ptr);
    if (!SWIG_IsOK(res2)) {
      SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "API_getDiffBoundaryDiffusionActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getDiffBoundaryDiffusionActive" "', argument " "2"" of type '" "std::string const &""'"); 
    }
    arg2 = ptr;
  }
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getDiffBoundaryDiffusionActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getDiffBoundaryDiffusionActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getDiffBoundaryDiffusionActive((std::string const &)*arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res2)) delete arg2;
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriArea(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:API_getTriArea",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriArea" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriArea" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriArea(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriSpecDefined(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriSpecDefined",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriSpecDefined" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriSpecDefined" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriSpecDefined" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriSpecDefined" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTriSpecDefined(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriCount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriCount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriCount(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTriCount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTriCount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTriCount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTriCount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTriCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTriCount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTriCount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTriCount(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriAmount",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriAmount" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriAmount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriAmount(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTriAmount(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTriAmount",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTriAmount" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTriAmount" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTriAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTriAmount" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTriAmount" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTriAmount(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriClamped",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriClamped" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriClamped" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTriClamped(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTriClamped(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTriClamped",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTriClamped" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTriClamped" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTriClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTriClamped" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTriClamped" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setTriClamped(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriSReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriSReacK",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriSReacK" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriSReacK" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriSReacK(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTriSReacK(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  double val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTriSReacK",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTriSReacK" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTriSReacK" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTriSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTriSReacK" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_double(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTriSReacK" "', argument " "4"" of type '" "double""'");
  } 
  arg4 = static_cast< double >(val4);
  {
    try {
      (arg1)->setTriSReacK(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriSReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriSReacActive",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriSReacActive" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriSReacActive" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (bool)((steps::solver::API const *)arg1)->getTriSReacActive(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_bool(static_cast< bool >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_setTriSReacActive(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  bool arg4 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  bool val4 ;
  int ecode4 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  PyObject * obj3 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOOO:API_setTriSReacActive",&obj0,&obj1,&obj2,&obj3)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_setTriSReacActive" "', argument " "1"" of type '" "steps::solver::API *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_setTriSReacActive" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_setTriSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_setTriSReacActive" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  ecode4 = SWIG_AsVal_bool(obj3, &val4);
  if (!SWIG_IsOK(ecode4)) {
    SWIG_exception_fail(SWIG_ArgError(ecode4), "in method '" "API_setTriSReacActive" "', argument " "4"" of type '" "bool""'");
  } 
  arg4 = static_cast< bool >(val4);
  {
    try {
      (arg1)->setTriSReacActive(arg2,(std::string const &)*arg3,arg4);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriSReacC(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriSReacC",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriSReacC" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriSReacC" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriSReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriSReacC" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriSReacC(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriSReacH(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriSReacH",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriSReacH" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriSReacH" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriSReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriSReacH" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriSReacH(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *_wrap_API_getTriSReacA(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::solver::API *arg1 = (steps::solver::API *) 0 ;
  unsigned int arg2 ;
  std::string *arg3 = 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  unsigned int val2 ;
  int ecode2 = 0 ;
  int res3 = SWIG_OLDOBJ ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:API_getTriSReacA",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__solver__API, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "API_getTriSReacA" "', argument " "1"" of type '" "steps::solver::API const *""'"); 
  }
  arg1 = reinterpret_cast< steps::solver::API * >(argp1);
  ecode2 = SWIG_AsVal_unsigned_SS_int(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "API_getTriSReacA" "', argument " "2"" of type '" "unsigned int""'");
  } 
  arg2 = static_cast< unsigned int >(val2);
  {
    std::string *ptr = (std::string *)0;
    res3 = SWIG_AsPtr_std_string(obj2, &ptr);
    if (!SWIG_IsOK(res3)) {
      SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "API_getTriSReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    if (!ptr) {
      SWIG_exception_fail(SWIG_ValueError, "invalid null reference " "in method '" "API_getTriSReacA" "', argument " "3"" of type '" "std::string const &""'"); 
    }
    arg3 = ptr;
  }
  {
    try {
      result = (double)((steps::solver::API const *)arg1)->getTriSReacA(arg2,(std::string const &)*arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  if (SWIG_IsNewObj(res3)) delete arg3;
  return resultobj;
fail:
  if (SWIG_IsNewObj(res3)) delete arg3;
  return NULL;
}


SWIGINTERN PyObject *API_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__solver__API, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Wmrk4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  steps::wm::Geom *arg2 = (steps::wm::Geom *) 0 ;
  steps::rng::RNG *arg3 = (steps::rng::RNG *) 0 ;
  steps::wmrk4::Wmrk4 *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:new_Wmrk4",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Wmrk4" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Wmrk4" "', argument " "2"" of type '" "steps::wm::Geom *""'"); 
  }
  arg2 = reinterpret_cast< steps::wm::Geom * >(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Wmrk4" "', argument " "3"" of type '" "steps::rng::RNG *""'"); 
  }
  arg3 = reinterpret_cast< steps::rng::RNG * >(argp3);
  {
    try {
      result = (steps::wmrk4::Wmrk4 *)new steps::wmrk4::Wmrk4(arg1,arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wmrk4__Wmrk4, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Wmrk4(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Wmrk4",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Wmrk4" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_getSolverName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmrk4_getSolverName",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_getSolverName" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      result = ((steps::wmrk4::Wmrk4 const *)arg1)->getSolverName();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_getSolverDesc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmrk4_getSolverDesc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_getSolverDesc" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      result = ((steps::wmrk4::Wmrk4 const *)arg1)->getSolverDesc();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_getSolverAuthors(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmrk4_getSolverAuthors",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_getSolverAuthors" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      result = ((steps::wmrk4::Wmrk4 const *)arg1)->getSolverAuthors();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_getSolverEmail(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmrk4_getSolverEmail",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_getSolverEmail" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      result = ((steps::wmrk4::Wmrk4 const *)arg1)->getSolverEmail();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_reset(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmrk4_reset",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_reset" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      (arg1)->reset();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_run(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Wmrk4_run",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_run" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Wmrk4_run" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->run(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmrk4_getTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmrk4::Wmrk4 *arg1 = (steps::wmrk4::Wmrk4 *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmrk4_getTime",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmrk4__Wmrk4, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmrk4_getTime" "', argument " "1"" of type '" "steps::wmrk4::Wmrk4 const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmrk4::Wmrk4 * >(argp1);
  {
    try {
      result = (double)((steps::wmrk4::Wmrk4 const *)arg1)->getTime();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Wmrk4_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__wmrk4__Wmrk4, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Wmdirect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  steps::wm::Geom *arg2 = (steps::wm::Geom *) 0 ;
  steps::rng::RNG *arg3 = (steps::rng::RNG *) 0 ;
  steps::wmdirect::Wmdirect *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:new_Wmdirect",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Wmdirect" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Wmdirect" "', argument " "2"" of type '" "steps::wm::Geom *""'"); 
  }
  arg2 = reinterpret_cast< steps::wm::Geom * >(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Wmdirect" "', argument " "3"" of type '" "steps::rng::RNG *""'"); 
  }
  arg3 = reinterpret_cast< steps::rng::RNG * >(argp3);
  {
    try {
      result = (steps::wmdirect::Wmdirect *)new steps::wmdirect::Wmdirect(arg1,arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__wmdirect__Wmdirect, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Wmdirect(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Wmdirect",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Wmdirect" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_getSolverName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmdirect_getSolverName",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_getSolverName" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      result = ((steps::wmdirect::Wmdirect const *)arg1)->getSolverName();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_getSolverDesc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmdirect_getSolverDesc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_getSolverDesc" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      result = ((steps::wmdirect::Wmdirect const *)arg1)->getSolverDesc();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_getSolverAuthors(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmdirect_getSolverAuthors",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_getSolverAuthors" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      result = ((steps::wmdirect::Wmdirect const *)arg1)->getSolverAuthors();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_getSolverEmail(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmdirect_getSolverEmail",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_getSolverEmail" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      result = ((steps::wmdirect::Wmdirect const *)arg1)->getSolverEmail();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_reset(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmdirect_reset",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_reset" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      (arg1)->reset();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_run(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Wmdirect_run",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_run" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Wmdirect_run" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->run(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Wmdirect_getTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::wmdirect::Wmdirect *arg1 = (steps::wmdirect::Wmdirect *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Wmdirect_getTime",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__wmdirect__Wmdirect, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Wmdirect_getTime" "', argument " "1"" of type '" "steps::wmdirect::Wmdirect const *""'"); 
  }
  arg1 = reinterpret_cast< steps::wmdirect::Wmdirect * >(argp1);
  {
    try {
      result = (double)((steps::wmdirect::Wmdirect const *)arg1)->getTime();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Wmdirect_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__wmdirect__Wmdirect, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

SWIGINTERN PyObject *_wrap_new_Tetexact(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::model::Model *arg1 = (steps::model::Model *) 0 ;
  steps::wm::Geom *arg2 = (steps::wm::Geom *) 0 ;
  steps::rng::RNG *arg3 = (steps::rng::RNG *) 0 ;
  steps::tetexact::Tetexact *result = 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  void *argp2 = 0 ;
  int res2 = 0 ;
  void *argp3 = 0 ;
  int res3 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  PyObject * obj2 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OOO:new_Tetexact",&obj0,&obj1,&obj2)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__model__Model, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "new_Tetexact" "', argument " "1"" of type '" "steps::model::Model *""'"); 
  }
  arg1 = reinterpret_cast< steps::model::Model * >(argp1);
  res2 = SWIG_ConvertPtr(obj1, &argp2,SWIGTYPE_p_steps__wm__Geom, 0 |  0 );
  if (!SWIG_IsOK(res2)) {
    SWIG_exception_fail(SWIG_ArgError(res2), "in method '" "new_Tetexact" "', argument " "2"" of type '" "steps::wm::Geom *""'"); 
  }
  arg2 = reinterpret_cast< steps::wm::Geom * >(argp2);
  res3 = SWIG_ConvertPtr(obj2, &argp3,SWIGTYPE_p_steps__rng__RNG, 0 |  0 );
  if (!SWIG_IsOK(res3)) {
    SWIG_exception_fail(SWIG_ArgError(res3), "in method '" "new_Tetexact" "', argument " "3"" of type '" "steps::rng::RNG *""'"); 
  }
  arg3 = reinterpret_cast< steps::rng::RNG * >(argp3);
  {
    try {
      result = (steps::tetexact::Tetexact *)new steps::tetexact::Tetexact(arg1,arg2,arg3);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_NewPointerObj(SWIG_as_voidptr(result), SWIGTYPE_p_steps__tetexact__Tetexact, SWIG_POINTER_NEW |  0 );
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_delete_Tetexact(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:delete_Tetexact",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, SWIG_POINTER_DISOWN |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "delete_Tetexact" "', argument " "1"" of type '" "steps::tetexact::Tetexact *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      delete arg1;
      
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_getSolverName(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetexact_getSolverName",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_getSolverName" "', argument " "1"" of type '" "steps::tetexact::Tetexact const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      result = ((steps::tetexact::Tetexact const *)arg1)->getSolverName();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_getSolverDesc(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetexact_getSolverDesc",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_getSolverDesc" "', argument " "1"" of type '" "steps::tetexact::Tetexact const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      result = ((steps::tetexact::Tetexact const *)arg1)->getSolverDesc();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_getSolverAuthors(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetexact_getSolverAuthors",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_getSolverAuthors" "', argument " "1"" of type '" "steps::tetexact::Tetexact const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      result = ((steps::tetexact::Tetexact const *)arg1)->getSolverAuthors();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_getSolverEmail(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  std::string result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetexact_getSolverEmail",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_getSolverEmail" "', argument " "1"" of type '" "steps::tetexact::Tetexact const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      result = ((steps::tetexact::Tetexact const *)arg1)->getSolverEmail();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_std_string(static_cast< std::string >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_reset(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetexact_reset",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_reset" "', argument " "1"" of type '" "steps::tetexact::Tetexact *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      (arg1)->reset();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_run(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  double arg2 ;
  void *argp1 = 0 ;
  int res1 = 0 ;
  double val2 ;
  int ecode2 = 0 ;
  PyObject * obj0 = 0 ;
  PyObject * obj1 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"OO:Tetexact_run",&obj0,&obj1)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_run" "', argument " "1"" of type '" "steps::tetexact::Tetexact *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  ecode2 = SWIG_AsVal_double(obj1, &val2);
  if (!SWIG_IsOK(ecode2)) {
    SWIG_exception_fail(SWIG_ArgError(ecode2), "in method '" "Tetexact_run" "', argument " "2"" of type '" "double""'");
  } 
  arg2 = static_cast< double >(val2);
  {
    try {
      (arg1)->run(arg2);
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_Py_Void();
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *_wrap_Tetexact_getTime(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *resultobj = 0;
  steps::tetexact::Tetexact *arg1 = (steps::tetexact::Tetexact *) 0 ;
  double result;
  void *argp1 = 0 ;
  int res1 = 0 ;
  PyObject * obj0 = 0 ;
  
  if (!PyArg_ParseTuple(args,(char *)"O:Tetexact_getTime",&obj0)) SWIG_fail;
  res1 = SWIG_ConvertPtr(obj0, &argp1,SWIGTYPE_p_steps__tetexact__Tetexact, 0 |  0 );
  if (!SWIG_IsOK(res1)) {
    SWIG_exception_fail(SWIG_ArgError(res1), "in method '" "Tetexact_getTime" "', argument " "1"" of type '" "steps::tetexact::Tetexact const *""'"); 
  }
  arg1 = reinterpret_cast< steps::tetexact::Tetexact * >(argp1);
  {
    try {
      result = (double)((steps::tetexact::Tetexact const *)arg1)->getTime();
    } catch (steps::ArgErr & ae) {
      PyErr_SetString(PyExc_NameError, ae.getMsg());
      return NULL;
    } catch (steps::NotImplErr & nie) {
      PyErr_SetString(PyExc_NotImplementedError, nie.getMsg());
      return NULL;
    }
  }
  resultobj = SWIG_From_double(static_cast< double >(result));
  return resultobj;
fail:
  return NULL;
}


SWIGINTERN PyObject *Tetexact_swigregister(PyObject *SWIGUNUSEDPARM(self), PyObject *args) {
  PyObject *obj;
  if (!PyArg_ParseTuple(args,(char*)"O|swigregister", &obj)) return NULL;
  SWIG_TypeNewClientData(SWIGTYPE_p_steps__tetexact__Tetexact, SWIG_NewClientData(obj));
  return SWIG_Py_Void();
}

static PyMethodDef SwigMethods[] = {
	 { (char *)"delete_PySwigIterator", _wrap_delete_PySwigIterator, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_value", _wrap_PySwigIterator_value, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_incr", _wrap_PySwigIterator_incr, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_decr", _wrap_PySwigIterator_decr, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_distance", _wrap_PySwigIterator_distance, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_equal", _wrap_PySwigIterator_equal, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_copy", _wrap_PySwigIterator_copy, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_next", _wrap_PySwigIterator_next, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_previous", _wrap_PySwigIterator_previous, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_advance", _wrap_PySwigIterator_advance, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator___eq__", _wrap_PySwigIterator___eq__, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator___ne__", _wrap_PySwigIterator___ne__, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator___iadd__", _wrap_PySwigIterator___iadd__, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator___isub__", _wrap_PySwigIterator___isub__, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator___add__", _wrap_PySwigIterator___add__, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator___sub__", _wrap_PySwigIterator___sub__, METH_VARARGS, NULL},
	 { (char *)"PySwigIterator_swigregister", PySwigIterator_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Err", _wrap_new_Err, METH_VARARGS, NULL},
	 { (char *)"Err_getMsg", _wrap_Err_getMsg, METH_VARARGS, NULL},
	 { (char *)"delete_Err", _wrap_delete_Err, METH_VARARGS, NULL},
	 { (char *)"Err_swigregister", Err_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_ArgErr", _wrap_new_ArgErr, METH_VARARGS, NULL},
	 { (char *)"delete_ArgErr", _wrap_delete_ArgErr, METH_VARARGS, NULL},
	 { (char *)"ArgErr_swigregister", ArgErr_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_NotImplErr", _wrap_new_NotImplErr, METH_VARARGS, NULL},
	 { (char *)"delete_NotImplErr", _wrap_delete_NotImplErr, METH_VARARGS, NULL},
	 { (char *)"NotImplErr_swigregister", NotImplErr_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_str_iterator", _wrap_vector_str_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_str___nonzero__", _wrap_vector_str___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_str___len__", _wrap_vector_str___len__, METH_VARARGS, NULL},
	 { (char *)"vector_str_pop", _wrap_vector_str_pop, METH_VARARGS, NULL},
	 { (char *)"vector_str___getslice__", _wrap_vector_str___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_str___setslice__", _wrap_vector_str___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_str___delslice__", _wrap_vector_str___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_str___delitem__", _wrap_vector_str___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_str___getitem__", _wrap_vector_str___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_str___setitem__", _wrap_vector_str___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_str_append", _wrap_vector_str_append, METH_VARARGS, NULL},
	 { (char *)"vector_str_empty", _wrap_vector_str_empty, METH_VARARGS, NULL},
	 { (char *)"vector_str_size", _wrap_vector_str_size, METH_VARARGS, NULL},
	 { (char *)"vector_str_clear", _wrap_vector_str_clear, METH_VARARGS, NULL},
	 { (char *)"vector_str_swap", _wrap_vector_str_swap, METH_VARARGS, NULL},
	 { (char *)"vector_str_get_allocator", _wrap_vector_str_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_str_begin", _wrap_vector_str_begin, METH_VARARGS, NULL},
	 { (char *)"vector_str_end", _wrap_vector_str_end, METH_VARARGS, NULL},
	 { (char *)"vector_str_rbegin", _wrap_vector_str_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_str_rend", _wrap_vector_str_rend, METH_VARARGS, NULL},
	 { (char *)"vector_str_pop_back", _wrap_vector_str_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_str_erase", _wrap_vector_str_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_str", _wrap_new_vector_str, METH_VARARGS, NULL},
	 { (char *)"vector_str_push_back", _wrap_vector_str_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_str_front", _wrap_vector_str_front, METH_VARARGS, NULL},
	 { (char *)"vector_str_back", _wrap_vector_str_back, METH_VARARGS, NULL},
	 { (char *)"vector_str_assign", _wrap_vector_str_assign, METH_VARARGS, NULL},
	 { (char *)"vector_str_resize", _wrap_vector_str_resize, METH_VARARGS, NULL},
	 { (char *)"vector_str_insert", _wrap_vector_str_insert, METH_VARARGS, NULL},
	 { (char *)"vector_str_reserve", _wrap_vector_str_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_str_capacity", _wrap_vector_str_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_str", _wrap_delete_vector_str, METH_VARARGS, NULL},
	 { (char *)"vector_str_swigregister", vector_str_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_spc_iterator", _wrap_vector_spc_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_spc___nonzero__", _wrap_vector_spc___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_spc___len__", _wrap_vector_spc___len__, METH_VARARGS, NULL},
	 { (char *)"vector_spc_pop", _wrap_vector_spc_pop, METH_VARARGS, NULL},
	 { (char *)"vector_spc___getslice__", _wrap_vector_spc___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_spc___setslice__", _wrap_vector_spc___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_spc___delslice__", _wrap_vector_spc___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_spc___delitem__", _wrap_vector_spc___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_spc___getitem__", _wrap_vector_spc___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_spc___setitem__", _wrap_vector_spc___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_spc_append", _wrap_vector_spc_append, METH_VARARGS, NULL},
	 { (char *)"vector_spc_empty", _wrap_vector_spc_empty, METH_VARARGS, NULL},
	 { (char *)"vector_spc_size", _wrap_vector_spc_size, METH_VARARGS, NULL},
	 { (char *)"vector_spc_clear", _wrap_vector_spc_clear, METH_VARARGS, NULL},
	 { (char *)"vector_spc_swap", _wrap_vector_spc_swap, METH_VARARGS, NULL},
	 { (char *)"vector_spc_get_allocator", _wrap_vector_spc_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_spc_begin", _wrap_vector_spc_begin, METH_VARARGS, NULL},
	 { (char *)"vector_spc_end", _wrap_vector_spc_end, METH_VARARGS, NULL},
	 { (char *)"vector_spc_rbegin", _wrap_vector_spc_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_spc_rend", _wrap_vector_spc_rend, METH_VARARGS, NULL},
	 { (char *)"vector_spc_pop_back", _wrap_vector_spc_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_spc_erase", _wrap_vector_spc_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_spc", _wrap_new_vector_spc, METH_VARARGS, NULL},
	 { (char *)"vector_spc_push_back", _wrap_vector_spc_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_spc_front", _wrap_vector_spc_front, METH_VARARGS, NULL},
	 { (char *)"vector_spc_back", _wrap_vector_spc_back, METH_VARARGS, NULL},
	 { (char *)"vector_spc_assign", _wrap_vector_spc_assign, METH_VARARGS, NULL},
	 { (char *)"vector_spc_resize", _wrap_vector_spc_resize, METH_VARARGS, NULL},
	 { (char *)"vector_spc_insert", _wrap_vector_spc_insert, METH_VARARGS, NULL},
	 { (char *)"vector_spc_reserve", _wrap_vector_spc_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_spc_capacity", _wrap_vector_spc_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_spc", _wrap_delete_vector_spc, METH_VARARGS, NULL},
	 { (char *)"vector_spc_swigregister", vector_spc_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_rec_iterator", _wrap_vector_rec_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_rec___nonzero__", _wrap_vector_rec___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_rec___len__", _wrap_vector_rec___len__, METH_VARARGS, NULL},
	 { (char *)"vector_rec_pop", _wrap_vector_rec_pop, METH_VARARGS, NULL},
	 { (char *)"vector_rec___getslice__", _wrap_vector_rec___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_rec___setslice__", _wrap_vector_rec___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_rec___delslice__", _wrap_vector_rec___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_rec___delitem__", _wrap_vector_rec___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_rec___getitem__", _wrap_vector_rec___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_rec___setitem__", _wrap_vector_rec___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_rec_append", _wrap_vector_rec_append, METH_VARARGS, NULL},
	 { (char *)"vector_rec_empty", _wrap_vector_rec_empty, METH_VARARGS, NULL},
	 { (char *)"vector_rec_size", _wrap_vector_rec_size, METH_VARARGS, NULL},
	 { (char *)"vector_rec_clear", _wrap_vector_rec_clear, METH_VARARGS, NULL},
	 { (char *)"vector_rec_swap", _wrap_vector_rec_swap, METH_VARARGS, NULL},
	 { (char *)"vector_rec_get_allocator", _wrap_vector_rec_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_rec_begin", _wrap_vector_rec_begin, METH_VARARGS, NULL},
	 { (char *)"vector_rec_end", _wrap_vector_rec_end, METH_VARARGS, NULL},
	 { (char *)"vector_rec_rbegin", _wrap_vector_rec_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_rec_rend", _wrap_vector_rec_rend, METH_VARARGS, NULL},
	 { (char *)"vector_rec_pop_back", _wrap_vector_rec_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_rec_erase", _wrap_vector_rec_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_rec", _wrap_new_vector_rec, METH_VARARGS, NULL},
	 { (char *)"vector_rec_push_back", _wrap_vector_rec_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_rec_front", _wrap_vector_rec_front, METH_VARARGS, NULL},
	 { (char *)"vector_rec_back", _wrap_vector_rec_back, METH_VARARGS, NULL},
	 { (char *)"vector_rec_assign", _wrap_vector_rec_assign, METH_VARARGS, NULL},
	 { (char *)"vector_rec_resize", _wrap_vector_rec_resize, METH_VARARGS, NULL},
	 { (char *)"vector_rec_insert", _wrap_vector_rec_insert, METH_VARARGS, NULL},
	 { (char *)"vector_rec_reserve", _wrap_vector_rec_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_rec_capacity", _wrap_vector_rec_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_rec", _wrap_delete_vector_rec, METH_VARARGS, NULL},
	 { (char *)"vector_rec_swigregister", vector_rec_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_src_iterator", _wrap_vector_src_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_src___nonzero__", _wrap_vector_src___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_src___len__", _wrap_vector_src___len__, METH_VARARGS, NULL},
	 { (char *)"vector_src_pop", _wrap_vector_src_pop, METH_VARARGS, NULL},
	 { (char *)"vector_src___getslice__", _wrap_vector_src___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_src___setslice__", _wrap_vector_src___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_src___delslice__", _wrap_vector_src___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_src___delitem__", _wrap_vector_src___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_src___getitem__", _wrap_vector_src___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_src___setitem__", _wrap_vector_src___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_src_append", _wrap_vector_src_append, METH_VARARGS, NULL},
	 { (char *)"vector_src_empty", _wrap_vector_src_empty, METH_VARARGS, NULL},
	 { (char *)"vector_src_size", _wrap_vector_src_size, METH_VARARGS, NULL},
	 { (char *)"vector_src_clear", _wrap_vector_src_clear, METH_VARARGS, NULL},
	 { (char *)"vector_src_swap", _wrap_vector_src_swap, METH_VARARGS, NULL},
	 { (char *)"vector_src_get_allocator", _wrap_vector_src_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_src_begin", _wrap_vector_src_begin, METH_VARARGS, NULL},
	 { (char *)"vector_src_end", _wrap_vector_src_end, METH_VARARGS, NULL},
	 { (char *)"vector_src_rbegin", _wrap_vector_src_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_src_rend", _wrap_vector_src_rend, METH_VARARGS, NULL},
	 { (char *)"vector_src_pop_back", _wrap_vector_src_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_src_erase", _wrap_vector_src_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_src", _wrap_new_vector_src, METH_VARARGS, NULL},
	 { (char *)"vector_src_push_back", _wrap_vector_src_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_src_front", _wrap_vector_src_front, METH_VARARGS, NULL},
	 { (char *)"vector_src_back", _wrap_vector_src_back, METH_VARARGS, NULL},
	 { (char *)"vector_src_assign", _wrap_vector_src_assign, METH_VARARGS, NULL},
	 { (char *)"vector_src_resize", _wrap_vector_src_resize, METH_VARARGS, NULL},
	 { (char *)"vector_src_insert", _wrap_vector_src_insert, METH_VARARGS, NULL},
	 { (char *)"vector_src_reserve", _wrap_vector_src_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_src_capacity", _wrap_vector_src_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_src", _wrap_delete_vector_src, METH_VARARGS, NULL},
	 { (char *)"vector_src_swigregister", vector_src_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_dif_iterator", _wrap_vector_dif_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_dif___nonzero__", _wrap_vector_dif___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_dif___len__", _wrap_vector_dif___len__, METH_VARARGS, NULL},
	 { (char *)"vector_dif_pop", _wrap_vector_dif_pop, METH_VARARGS, NULL},
	 { (char *)"vector_dif___getslice__", _wrap_vector_dif___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_dif___setslice__", _wrap_vector_dif___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_dif___delslice__", _wrap_vector_dif___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_dif___delitem__", _wrap_vector_dif___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_dif___getitem__", _wrap_vector_dif___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_dif___setitem__", _wrap_vector_dif___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_dif_append", _wrap_vector_dif_append, METH_VARARGS, NULL},
	 { (char *)"vector_dif_empty", _wrap_vector_dif_empty, METH_VARARGS, NULL},
	 { (char *)"vector_dif_size", _wrap_vector_dif_size, METH_VARARGS, NULL},
	 { (char *)"vector_dif_clear", _wrap_vector_dif_clear, METH_VARARGS, NULL},
	 { (char *)"vector_dif_swap", _wrap_vector_dif_swap, METH_VARARGS, NULL},
	 { (char *)"vector_dif_get_allocator", _wrap_vector_dif_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_dif_begin", _wrap_vector_dif_begin, METH_VARARGS, NULL},
	 { (char *)"vector_dif_end", _wrap_vector_dif_end, METH_VARARGS, NULL},
	 { (char *)"vector_dif_rbegin", _wrap_vector_dif_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_dif_rend", _wrap_vector_dif_rend, METH_VARARGS, NULL},
	 { (char *)"vector_dif_pop_back", _wrap_vector_dif_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_dif_erase", _wrap_vector_dif_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_dif", _wrap_new_vector_dif, METH_VARARGS, NULL},
	 { (char *)"vector_dif_push_back", _wrap_vector_dif_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_dif_front", _wrap_vector_dif_front, METH_VARARGS, NULL},
	 { (char *)"vector_dif_back", _wrap_vector_dif_back, METH_VARARGS, NULL},
	 { (char *)"vector_dif_assign", _wrap_vector_dif_assign, METH_VARARGS, NULL},
	 { (char *)"vector_dif_resize", _wrap_vector_dif_resize, METH_VARARGS, NULL},
	 { (char *)"vector_dif_insert", _wrap_vector_dif_insert, METH_VARARGS, NULL},
	 { (char *)"vector_dif_reserve", _wrap_vector_dif_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_dif_capacity", _wrap_vector_dif_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_dif", _wrap_delete_vector_dif, METH_VARARGS, NULL},
	 { (char *)"vector_dif_swigregister", vector_dif_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_iterator", _wrap_vector_vsys_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___nonzero__", _wrap_vector_vsys___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___len__", _wrap_vector_vsys___len__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_pop", _wrap_vector_vsys_pop, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___getslice__", _wrap_vector_vsys___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___setslice__", _wrap_vector_vsys___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___delslice__", _wrap_vector_vsys___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___delitem__", _wrap_vector_vsys___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___getitem__", _wrap_vector_vsys___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys___setitem__", _wrap_vector_vsys___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_append", _wrap_vector_vsys_append, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_empty", _wrap_vector_vsys_empty, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_size", _wrap_vector_vsys_size, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_clear", _wrap_vector_vsys_clear, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_swap", _wrap_vector_vsys_swap, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_get_allocator", _wrap_vector_vsys_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_begin", _wrap_vector_vsys_begin, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_end", _wrap_vector_vsys_end, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_rbegin", _wrap_vector_vsys_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_rend", _wrap_vector_vsys_rend, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_pop_back", _wrap_vector_vsys_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_erase", _wrap_vector_vsys_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_vsys", _wrap_new_vector_vsys, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_push_back", _wrap_vector_vsys_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_front", _wrap_vector_vsys_front, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_back", _wrap_vector_vsys_back, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_assign", _wrap_vector_vsys_assign, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_resize", _wrap_vector_vsys_resize, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_insert", _wrap_vector_vsys_insert, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_reserve", _wrap_vector_vsys_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_capacity", _wrap_vector_vsys_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_vsys", _wrap_delete_vector_vsys, METH_VARARGS, NULL},
	 { (char *)"vector_vsys_swigregister", vector_vsys_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_iterator", _wrap_vector_ssys_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___nonzero__", _wrap_vector_ssys___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___len__", _wrap_vector_ssys___len__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_pop", _wrap_vector_ssys_pop, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___getslice__", _wrap_vector_ssys___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___setslice__", _wrap_vector_ssys___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___delslice__", _wrap_vector_ssys___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___delitem__", _wrap_vector_ssys___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___getitem__", _wrap_vector_ssys___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys___setitem__", _wrap_vector_ssys___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_append", _wrap_vector_ssys_append, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_empty", _wrap_vector_ssys_empty, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_size", _wrap_vector_ssys_size, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_clear", _wrap_vector_ssys_clear, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_swap", _wrap_vector_ssys_swap, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_get_allocator", _wrap_vector_ssys_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_begin", _wrap_vector_ssys_begin, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_end", _wrap_vector_ssys_end, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_rbegin", _wrap_vector_ssys_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_rend", _wrap_vector_ssys_rend, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_pop_back", _wrap_vector_ssys_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_erase", _wrap_vector_ssys_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_ssys", _wrap_new_vector_ssys, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_push_back", _wrap_vector_ssys_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_front", _wrap_vector_ssys_front, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_back", _wrap_vector_ssys_back, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_assign", _wrap_vector_ssys_assign, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_resize", _wrap_vector_ssys_resize, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_insert", _wrap_vector_ssys_insert, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_reserve", _wrap_vector_ssys_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_capacity", _wrap_vector_ssys_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_ssys", _wrap_delete_vector_ssys, METH_VARARGS, NULL},
	 { (char *)"vector_ssys_swigregister", vector_ssys_swigregister, METH_VARARGS, NULL},
	 { (char *)"isValidID", _wrap_isValidID, METH_VARARGS, NULL},
	 { (char *)"checkID", _wrap_checkID, METH_VARARGS, NULL},
	 { (char *)"new_Model", _wrap_new_Model, METH_VARARGS, NULL},
	 { (char *)"delete_Model", _wrap_delete_Model, METH_VARARGS, NULL},
	 { (char *)"Model_getSpec", _wrap_Model_getSpec, METH_VARARGS, NULL},
	 { (char *)"Model_delSpec", _wrap_Model_delSpec, METH_VARARGS, NULL},
	 { (char *)"Model_getAllSpecs", _wrap_Model_getAllSpecs, METH_VARARGS, NULL},
	 { (char *)"Model_getVolsys", _wrap_Model_getVolsys, METH_VARARGS, NULL},
	 { (char *)"Model_delVolsys", _wrap_Model_delVolsys, METH_VARARGS, NULL},
	 { (char *)"Model_getAllVolsyss", _wrap_Model_getAllVolsyss, METH_VARARGS, NULL},
	 { (char *)"Model_getSurfsys", _wrap_Model_getSurfsys, METH_VARARGS, NULL},
	 { (char *)"Model_delSurfsys", _wrap_Model_delSurfsys, METH_VARARGS, NULL},
	 { (char *)"Model_getAllSurfsyss", _wrap_Model_getAllSurfsyss, METH_VARARGS, NULL},
	 { (char *)"Model_swigregister", Model_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Spec", _wrap_new_Spec, METH_VARARGS, NULL},
	 { (char *)"delete_Spec", _wrap_delete_Spec, METH_VARARGS, NULL},
	 { (char *)"Spec_getID", _wrap_Spec_getID, METH_VARARGS, NULL},
	 { (char *)"Spec_setID", _wrap_Spec_setID, METH_VARARGS, NULL},
	 { (char *)"Spec_getModel", _wrap_Spec_getModel, METH_VARARGS, NULL},
	 { (char *)"Spec_swigregister", Spec_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Surfsys", _wrap_new_Surfsys, METH_VARARGS, NULL},
	 { (char *)"delete_Surfsys", _wrap_delete_Surfsys, METH_VARARGS, NULL},
	 { (char *)"Surfsys_getID", _wrap_Surfsys_getID, METH_VARARGS, NULL},
	 { (char *)"Surfsys_setID", _wrap_Surfsys_setID, METH_VARARGS, NULL},
	 { (char *)"Surfsys_getModel", _wrap_Surfsys_getModel, METH_VARARGS, NULL},
	 { (char *)"Surfsys_getSReac", _wrap_Surfsys_getSReac, METH_VARARGS, NULL},
	 { (char *)"Surfsys_delSReac", _wrap_Surfsys_delSReac, METH_VARARGS, NULL},
	 { (char *)"Surfsys_getAllSReacs", _wrap_Surfsys_getAllSReacs, METH_VARARGS, NULL},
	 { (char *)"Surfsys_getAllSpecs", _wrap_Surfsys_getAllSpecs, METH_VARARGS, NULL},
	 { (char *)"Surfsys_swigregister", Surfsys_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Volsys", _wrap_new_Volsys, METH_VARARGS, NULL},
	 { (char *)"delete_Volsys", _wrap_delete_Volsys, METH_VARARGS, NULL},
	 { (char *)"Volsys_getID", _wrap_Volsys_getID, METH_VARARGS, NULL},
	 { (char *)"Volsys_setID", _wrap_Volsys_setID, METH_VARARGS, NULL},
	 { (char *)"Volsys_getModel", _wrap_Volsys_getModel, METH_VARARGS, NULL},
	 { (char *)"Volsys_getReac", _wrap_Volsys_getReac, METH_VARARGS, NULL},
	 { (char *)"Volsys_delReac", _wrap_Volsys_delReac, METH_VARARGS, NULL},
	 { (char *)"Volsys_getAllReacs", _wrap_Volsys_getAllReacs, METH_VARARGS, NULL},
	 { (char *)"Volsys_getDiff", _wrap_Volsys_getDiff, METH_VARARGS, NULL},
	 { (char *)"Volsys_delDiff", _wrap_Volsys_delDiff, METH_VARARGS, NULL},
	 { (char *)"Volsys_getAllDiffs", _wrap_Volsys_getAllDiffs, METH_VARARGS, NULL},
	 { (char *)"Volsys_getAllSpecs", _wrap_Volsys_getAllSpecs, METH_VARARGS, NULL},
	 { (char *)"Volsys_swigregister", Volsys_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Diff", (PyCFunction) _wrap_new_Diff, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_Diff", _wrap_delete_Diff, METH_VARARGS, NULL},
	 { (char *)"Diff_getID", _wrap_Diff_getID, METH_VARARGS, NULL},
	 { (char *)"Diff_setID", _wrap_Diff_setID, METH_VARARGS, NULL},
	 { (char *)"Diff_getVolsys", _wrap_Diff_getVolsys, METH_VARARGS, NULL},
	 { (char *)"Diff_getModel", _wrap_Diff_getModel, METH_VARARGS, NULL},
	 { (char *)"Diff_getLig", _wrap_Diff_getLig, METH_VARARGS, NULL},
	 { (char *)"Diff_setLig", _wrap_Diff_setLig, METH_VARARGS, NULL},
	 { (char *)"Diff_getDcst", _wrap_Diff_getDcst, METH_VARARGS, NULL},
	 { (char *)"Diff_setDcst", _wrap_Diff_setDcst, METH_VARARGS, NULL},
	 { (char *)"Diff_swigregister", Diff_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Reac", (PyCFunction) _wrap_new_Reac, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_Reac", _wrap_delete_Reac, METH_VARARGS, NULL},
	 { (char *)"Reac_getID", _wrap_Reac_getID, METH_VARARGS, NULL},
	 { (char *)"Reac_setID", _wrap_Reac_setID, METH_VARARGS, NULL},
	 { (char *)"Reac_getVolsys", _wrap_Reac_getVolsys, METH_VARARGS, NULL},
	 { (char *)"Reac_getModel", _wrap_Reac_getModel, METH_VARARGS, NULL},
	 { (char *)"Reac_getLHS", _wrap_Reac_getLHS, METH_VARARGS, NULL},
	 { (char *)"Reac_setLHS", _wrap_Reac_setLHS, METH_VARARGS, NULL},
	 { (char *)"Reac_getRHS", _wrap_Reac_getRHS, METH_VARARGS, NULL},
	 { (char *)"Reac_setRHS", _wrap_Reac_setRHS, METH_VARARGS, NULL},
	 { (char *)"Reac_getOrder", _wrap_Reac_getOrder, METH_VARARGS, NULL},
	 { (char *)"Reac_getKcst", _wrap_Reac_getKcst, METH_VARARGS, NULL},
	 { (char *)"Reac_setKcst", _wrap_Reac_setKcst, METH_VARARGS, NULL},
	 { (char *)"Reac_getAllSpecs", _wrap_Reac_getAllSpecs, METH_VARARGS, NULL},
	 { (char *)"Reac_swigregister", Reac_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_SReac", (PyCFunction) _wrap_new_SReac, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_SReac", _wrap_delete_SReac, METH_VARARGS, NULL},
	 { (char *)"SReac_getID", _wrap_SReac_getID, METH_VARARGS, NULL},
	 { (char *)"SReac_setID", _wrap_SReac_setID, METH_VARARGS, NULL},
	 { (char *)"SReac_getSurfsys", _wrap_SReac_getSurfsys, METH_VARARGS, NULL},
	 { (char *)"SReac_getModel", _wrap_SReac_getModel, METH_VARARGS, NULL},
	 { (char *)"SReac_getInner", _wrap_SReac_getInner, METH_VARARGS, NULL},
	 { (char *)"SReac_getOuter", _wrap_SReac_getOuter, METH_VARARGS, NULL},
	 { (char *)"SReac_getOLHS", _wrap_SReac_getOLHS, METH_VARARGS, NULL},
	 { (char *)"SReac_setOLHS", _wrap_SReac_setOLHS, METH_VARARGS, NULL},
	 { (char *)"SReac_getILHS", _wrap_SReac_getILHS, METH_VARARGS, NULL},
	 { (char *)"SReac_setILHS", _wrap_SReac_setILHS, METH_VARARGS, NULL},
	 { (char *)"SReac_getSLHS", _wrap_SReac_getSLHS, METH_VARARGS, NULL},
	 { (char *)"SReac_setSLHS", _wrap_SReac_setSLHS, METH_VARARGS, NULL},
	 { (char *)"SReac_getIRHS", _wrap_SReac_getIRHS, METH_VARARGS, NULL},
	 { (char *)"SReac_setIRHS", _wrap_SReac_setIRHS, METH_VARARGS, NULL},
	 { (char *)"SReac_getSRHS", _wrap_SReac_getSRHS, METH_VARARGS, NULL},
	 { (char *)"SReac_setSRHS", _wrap_SReac_setSRHS, METH_VARARGS, NULL},
	 { (char *)"SReac_getORHS", _wrap_SReac_getORHS, METH_VARARGS, NULL},
	 { (char *)"SReac_setORHS", _wrap_SReac_setORHS, METH_VARARGS, NULL},
	 { (char *)"SReac_getOrder", _wrap_SReac_getOrder, METH_VARARGS, NULL},
	 { (char *)"SReac_getKcst", _wrap_SReac_getKcst, METH_VARARGS, NULL},
	 { (char *)"SReac_setKcst", _wrap_SReac_setKcst, METH_VARARGS, NULL},
	 { (char *)"SReac_getAllSpecs", _wrap_SReac_getAllSpecs, METH_VARARGS, NULL},
	 { (char *)"SReac_swigregister", SReac_swigregister, METH_VARARGS, NULL},
	 { (char *)"set_str_iterator", _wrap_set_str_iterator, METH_VARARGS, NULL},
	 { (char *)"set_str___nonzero__", _wrap_set_str___nonzero__, METH_VARARGS, NULL},
	 { (char *)"set_str___len__", _wrap_set_str___len__, METH_VARARGS, NULL},
	 { (char *)"set_str_append", _wrap_set_str_append, METH_VARARGS, NULL},
	 { (char *)"set_str___contains__", _wrap_set_str___contains__, METH_VARARGS, NULL},
	 { (char *)"set_str___getitem__", _wrap_set_str___getitem__, METH_VARARGS, NULL},
	 { (char *)"new_set_str", _wrap_new_set_str, METH_VARARGS, NULL},
	 { (char *)"set_str_empty", _wrap_set_str_empty, METH_VARARGS, NULL},
	 { (char *)"set_str_size", _wrap_set_str_size, METH_VARARGS, NULL},
	 { (char *)"set_str_clear", _wrap_set_str_clear, METH_VARARGS, NULL},
	 { (char *)"set_str_swap", _wrap_set_str_swap, METH_VARARGS, NULL},
	 { (char *)"set_str_get_allocator", _wrap_set_str_get_allocator, METH_VARARGS, NULL},
	 { (char *)"set_str_begin", _wrap_set_str_begin, METH_VARARGS, NULL},
	 { (char *)"set_str_end", _wrap_set_str_end, METH_VARARGS, NULL},
	 { (char *)"set_str_rbegin", _wrap_set_str_rbegin, METH_VARARGS, NULL},
	 { (char *)"set_str_rend", _wrap_set_str_rend, METH_VARARGS, NULL},
	 { (char *)"set_str_count", _wrap_set_str_count, METH_VARARGS, NULL},
	 { (char *)"set_str_erase", _wrap_set_str_erase, METH_VARARGS, NULL},
	 { (char *)"set_str_find", _wrap_set_str_find, METH_VARARGS, NULL},
	 { (char *)"set_str_lower_bound", _wrap_set_str_lower_bound, METH_VARARGS, NULL},
	 { (char *)"set_str_upper_bound", _wrap_set_str_upper_bound, METH_VARARGS, NULL},
	 { (char *)"set_str_equal_range", _wrap_set_str_equal_range, METH_VARARGS, NULL},
	 { (char *)"set_str_insert", _wrap_set_str_insert, METH_VARARGS, NULL},
	 { (char *)"delete_set_str", _wrap_delete_set_str, METH_VARARGS, NULL},
	 { (char *)"set_str_swigregister", set_str_swigregister, METH_VARARGS, NULL},
	 { (char *)"set_ptc_iterator", _wrap_set_ptc_iterator, METH_VARARGS, NULL},
	 { (char *)"set_ptc___nonzero__", _wrap_set_ptc___nonzero__, METH_VARARGS, NULL},
	 { (char *)"set_ptc___len__", _wrap_set_ptc___len__, METH_VARARGS, NULL},
	 { (char *)"set_ptc_append", _wrap_set_ptc_append, METH_VARARGS, NULL},
	 { (char *)"set_ptc___contains__", _wrap_set_ptc___contains__, METH_VARARGS, NULL},
	 { (char *)"set_ptc___getitem__", _wrap_set_ptc___getitem__, METH_VARARGS, NULL},
	 { (char *)"new_set_ptc", _wrap_new_set_ptc, METH_VARARGS, NULL},
	 { (char *)"set_ptc_empty", _wrap_set_ptc_empty, METH_VARARGS, NULL},
	 { (char *)"set_ptc_size", _wrap_set_ptc_size, METH_VARARGS, NULL},
	 { (char *)"set_ptc_clear", _wrap_set_ptc_clear, METH_VARARGS, NULL},
	 { (char *)"set_ptc_swap", _wrap_set_ptc_swap, METH_VARARGS, NULL},
	 { (char *)"set_ptc_get_allocator", _wrap_set_ptc_get_allocator, METH_VARARGS, NULL},
	 { (char *)"set_ptc_begin", _wrap_set_ptc_begin, METH_VARARGS, NULL},
	 { (char *)"set_ptc_end", _wrap_set_ptc_end, METH_VARARGS, NULL},
	 { (char *)"set_ptc_rbegin", _wrap_set_ptc_rbegin, METH_VARARGS, NULL},
	 { (char *)"set_ptc_rend", _wrap_set_ptc_rend, METH_VARARGS, NULL},
	 { (char *)"set_ptc_count", _wrap_set_ptc_count, METH_VARARGS, NULL},
	 { (char *)"set_ptc_erase", _wrap_set_ptc_erase, METH_VARARGS, NULL},
	 { (char *)"set_ptc_find", _wrap_set_ptc_find, METH_VARARGS, NULL},
	 { (char *)"set_ptc_lower_bound", _wrap_set_ptc_lower_bound, METH_VARARGS, NULL},
	 { (char *)"set_ptc_upper_bound", _wrap_set_ptc_upper_bound, METH_VARARGS, NULL},
	 { (char *)"set_ptc_equal_range", _wrap_set_ptc_equal_range, METH_VARARGS, NULL},
	 { (char *)"set_ptc_insert", _wrap_set_ptc_insert, METH_VARARGS, NULL},
	 { (char *)"delete_set_ptc", _wrap_delete_set_ptc, METH_VARARGS, NULL},
	 { (char *)"set_ptc_swigregister", set_ptc_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_iterator", _wrap_vector_ptc_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___nonzero__", _wrap_vector_ptc___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___len__", _wrap_vector_ptc___len__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_pop", _wrap_vector_ptc_pop, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___getslice__", _wrap_vector_ptc___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___setslice__", _wrap_vector_ptc___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___delslice__", _wrap_vector_ptc___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___delitem__", _wrap_vector_ptc___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___getitem__", _wrap_vector_ptc___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc___setitem__", _wrap_vector_ptc___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_append", _wrap_vector_ptc_append, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_empty", _wrap_vector_ptc_empty, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_size", _wrap_vector_ptc_size, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_clear", _wrap_vector_ptc_clear, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_swap", _wrap_vector_ptc_swap, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_get_allocator", _wrap_vector_ptc_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_begin", _wrap_vector_ptc_begin, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_end", _wrap_vector_ptc_end, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_rbegin", _wrap_vector_ptc_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_rend", _wrap_vector_ptc_rend, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_pop_back", _wrap_vector_ptc_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_erase", _wrap_vector_ptc_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_ptc", _wrap_new_vector_ptc, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_push_back", _wrap_vector_ptc_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_front", _wrap_vector_ptc_front, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_back", _wrap_vector_ptc_back, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_assign", _wrap_vector_ptc_assign, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_resize", _wrap_vector_ptc_resize, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_insert", _wrap_vector_ptc_insert, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_reserve", _wrap_vector_ptc_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_capacity", _wrap_vector_ptc_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_ptc", _wrap_delete_vector_ptc, METH_VARARGS, NULL},
	 { (char *)"vector_ptc_swigregister", vector_ptc_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_iterator", _wrap_vector_cmp_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___nonzero__", _wrap_vector_cmp___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___len__", _wrap_vector_cmp___len__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_pop", _wrap_vector_cmp_pop, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___getslice__", _wrap_vector_cmp___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___setslice__", _wrap_vector_cmp___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___delslice__", _wrap_vector_cmp___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___delitem__", _wrap_vector_cmp___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___getitem__", _wrap_vector_cmp___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp___setitem__", _wrap_vector_cmp___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_append", _wrap_vector_cmp_append, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_empty", _wrap_vector_cmp_empty, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_size", _wrap_vector_cmp_size, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_clear", _wrap_vector_cmp_clear, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_swap", _wrap_vector_cmp_swap, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_get_allocator", _wrap_vector_cmp_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_begin", _wrap_vector_cmp_begin, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_end", _wrap_vector_cmp_end, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_rbegin", _wrap_vector_cmp_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_rend", _wrap_vector_cmp_rend, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_pop_back", _wrap_vector_cmp_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_erase", _wrap_vector_cmp_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_cmp", _wrap_new_vector_cmp, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_push_back", _wrap_vector_cmp_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_front", _wrap_vector_cmp_front, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_back", _wrap_vector_cmp_back, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_assign", _wrap_vector_cmp_assign, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_resize", _wrap_vector_cmp_resize, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_insert", _wrap_vector_cmp_insert, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_reserve", _wrap_vector_cmp_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_capacity", _wrap_vector_cmp_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_cmp", _wrap_delete_vector_cmp, METH_VARARGS, NULL},
	 { (char *)"vector_cmp_swigregister", vector_cmp_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_int_iterator", _wrap_vector_int_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_int___nonzero__", _wrap_vector_int___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_int___len__", _wrap_vector_int___len__, METH_VARARGS, NULL},
	 { (char *)"vector_int_pop", _wrap_vector_int_pop, METH_VARARGS, NULL},
	 { (char *)"vector_int___getslice__", _wrap_vector_int___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_int___setslice__", _wrap_vector_int___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_int___delslice__", _wrap_vector_int___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_int___delitem__", _wrap_vector_int___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_int___getitem__", _wrap_vector_int___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_int___setitem__", _wrap_vector_int___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_int_append", _wrap_vector_int_append, METH_VARARGS, NULL},
	 { (char *)"vector_int_empty", _wrap_vector_int_empty, METH_VARARGS, NULL},
	 { (char *)"vector_int_size", _wrap_vector_int_size, METH_VARARGS, NULL},
	 { (char *)"vector_int_clear", _wrap_vector_int_clear, METH_VARARGS, NULL},
	 { (char *)"vector_int_swap", _wrap_vector_int_swap, METH_VARARGS, NULL},
	 { (char *)"vector_int_get_allocator", _wrap_vector_int_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_int_begin", _wrap_vector_int_begin, METH_VARARGS, NULL},
	 { (char *)"vector_int_end", _wrap_vector_int_end, METH_VARARGS, NULL},
	 { (char *)"vector_int_rbegin", _wrap_vector_int_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_int_rend", _wrap_vector_int_rend, METH_VARARGS, NULL},
	 { (char *)"vector_int_pop_back", _wrap_vector_int_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_int_erase", _wrap_vector_int_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_int", _wrap_new_vector_int, METH_VARARGS, NULL},
	 { (char *)"vector_int_push_back", _wrap_vector_int_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_int_front", _wrap_vector_int_front, METH_VARARGS, NULL},
	 { (char *)"vector_int_back", _wrap_vector_int_back, METH_VARARGS, NULL},
	 { (char *)"vector_int_assign", _wrap_vector_int_assign, METH_VARARGS, NULL},
	 { (char *)"vector_int_resize", _wrap_vector_int_resize, METH_VARARGS, NULL},
	 { (char *)"vector_int_insert", _wrap_vector_int_insert, METH_VARARGS, NULL},
	 { (char *)"vector_int_reserve", _wrap_vector_int_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_int_capacity", _wrap_vector_int_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_int", _wrap_delete_vector_int, METH_VARARGS, NULL},
	 { (char *)"vector_int_swigregister", vector_int_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_uint_iterator", _wrap_vector_uint_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_uint___nonzero__", _wrap_vector_uint___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_uint___len__", _wrap_vector_uint___len__, METH_VARARGS, NULL},
	 { (char *)"vector_uint_pop", _wrap_vector_uint_pop, METH_VARARGS, NULL},
	 { (char *)"vector_uint___getslice__", _wrap_vector_uint___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_uint___setslice__", _wrap_vector_uint___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_uint___delslice__", _wrap_vector_uint___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_uint___delitem__", _wrap_vector_uint___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_uint___getitem__", _wrap_vector_uint___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_uint___setitem__", _wrap_vector_uint___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_uint_append", _wrap_vector_uint_append, METH_VARARGS, NULL},
	 { (char *)"vector_uint_empty", _wrap_vector_uint_empty, METH_VARARGS, NULL},
	 { (char *)"vector_uint_size", _wrap_vector_uint_size, METH_VARARGS, NULL},
	 { (char *)"vector_uint_clear", _wrap_vector_uint_clear, METH_VARARGS, NULL},
	 { (char *)"vector_uint_swap", _wrap_vector_uint_swap, METH_VARARGS, NULL},
	 { (char *)"vector_uint_get_allocator", _wrap_vector_uint_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_uint_begin", _wrap_vector_uint_begin, METH_VARARGS, NULL},
	 { (char *)"vector_uint_end", _wrap_vector_uint_end, METH_VARARGS, NULL},
	 { (char *)"vector_uint_rbegin", _wrap_vector_uint_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_uint_rend", _wrap_vector_uint_rend, METH_VARARGS, NULL},
	 { (char *)"vector_uint_pop_back", _wrap_vector_uint_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_uint_erase", _wrap_vector_uint_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_uint", _wrap_new_vector_uint, METH_VARARGS, NULL},
	 { (char *)"vector_uint_push_back", _wrap_vector_uint_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_uint_front", _wrap_vector_uint_front, METH_VARARGS, NULL},
	 { (char *)"vector_uint_back", _wrap_vector_uint_back, METH_VARARGS, NULL},
	 { (char *)"vector_uint_assign", _wrap_vector_uint_assign, METH_VARARGS, NULL},
	 { (char *)"vector_uint_resize", _wrap_vector_uint_resize, METH_VARARGS, NULL},
	 { (char *)"vector_uint_insert", _wrap_vector_uint_insert, METH_VARARGS, NULL},
	 { (char *)"vector_uint_reserve", _wrap_vector_uint_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_uint_capacity", _wrap_vector_uint_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_uint", _wrap_delete_vector_uint, METH_VARARGS, NULL},
	 { (char *)"vector_uint_swigregister", vector_uint_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_iterator", _wrap_vector_dbl_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___nonzero__", _wrap_vector_dbl___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___len__", _wrap_vector_dbl___len__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_pop", _wrap_vector_dbl_pop, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___getslice__", _wrap_vector_dbl___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___setslice__", _wrap_vector_dbl___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___delslice__", _wrap_vector_dbl___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___delitem__", _wrap_vector_dbl___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___getitem__", _wrap_vector_dbl___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl___setitem__", _wrap_vector_dbl___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_append", _wrap_vector_dbl_append, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_empty", _wrap_vector_dbl_empty, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_size", _wrap_vector_dbl_size, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_clear", _wrap_vector_dbl_clear, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_swap", _wrap_vector_dbl_swap, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_get_allocator", _wrap_vector_dbl_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_begin", _wrap_vector_dbl_begin, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_end", _wrap_vector_dbl_end, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_rbegin", _wrap_vector_dbl_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_rend", _wrap_vector_dbl_rend, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_pop_back", _wrap_vector_dbl_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_erase", _wrap_vector_dbl_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_dbl", _wrap_new_vector_dbl, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_push_back", _wrap_vector_dbl_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_front", _wrap_vector_dbl_front, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_back", _wrap_vector_dbl_back, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_assign", _wrap_vector_dbl_assign, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_resize", _wrap_vector_dbl_resize, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_insert", _wrap_vector_dbl_insert, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_reserve", _wrap_vector_dbl_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_capacity", _wrap_vector_dbl_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_dbl", _wrap_delete_vector_dbl, METH_VARARGS, NULL},
	 { (char *)"vector_dbl_swigregister", vector_dbl_swigregister, METH_VARARGS, NULL},
	 { (char *)"vector_bool_iterator", _wrap_vector_bool_iterator, METH_VARARGS, NULL},
	 { (char *)"vector_bool___nonzero__", _wrap_vector_bool___nonzero__, METH_VARARGS, NULL},
	 { (char *)"vector_bool___len__", _wrap_vector_bool___len__, METH_VARARGS, NULL},
	 { (char *)"vector_bool_pop", _wrap_vector_bool_pop, METH_VARARGS, NULL},
	 { (char *)"vector_bool___getslice__", _wrap_vector_bool___getslice__, METH_VARARGS, NULL},
	 { (char *)"vector_bool___setslice__", _wrap_vector_bool___setslice__, METH_VARARGS, NULL},
	 { (char *)"vector_bool___delslice__", _wrap_vector_bool___delslice__, METH_VARARGS, NULL},
	 { (char *)"vector_bool___delitem__", _wrap_vector_bool___delitem__, METH_VARARGS, NULL},
	 { (char *)"vector_bool___getitem__", _wrap_vector_bool___getitem__, METH_VARARGS, NULL},
	 { (char *)"vector_bool___setitem__", _wrap_vector_bool___setitem__, METH_VARARGS, NULL},
	 { (char *)"vector_bool_append", _wrap_vector_bool_append, METH_VARARGS, NULL},
	 { (char *)"vector_bool_empty", _wrap_vector_bool_empty, METH_VARARGS, NULL},
	 { (char *)"vector_bool_size", _wrap_vector_bool_size, METH_VARARGS, NULL},
	 { (char *)"vector_bool_clear", _wrap_vector_bool_clear, METH_VARARGS, NULL},
	 { (char *)"vector_bool_swap", _wrap_vector_bool_swap, METH_VARARGS, NULL},
	 { (char *)"vector_bool_get_allocator", _wrap_vector_bool_get_allocator, METH_VARARGS, NULL},
	 { (char *)"vector_bool_begin", _wrap_vector_bool_begin, METH_VARARGS, NULL},
	 { (char *)"vector_bool_end", _wrap_vector_bool_end, METH_VARARGS, NULL},
	 { (char *)"vector_bool_rbegin", _wrap_vector_bool_rbegin, METH_VARARGS, NULL},
	 { (char *)"vector_bool_rend", _wrap_vector_bool_rend, METH_VARARGS, NULL},
	 { (char *)"vector_bool_pop_back", _wrap_vector_bool_pop_back, METH_VARARGS, NULL},
	 { (char *)"vector_bool_erase", _wrap_vector_bool_erase, METH_VARARGS, NULL},
	 { (char *)"new_vector_bool", _wrap_new_vector_bool, METH_VARARGS, NULL},
	 { (char *)"vector_bool_push_back", _wrap_vector_bool_push_back, METH_VARARGS, NULL},
	 { (char *)"vector_bool_front", _wrap_vector_bool_front, METH_VARARGS, NULL},
	 { (char *)"vector_bool_back", _wrap_vector_bool_back, METH_VARARGS, NULL},
	 { (char *)"vector_bool_assign", _wrap_vector_bool_assign, METH_VARARGS, NULL},
	 { (char *)"vector_bool_resize", _wrap_vector_bool_resize, METH_VARARGS, NULL},
	 { (char *)"vector_bool_insert", _wrap_vector_bool_insert, METH_VARARGS, NULL},
	 { (char *)"vector_bool_reserve", _wrap_vector_bool_reserve, METH_VARARGS, NULL},
	 { (char *)"vector_bool_capacity", _wrap_vector_bool_capacity, METH_VARARGS, NULL},
	 { (char *)"delete_vector_bool", _wrap_delete_vector_bool, METH_VARARGS, NULL},
	 { (char *)"vector_bool_swigregister", vector_bool_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Geom", _wrap_new_Geom, METH_VARARGS, NULL},
	 { (char *)"delete_Geom", _wrap_delete_Geom, METH_VARARGS, NULL},
	 { (char *)"Geom_getComp", _wrap_Geom_getComp, METH_VARARGS, NULL},
	 { (char *)"Geom_delComp", _wrap_Geom_delComp, METH_VARARGS, NULL},
	 { (char *)"Geom_getAllComps", _wrap_Geom_getAllComps, METH_VARARGS, NULL},
	 { (char *)"Geom_getPatch", _wrap_Geom_getPatch, METH_VARARGS, NULL},
	 { (char *)"Geom_delPatch", _wrap_Geom_delPatch, METH_VARARGS, NULL},
	 { (char *)"Geom_getAllPatches", _wrap_Geom_getAllPatches, METH_VARARGS, NULL},
	 { (char *)"Geom_swigregister", Geom_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Patch", (PyCFunction) _wrap_new_Patch, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_Patch", _wrap_delete_Patch, METH_VARARGS, NULL},
	 { (char *)"Patch_getID", _wrap_Patch_getID, METH_VARARGS, NULL},
	 { (char *)"Patch_setID", _wrap_Patch_setID, METH_VARARGS, NULL},
	 { (char *)"Patch_getContainer", _wrap_Patch_getContainer, METH_VARARGS, NULL},
	 { (char *)"Patch_getArea", _wrap_Patch_getArea, METH_VARARGS, NULL},
	 { (char *)"Patch_setArea", _wrap_Patch_setArea, METH_VARARGS, NULL},
	 { (char *)"Patch_addSurfsys", _wrap_Patch_addSurfsys, METH_VARARGS, NULL},
	 { (char *)"Patch_getSurfsys", _wrap_Patch_getSurfsys, METH_VARARGS, NULL},
	 { (char *)"Patch_delSurfsys", _wrap_Patch_delSurfsys, METH_VARARGS, NULL},
	 { (char *)"Patch_getIComp", _wrap_Patch_getIComp, METH_VARARGS, NULL},
	 { (char *)"Patch_getOComp", _wrap_Patch_getOComp, METH_VARARGS, NULL},
	 { (char *)"Patch_swigregister", Patch_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Comp", (PyCFunction) _wrap_new_Comp, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_Comp", _wrap_delete_Comp, METH_VARARGS, NULL},
	 { (char *)"Comp_getID", _wrap_Comp_getID, METH_VARARGS, NULL},
	 { (char *)"Comp_setID", _wrap_Comp_setID, METH_VARARGS, NULL},
	 { (char *)"Comp_getContainer", _wrap_Comp_getContainer, METH_VARARGS, NULL},
	 { (char *)"Comp_getVol", _wrap_Comp_getVol, METH_VARARGS, NULL},
	 { (char *)"Comp_setVol", _wrap_Comp_setVol, METH_VARARGS, NULL},
	 { (char *)"Comp_addVolsys", _wrap_Comp_addVolsys, METH_VARARGS, NULL},
	 { (char *)"Comp_getVolsys", _wrap_Comp_getVolsys, METH_VARARGS, NULL},
	 { (char *)"Comp_delVolsys", _wrap_Comp_delVolsys, METH_VARARGS, NULL},
	 { (char *)"Comp_getIPatches", _wrap_Comp_getIPatches, METH_VARARGS, NULL},
	 { (char *)"Comp_getOPatches", _wrap_Comp_getOPatches, METH_VARARGS, NULL},
	 { (char *)"Comp_swigregister", Comp_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Tetmesh", _wrap_new_Tetmesh, METH_VARARGS, NULL},
	 { (char *)"delete_Tetmesh", _wrap_delete_Tetmesh, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_setVertex", _wrap_Tetmesh_setVertex, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_setTri", _wrap_Tetmesh_setTri, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_setTet", _wrap_Tetmesh_setTet, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_setup", _wrap_Tetmesh_setup, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_isSetupDone", _wrap_Tetmesh_isSetupDone, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getVertex", _wrap_Tetmesh_getVertex, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_countVertices", _wrap_Tetmesh_countVertices, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTri", _wrap_Tetmesh_getTri, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_countTris", _wrap_Tetmesh_countTris, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriArea", _wrap_Tetmesh_getTriArea, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriBarycenter", _wrap_Tetmesh_getTriBarycenter, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTetQualityRER", _wrap_Tetmesh_getTetQualityRER, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriNorm", _wrap_Tetmesh_getTriNorm, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriPatch", _wrap_Tetmesh_getTriPatch, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriDiffBoundary", _wrap_Tetmesh_getTriDiffBoundary, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriTetNeighb", _wrap_Tetmesh_getTriTetNeighb, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTriBoundary", _wrap_Tetmesh_getTriBoundary, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getSurfTris", _wrap_Tetmesh_getSurfTris, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTet", _wrap_Tetmesh_getTet, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_countTets", _wrap_Tetmesh_countTets, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTetVol", _wrap_Tetmesh_getTetVol, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTetBarycenter", _wrap_Tetmesh_getTetBarycenter, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTetComp", _wrap_Tetmesh_getTetComp, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTetTriNeighb", _wrap_Tetmesh_getTetTriNeighb, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getTetTetNeighb", _wrap_Tetmesh_getTetTetNeighb, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_findTetByPoint", _wrap_Tetmesh_findTetByPoint, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getBoundMin", _wrap_Tetmesh_getBoundMin, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getBoundMax", _wrap_Tetmesh_getBoundMax, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_getMeshVolume", _wrap_Tetmesh_getMeshVolume, METH_VARARGS, NULL},
	 { (char *)"Tetmesh_swigregister", Tetmesh_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TmComp", (PyCFunction) _wrap_new_TmComp, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_TmComp", _wrap_delete_TmComp, METH_VARARGS, NULL},
	 { (char *)"TmComp_setVol", _wrap_TmComp_setVol, METH_VARARGS, NULL},
	 { (char *)"TmComp_getAllTetIndices", _wrap_TmComp_getAllTetIndices, METH_VARARGS, NULL},
	 { (char *)"TmComp_countTets", _wrap_TmComp_countTets, METH_VARARGS, NULL},
	 { (char *)"TmComp_isTetInside", _wrap_TmComp_isTetInside, METH_VARARGS, NULL},
	 { (char *)"TmComp_getBoundMin", _wrap_TmComp_getBoundMin, METH_VARARGS, NULL},
	 { (char *)"TmComp_getBoundMax", _wrap_TmComp_getBoundMax, METH_VARARGS, NULL},
	 { (char *)"TmComp_swigregister", TmComp_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_TmPatch", (PyCFunction) _wrap_new_TmPatch, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_TmPatch", _wrap_delete_TmPatch, METH_VARARGS, NULL},
	 { (char *)"TmPatch_setArea", _wrap_TmPatch_setArea, METH_VARARGS, NULL},
	 { (char *)"TmPatch_getAllTriIndices", _wrap_TmPatch_getAllTriIndices, METH_VARARGS, NULL},
	 { (char *)"TmPatch_isTriInside", _wrap_TmPatch_isTriInside, METH_VARARGS, NULL},
	 { (char *)"TmPatch_swigregister", TmPatch_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_DiffBoundary", (PyCFunction) _wrap_new_DiffBoundary, METH_VARARGS | METH_KEYWORDS, NULL},
	 { (char *)"delete_DiffBoundary", _wrap_delete_DiffBoundary, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_getID", _wrap_DiffBoundary_getID, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_setID", _wrap_DiffBoundary_setID, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_getContainer", _wrap_DiffBoundary_getContainer, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_getComps", _wrap_DiffBoundary_getComps, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_isTriInside", _wrap_DiffBoundary_isTriInside, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_getAllTriIndices", _wrap_DiffBoundary_getAllTriIndices, METH_VARARGS, NULL},
	 { (char *)"DiffBoundary_swigregister", DiffBoundary_swigregister, METH_VARARGS, NULL},
	 { (char *)"delete_RNG", _wrap_delete_RNG, METH_VARARGS, NULL},
	 { (char *)"RNG_initialize", _wrap_RNG_initialize, METH_VARARGS, NULL},
	 { (char *)"RNG_get", _wrap_RNG_get, METH_VARARGS, NULL},
	 { (char *)"RNG_getUnfII", _wrap_RNG_getUnfII, METH_VARARGS, NULL},
	 { (char *)"RNG_getUnfIE", _wrap_RNG_getUnfIE, METH_VARARGS, NULL},
	 { (char *)"RNG_getUnfEE", _wrap_RNG_getUnfEE, METH_VARARGS, NULL},
	 { (char *)"RNG_getUnfIE53", _wrap_RNG_getUnfIE53, METH_VARARGS, NULL},
	 { (char *)"RNG_getStdExp", _wrap_RNG_getStdExp, METH_VARARGS, NULL},
	 { (char *)"RNG_getExp", _wrap_RNG_getExp, METH_VARARGS, NULL},
	 { (char *)"RNG_getPsn", _wrap_RNG_getPsn, METH_VARARGS, NULL},
	 { (char *)"RNG_getStdNrm", _wrap_RNG_getStdNrm, METH_VARARGS, NULL},
	 { (char *)"RNG_swigregister", RNG_swigregister, METH_VARARGS, NULL},
	 { (char *)"create_mt19937", _wrap_create_mt19937, METH_VARARGS, NULL},
	 { (char *)"create", _wrap_create, METH_VARARGS, NULL},
	 { (char *)"delete_API", _wrap_delete_API, METH_VARARGS, NULL},
	 { (char *)"API_getSolverName", _wrap_API_getSolverName, METH_VARARGS, NULL},
	 { (char *)"API_getSolverDesc", _wrap_API_getSolverDesc, METH_VARARGS, NULL},
	 { (char *)"API_getSolverAuthors", _wrap_API_getSolverAuthors, METH_VARARGS, NULL},
	 { (char *)"API_getSolverEmail", _wrap_API_getSolverEmail, METH_VARARGS, NULL},
	 { (char *)"API_reset", _wrap_API_reset, METH_VARARGS, NULL},
	 { (char *)"API_checkpoint", _wrap_API_checkpoint, METH_VARARGS, NULL},
	 { (char *)"API_restore", _wrap_API_restore, METH_VARARGS, NULL},
	 { (char *)"API_run", _wrap_API_run, METH_VARARGS, NULL},
	 { (char *)"API_getTime", _wrap_API_getTime, METH_VARARGS, NULL},
	 { (char *)"API_advance", _wrap_API_advance, METH_VARARGS, NULL},
	 { (char *)"API_step", _wrap_API_step, METH_VARARGS, NULL},
	 { (char *)"API_setRk4DT", _wrap_API_setRk4DT, METH_VARARGS, NULL},
	 { (char *)"API_setDT", _wrap_API_setDT, METH_VARARGS, NULL},
	 { (char *)"API_getRk4DT", _wrap_API_getRk4DT, METH_VARARGS, NULL},
	 { (char *)"API_getDT", _wrap_API_getDT, METH_VARARGS, NULL},
	 { (char *)"API_setTime", _wrap_API_setTime, METH_VARARGS, NULL},
	 { (char *)"API_getA0", _wrap_API_getA0, METH_VARARGS, NULL},
	 { (char *)"API_getNSteps", _wrap_API_getNSteps, METH_VARARGS, NULL},
	 { (char *)"API_setNSteps", _wrap_API_setNSteps, METH_VARARGS, NULL},
	 { (char *)"API_getCompVol", _wrap_API_getCompVol, METH_VARARGS, NULL},
	 { (char *)"API_setCompVol", _wrap_API_setCompVol, METH_VARARGS, NULL},
	 { (char *)"API_getCompCount", _wrap_API_getCompCount, METH_VARARGS, NULL},
	 { (char *)"API_setCompCount", _wrap_API_setCompCount, METH_VARARGS, NULL},
	 { (char *)"API_getCompAmount", _wrap_API_getCompAmount, METH_VARARGS, NULL},
	 { (char *)"API_setCompAmount", _wrap_API_setCompAmount, METH_VARARGS, NULL},
	 { (char *)"API_getCompConc", _wrap_API_getCompConc, METH_VARARGS, NULL},
	 { (char *)"API_setCompConc", _wrap_API_setCompConc, METH_VARARGS, NULL},
	 { (char *)"API_getCompClamped", _wrap_API_getCompClamped, METH_VARARGS, NULL},
	 { (char *)"API_setCompClamped", _wrap_API_setCompClamped, METH_VARARGS, NULL},
	 { (char *)"API_getCompReacK", _wrap_API_getCompReacK, METH_VARARGS, NULL},
	 { (char *)"API_setCompReacK", _wrap_API_setCompReacK, METH_VARARGS, NULL},
	 { (char *)"API_getCompReacActive", _wrap_API_getCompReacActive, METH_VARARGS, NULL},
	 { (char *)"API_setCompReacActive", _wrap_API_setCompReacActive, METH_VARARGS, NULL},
	 { (char *)"API_getCompDiffD", _wrap_API_getCompDiffD, METH_VARARGS, NULL},
	 { (char *)"API_setCompDiffD", _wrap_API_setCompDiffD, METH_VARARGS, NULL},
	 { (char *)"API_getCompDiffActive", _wrap_API_getCompDiffActive, METH_VARARGS, NULL},
	 { (char *)"API_setCompDiffActive", _wrap_API_setCompDiffActive, METH_VARARGS, NULL},
	 { (char *)"API_getCompReacC", _wrap_API_getCompReacC, METH_VARARGS, NULL},
	 { (char *)"API_getCompReacH", _wrap_API_getCompReacH, METH_VARARGS, NULL},
	 { (char *)"API_getCompReacA", _wrap_API_getCompReacA, METH_VARARGS, NULL},
	 { (char *)"API_getCompReacExtent", _wrap_API_getCompReacExtent, METH_VARARGS, NULL},
	 { (char *)"API_resetCompReacExtent", _wrap_API_resetCompReacExtent, METH_VARARGS, NULL},
	 { (char *)"API_getTetVol", _wrap_API_getTetVol, METH_VARARGS, NULL},
	 { (char *)"API_getTetSpecDefined", _wrap_API_getTetSpecDefined, METH_VARARGS, NULL},
	 { (char *)"API_getTetCount", _wrap_API_getTetCount, METH_VARARGS, NULL},
	 { (char *)"API_setTetCount", _wrap_API_setTetCount, METH_VARARGS, NULL},
	 { (char *)"API_getTetAmount", _wrap_API_getTetAmount, METH_VARARGS, NULL},
	 { (char *)"API_setTetAmount", _wrap_API_setTetAmount, METH_VARARGS, NULL},
	 { (char *)"API_getTetConc", _wrap_API_getTetConc, METH_VARARGS, NULL},
	 { (char *)"API_setTetConc", _wrap_API_setTetConc, METH_VARARGS, NULL},
	 { (char *)"API_getTetClamped", _wrap_API_getTetClamped, METH_VARARGS, NULL},
	 { (char *)"API_setTetClamped", _wrap_API_setTetClamped, METH_VARARGS, NULL},
	 { (char *)"API_getTetReacK", _wrap_API_getTetReacK, METH_VARARGS, NULL},
	 { (char *)"API_setTetReacK", _wrap_API_setTetReacK, METH_VARARGS, NULL},
	 { (char *)"API_getTetReacActive", _wrap_API_getTetReacActive, METH_VARARGS, NULL},
	 { (char *)"API_setTetReacActive", _wrap_API_setTetReacActive, METH_VARARGS, NULL},
	 { (char *)"API_getTetDiffD", _wrap_API_getTetDiffD, METH_VARARGS, NULL},
	 { (char *)"API_setTetDiffD", _wrap_API_setTetDiffD, METH_VARARGS, NULL},
	 { (char *)"API_getTetDiffActive", _wrap_API_getTetDiffActive, METH_VARARGS, NULL},
	 { (char *)"API_setTetDiffActive", _wrap_API_setTetDiffActive, METH_VARARGS, NULL},
	 { (char *)"API_getTetReacC", _wrap_API_getTetReacC, METH_VARARGS, NULL},
	 { (char *)"API_getTetReacH", _wrap_API_getTetReacH, METH_VARARGS, NULL},
	 { (char *)"API_getTetReacA", _wrap_API_getTetReacA, METH_VARARGS, NULL},
	 { (char *)"API_getTetDiffA", _wrap_API_getTetDiffA, METH_VARARGS, NULL},
	 { (char *)"API_getPatchArea", _wrap_API_getPatchArea, METH_VARARGS, NULL},
	 { (char *)"API_setPatchArea", _wrap_API_setPatchArea, METH_VARARGS, NULL},
	 { (char *)"API_getPatchCount", _wrap_API_getPatchCount, METH_VARARGS, NULL},
	 { (char *)"API_setPatchCount", _wrap_API_setPatchCount, METH_VARARGS, NULL},
	 { (char *)"API_getPatchAmount", _wrap_API_getPatchAmount, METH_VARARGS, NULL},
	 { (char *)"API_setPatchAmount", _wrap_API_setPatchAmount, METH_VARARGS, NULL},
	 { (char *)"API_getPatchClamped", _wrap_API_getPatchClamped, METH_VARARGS, NULL},
	 { (char *)"API_setPatchClamped", _wrap_API_setPatchClamped, METH_VARARGS, NULL},
	 { (char *)"API_getPatchSReacK", _wrap_API_getPatchSReacK, METH_VARARGS, NULL},
	 { (char *)"API_setPatchSReacK", _wrap_API_setPatchSReacK, METH_VARARGS, NULL},
	 { (char *)"API_getPatchSReacActive", _wrap_API_getPatchSReacActive, METH_VARARGS, NULL},
	 { (char *)"API_setPatchSReacActive", _wrap_API_setPatchSReacActive, METH_VARARGS, NULL},
	 { (char *)"API_getPatchSReacC", _wrap_API_getPatchSReacC, METH_VARARGS, NULL},
	 { (char *)"API_getPatchSReacH", _wrap_API_getPatchSReacH, METH_VARARGS, NULL},
	 { (char *)"API_getPatchSReacA", _wrap_API_getPatchSReacA, METH_VARARGS, NULL},
	 { (char *)"API_getPatchSReacExtent", _wrap_API_getPatchSReacExtent, METH_VARARGS, NULL},
	 { (char *)"API_resetPatchSReacExtent", _wrap_API_resetPatchSReacExtent, METH_VARARGS, NULL},
	 { (char *)"API_setDiffBoundaryDiffusionActive", _wrap_API_setDiffBoundaryDiffusionActive, METH_VARARGS, NULL},
	 { (char *)"API_getDiffBoundaryDiffusionActive", _wrap_API_getDiffBoundaryDiffusionActive, METH_VARARGS, NULL},
	 { (char *)"API_getTriArea", _wrap_API_getTriArea, METH_VARARGS, NULL},
	 { (char *)"API_getTriSpecDefined", _wrap_API_getTriSpecDefined, METH_VARARGS, NULL},
	 { (char *)"API_getTriCount", _wrap_API_getTriCount, METH_VARARGS, NULL},
	 { (char *)"API_setTriCount", _wrap_API_setTriCount, METH_VARARGS, NULL},
	 { (char *)"API_getTriAmount", _wrap_API_getTriAmount, METH_VARARGS, NULL},
	 { (char *)"API_setTriAmount", _wrap_API_setTriAmount, METH_VARARGS, NULL},
	 { (char *)"API_getTriClamped", _wrap_API_getTriClamped, METH_VARARGS, NULL},
	 { (char *)"API_setTriClamped", _wrap_API_setTriClamped, METH_VARARGS, NULL},
	 { (char *)"API_getTriSReacK", _wrap_API_getTriSReacK, METH_VARARGS, NULL},
	 { (char *)"API_setTriSReacK", _wrap_API_setTriSReacK, METH_VARARGS, NULL},
	 { (char *)"API_getTriSReacActive", _wrap_API_getTriSReacActive, METH_VARARGS, NULL},
	 { (char *)"API_setTriSReacActive", _wrap_API_setTriSReacActive, METH_VARARGS, NULL},
	 { (char *)"API_getTriSReacC", _wrap_API_getTriSReacC, METH_VARARGS, NULL},
	 { (char *)"API_getTriSReacH", _wrap_API_getTriSReacH, METH_VARARGS, NULL},
	 { (char *)"API_getTriSReacA", _wrap_API_getTriSReacA, METH_VARARGS, NULL},
	 { (char *)"API_swigregister", API_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Wmrk4", _wrap_new_Wmrk4, METH_VARARGS, NULL},
	 { (char *)"delete_Wmrk4", _wrap_delete_Wmrk4, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_getSolverName", _wrap_Wmrk4_getSolverName, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_getSolverDesc", _wrap_Wmrk4_getSolverDesc, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_getSolverAuthors", _wrap_Wmrk4_getSolverAuthors, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_getSolverEmail", _wrap_Wmrk4_getSolverEmail, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_reset", _wrap_Wmrk4_reset, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_run", _wrap_Wmrk4_run, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_getTime", _wrap_Wmrk4_getTime, METH_VARARGS, NULL},
	 { (char *)"Wmrk4_swigregister", Wmrk4_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Wmdirect", _wrap_new_Wmdirect, METH_VARARGS, NULL},
	 { (char *)"delete_Wmdirect", _wrap_delete_Wmdirect, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_getSolverName", _wrap_Wmdirect_getSolverName, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_getSolverDesc", _wrap_Wmdirect_getSolverDesc, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_getSolverAuthors", _wrap_Wmdirect_getSolverAuthors, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_getSolverEmail", _wrap_Wmdirect_getSolverEmail, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_reset", _wrap_Wmdirect_reset, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_run", _wrap_Wmdirect_run, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_getTime", _wrap_Wmdirect_getTime, METH_VARARGS, NULL},
	 { (char *)"Wmdirect_swigregister", Wmdirect_swigregister, METH_VARARGS, NULL},
	 { (char *)"new_Tetexact", _wrap_new_Tetexact, METH_VARARGS, NULL},
	 { (char *)"delete_Tetexact", _wrap_delete_Tetexact, METH_VARARGS, NULL},
	 { (char *)"Tetexact_getSolverName", _wrap_Tetexact_getSolverName, METH_VARARGS, NULL},
	 { (char *)"Tetexact_getSolverDesc", _wrap_Tetexact_getSolverDesc, METH_VARARGS, NULL},
	 { (char *)"Tetexact_getSolverAuthors", _wrap_Tetexact_getSolverAuthors, METH_VARARGS, NULL},
	 { (char *)"Tetexact_getSolverEmail", _wrap_Tetexact_getSolverEmail, METH_VARARGS, NULL},
	 { (char *)"Tetexact_reset", _wrap_Tetexact_reset, METH_VARARGS, NULL},
	 { (char *)"Tetexact_run", _wrap_Tetexact_run, METH_VARARGS, NULL},
	 { (char *)"Tetexact_getTime", _wrap_Tetexact_getTime, METH_VARARGS, NULL},
	 { (char *)"Tetexact_swigregister", Tetexact_swigregister, METH_VARARGS, NULL},
	 { NULL, NULL, 0, NULL }
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (BEGIN) -------- */

static void *_p_steps__ArgErrTo_p_steps__Err(void *x) {
    return (void *)((steps::Err *)  ((steps::ArgErr *) x));
}
static void *_p_steps__NotImplErrTo_p_steps__Err(void *x) {
    return (void *)((steps::Err *)  ((steps::NotImplErr *) x));
}
static void *_p_steps__tetmesh__TmCompTo_p_steps__wm__Comp(void *x) {
    return (void *)((steps::wm::Comp *)  ((steps::tetmesh::TmComp *) x));
}
static void *_p_steps__wmrk4__Wmrk4To_p_steps__solver__API(void *x) {
    return (void *)((steps::solver::API *)  ((steps::wmrk4::Wmrk4 *) x));
}
static void *_p_steps__wmdirect__WmdirectTo_p_steps__solver__API(void *x) {
    return (void *)((steps::solver::API *)  ((steps::wmdirect::Wmdirect *) x));
}
static void *_p_steps__tetexact__TetexactTo_p_steps__solver__API(void *x) {
    return (void *)((steps::solver::API *)  ((steps::tetexact::Tetexact *) x));
}
static void *_p_steps__tetmesh__TetmeshTo_p_steps__wm__Geom(void *x) {
    return (void *)((steps::wm::Geom *)  ((steps::tetmesh::Tetmesh *) x));
}
static void *_p_p_steps__tetmesh__TmPatchTo_p_p_steps__wm__Patch(void *x) {
    return (void *)((steps::wm::Patch **)  ((steps::tetmesh::TmPatch **) x));
}
static void *_p_steps__tetmesh__TmPatchTo_p_steps__wm__Patch(void *x) {
    return (void *)((steps::wm::Patch *)  ((steps::tetmesh::TmPatch *) x));
}
static swig_type_info _swigt__p_allocator_type = {"_p_allocator_type", "allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_char = {"_p_char", "char *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_const_reference = {"_p_const_reference", "const_reference *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_difference_type = {"_p_difference_type", "difference_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_key_type = {"_p_key_type", "key_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_PyObject = {"_p_p_PyObject", "PyObject **", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_reference = {"_p_reference", "reference *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_size_type = {"_p_size_type", "size_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__invalid_argument = {"_p_std__invalid_argument", "std::invalid_argument *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t = {"_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t", "std::set<std::string > *|std::set<std::string,std::less<std::string >,std::allocator<std::string > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type = {"_p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type", "std::allocator<std::string > *|std::set<std::string >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t = {"_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t", "std::set<steps::wm::Patch * > *|std::set<steps::wm::Patch *,std::less<steps::wm::Patch * >,std::allocator<steps::wm::Patch * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type = {"_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type", "std::allocator<steps::wm::Patch * > *|std::set<steps::wm::Patch * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type = {"_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type", "steps::wm::Patch **|std::set<steps::wm::Patch * >::key_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_p_steps__tetmesh__TmPatch = {"_p_p_steps__tetmesh__TmPatch", 0, 0, 0, 0, 0};
static swig_type_info _swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type = {"_p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type", "steps::wm::Patch **|std::set<steps::wm::Patch * >::value_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTbool_std__allocatorTbool_t_t = {"_p_std__vectorTbool_std__allocatorTbool_t_t", "std::vector<bool > *|std::vector<bool,std::allocator<bool > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type = {"_p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type", "std::allocator<bool > *|std::vector<bool >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTdouble_std__allocatorTdouble_t_t = {"_p_std__vectorTdouble_std__allocatorTdouble_t_t", "std::vector<double > *|std::vector<double,std::allocator<double > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type = {"_p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type", "std::allocator<double > *|std::vector<double >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTint_std__allocatorTint_t_t = {"_p_std__vectorTint_std__allocatorTint_t_t", "std::vector<int > *|std::vector<int,std::allocator<int > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTint_std__allocatorTint_t_t__allocator_type = {"_p_std__vectorTint_std__allocatorTint_t_t__allocator_type", "std::allocator<int > *|std::vector<int >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t = {"_p_std__vectorTstd__string_std__allocatorTstd__string_t_t", "std::vector<std::string > *|std::vector<std::string,std::allocator<std::string > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type = {"_p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type", "std::allocator<std::string > *|std::vector<std::string >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t = {"_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t", "std::vector<steps::model::Diff * > *|std::vector<steps::model::Diff *,std::allocator<steps::model::Diff * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type = {"_p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type", "std::allocator<steps::model::Diff * > *|std::vector<steps::model::Diff * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t = {"_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t", "std::vector<steps::model::Reac * > *|std::vector<steps::model::Reac *,std::allocator<steps::model::Reac * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type = {"_p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type", "std::allocator<steps::model::Reac * > *|std::vector<steps::model::Reac * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t = {"_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t", "std::vector<steps::model::SReac * > *|std::vector<steps::model::SReac *,std::allocator<steps::model::SReac * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type = {"_p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type", "std::allocator<steps::model::SReac * > *|std::vector<steps::model::SReac * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t = {"_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t", "std::vector<steps::model::Spec * > *|std::vector<steps::model::Spec *,std::allocator<steps::model::Spec * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type = {"_p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type", "std::allocator<steps::model::Spec * > *|std::vector<steps::model::Spec * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t = {"_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t", "std::vector<steps::model::Surfsys * > *|std::vector<steps::model::Surfsys *,std::allocator<steps::model::Surfsys * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type = {"_p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type", "std::allocator<steps::model::Surfsys * > *|std::vector<steps::model::Surfsys * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t = {"_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t", "std::vector<steps::model::Volsys * > *|std::vector<steps::model::Volsys *,std::allocator<steps::model::Volsys * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type = {"_p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type", "std::allocator<steps::model::Volsys * > *|std::vector<steps::model::Volsys * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t = {"_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t", "std::vector<steps::wm::Comp * > *|std::vector<steps::wm::Comp *,std::allocator<steps::wm::Comp * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type = {"_p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type", "std::allocator<steps::wm::Comp * > *|std::vector<steps::wm::Comp * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t = {"_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t", "std::vector<steps::wm::Patch * > *|std::vector<steps::wm::Patch *,std::allocator<steps::wm::Patch * > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type = {"_p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type", "std::allocator<steps::wm::Patch * > *|std::vector<steps::wm::Patch * >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t = {"_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t", "std::vector<unsigned int > *|std::vector<unsigned int,std::allocator<unsigned int > > *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type = {"_p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type", "std::allocator<unsigned int > *|std::vector<unsigned int >::allocator_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__ArgErr = {"_p_steps__ArgErr", "steps::ArgErr *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__Err = {"_p_steps__Err", "steps::Err *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__NotImplErr = {"_p_steps__NotImplErr", "steps::NotImplErr *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__Diff = {"_p_steps__model__Diff", "steps::model::Diff *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__Model = {"_p_steps__model__Model", "steps::model::Model *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__Reac = {"_p_steps__model__Reac", "steps::model::Reac *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__SReac = {"_p_steps__model__SReac", "steps::model::SReac *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__Spec = {"_p_steps__model__Spec", "steps::model::Spec *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__Surfsys = {"_p_steps__model__Surfsys", "steps::model::Surfsys *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__model__Volsys = {"_p_steps__model__Volsys", "steps::model::Volsys *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__rng__RNG = {"_p_steps__rng__RNG", "steps::rng::RNG *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__solver__API = {"_p_steps__solver__API", "steps::solver::API *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__tetexact__Tetexact = {"_p_steps__tetexact__Tetexact", "steps::tetexact::Tetexact *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__tetmesh__DiffBoundary = {"_p_steps__tetmesh__DiffBoundary", "steps::tetmesh::DiffBoundary *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__tetmesh__Tetmesh = {"_p_steps__tetmesh__Tetmesh", "steps::tetmesh::Tetmesh *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__tetmesh__TmComp = {"_p_steps__tetmesh__TmComp", "steps::tetmesh::TmComp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__tetmesh__TmPatch = {"_p_steps__tetmesh__TmPatch", "steps::tetmesh::TmPatch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__wm__Comp = {"_p_steps__wm__Comp", "steps::wm::Comp *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__wm__Geom = {"_p_steps__wm__Geom", "steps::wm::Geom *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__wm__Patch = {"_p_steps__wm__Patch", "steps::wm::Patch *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__wmdirect__Wmdirect = {"_p_steps__wmdirect__Wmdirect", "steps::wmdirect::Wmdirect *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_steps__wmrk4__Wmrk4 = {"_p_steps__wmrk4__Wmrk4", "steps::wmrk4::Wmrk4 *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_swig__PySwigIterator = {"_p_swig__PySwigIterator", "swig::PySwigIterator *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_char = {"_p_unsigned_char", "unsigned char *|uchar *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_int = {"_p_unsigned_int", "unsigned int *|uint *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_long = {"_p_unsigned_long", "unsigned long *|ulong *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_unsigned_short = {"_p_unsigned_short", "unsigned short *|ushort *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__p_value_type = {"_p_value_type", "value_type *", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type = {"_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type", "steps::wm::Patch *|std::set<steps::wm::Patch * >::key_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type = {"_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type", "steps::wm::Patch *|std::set<steps::wm::Patch * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type = {"_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type", "steps::model::Diff *|std::vector<steps::model::Diff * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type = {"_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type", "steps::model::Reac *|std::vector<steps::model::Reac * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type = {"_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type", "steps::model::SReac *|std::vector<steps::model::SReac * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type = {"_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type", "steps::model::Spec *|std::vector<steps::model::Spec * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type = {"_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type", "steps::model::Surfsys *|std::vector<steps::model::Surfsys * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type = {"_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type", "steps::model::Volsys *|std::vector<steps::model::Volsys * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type = {"_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type", "steps::wm::Comp *|std::vector<steps::wm::Comp * >::value_type", 0, 0, (void*)0, 0};
static swig_type_info _swigt__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type = {"_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type", "steps::wm::Patch *|std::vector<steps::wm::Patch * >::value_type", 0, 0, (void*)0, 0};

static swig_type_info *swig_type_initial[] = {
  &_swigt__p_allocator_type,
  &_swigt__p_char,
  &_swigt__p_const_reference,
  &_swigt__p_difference_type,
  &_swigt__p_key_type,
  &_swigt__p_p_PyObject,
  &_swigt__p_p_steps__tetmesh__TmPatch,
  &_swigt__p_reference,
  &_swigt__p_size_type,
  &_swigt__p_std__invalid_argument,
  &_swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t,
  &_swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type,
  &_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t,
  &_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type,
  &_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type,
  &_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type,
  &_swigt__p_std__vectorTbool_std__allocatorTbool_t_t,
  &_swigt__p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type,
  &_swigt__p_std__vectorTdouble_std__allocatorTdouble_t_t,
  &_swigt__p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type,
  &_swigt__p_std__vectorTint_std__allocatorTint_t_t,
  &_swigt__p_std__vectorTint_std__allocatorTint_t_t__allocator_type,
  &_swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t,
  &_swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t,
  &_swigt__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t,
  &_swigt__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t,
  &_swigt__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t,
  &_swigt__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t,
  &_swigt__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t,
  &_swigt__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t,
  &_swigt__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type,
  &_swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t,
  &_swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type,
  &_swigt__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t,
  &_swigt__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type,
  &_swigt__p_steps__ArgErr,
  &_swigt__p_steps__Err,
  &_swigt__p_steps__NotImplErr,
  &_swigt__p_steps__model__Diff,
  &_swigt__p_steps__model__Model,
  &_swigt__p_steps__model__Reac,
  &_swigt__p_steps__model__SReac,
  &_swigt__p_steps__model__Spec,
  &_swigt__p_steps__model__Surfsys,
  &_swigt__p_steps__model__Volsys,
  &_swigt__p_steps__rng__RNG,
  &_swigt__p_steps__solver__API,
  &_swigt__p_steps__tetexact__Tetexact,
  &_swigt__p_steps__tetmesh__DiffBoundary,
  &_swigt__p_steps__tetmesh__Tetmesh,
  &_swigt__p_steps__tetmesh__TmComp,
  &_swigt__p_steps__tetmesh__TmPatch,
  &_swigt__p_steps__wm__Comp,
  &_swigt__p_steps__wm__Geom,
  &_swigt__p_steps__wm__Patch,
  &_swigt__p_steps__wmdirect__Wmdirect,
  &_swigt__p_steps__wmrk4__Wmrk4,
  &_swigt__p_swig__PySwigIterator,
  &_swigt__p_unsigned_char,
  &_swigt__p_unsigned_int,
  &_swigt__p_unsigned_long,
  &_swigt__p_unsigned_short,
  &_swigt__p_value_type,
  &_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type,
  &_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type,
  &_swigt__std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type,
  &_swigt__std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type,
  &_swigt__std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type,
  &_swigt__std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type,
  &_swigt__std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type,
  &_swigt__std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type,
  &_swigt__std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type,
  &_swigt__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type,
};

static swig_cast_info _swigc__p_allocator_type[] = {  {&_swigt__p_allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_char[] = {  {&_swigt__p_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_const_reference[] = {  {&_swigt__p_const_reference, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_difference_type[] = {  {&_swigt__p_difference_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_key_type[] = {  {&_swigt__p_key_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_PyObject[] = {  {&_swigt__p_p_PyObject, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_reference[] = {  {&_swigt__p_reference, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_size_type[] = {  {&_swigt__p_size_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__invalid_argument[] = {  {&_swigt__p_std__invalid_argument, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t[] = {  {&_swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type[] = {  {&_swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type, 0, 0, 0},  {&_swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t[] = {  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type, 0, 0, 0},  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_p_steps__tetmesh__TmPatch[] = {{&_swigt__p_p_steps__tetmesh__TmPatch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type[] = {  {&_swigt__p_p_steps__tetmesh__TmPatch, 0, 0, 0},  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0, 0, 0},  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type[] = {  {&_swigt__p_p_steps__tetmesh__TmPatch, 0, 0, 0},  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTbool_std__allocatorTbool_t_t[] = {  {&_swigt__p_std__vectorTbool_std__allocatorTbool_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTdouble_std__allocatorTdouble_t_t[] = {  {&_swigt__p_std__vectorTdouble_std__allocatorTdouble_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTint_std__allocatorTint_t_t[] = {  {&_swigt__p_std__vectorTint_std__allocatorTint_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTint_std__allocatorTint_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTint_std__allocatorTint_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTstd__string_std__allocatorTstd__string_t_t[] = {  {&_swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type[] = {  {&_swigt__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type, 0, 0, 0},  {&_swigt__p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t[] = {  {&_swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type, 0, 0, 0},  {&_swigt__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t[] = {  {&_swigt__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type[] = {  {&_swigt__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__ArgErr[] = {  {&_swigt__p_steps__ArgErr, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__Err[] = {  {&_swigt__p_steps__NotImplErr, _p_steps__NotImplErrTo_p_steps__Err, 0, 0},  {&_swigt__p_steps__ArgErr, _p_steps__ArgErrTo_p_steps__Err, 0, 0},  {&_swigt__p_steps__Err, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__NotImplErr[] = {  {&_swigt__p_steps__NotImplErr, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__Diff[] = {  {&_swigt__std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Diff, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__Model[] = {  {&_swigt__p_steps__model__Model, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__Reac[] = {  {&_swigt__p_steps__model__Reac, 0, 0, 0},  {&_swigt__std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__SReac[] = {  {&_swigt__p_steps__model__SReac, 0, 0, 0},  {&_swigt__std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__Spec[] = {  {&_swigt__std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Spec, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__Surfsys[] = {  {&_swigt__std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Surfsys, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__model__Volsys[] = {  {&_swigt__p_steps__model__Volsys, 0, 0, 0},  {&_swigt__std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__rng__RNG[] = {  {&_swigt__p_steps__rng__RNG, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__solver__API[] = {  {&_swigt__p_steps__wmrk4__Wmrk4, _p_steps__wmrk4__Wmrk4To_p_steps__solver__API, 0, 0},  {&_swigt__p_steps__wmdirect__Wmdirect, _p_steps__wmdirect__WmdirectTo_p_steps__solver__API, 0, 0},  {&_swigt__p_steps__tetexact__Tetexact, _p_steps__tetexact__TetexactTo_p_steps__solver__API, 0, 0},  {&_swigt__p_steps__solver__API, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__tetexact__Tetexact[] = {  {&_swigt__p_steps__tetexact__Tetexact, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__tetmesh__DiffBoundary[] = {  {&_swigt__p_steps__tetmesh__DiffBoundary, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__tetmesh__Tetmesh[] = {  {&_swigt__p_steps__tetmesh__Tetmesh, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__tetmesh__TmComp[] = {  {&_swigt__p_steps__tetmesh__TmComp, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__tetmesh__TmPatch[] = {  {&_swigt__p_steps__tetmesh__TmPatch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__wm__Comp[] = {  {&_swigt__std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__wm__Comp, 0, 0, 0},  {&_swigt__p_steps__tetmesh__TmComp, _p_steps__tetmesh__TmCompTo_p_steps__wm__Comp, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__wm__Geom[] = {  {&_swigt__p_steps__wm__Geom, 0, 0, 0},  {&_swigt__p_steps__tetmesh__Tetmesh, _p_steps__tetmesh__TetmeshTo_p_steps__wm__Geom, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__wm__Patch[] = {  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0, 0, 0},  {&_swigt__p_steps__tetmesh__TmPatch, _p_steps__tetmesh__TmPatchTo_p_steps__wm__Patch, 0, 0},  {&_swigt__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__wm__Patch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__wmdirect__Wmdirect[] = {  {&_swigt__p_steps__wmdirect__Wmdirect, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_steps__wmrk4__Wmrk4[] = {  {&_swigt__p_steps__wmrk4__Wmrk4, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_swig__PySwigIterator[] = {  {&_swigt__p_swig__PySwigIterator, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_char[] = {  {&_swigt__p_unsigned_char, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_int[] = {  {&_swigt__p_unsigned_int, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_long[] = {  {&_swigt__p_unsigned_long, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_unsigned_short[] = {  {&_swigt__p_unsigned_short, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__p_value_type[] = {  {&_swigt__p_value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type[] = {  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0, 0, 0},  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__tetmesh__TmPatch, _p_steps__tetmesh__TmPatchTo_p_steps__wm__Patch, 0, 0},  {&_swigt__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__wm__Patch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type[] = {  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0, 0, 0},  {&_swigt__p_steps__tetmesh__TmPatch, _p_steps__tetmesh__TmPatchTo_p_steps__wm__Patch, 0, 0},  {&_swigt__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__wm__Patch, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type[] = {  {&_swigt__std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Diff, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type[] = {  {&_swigt__p_steps__model__Reac, 0, 0, 0},  {&_swigt__std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type[] = {  {&_swigt__p_steps__model__SReac, 0, 0, 0},  {&_swigt__std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type[] = {  {&_swigt__std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Spec, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type[] = {  {&_swigt__std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Surfsys, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type[] = {  {&_swigt__std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__model__Volsys, 0, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type[] = {  {&_swigt__std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__wm__Comp, 0, 0, 0},  {&_swigt__p_steps__tetmesh__TmComp, _p_steps__tetmesh__TmCompTo_p_steps__wm__Comp, 0, 0},{0, 0, 0, 0}};
static swig_cast_info _swigc__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type[] = {  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type, 0, 0, 0},  {&_swigt__p_steps__tetmesh__TmPatch, _p_steps__tetmesh__TmPatchTo_p_steps__wm__Patch, 0, 0},  {&_swigt__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type, 0, 0, 0},  {&_swigt__p_steps__wm__Patch, 0, 0, 0},{0, 0, 0, 0}};

static swig_cast_info *swig_cast_initial[] = {
  _swigc__p_allocator_type,
  _swigc__p_char,
  _swigc__p_const_reference,
  _swigc__p_difference_type,
  _swigc__p_key_type,
  _swigc__p_p_PyObject,
  _swigc__p_p_steps__tetmesh__TmPatch,
  _swigc__p_reference,
  _swigc__p_size_type,
  _swigc__p_std__invalid_argument,
  _swigc__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t,
  _swigc__p_std__setTstd__string_std__lessTstd__string_t_std__allocatorTstd__string_t_t__allocator_type,
  _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t,
  _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type,
  _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type,
  _swigc__p_std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type,
  _swigc__p_std__vectorTbool_std__allocatorTbool_t_t,
  _swigc__p_std__vectorTbool_std__allocatorTbool_t_t__allocator_type,
  _swigc__p_std__vectorTdouble_std__allocatorTdouble_t_t,
  _swigc__p_std__vectorTdouble_std__allocatorTdouble_t_t__allocator_type,
  _swigc__p_std__vectorTint_std__allocatorTint_t_t,
  _swigc__p_std__vectorTint_std__allocatorTint_t_t__allocator_type,
  _swigc__p_std__vectorTstd__string_std__allocatorTstd__string_t_t,
  _swigc__p_std__vectorTstd__string_std__allocatorTstd__string_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t,
  _swigc__p_std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t,
  _swigc__p_std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t,
  _swigc__p_std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t,
  _swigc__p_std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t,
  _swigc__p_std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t,
  _swigc__p_std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t,
  _swigc__p_std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__allocator_type,
  _swigc__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t,
  _swigc__p_std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__allocator_type,
  _swigc__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t,
  _swigc__p_std__vectorTunsigned_int_std__allocatorTunsigned_int_t_t__allocator_type,
  _swigc__p_steps__ArgErr,
  _swigc__p_steps__Err,
  _swigc__p_steps__NotImplErr,
  _swigc__p_steps__model__Diff,
  _swigc__p_steps__model__Model,
  _swigc__p_steps__model__Reac,
  _swigc__p_steps__model__SReac,
  _swigc__p_steps__model__Spec,
  _swigc__p_steps__model__Surfsys,
  _swigc__p_steps__model__Volsys,
  _swigc__p_steps__rng__RNG,
  _swigc__p_steps__solver__API,
  _swigc__p_steps__tetexact__Tetexact,
  _swigc__p_steps__tetmesh__DiffBoundary,
  _swigc__p_steps__tetmesh__Tetmesh,
  _swigc__p_steps__tetmesh__TmComp,
  _swigc__p_steps__tetmesh__TmPatch,
  _swigc__p_steps__wm__Comp,
  _swigc__p_steps__wm__Geom,
  _swigc__p_steps__wm__Patch,
  _swigc__p_steps__wmdirect__Wmdirect,
  _swigc__p_steps__wmrk4__Wmrk4,
  _swigc__p_swig__PySwigIterator,
  _swigc__p_unsigned_char,
  _swigc__p_unsigned_int,
  _swigc__p_unsigned_long,
  _swigc__p_unsigned_short,
  _swigc__p_value_type,
  _swigc__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__key_type,
  _swigc__std__setTsteps__wm__Patch_p_std__lessTsteps__wm__Patch_p_t_std__allocatorTsteps__wm__Patch_p_t_t__value_type,
  _swigc__std__vectorTsteps__model__Diff_p_std__allocatorTsteps__model__Diff_p_t_t__value_type,
  _swigc__std__vectorTsteps__model__Reac_p_std__allocatorTsteps__model__Reac_p_t_t__value_type,
  _swigc__std__vectorTsteps__model__SReac_p_std__allocatorTsteps__model__SReac_p_t_t__value_type,
  _swigc__std__vectorTsteps__model__Spec_p_std__allocatorTsteps__model__Spec_p_t_t__value_type,
  _swigc__std__vectorTsteps__model__Surfsys_p_std__allocatorTsteps__model__Surfsys_p_t_t__value_type,
  _swigc__std__vectorTsteps__model__Volsys_p_std__allocatorTsteps__model__Volsys_p_t_t__value_type,
  _swigc__std__vectorTsteps__wm__Comp_p_std__allocatorTsteps__wm__Comp_p_t_t__value_type,
  _swigc__std__vectorTsteps__wm__Patch_p_std__allocatorTsteps__wm__Patch_p_t_t__value_type,
};


/* -------- TYPE CONVERSION AND EQUIVALENCE RULES (END) -------- */

static swig_const_info swig_const_table[] = {
{0, 0, 0, 0.0, 0, 0}};

#ifdef __cplusplus
}
#endif
/* -----------------------------------------------------------------------------
 * Type initialization:
 * This problem is tough by the requirement that no dynamic 
 * memory is used. Also, since swig_type_info structures store pointers to 
 * swig_cast_info structures and swig_cast_info structures store pointers back
 * to swig_type_info structures, we need some lookup code at initialization. 
 * The idea is that swig generates all the structures that are needed. 
 * The runtime then collects these partially filled structures. 
 * The SWIG_InitializeModule function takes these initial arrays out of 
 * swig_module, and does all the lookup, filling in the swig_module.types
 * array with the correct data and linking the correct swig_cast_info
 * structures together.
 *
 * The generated swig_type_info structures are assigned staticly to an initial 
 * array. We just loop through that array, and handle each type individually.
 * First we lookup if this type has been already loaded, and if so, use the
 * loaded structure instead of the generated one. Then we have to fill in the
 * cast linked list. The cast data is initially stored in something like a
 * two-dimensional array. Each row corresponds to a type (there are the same
 * number of rows as there are in the swig_type_initial array). Each entry in
 * a column is one of the swig_cast_info structures for that type.
 * The cast_initial array is actually an array of arrays, because each row has
 * a variable number of columns. So to actually build the cast linked list,
 * we find the array of casts associated with the type, and loop through it 
 * adding the casts to the list. The one last trick we need to do is making
 * sure the type pointer in the swig_cast_info struct is correct.
 *
 * First off, we lookup the cast->type name to see if it is already loaded. 
 * There are three cases to handle:
 *  1) If the cast->type has already been loaded AND the type we are adding
 *     casting info to has not been loaded (it is in this module), THEN we
 *     replace the cast->type pointer with the type pointer that has already
 *     been loaded.
 *  2) If BOTH types (the one we are adding casting info to, and the 
 *     cast->type) are loaded, THEN the cast info has already been loaded by
 *     the previous module so we just ignore it.
 *  3) Finally, if cast->type has not already been loaded, then we add that
 *     swig_cast_info to the linked list (because the cast->type) pointer will
 *     be correct.
 * ----------------------------------------------------------------------------- */

#ifdef __cplusplus
extern "C" {
#if 0
} /* c-mode */
#endif
#endif

#if 0
#define SWIGRUNTIME_DEBUG
#endif


SWIGRUNTIME void
SWIG_InitializeModule(void *clientdata) {
  size_t i;
  swig_module_info *module_head, *iter;
  int found;
  
  clientdata = clientdata;
  
  /* check to see if the circular list has been setup, if not, set it up */
  if (swig_module.next==0) {
    /* Initialize the swig_module */
    swig_module.type_initial = swig_type_initial;
    swig_module.cast_initial = swig_cast_initial;
    swig_module.next = &swig_module;
  }
  
  /* Try and load any already created modules */
  module_head = SWIG_GetModule(clientdata);
  if (!module_head) {
    /* This is the first module loaded for this interpreter */
    /* so set the swig module into the interpreter */
    SWIG_SetModule(clientdata, &swig_module);
    module_head = &swig_module;
  } else {
    /* the interpreter has loaded a SWIG module, but has it loaded this one? */
    found=0;
    iter=module_head;
    do {
      if (iter==&swig_module) {
        found=1;
        break;
      }
      iter=iter->next;
    } while (iter!= module_head);
    
    /* if the is found in the list, then all is done and we may leave */
    if (found) return;
    /* otherwise we must add out module into the list */
    swig_module.next = module_head->next;
    module_head->next = &swig_module;
  }
  
  /* Now work on filling in swig_module.types */
#ifdef SWIGRUNTIME_DEBUG
  printf("SWIG_InitializeModule: size %d\n", swig_module.size);
#endif
  for (i = 0; i < swig_module.size; ++i) {
    swig_type_info *type = 0;
    swig_type_info *ret;
    swig_cast_info *cast;
    
#ifdef SWIGRUNTIME_DEBUG
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
#endif
    
    /* if there is another module already loaded */
    if (swig_module.next != &swig_module) {
      type = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, swig_module.type_initial[i]->name);
    }
    if (type) {
      /* Overwrite clientdata field */
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: found type %s\n", type->name);
#endif
      if (swig_module.type_initial[i]->clientdata) {
        type->clientdata = swig_module.type_initial[i]->clientdata;
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: found and overwrite type %s \n", type->name);
#endif
      }
    } else {
      type = swig_module.type_initial[i];
    }
    
    /* Insert casting types */
    cast = swig_module.cast_initial[i];
    while (cast->type) {
      /* Don't need to add information already in the list */
      ret = 0;
#ifdef SWIGRUNTIME_DEBUG
      printf("SWIG_InitializeModule: look cast %s\n", cast->type->name);
#endif
      if (swig_module.next != &swig_module) {
        ret = SWIG_MangledTypeQueryModule(swig_module.next, &swig_module, cast->type->name);
#ifdef SWIGRUNTIME_DEBUG
        if (ret) printf("SWIG_InitializeModule: found cast %s\n", ret->name);
#endif
      }
      if (ret) {
        if (type == swig_module.type_initial[i]) {
#ifdef SWIGRUNTIME_DEBUG
          printf("SWIG_InitializeModule: skip old type %s\n", ret->name);
#endif
          cast->type = ret;
          ret = 0;
        } else {
          /* Check for casting already in the list */
          swig_cast_info *ocast = SWIG_TypeCheck(ret->name, type);
#ifdef SWIGRUNTIME_DEBUG
          if (ocast) printf("SWIG_InitializeModule: skip old cast %s\n", ret->name);
#endif
          if (!ocast) ret = 0;
        }
      }
      
      if (!ret) {
#ifdef SWIGRUNTIME_DEBUG
        printf("SWIG_InitializeModule: adding cast %s\n", cast->type->name);
#endif
        if (type->cast) {
          type->cast->prev = cast;
          cast->next = type->cast;
        }
        type->cast = cast;
      }
      cast++;
    }
    /* Set entry in modules->types array equal to the type */
    swig_module.types[i] = type;
  }
  swig_module.types[i] = 0;
  
#ifdef SWIGRUNTIME_DEBUG
  printf("**** SWIG_InitializeModule: Cast List ******\n");
  for (i = 0; i < swig_module.size; ++i) {
    int j = 0;
    swig_cast_info *cast = swig_module.cast_initial[i];
    printf("SWIG_InitializeModule: type %d %s\n", i, swig_module.type_initial[i]->name);
    while (cast->type) {
      printf("SWIG_InitializeModule: cast type %s\n", cast->type->name);
      cast++;
      ++j;
    }
    printf("---- Total casts: %d\n",j);
  }
  printf("**** SWIG_InitializeModule: Cast List ******\n");
#endif
}

/* This function will propagate the clientdata field of type to
* any new swig_type_info structures that have been added into the list
* of equivalent types.  It is like calling
* SWIG_TypeClientData(type, clientdata) a second time.
*/
SWIGRUNTIME void
SWIG_PropagateClientData(void) {
  size_t i;
  swig_cast_info *equiv;
  static int init_run = 0;
  
  if (init_run) return;
  init_run = 1;
  
  for (i = 0; i < swig_module.size; i++) {
    if (swig_module.types[i]->clientdata) {
      equiv = swig_module.types[i]->cast;
      while (equiv) {
        if (!equiv->converter) {
          if (equiv->type && !equiv->type->clientdata)
          SWIG_TypeClientData(equiv->type, swig_module.types[i]->clientdata);
        }
        equiv = equiv->next;
      }
    }
  }
}

#ifdef __cplusplus
#if 0
{
  /* c-mode */
#endif
}
#endif



#ifdef __cplusplus
extern "C" {
#endif
  
  /* Python-specific SWIG API */
#define SWIG_newvarlink()                             SWIG_Python_newvarlink()
#define SWIG_addvarlink(p, name, get_attr, set_attr)  SWIG_Python_addvarlink(p, name, get_attr, set_attr)
#define SWIG_InstallConstants(d, constants)           SWIG_Python_InstallConstants(d, constants)
  
  /* -----------------------------------------------------------------------------
   * global variable support code.
   * ----------------------------------------------------------------------------- */
  
  typedef struct swig_globalvar {
    char       *name;                  /* Name of global variable */
    PyObject *(*get_attr)(void);       /* Return the current value */
    int       (*set_attr)(PyObject *); /* Set the value */
    struct swig_globalvar *next;
  } swig_globalvar;
  
  typedef struct swig_varlinkobject {
    PyObject_HEAD
    swig_globalvar *vars;
  } swig_varlinkobject;
  
  SWIGINTERN PyObject *
  swig_varlink_repr(swig_varlinkobject *SWIGUNUSEDPARM(v)) {
    return PyString_FromString("<Swig global variables>");
  }
  
  SWIGINTERN PyObject *
  swig_varlink_str(swig_varlinkobject *v) {
    PyObject *str = PyString_FromString("(");
    swig_globalvar  *var;
    for (var = v->vars; var; var=var->next) {
      PyString_ConcatAndDel(&str,PyString_FromString(var->name));
      if (var->next) PyString_ConcatAndDel(&str,PyString_FromString(", "));
    }
    PyString_ConcatAndDel(&str,PyString_FromString(")"));
    return str;
  }
  
  SWIGINTERN int
  swig_varlink_print(swig_varlinkobject *v, FILE *fp, int SWIGUNUSEDPARM(flags)) {
    PyObject *str = swig_varlink_str(v);
    fprintf(fp,"Swig global variables ");
    fprintf(fp,"%s\n", PyString_AsString(str));
    Py_DECREF(str);
    return 0;
  }
  
  SWIGINTERN void
  swig_varlink_dealloc(swig_varlinkobject *v) {
    swig_globalvar *var = v->vars;
    while (var) {
      swig_globalvar *n = var->next;
      free(var->name);
      free(var);
      var = n;
    }
  }
  
  SWIGINTERN PyObject *
  swig_varlink_getattr(swig_varlinkobject *v, char *n) {
    PyObject *res = NULL;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->get_attr)();
        break;
      }
      var = var->next;
    }
    if (res == NULL && !PyErr_Occurred()) {
      PyErr_SetString(PyExc_NameError,"Unknown C global variable");
    }
    return res;
  }
  
  SWIGINTERN int
  swig_varlink_setattr(swig_varlinkobject *v, char *n, PyObject *p) {
    int res = 1;
    swig_globalvar *var = v->vars;
    while (var) {
      if (strcmp(var->name,n) == 0) {
        res = (*var->set_attr)(p);
        break;
      }
      var = var->next;
    }
    if (res == 1 && !PyErr_Occurred()) {
      PyErr_SetString(PyExc_NameError,"Unknown C global variable");
    }
    return res;
  }
  
  SWIGINTERN PyTypeObject*
  swig_varlink_type(void) {
    static char varlink__doc__[] = "Swig var link object";
    static PyTypeObject varlink_type;
    static int type_init = 0;  
    if (!type_init) {
      const PyTypeObject tmp
      = {
        PyObject_HEAD_INIT(NULL)
        0,                                  /* Number of items in variable part (ob_size) */
        (char *)"swigvarlink",              /* Type name (tp_name) */
        sizeof(swig_varlinkobject),         /* Basic size (tp_basicsize) */
        0,                                  /* Itemsize (tp_itemsize) */
        (destructor) swig_varlink_dealloc,   /* Deallocator (tp_dealloc) */ 
        (printfunc) swig_varlink_print,     /* Print (tp_print) */
        (getattrfunc) swig_varlink_getattr, /* get attr (tp_getattr) */
        (setattrfunc) swig_varlink_setattr, /* Set attr (tp_setattr) */
        0,                                  /* tp_compare */
        (reprfunc) swig_varlink_repr,       /* tp_repr */
        0,                                  /* tp_as_number */
        0,                                  /* tp_as_sequence */
        0,                                  /* tp_as_mapping */
        0,                                  /* tp_hash */
        0,                                  /* tp_call */
        (reprfunc)swig_varlink_str,        /* tp_str */
        0,                                  /* tp_getattro */
        0,                                  /* tp_setattro */
        0,                                  /* tp_as_buffer */
        0,                                  /* tp_flags */
        varlink__doc__,                     /* tp_doc */
        0,                                  /* tp_traverse */
        0,                                  /* tp_clear */
        0,                                  /* tp_richcompare */
        0,                                  /* tp_weaklistoffset */
#if PY_VERSION_HEX >= 0x02020000
        0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0, /* tp_iter -> tp_weaklist */
#endif
#if PY_VERSION_HEX >= 0x02030000
        0,                                  /* tp_del */
#endif
#ifdef COUNT_ALLOCS
        0,0,0,0                             /* tp_alloc -> tp_next */
#endif
      };
      varlink_type = tmp;
      varlink_type.ob_type = &PyType_Type;
      type_init = 1;
    }
    return &varlink_type;
  }
  
  /* Create a variable linking object for use later */
  SWIGINTERN PyObject *
  SWIG_Python_newvarlink(void) {
    swig_varlinkobject *result = PyObject_NEW(swig_varlinkobject, swig_varlink_type());
    if (result) {
      result->vars = 0;
    }
    return ((PyObject*) result);
  }
  
  SWIGINTERN void 
  SWIG_Python_addvarlink(PyObject *p, char *name, PyObject *(*get_attr)(void), int (*set_attr)(PyObject *p)) {
    swig_varlinkobject *v = (swig_varlinkobject *) p;
    swig_globalvar *gv = (swig_globalvar *) malloc(sizeof(swig_globalvar));
    if (gv) {
      size_t size = strlen(name)+1;
      gv->name = (char *)malloc(size);
      if (gv->name) {
        strncpy(gv->name,name,size);
        gv->get_attr = get_attr;
        gv->set_attr = set_attr;
        gv->next = v->vars;
      }
    }
    v->vars = gv;
  }
  
  SWIGINTERN PyObject *
  SWIG_globals(void) {
    static PyObject *_SWIG_globals = 0; 
    if (!_SWIG_globals) _SWIG_globals = SWIG_newvarlink();  
    return _SWIG_globals;
  }
  
  /* -----------------------------------------------------------------------------
   * constants/methods manipulation
   * ----------------------------------------------------------------------------- */
  
  /* Install Constants */
  SWIGINTERN void
  SWIG_Python_InstallConstants(PyObject *d, swig_const_info constants[]) {
    PyObject *obj = 0;
    size_t i;
    for (i = 0; constants[i].type; ++i) {
      switch(constants[i].type) {
      case SWIG_PY_POINTER:
        obj = SWIG_NewPointerObj(constants[i].pvalue, *(constants[i]).ptype,0);
        break;
      case SWIG_PY_BINARY:
        obj = SWIG_NewPackedObj(constants[i].pvalue, constants[i].lvalue, *(constants[i].ptype));
        break;
      default:
        obj = 0;
        break;
      }
      if (obj) {
        PyDict_SetItemString(d, constants[i].name, obj);
        Py_DECREF(obj);
      }
    }
  }
  
  /* -----------------------------------------------------------------------------*/
  /* Fix SwigMethods to carry the callback ptrs when needed */
  /* -----------------------------------------------------------------------------*/
  
  SWIGINTERN void
  SWIG_Python_FixMethods(PyMethodDef *methods,
    swig_const_info *const_table,
    swig_type_info **types,
    swig_type_info **types_initial) {
    size_t i;
    for (i = 0; methods[i].ml_name; ++i) {
      const char *c = methods[i].ml_doc;
      if (c && (c = strstr(c, "swig_ptr: "))) {
        int j;
        swig_const_info *ci = 0;
        const char *name = c + 10;
        for (j = 0; const_table[j].type; ++j) {
          if (strncmp(const_table[j].name, name, 
              strlen(const_table[j].name)) == 0) {
            ci = &(const_table[j]);
            break;
          }
        }
        if (ci) {
          size_t shift = (ci->ptype) - types;
          swig_type_info *ty = types_initial[shift];
          size_t ldoc = (c - methods[i].ml_doc);
          size_t lptr = strlen(ty->name)+2*sizeof(void*)+2;
          char *ndoc = (char*)malloc(ldoc + lptr + 10);
          if (ndoc) {
            char *buff = ndoc;
            void *ptr = (ci->type == SWIG_PY_POINTER) ? ci->pvalue : 0;
            if (ptr) {
              strncpy(buff, methods[i].ml_doc, ldoc);
              buff += ldoc;
              strncpy(buff, "swig_ptr: ", 10);
              buff += 10;
              SWIG_PackVoidPtr(buff, ptr, ty->name, lptr);
              methods[i].ml_doc = ndoc;
            }
          }
        }
      }
    }
  } 
  
#ifdef __cplusplus
}
#endif

/* -----------------------------------------------------------------------------*
 *  Partial Init method
 * -----------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C"
#endif
SWIGEXPORT void SWIG_init(void) {
  PyObject *m, *d;
  
  /* Fix SwigMethods to carry the callback ptrs when needed */
  SWIG_Python_FixMethods(SwigMethods, swig_const_table, swig_types, swig_type_initial);
  
  m = Py_InitModule((char *) SWIG_name, SwigMethods);
  d = PyModule_GetDict(m);
  
  SWIG_InitializeModule(0);
  SWIG_InstallConstants(d,swig_const_table);
  
  
}

